Package: PTXQC
Type: Package
Title: Quality Report Generation for MaxQuant and mzTab Results
Version: 1.1.6
Date: 2026-08-20
Authors@R: c(person(given = "Chris",
                      family = "Bielow",
                      role = c("aut", "cre"),
                      email = "chris.bielow@bsc.fu-berlin.de"),
               person(given = "Juliane",
                      family = "Schmachtenberg",
                      role = "ctb"),
               person(given = "Swenja",
                      family = "Wagner",
                      role = "ctb"),
               person(given = "Patricia",
                      family = "Scheil",
                      role = "ctb"),
               person(given = "Tom",
                      family = "Waschischek",
                      role = "ctb"),
               person(given = "Guido",
                      family = "Mastrobuoni",
                      role = c("dtc", "rev")))
Description: Generates Proteomics (PTX) quality control (QC) reports for shotgun LC-MS data analyzed with the 
             MaxQuant software suite (from .txt files) or mzTab files (ideally from OpenMS 'QualityControl' tool).
             Reports are customizable (target thresholds, subsetting) and available in HTML or PDF format.
             Published in J. Proteome Res., Proteomics Quality Control: Quality Control Software for MaxQuant Results (2015)
             <doi:10.1021/acs.jproteome.5b00780>.
SystemRequirements: pandoc (http://pandoc.org) for building Vignettes
        and output reports as HTML
Depends: R (>= 3.3.0)
Imports: data.table, ggplot2 (>= 3.4), ggdendro, grid, gridExtra,
        grDevices, gtable, htmlTable, knitr (>= 1.10), magrittr,
        methods, plyr, R6, R6P, RColorBrewer, reshape2, rmarkdown,
        rmzqc (>= 0.5.0), seqinr, stats, utils, UpSetR, xml2, yaml
Suggests: testthat
VignetteBuilder: knitr
License: BSD_3_clause + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.2
URL: https://github.com/cbielow/PTXQC
BugReports: https://github.com/cbielow/PTXQC/issues
NeedsCompilation: no
Packaged: 2026-08-21 20:31:07 UTC; bielow
Author: Chris Bielow [aut, cre],
  Juliane Schmachtenberg [ctb],
  Swenja Wagner [ctb],
  Patricia Scheil [ctb],
  Tom Waschischek [ctb],
  Guido Mastrobuoni [dtc, rev]
Maintainer: Chris Bielow <chris.bielow@bsc.fu-berlin.de>
Repository: CRAN
Date/Publication: 2026-08-21 23:50:07 UTC

# Generated by roxygen2: do not edit by hand

S3method(print,PTXQC_table)
export(FilenameMapper)
export(LCS)
export(LCSn)
export(QCMetaFilenames)
export(RTalignmentTree)
export(YAMLClass)
export(assembleMZQC)
export(assignBlocks)
export(boxplotCompare)
export(byX)
export(byXflex)
export(correctSetSize)
export(createReport)
export(createYaml)
export(delLCP)
export(delLCS)
export(getECDF)
export(getFragmentErrors)
export(getHTMLTable)
export(getMQPARValue)
export(getMaxima)
export(getPCA)
export(getReportFilenames)
export(grepv)
export(lcpCount)
export(lcsCount)
export(longestCommonPrefix)
export(longestCommonSuffix)
export(modsToTable)
export(modsToTableByRaw)
export(mosaicize)
export(pasten)
export(pastet)
export(peakWidthOverTime)
export(plotTable)
export(plot_CalibratedMSErr)
export(plot_Charge)
export(plot_ContEVD)
export(plot_ContUser)
export(plot_ContUserScore)
export(plot_ContsPG)
export(plot_CountData)
export(plot_DataOverRT)
export(plot_IDRate)
export(plot_IDsOverRT)
export(plot_IonInjectionTimeOverRT)
export(plot_MBRAlign)
export(plot_MBRIDtransfer)
export(plot_MBRgain)
export(plot_MS2Decal)
export(plot_MS2Oversampling)
export(plot_MissedCleavages)
export(plot_RTPeakWidth)
export(plot_RatiosPG)
export(plot_ScanIDRate)
export(plot_TIC)
export(plot_TopN)
export(plot_TopNoverRT)
export(plot_UncalibratedMSErr)
export(plot_peptideMods)
export(pointsPutX)
export(qcMetric)
export(qualBestKS)
export(qualCentered)
export(qualCenteredRef)
export(qualGaussDev)
export(qualHighest)
export(qualLinThresh)
export(qualMedianDist)
export(qualUniform)
export(read.MQ)
export(renameFile)
export(repEach)
export(shortenStrings)
export(simplifyNames)
export(supCount)
export(theme_blank)
exportClasses(FilenameMapper)
exportClasses(YAMLClass)
exportClasses(qcMetric)
import(R6)
import(R6P)
import(RColorBrewer)
import(UpSetR)
import(data.table)
import(ggdendro)
import(ggplot2)
import(grDevices)
import(grid)
import(gridExtra)
import(gtable)
import(htmlTable)
import(knitr)
import(magrittr)
import(methods)
import(plyr)
import(reshape2, except = c(dcast, melt))
import(rmarkdown)
import(rmzqc)
import(stats)
import(utils)
import(xml2)
import(yaml)
importFrom(seqinr,circle)

PTXQC
---------------

[![License (3-Clause BSD)](https://img.shields.io/badge/license-BSD%203--Clause-blue.svg?style=flat-square)](https://opensource.org/license/bsd-3-clause)
[![Project Stats](https://openhub.net/p/PTXQC/widgets/project_thin_badge.gif)](https://openhub.net/p/PTXQC)

**This package allows users of MaxQuant (from .txt files) and OpenMS (from mzTab files) to generate quality control reports in Html/PDF format.**

### Latest changes / ChangeLog

latest Release: v1.1.6 - August 2026<br>
latest Release on CRAN: same
  
See [NEWS][News_File] file for a version history.

### Platform support
  - Windows (recommended for convenience to make use of the [drag'n'drop][1] batch file provided)
  - Linux
  - MacOSX
  
### Citation

Please cite PTXQC when using it to check data in your publications:

**Proteomics Quality Control: Quality Control Software for MaxQuant Results**
Chris Bielow, Guido Mastrobuoni, and Stefan Kempa
*J. Proteome Res.*, 2016, 15 (3), pp 777-787.
DOI: [10.1021/acs.jproteome.5b00780][JPR_paper]
  
### Features
  - plethora of quality metrics
    - intensity distributions
    - digestion efficiency
    - contaminant visualizations
    - identification performance
    - Match-between-runs performance
  - easy usage ([Windows OS only] `drag'n'drop` your `txt output folder` onto a `batch file`)
    - 10 min [Installation](#installation)
  - MaxQuant locale check, i.e. dot as decimal separator (since PTXQC 1.0.10; see https://github.com/cbielow/PTXQC/pull/99 for details)
  - Html/PDF report will be generated within your MaxQuant-txt folder or next to the mzTab file
  - writes a mockup mzQC file (https://github.com/HUPO-PSI/mzQC/) for archiving or downstream processing (actual metrics still require exporting; PR's welcome)
  - optional configuration file *in YAML format* for generation of shorter/customized reports

### Target audience
  - MaxQuant users (no knowledge of R required)
  - OpenMS users (or any other software which can write an mzTab)
  - bioinformaticians (who want to contribute or customize)


### Documentation

A short **overview video** on PTX-QC] can be found [here](https://youtu.be/sb-mydbNRS4). 
If you want to jump to certain sections:

  - [Motivation for doing QC](https://youtu.be/sb-mydbNRS4?t=75)
  - [Installation demo](https://youtu.be/sb-mydbNRS4?t=522)
  - [Report Generation using Drag'n'Drop](https://youtu.be/sb-mydbNRS4?t=880)
  - [Inspecting and configuring a Report](https://youtu.be/sb-mydbNRS4?t=1043)
  
We use [pgkdown](https://pkgdown.r-lib.org/) to create a HTML documentation, which includes the Vignettes, our function documentation etc.
You can create the documentation locally: `pkgdown::build_site()` or visit the online-version on Github at [./docs/index.html](http://htmlpreview.github.io/?https://github.com/cbielow/PTXQC/blob/master/docs/index.html).

If you do not know where to start, look at the package vignettes first.

Our Vignettes give details on:

  - Full List of Quality Metrics with help text ('List of Metrics')
  - Input and Output
  - Report customization
  - (for MaxQuant/OpenMS users) Usage of Drag'n'drop
  - (for R users) Code examples in R

The 'List of Metrics' vignette contains a full description for each metric (the same as seen in the Help section of each Html report).

Within R, you can browse the vignettes using either of these commands (*after* the package is installed (see below)):
```
    help(package="PTXQC")
    browseVignettes(package = 'PTXQC')
```

Of course, you can also look at the Vignettes on CRAN:
[latest online vignette at CRAN](https://cran.r-project.org/package=PTXQC)

  
### Installation

**You can use PTX-QC without installing it, by using our webserver:**
Visit http://ptxqc.bsc.fu-berlin.de and simply upload your data.
This service is not suited for large-scale data analysis, but should be fine for the occasional analysis.

**If you want to generate QC reports without actually getting involved in R:**

We offer a Batch-file based Drag'n'drop mechanism to trigger PTXQC on any MaxQuant output folder.
This only works for Windows (not Linux or MacOS) at the moment -- but you have a Windows anyway to run MaxQuant, right?!
See [drag'n'drop][1] for details. It takes 10 minutes and you are done!

**If you just want to use PTXQC (and maybe even modify) it:**

*First*, install [pandoc](https://github.com/jgm/pandoc/releases) (see bottom of linked page). Pandoc is required in order to locally build the package vignettes (documentation),
but you can also read the [vignettes][Ref_Vign] online from the PTXQC GitHub page. More importantly, Pandoc enables PTXQC to write QC reports in HTML format (which come
with a help text for each plot and are interactive). PDF reports only contain plots!
The reports are printed as PDF by default and additionally as HTML if Pandoc is found.
**If you install Pandoc later while your R session is already open, you need to close and re-open R in order to make R aware of Pandoc!**
   
You can grab PTXQC from either [CRAN](https://CRAN.R-project.org/package=PTXQC) *or* [GitHub](https://github.com/cbielow/PTXQC#installation).
GitHub installation will give you the latest package; the CRAN version might be a little older, but is faster to install. Check the [NEWS][News_File] file for CRAN submissions and version.
> For the code blocks below: Run **each line** separately in your R console, i.e. do not copy and paste the whole block.
> If an error should occur, this allows to track it down more easily. See [FAQ - Installation][Ref_VignFAQ]
> how to resolve them.

    ## CRAN
    install.packages("PTXQC")
or

    ## GitHub
    if (!require(devtools, quietly = TRUE)) install.packages("devtools")
    library("devtools")             ## this might give a warning like 'WARNING: Rtools is required ...'. Ignore it.
    
    ## use build_vignettes = FALSE if you did not install pandoc or if you encounter errors when building vignettes (e.g. PRIDE ftp unavailable)!
    install_github("cbielow/PTXQC", build_vignettes = TRUE, dependencies = TRUE)


### Usage

To get started, see the help and/or vignettes:

    browseVignettes(package = 'PTXQC')
    help(package="PTXQC")

The simplest and fastest way to create a report is

    library(PTXQC)
    createReport("c:/PATH/TO/DATAs/COMBINED/TXT")


### Report Examples

An overview chart at the beginning of the report will give you a first impression.
<img src="./inst/examples/example_heatmap.png?raw=true" width="500" /><br>
Detailed plots can be found in the remainder of each report.

For example input data and full reports, see the ['inst/examples' subfolder][2].

### Bug reporting / Feature requests

If you encounter a bug, e.g. error message, wrong figures, missing axis annotation or anything which looks
suspicious, please use the [GitHub issue tracker][issuetracker] and file a report.

You should include
  - **stage** you encounter the bug, e.g. during installation, report creation, or after report creation (i.e. a bug in the report itself).
  - **PDF/Html report** itself (if one was generated).
  - **version of PTXQC**, e.g. see the report_XXX.pdf/html (where XXX will be the version) or see the DESCRIPTION file of the PTXQC package or call `help(package="PTXQC")` within R
  - **error message** (very important!). Either copy it or provide a screen shot.

Please be as precise as possible when providing the bug report - just imagine what kind of information you would like to have in order
to track down the issue.
In certain situations, the whole txt-folder or a single MaxQuant/mzTab file might be helpful to solve the problem.

### Contributing - Get Involved!

We welcome input from our user base!
PTX-QC has a very permissive **BSD-3 clause License** (see [DESCRIPTION](DESCRIPTION) file), so feel free to fork, patch and contribute!
There are many ways to get involved, _you do not need to be a developer_!
  - suggest a new metric (and why you think it's useful) by opening [a new ticket][issuetracker] here on GitHub.
  - suggest changes to existing metrics (improvements or bugfixes), see above.
  - suggest improvements to our documentation (e.g. [additional vignettes][Ref_Vign])
  - write code (in R) and submit a [Pull Request (PR)][PullRequest].

### PTX-QC user forum

Come and interact with developers and users at https://discord.gg/MB6PvpctUY


### Misc

Use [PTXQC v0.69.3][JPR_PTXQC] if you want the version which was used in the paper, i.e.
use `install_github(..., ref="v0.69.3")` when following the [Installation](#installation) procedure.

The input data used in the original publication is available in the ['inst/examples' subfolder][2].

We recommend to use the most recent PTXQC for the best user experience.
  
  [1]: https://github.com/cbielow/PTXQC/tree/master/inst/dragNdrop
  [2]: https://github.com/cbielow/PTXQC/tree/master/inst/examples
  [issuetracker]: https://github.com/cbielow/PTXQC/issues
  [PullRequest]: https://github.com/cbielow/PTXQC/pulls
  [JPR_PTXQC]: https://github.com/cbielow/PTXQC/releases/tag/v0.69.3
  [Ref_VignFAQ]: https://github.com/cbielow/PTXQC/blob/master/vignettes/PTXQC-FAQ.Rmd
  [Ref_Vign]: https://github.com/cbielow/PTXQC/tree/master/vignettes
  [News_File]: https://github.com/cbielow/PTXQC/blob/master/NEWS
  [JPR_paper]: https://doi.org/10.1021/acs.jproteome.5b00780

This file is for admins/support-staff who want to configure and install the PTXQC package
and provide an automated drag'n'drop solution to their MaxQuant users (incl. themselves) for creating QC reports.
Please contact an admin or proficient R user, if you are new to installing R packages.

**Note:** the drag'n'drop currently only works for Windows, not MacOSX or Linux. 
          For the latter, please refer to the [vignettes][Ref_Vign] of this package (after installation), which explain how to invoke PTXQC from within R.

We recommend using a shared network drive as installation folder to which all users have read access. Alternatively you can copy the folder structure we are about to create to a folder on your local machine. However, if you want to use the QC reporting from multiple PC's just use a network folder.

When you are done, provide the users with the [PTXQC-DragNDrop][Ref_VignDrag] vignette (or write your own).
This will show them how to invoke the QC (it's very easy - see 'Usage' below).

### Installation
  
In order to make PTXQC available as a standalone application on a network drive, we will need to make a copy of the whole R installation
directory as one of the last steps. If you use an existing R installation which already contains a large package library, you might want to find a fresh PC without R
and install R from scratch, such that you do not end up copying useless packages which are not relevant for PTXQC.
If you are a knowledgeable R user already, you can probably find other ways around it (e.g. a temporary library, or selective copying of packages, but
this is out of scope of this manual).

 - create a clean R installation by downloading R from [http://www.r-project.org] and installing it. We highly recommend the 64bit version to avoid out-of-memory errors.
   If you already have R installed, you may skip this step.
 - install [pandoc](https://github.com/jgm/pandoc/releases) (see bottom of linked page). Pandoc is required in order to locally build the package vignettes (documentation),
but you can also read the [vignettes][Ref_Vign] online from the PTXQC GitHub page. More importantly, Pandoc enables PTXQC to write QC reports in HTML format (which come
with a help text for each plot and are interactive). PDF reports only contain plots!
The reports are printed as PDF by default and additionally as HTML if Pandoc is found.
   **If you install Pandoc later while your R session is already open, you need to close and re-open R in order to make R aware of Pandoc!**
 - Now, we install PTXQC and its dependency packages.
   Start the R GUI (64bit!) and execute the following commands to install the `PTXQC` package (dependencies are installed automatically).
   
You can grab PTXQC from either [CRAN](https://cran.r-project.org/web/packages/PTXQC/index.html) *or* [GitHub](https://github.com/cbielow/PTXQC#installation).
GitHub installation will give you the latest package; the CRAN version might be a little older, but is faster to install. Check the [NEWS][News_File] file for CRAN submissions and version.
> For the code blocks below: Run **each line** separately in your R console, i.e. do not copy and paste the whole block.
> If an error should occur, this allows to track it down more easily. See [FAQ - Installation][Ref_VignFAQ]
> how to resolve them.

    ## CRAN
    install.packages("PTXQC")
or

    ## GitHub
    if (!require(devtools, quietly = TRUE)) install.packages("devtools")
    library("devtools")             ## this might give a warning like 'WARNING: Rtools is required ...'. Ignore it.
    
	## use build_vignettes = FALSE if you did not install pandoc or if you encounter errors when building vignettes (e.g. PRIDE ftp unavailable)!
	install_github("cbielow/PTXQC", build_vignettes = TRUE, dependencies = TRUE)

Finally, always run:

    cat(paste0("\nPTXQC was installed to '", .libPaths()[1], "'.\n\n"))

> **Here's some intuition about the next steps**   
> To make PTXQC Drag'n'drop a standalone application which can be run from any user and even from a network drive,
> we need a copy of the whole R installation. Within the PTXQC package (which is part of the R installation) we have prepared a 
> drag'n'drop batch script which invokes R and calls PTXQC. We need to extract this script (i.e. copy it to a top-level folder)
> and then place the whole R installation beneath it. This includes all your R-libraries, which might reside in a different place.

Almost done, we just need to copy some folders.
        
 1. The last command will tell you where R installed PTXQC, i.e. `PTXQC was installed to '<libR>'.`
    **Open** this **`<libR>`** folder in your file explorer. It will either be your default R library 
    folder which comes with R (e.g. `C:\Program Files\R\R-3.1.0\library`), or a user-specific library folder like 
    this `C:/Users/cbielow/Documents/R/win-library/3.1`. Keep this directory in mind. We will need it later.
 2. **Copy** the folder `<libR>\PTXQC\dragNdrop\QC-dragdrop` to a custom target location of your choice (let's call this target folder **`<QCdir>`**) where you want PTXQC to reside.
    Usually, that's some network or local drive (e.g. **`<QCdir>`** = `Z:\my-proteomics`). Now you should have `<QCdir>\QC-dragdrop\...`.
 3. **Copy** the whole R installation directory (e.g. `c:\program files\R\R-3.1.0`) into the `<QCdir>\QC-dragdrop\_internal` sub-folder,
    such that you end up with `<QCdir>\QC-dragdrop\_internal\R-3.1.0\` (your R version number might differ)
 4. **Copy** the Pandoc.exe (which you installed earlier) to `<QCdir>\QC-dragdrop\_internal\R-3.1.0\bin\pandoc\pandoc.exe` (you will need to create the pandoc folder).
    The batch file expects to find Pandoc there and will throw an error if Pandoc is missing.
 5. If your newly copied R installation directory is _not_ named `R-3.1.0`, **rename** it to `R-3.1.0` 
    (or edit the `<QCdir>\QC-dragdrop\createQC_dragNdrop.bat` to match your R version)
 6. If the `<libR>` folder was not your default R library (see step 1), but a user library, then **copy** the packages it contains (PTXQC among others),
    into `<QCdir>\QC-dragdrop\_internal\R-3.1.0\library\`.
 
Now, you should have the following structure

    QCdir\QC-dragdrop\
                     \createQC_dragNdrop.bat
                     \createQC_dragNdrop_withYAML.bat
                     \_internal\
                               \R-3.1.0\...  (version may differ)
                               \compute_QC_report.R
 
You can rename the `QC-dragdrop` folder to anything you like (try to avoid spaces - they usually cause trouble).

To make PTXQC accessible more easily, you can create a shortcut on your Windows Desktop pointing to `createQC_dragNdrop.bat`. Use this shortcut as if it were the original batch file (see usage below).


### Updating the PTXQC Drag'n'drop installation

Start the R GUI within in PTXQC standalone installation at `<QCdir>\QC-dragdrop\_internal\R-3.1.0\bin\x64\rgui.exe` and
run the block of code from the installation (see above) again. Then, copy the packages from the `<libR>` folder to the main R library (see the last bullet point above).

**Important:** make sure that you start with a clean R-Session and that no other R-Sessions are open, since they might prevent the update of package files where
R will only give you warnings like `Warning: cannot remove prior installation of package "digest"`, but will fail hard with a misleading error message afterwards. 
 
### Usage (short version)

  For the long version, see the [PTXQC-DragNDrop][Ref_VignDrag] vignette of this package.
  
  You can create a QC report by dragging a **txt-folder** (or any file within the txt-folder)
  onto the `createQC_dragNdrop.bat` file which resides on your (network) drive.


  [Ref_Vign]: https://github.com/cbielow/PTXQC/tree/master/vignettes
  [Ref_VignDrag]: https://github.com/cbielow/PTXQC/blob/master/vignettes/PTXQC-DragNDrop.Rmd
  [Ref_VignFAQ]: https://github.com/cbielow/PTXQC/blob/master/vignettes/PTXQC-FAQ.Rmd

This folder contains the package vignettes in Markdown (.Rmd files).

When the package is build, R/knitr will automatically create 'proper' vignette files,
which end up as HTML in PTXQC/doc/*.html.

On GitHub, you can just open any of the vignettes and they will be displayed as HTML in your browser.


Technical thoughts on GitHub:

Since this package is also hosted on GitHub, vignettes can be directly viewed 'raw' by just browsing to this directory (PTXQC/vignettes).
GitHub has a nice feature, which renders the Markdown header as table (instead of displaying the YAML text or ugly LaTex commands).

Currently, GitHub will apply this nice formatting only if the vignette is encoded as ANSI. Using UTF-8 encoding showed the ugly header formatting.
Despite the fact that we promise UTF-8 to LaTex/Knitr in the header, ANSI seems to work nicely.

A markdown header is basically YAML, mixable with LaTex and might look like this:

---
title: "Basic R-Usage Guide for PTXQC"
author: "Chris Bielow <chris.bielow@mdc-berlin.de>"
date: '`r Sys.Date()`'
output:
  html_document: default
  pdf_document: null
vignette: >
  %\VignetteIndexEntry{Basic R-Usage Guide for PTXQC}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

See our vignettes for complete examples.

# Examples

PTXQC will provide you with a report in Html/PDF format, summarizing the quality of your data.

Please refer to the publication (currently under review; link is coming soon) for details on which quality metrics are available and how they are scored.

## Obtaining example data
If you want to generate a report yourself, example input data can be found on the **PRIDE** repository under the
following accession numbers: [PXD003133 (MBR)][1], [PXD003134 (Myco)][2], and [PXD000427 (TMT)][3].
This data corresponds to the MaxQuant data used in the publication.

The [PTXQC-Basic_Guide_for_R_users][4] vignette has R code snippets to automatically download the data. If you are not an R person and have Windows as operating system, just download the zip files from the PRIDE link above, and run PTXQC via drag'n'drop (as described in the [PTXQC-DragNDrop][5] vignette).

## Report Layout

The first page shows an overview heatmap, representing scores for each Raw file for a set of quality metrics.
The screenshots given here, a from a rather small run of just 5 files. You can easily analyse a study with more than 100 files, but the report will get bigger :)

![Overview Heatmap](./example_heatmap.png?raw=true "Overview heatmap showing quality criteria for each LC-MS file")
 
After that, there is a summary of MaxQuant parameters (i.e. which FASTA files were used, was Match-between-runs activated etc...).

In the remainder of the report you can follow up on each individual metric and explore the reason for failure. The Html output additionally contains a
helpful description of each metric in case you have not read the [publication][JPR_paper] or need a refresh on what the plot actually tells you.

For example, the alignment in MaxQuant via the Match-between-runs did not work optimal across all Raw files in this case,
since MaxQuant could not normalize the LC gradients at the start and end of the run:
![Alignment Performance](./example_MBRalignment.png?raw=true "Alignment of 5 raw files (the first file serves as reference here)")

Also, you could check how a new LC gradient influences the identifications over time:
![Identifications over Retention Time](./example_IDoverRT.png?raw=true "Identifications over Retention Time")

Find a full report as [html][example_html] or [pdf][example_pdf].

See the package vignettes for documentation on how to create and customize a report.

  [1]: http://www.ebi.ac.uk/pride/archive/projects/PXD003133
  [2]: http://www.ebi.ac.uk/pride/archive/projects/PXD003134
  [3]: http://www.ebi.ac.uk/pride/archive/projects/PXD000427
  [4]: https://github.com/cbielow/PTXQC/blob/master/vignettes/PTXQC-Basic_Guide_for_R_users.Rmd
  [5]: https://github.com/cbielow/PTXQC/blob/master/vignettes/PTXQC-DragNDrop.Rmd
  [JPR_paper]: https://doi.org/10.1021/acs.jproteome.5b00780
  [example_html]: http://htmlpreview.github.io/?https://github.com/cbielow/PTXQC/blob/master/inst/examples/report_v1.1.0__txt_5files_withMatch-100min.html
  [example_pdf]: report_v1.1.0__txt_5files_withMatch-100min.pdf

## ----eval=FALSE---------------------------------------------------------------
# vignette("PTXQC-InputData", package = "PTXQC")

## ----eval=FALSE---------------------------------------------------------------
# require(PTXQC)
# 
# ## the next require() is needed to prevent a spurious error in certain R versions (might be a bug in R or a package)
# ## error message is:
# ##    Error in Scales$new : could not find function "loadMethod"
# require(methods)
# 
# ## specify a path to a MaxQuant txt folder
# ## Note: This folder needs to be complete (see 'vignette("PTXQC-InputData", package = "PTXQC")')
# if (1) {
#   ## we will use an example dataset from PRIDE (dataset 2 of the PTXQC publication)
#   local_zip = tempfile(fileext=".zip")
#   tryCatch({
#     download.file("ftp://ftp.pride.ebi.ac.uk/pride/data/archive/2015/11/PXD003133/txt_20min.zip", destfile = local_zip)
#   }, error = function(err) {
#     ## on Windows, one can get a 'cannot open URL' using the default method. So we try another
#     download.file("ftp://ftp.pride.ebi.ac.uk/pride/data/archive/2015/11/PXD003133/txt_20min.zip", destfile = local_zip, method= "internal")
# 
#   })
#   unzip(local_zip, exdir = tempdir()) ## extracts content
#   txt_folder = file.path(tempdir(), "txt_20min")
# } else {
#   ## if you have local MaxQuant output, just use it
#   txt_folder = "c:/Proteomics/MouseLiver/combined/txt"
# }
# 
# r = createReport(txt_folder)
# 
# cat(paste0("\nReport generated as '", r$report_file, "'\n\n"))
# 

## ----eval=FALSE---------------------------------------------------------------
# vignette("PTXQC-CustomizeReport", package = "PTXQC")

## ----eval=FALSE---------------------------------------------------------------
# require(PTXQC)
# require(yaml)
# 
# ## the next require() is needed to prevent a spurious error in certain R versions (might be a bug in R or a package)
# ## error message is:
# ##    Error in Scales$new : could not find function "loadMethod"
# require(methods)
# 
# ## specify a path to a MaxQuant txt folder
# ## Note: This folder can be incomplete, depending on your YAML config
# if (1) {
#   ## we will use an example dataset from PRIDE (dataset 2 of the PTXQC publication)
#   local_zip = tempfile(fileext=".zip")
#   download.file("ftp://ftp.pride.ebi.ac.uk/pride/data/archive/2015/11/PXD003133/txt_20min.zip", destfile = local_zip)
#   unzip(local_zip, exdir = tempdir()) ## extracts content
#   txt_folder = file.path(tempdir(), "txt_20min")
# } else {
#   ## if you have local MaxQuant output, just use it
#   txt_folder = "c:/Proteomics/MouseLiver/combined/txt"
# }
# 
# ## use a YAML config inside the target directory if present
# fh_out = getReportFilenames(txt_folder)
# if (file.exists(fh_out$yaml_file))
# {
#   cat("\nUsing YAML config already present in target directory ...\n")
#   yaml_config = yaml.load_file(input = fh_out$yaml_file)
# } else {
#   cat("\nYAML config not found in folder '", txt_folder, "'. The first run of PTXQC will create one for you.", sep="")
#   yaml_config = list()
# }
# 
# r = createReport(txt_folder, mztab_file = NULL, yaml_obj = yaml_config)
# 
# cat(paste0("\nReport generated as '", r$report_file, "'\n\n"))
# 

## load packages
library(PTXQC)
library(yaml)
## the next library() is needed to prevent a spurious error in certain R versions (might be a bug in R or a package)
## error message is:
##    Error in Scales$new : could not find function "loadMethod"
library(methods)

argv = commandArgs(TRUE)
#argv = c('Z:\\projects\\QC\\PTXQC\\data\\_txt_withMBR_withFractions_MQ15')
cat("Command line args are:\n")
cat(paste(argv, collapse="\n", sep=""))
cat("\n")

if(!(length(argv) %in% 1:2))
{
  stop("Wrong number of parameters!\n",
    "Received: \n - ", paste(argv, collapse="\n - ", sep=""),
    "\n\nUsage: <thisScript.R> <PATH_TO_TXT> [<PATH_TO_YAML_CONFIG>]\n");
}

PATH_TO_TXT = argv[1]
fi = file.info(PATH_TO_TXT)
if (is.na(fi$isdir) || !fi$isdir)
{
  stop(paste0("Argument '", PATH_TO_TXT, "' is not a valid directory\n"));
}

YAML_CONFIG = list()
if (length(argv)==2 && nchar(argv[2])>0)
{ ## YAML was passed via command line
  cat("\nUsing YAML config provided via command line ...\n")
  YAML_CONFIG = yaml.load_file(input = argv[2])
} else {
  ## use a YAML config inside the target directory if present
  rprt_fns = getReportFilenames(PATH_TO_TXT)
  if (file.exists(rprt_fns$yaml_file))
  {
    cat("\nUsing YAML config already present in target directory ...\n")
    YAML_CONFIG = yaml.load_file(input = rprt_fns$yaml_file)
  }
}
## use YAML_CONFIG to get output-filenames (which contains the log-file name)
yc = YAMLClass$new(YAML_CONFIG)
use_extended_reportname = yc$getYAML("PTXQC$ReportFilename$extended", TRUE)
rprt_fns = getReportFilenames(PATH_TO_TXT, use_extended_reportname)

output_files = try(createReport(PATH_TO_TXT, NULL, YAML_CONFIG, rprt_fns, enable_log = TRUE))

## ----setup, include=TRUE, echo=FALSE, results="asis"--------------------------
## global options
knitr::opts_chunk$set(echo=FALSE, warning=FALSE, error=FALSE, message=FALSE, fig.width=10)

lst_qcMetrics_ord = PTXQC:::getMetricsObjects()

txt_TOC = "# Table of Contents
  - [Overview](#Overview)
  - [Metrics](#Metrics)
"
txt_BODY = ""
for (qcm in lst_qcMetrics_ord)
{
  newname = gsub("(.*)\\(.*" , "\\1", gsub("[\\*~%]" , " ", gsub("[\\^\">]" , "", qcm$qcName))) 
  ## remove weird symbols (to serve as link)
  newname_lnk = gsub("[^a-zA-Z0-9]", "", newname)
  txt_TOC = paste0(txt_TOC, paste0("    - [", newname, "](#", newname_lnk, ")\n"))
  txt_BODY = paste0(txt_BODY, '\n
## ', newname, ' <a name="', newname_lnk, '"/>
\n
  <div class="helpText">', qcm$helpTextTemplate, '</div>
<p style="text-align: right"><a href="#TOP">-- back to top --</a></p>\n\n')
}

## print the TOC
cat(txt_TOC)


## ----metrics, echo=FALSE, results="asis"--------------------------------------
  cat(txt_BODY)

#yamlConfig = "foo:\n bar: 456 \n"
#yamlObj = yaml.load(yamlConfig)
#eval(parse(text="yamlObj$bla$foo = 2"))
#yamlObj
#useP = getYAML(yamlObj, "SEC_Parameters$use", TRUE)

#' 
#' Query a YAML object for a certain parameter.
#' 
#' If the object has the param, then return it.
#' If the param is unknown, create it with the given default value and return the default.
#' 
#' @field yamlObj A Yaml object as created by \code{\link[yaml]{yaml.load}}
#'
#' @exportClass YAMLClass
#' @export YAMLClass
#' 
#' @examples 
#'     yc = YAMLClass$new(list())
#'     val = yc$getYAML("cat$subCat", "someDefault")
#'     val  ## someDefault
#'     val = yc$setYAML("cat$subCat", "someValue")
#'     val  ## someValue
#'     yc$getYAML("cat$subCat", "someDefault") ## still 'someValue' (since its set already)
#' 
YAMLClass = setRefClass(
  "YAMLClass",
  
  fields = list(yamlObj = "list" # A Yaml object as created by \code{\link[yaml]{yaml.load}}
  ),
  
  methods = list(
    
    ##
    ##  ctor
    ##
    initialize = function(yamlObj = list()) {
      .self$yamlObj = yamlObj;
      return(.self)
    },

    getYAML = function(param_name, default, min = NA, max = NA)
    {
      "Query this YAML object for a certain parameter and return its value. If it does not exist it is created with a default value.
       An optional min/max range can be specified and will be enforced if the value is known (default will be used upon violation)."
      cat(paste0("YAML: ", param_name, " def: ", paste(default, sep="", collapse=",")))
      pval = eval(parse(text=paste0(".self$yamlObj$", param_name))) 
      if (is.null(pval))
      { ## param not known yet --> add
        expr = paste0(".self$yamlObj$", param_name, " = ", quote(default))
        eval(parse(text=expr)) 
        cat("\n")
        return (default)
      } else {
        if (!is.na(min))
        { # check range
          if (!is.numeric(pval) || pval < min || max < pval)
          {
            cat(paste0("YAML value for '", param_name, "' is invalid ('", pval, "'). Using default of ", default, "."))
            pval = .self$setYAML(param_name, default)
          }
        }
        cat(paste0(" || new val: ", paste(pval, sep="", collapse=","), "\n"))
        return (pval)
      }
    },
    
    setYAML = function(param_name, value)
    {
      "Set a YAML parameter to a certain value. Overwrites the old value or creates a new entry if hithero unknown."
      expr = paste0(".self$yamlObj$", param_name, " = ", quote(value))
      eval(parse(text=expr)) 
      return (value)
    },
    
    
    writeYAML = function(filename)
    {
      "Write YAML config (including some documentation) to a YAML file. Returns TRUE on success (always), unless writing the file generates an error."
      yaml.user.warning = 
        "# This is a configuration file for PTXQC reporting.
# One such file is generated automatically every time a report PDF is created.
# You can make a copy of this file, then modify its values and use the copy as an input to another round of report generation,
# e.g., to exclude/include certain plots or change some global settings.
#
# It is recommended to work on a copy of this file, such that this original configuration reflects the content of the report.
# 
# Note that upon report generation, this file will be overwritten again (with potentially new values, if and depending on the YAML config you provided).
#
# This file has a certain structure, which should be *retained* when editing.
# Note that each parameter level has two more spaces for indentation
# A value is assigned using a colon followed by a space, i.e. ': '
#
# The values that can be taken by a parameter are indicated by the SUFFIX of the name.
# '<name>_num' :  a numeric value is expected
# '<name>_wA'  :  'no', 'yes', 'auto' (see below)
# '<name>      :  'no', 'yes'
#
# By default (no special ending) parameters are BINARY (yes or no).
# Parameters ending in 'wA' (== 'withAuto') offer a heuristic to decide if something should be plotted.
# Finally, numerical parameters... well... are numerical (usually integer).
#
#
# The subsection 'order' enables reordering of metrics by assigning other numbers.
# The order affects both the order of columns in the heatmap, and the order of plots in the report.
# If you set a metrics order number to a negative value, the metric will not be computed!
# This is useful to switch of uninteresting or broken metrics (until a fix is rolled out).
#
# Parameters whose name starts with 'MQpar_' should be matched to the parameters set in MaxQuant.
# E.g. 'File$Evidence$MQpar_MatchingTimeWindow_num' is the matching (not alignment!) tolerance
#      for Match-between-Runs, and has a default of 1 (minute).
#      Older MQ version prior to MaxQuant 1.4 allowed 2 minutes, very recent ones use 0.7 minutes.
# The parameter 'PTXQC$UseLocalMQPar' controls if the mqpar.xml file will be looked up or not. 
# It needs to be present within this txt folder (i.e. you need to copy it here!).
# If the mqpar.xml file is not found, a warning will be issued and the internal PTXQC default will
# be used (bearing the risk of being inaccurate).
#
# Furthermore, there is the SpecialContaminants section, where you can trigger the generation of a plot
# which just shows the fraction of proteins containing a certain string.
# By default we search for 'MYCOPLASMA' in the protein identifier, i.e. your txt should be derived from a 
# run using a database containing these identifiers (the FASTA description is not enough).
# Each of the special contaminants requires has its own section (e.g. 'cont_MYCO:') with two parameters: 
# - a string          = name within the protein identifier
# - an integer number = a threshold in % which will be plotted to visually ease interpretation
# If you do not want any contaminants search, just set
#     SpecialContaminants: no
#
# With the exception(!) of extra 'SpecialContaminants':
#   Do NOT add extra values (since they are ignored anyway and might even destroy the integrity of this configuration file).
#   Only modify existing values, but not their names. I.e. only change 123 but not 'test'
#   test: 123
#
#
"
      cat(paste0(yaml.user.warning, yaml::as.yaml(.self$yamlObj)), file=filename)
      
      return (TRUE);
    }
    
  )
)

#' PTXQC: A package for computing Quality Control (QC) metrics for Proteomics (PTX)
#'
#' The following sections describe the main components:
#'
#' @section Input:
#' Valid input data are either the files from MaxQuant's .txt folder (all versions from MaxQuant >= 1.0 upwards are supported)
#' or a single mzTab file. All mzTab files will work, but most metrics can be obtained from OpenMS' mzTab as produced
#' by the QualityControl TOPP tool (from OpenMS 2.5 onwards).
#'
#' @section Important functions:
#' The central function of this package is called \code{\link{createReport}} and it accepts either MaxQuant or mzTab data, along with
#' a configuration (optional).
#' There is a parser for mzTab \code{\link{MzTabReader}} and MaxQuant txt files \code{\link{MQDataReader}}, as well as a plethora of QC metrics
#' derived from a common \code{\link{qcMetric}} class and scoring functions \code{qual...}, e.g. \code{\link{qualGaussDev}}.
#'
#' @section Configuration:
#' The user can modify the behaviour of PTXQC, e.g. to enable/disable certain metrics or change scoring thresholds, via a YAML object/file.
#' By default a Yaml file is written automatically side-by-side to the input files upon running PTXQC for the first time on a particular input.
#' A custom Yaml object can be passed to the main \code{\link{createReport}} function for customization.
#' Use \code{yaml::yaml.load_file(input = 'myYAML.yaml')} to load an existing file and pass the Yaml object along.
#'
#' @section Output:
#' Either a PDF and/or Html report which contains QC plots and a description of the metrics.
#'
"_PACKAGE"
#' @name PTXQC
#' @title Quality control for proteomics data
#'
#' @import data.table
#' @import ggplot2
#' @import ggdendro
#' @import grid
#' @import gridExtra
#' @import grDevices
#' @import gtable
#' @import knitr
#' @import methods
#' @import plyr
#' @import R6
#' @import R6P
#' @import RColorBrewer
#' @rawNamespace import(reshape2, except = c(dcast, melt))
#' @import rmarkdown
#' @importFrom seqinr circle
#' @import stats
#' @import utils
#' @import UpSetR
#' @import xml2
#' @import yaml
#'
NULL

##
## All plotting functions for computeReport() [main function]
##

#'
#' Plot contaminants from proteinGroups.txt
#'
#' @param data A data.frame with columns 'group', 'cont_pc', 'logAbdClass'
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'
#'  data = data.frame( 'group' = letters[1:10], 'cont_pc' = 2:11, 'logAbdClass' = c("low","high"))
#'  plot_ContsPG(data)
#'
plot_ContsPG = function(data)
{
  data$section = as.integer(seq(0, nrow(data)/correctSetSize(nrow(data),30)*0.999, length.out=nrow(data)))
  p = ggplot(data=data, aes(x = .data$group, y = .data$cont_pc, alpha = .data$logAbdClass)) +
        suppressWarnings(## suppresses 'Using alpha for a discrete variable is not advised'
        scale_alpha_discrete(range = c(c(0.3, 1)[(length(unique(data$logAbdClass))==1) + 1], 1.0), ## ordering of range is critical!
                             name = "Abundance\nclass")) +
        geom_col() +
        theme(axis.text.x = element_text(angle = 90, vjust = 0.5, hjust = 1)) +
        xlab("")  +
        ggtitle("PG: Contaminant per condition") +
        ylab("contaminant (% intensity)") +
        geom_hline(aes(yintercept = 5), linetype = 'dashed')
  if (length(unique(data$section))>1) p = p + facet_wrap(~ section, ncol = 1, scales="free_x")
  #print(p)
  return (p)
}



#'
#' Plot user-defined contaminants from evidence.txt
#'
#' Kolmogorov-Smirnoff p-values are plotted on top of each group.
#' High p-values indicate that Andromeda scores for contaminant peptides
#' are equal or higher compared to sample peptide scores, i.e. the probability that
#' sample peptides scores are NOT greater than contaminant peptide scores.
#'
#' @param data A data.frame with columns 'fc.raw.file', 'variable', 'value'
#' @param name_contaminant Name of the contaminant shown in title
#' @param extra_limit Position where a h-line is plotted (for visual guidance)
#' @param subtitle Optional subtitle for plot
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'
#'  data = data.frame(fc.raw.file = letters[1:3],
#'                    variable = c(rep("spectralCount", 3),
#'                                  rep("intensity", 3),
#'                                  rep("above.thresh", 3),
#'                                  rep("score_KS", 3)),
#'                    value = c(10, 20, 15, 9, 21, 14, 0, 1, 1, 0.3, 0.01, 0.04))
#'  plot_ContUser(data, "myco", 5, "subtitle")
#'
plot_ContUser = function(data, name_contaminant, extra_limit, subtitle = NULL)
{
  datav = subset(data, data$variable %in% c('spectralCount', "intensity"))
  datav$section = assignBlocks(datav$fc.raw.file, set_size = 40, sort_values = TRUE)
  dataAT = subset(data, data$variable %in% c('above.thresh'))
  ## contRaws might be empty
  contRaws = dataAT$fc.raw.file[ dataAT$value == TRUE]
  dataKS = subset(data, data$variable == 'score_KS' & (data$fc.raw.file %in% contRaws))
  if (nrow(dataKS)>0) {
    dataKS$value = paste0("p = ", round(dataKS$value,2))
    ## use the same section, so ggplot knows how to subset the data
    dataKS$section = datav$section[match(dataKS$fc.raw.file, datav$fc.raw.file)]
  }
  #cat(paste0("CA entry is ", extra_limit, "\n"))
  maxY = max(datav$value, extra_limit)
  p = ggplot(datav, aes(x = .data$fc.raw.file, y = .data$value)) +
        geom_col(aes(fill = .data$variable), position = "dodge", width=.7) +
        ggtitle(paste0("EVD: Contaminant '", name_contaminant, "'"), subtitle) +
        xlab("")  +
        ylab("abundance fraction (%)") +
        ylim(c(0, maxY * 1.1)) +
        theme(plot.title = element_text(colour = "red"),
              axis.text.x = element_text(angle = 90, vjust = 0.5, hjust = 1)) +
        scale_fill_discrete(name = "Method") +
        geom_hline(yintercept = extra_limit, linetype = 'dashed')
  ## group(NULL) seems important in geom_text()
  if (nrow(dataKS)>0) p = p + geom_text(data = dataKS, aes(label = .data$value, y = maxY * 1.05, group = NULL))
  p = p + facet_wrap(~ section, ncol = 1, scales = "free_x")
  #print(p)
  return(p)
}



#'
#' Plot Andromeda score distribution of contaminant peptide vs. matrix peptides.
#'
#' The data is expected to be an ECDF already, x being the Andromeda score, y being the culmulative probability.
#' The Score is the probability of a Kolm.-Smirnoff test that the contaminant scores are larger (i.e.
#' large p-values indicate true contamination).
#' You will only see this plot if the %-threshold (YAML config) was reached. This is a saveguard against false-positive,
#' but high-scoring contaminant peptides, which would erroneously give you a large p-value and make you believe
#' your sample is contaminated although that's not the case.
#'
#' @param data A data.frame with columns 'x', 'y', 'condition'
#' @param raw.file Name of Raw file for which the data is displayed (will become part of the plot title)
#' @param score Score of how distinct the distributions are (will become part of the title)
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'
#'  data = data.frame(x = 10:60,
#'                    y = c(seq(0,1,length=51), seq(0.1, 1, length=51)),
#'                    condition = rep(c("sample","contaminant"), each=51))
#'  plot_ContUserScore(data, 'test file', 0.96)
#'
plot_ContUserScore = function(data, raw.file, score) {
  p = ggplot(data) +
    geom_line(aes(x = .data$x, y = .data$y, col = .data$condition)) +
    ggtitle(paste0("Empirical CDF of '", raw.file, "'\np = ", round(score, 2))) +
    ylab("Pr") +
    xlab("Andromeda score")
  return(p)
}


#'
#' Plot contaminants from evidence.txt, broken down into top5-proteins.
#'
#' @param data A data.frame with columns 'fc.raw.file', 'contaminant', 'pname', 'intensity'
#' @param top5 Name of the Top-5 Proteins (by relative intensity or whatever seems relevant)
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'
#'  data = data.frame(intensity = 1:12,
#'                    pname = rep(letters[1:3], 4),
#'                    fc.raw.file = rep(paste("f", 1:4), each=3),
#'                    contaminant = TRUE)
#'  ## providing more proteins than present... d,e will be ignored
#'  plot_ContEVD(data, top5 = letters[1:5])
#'  ## classify 'c' as 'other'
#'  plot_ContEVD(data, top5 = letters[1:2])
#'
plot_ContEVD = function(data, top5)
{
  #top5 = cont.top5.names
  if (is.null(top5)) stop("Function plot_ContEVD() called with invalid argument. Please report this bug.")
  if (length(top5) > 5) stop("Top5 protein list is longer than 5 (which is the maximum allowed).")

  intensity = NULL ## to make R CHECK happy...
  data.sub = data[data$contaminant > 0,]
  ## rewrite prot names, and subsume 6th and below as 'other'
  data.sub$pname = as.character(data.sub$pname)
  data.sub[!(data.sub$pname %in% top5), "pname"] = 'other'
  ## aggregate identical proteins
  ##  use sum(as.numeric(.)) to prevent overflow
  d_sum = plyr::ddply(data.sub[, c("intensity", "pname", "fc.raw.file")], c("pname", "fc.raw.file"),
                function(x) plyr::summarise(x, s.intensity=sum(as.numeric(intensity), na.rm = TRUE)))
  ## normalize by total intensity of raw file
  d_norm = plyr::ddply(data[, c("intensity", "fc.raw.file")],  "fc.raw.file",
                 function(x) plyr::summarise(x, total.intensity=sum(as.numeric(intensity), na.rm = TRUE)))

  d_sum$total.intensity = d_norm$total.intensity[match(d_sum$fc.raw.file, d_norm$fc.raw.file)]
  d_sum$Log10Diff = getAbundanceClass(log10(d_sum$total.intensity))
  d_sum$s.intensity = d_sum$s.intensity / d_sum$total.intensity * 100
  ## shorten protein-groups (at most two protein names)
  d_sum$pname = sapply(d_sum$pname, function(x) {
    p.split = unlist(strsplit(x, split=";"))
    ## shorten entries as well (at most 15 characters)
    p.split_s = sapply(p.split[1:(min(2, length(p.split)))], function(x) ifelse(nchar(x)>15, paste0(substr(x, start=1, stop=13), ".."), x))
    r = paste(p.split_s, sep="", collapse=";")
    if (length(p.split)>2) r=paste0(r, ";..")
    return(r)
  })
  ## order of pname determines order of bars
  d_sum = rbind(d_sum[d_sum$pname!="other",], d_sum[d_sum$pname=="other",])

  ## value of factors determines order in the legend
  ## --> make proteins a factor, with 'other' being the first
  d_sum$Protein = factor(d_sum$pname, levels = unique(c("other", d_sum$pname)), ordered = TRUE)
  head(d_sum)

  ## plot
  p = ggplot(d_sum, aes(  x = .data$fc.raw.file,
                          y = .data$s.intensity,
                       fill = .data$Protein)) +
        geom_col(aes(alpha = .data$Log10Diff)) +
        suppressWarnings(## suppresses 'Using alpha for a discrete variable is not advised'
          scale_alpha_discrete(range = c(c(0.3, 1)[(length(unique(d_sum$Log10Diff))==1) + 1], 1.0),
                               name = "Abundance\nclass")) +
        xlab("")  +
        theme_bw() +
        ggtitle("EVD: Top5 Contaminants per Raw file") +
        ylab("contaminant (% intensity)") +
        scale_fill_manual(values = RColorBrewer::brewer.pal(6,"Accent")) +
        scale_colour_manual(values = RColorBrewer::brewer.pal(6,"Accent")) +
        geom_hline(aes(yintercept = 5), linetype='dashed') +
        #guides(alpha=NULL, fill = guide_legend(nrow = 2, ncol = 3, byrow = TRUE, reverse = TRUE)) +
        #theme(legend.position="top", legend.title=element_blank()) +
        coord_flip() +
        scale_x_discrete_reverse(d_sum$fc.raw.file)

  #print(p)
  return(p)
}



#'
#' Plot ratios of labeled data (e.g. SILAC) from proteinGroups.txt
#'
#' The 'x' values are expected to be log2() transformed already.
#'
#' @param df_ratios A data.frame with columns 'x', 'y', 'col', 'ltype'
#' @param d_range X-axis range of plot
#' @param main_title Plot title
#' @param main_col Color of title
#' @param legend_title Legend text
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'
#'  x1 = seq(-3, 3, by = 0.1)
#'  y1 = dnorm(x1)
#'  x2 = seq(-5, 1, by = 0.1)
#'  y2 = dnorm(x2, mean = -1)
#'  data = data.frame( x = c(x1,x2),
#'                     y = c(y1,y2),
#'                     col = c(rep("ok", length(x1)), rep("shifted", length(x2))),
#'                     ltype = c(rep("solid", length(x1)), rep("dotted", length(x2))))
#'  plot_RatiosPG(data, range(data$x), "Ratio plot", "red", "group")
#'
plot_RatiosPG = function(df_ratios, d_range, main_title, main_col, legend_title)
{
  br = c(2, 5, 10, 20);

  p =
    ggplot(data = df_ratios, aes(x = .data$x, y = .data$y, colour = .data$col)) +
    facet_grid(col ~ ., scales = "free_y") +
    geom_line(linewidth = 1.2) +
    geom_area(aes(alpha = .data$ltype, fill = .data$col)) +
    xlab("ratio")  +
    ylab("density")  +
    scale_fill_manual(values = rep(RColorBrewer::brewer.pal(6,"Accent"), times=40), guide = guide_legend(legend_title)) +
    scale_colour_manual(values = rep(RColorBrewer::brewer.pal(6,"Accent"), times=40)) +
    suppressWarnings(scale_alpha_discrete(range = c(1, 0.2),
                         labels=c("dotted"="unimodal", "solid"="multimodal"),
                         guide = guide_legend("shape")
    )) +
    scale_x_continuous(limits = d_range, trans = "identity", breaks = c(-br, 0, br), labels=c(paste0("1/",2^(br)), 1, 2^br)) +
    guides(color = "none") +
    theme(plot.title = element_text(colour = main_col)) +
    theme_bw() +
    geom_vline(alpha = 0.5, xintercept = 0, colour = "green", linetype = "dashed", linewidth = 1.5) +
    ggtitle(main_title)
  #print(p)
  return (p)
}

#'
#' Plot Protein groups per Raw file
#'
#' The input is a data.frame with protein/peptide counts, where 'category' designates
#' the origin of information (genuine ID, transferred ID, or both).
#'
#' @param data A data.frame with columns 'fc.raw.file', 'counts', 'category'
#' @param y_max Plot limit of y-axis
#' @param thresh_line Position of a threshold line, indicating the usual target value
#' @param title Main title, and optional subtitle (if vector of length 2 is provided)
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'
#'  data = data.frame(fc.raw.file = rep(c("file A", "file B"), each=3),
#'                    counts = c(3674, 593, 1120, 2300, 400, 600),
#'                    category = c("genuine","genuine+transferred","transferred"))
#'  plot_CountData(data, 6000, 4000, c("EVD: Protein Groups count", "gain: 23%"))
#'
plot_CountData = function(data, y_max, thresh_line, title)
{
  title_main = title[1]
  title_sub = ifelse(length(title) > 1,  title[2], "")
  p = ggplot(data, aes(x = .data$fc.raw.file, y = .data$counts, fill = .data$category)) +
        geom_col(position = position_stack(reverse = TRUE)) +
        xlab("") +
        ylab("count") +
        scale_x_discrete_reverse(data$fc.raw.file) +
        ylim(0, y_max) +
        scale_fill_manual(values = c("green", "#BEAED4", "blue")) +
        ggtitle(title_main, title_sub) +
        geom_abline(alpha = 0.5, intercept = thresh_line, slope = 0, colour = "black", linetype = "dashed", linewidth = 1.5) +
        coord_flip()
  return(p)
}


#'
#' Plot RT peak width over time
#'
#' The input is a data.frame with already averaged counts over binned RT-slices.
#'
#' @param data A data.frame with columns 'fc.raw.file', 'RT', 'peakWidth'
#' @param x_lim Plot range of x-axis
#' @param y_lim Plot range of y-axis
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'
#'  data = data.frame(fc.raw.file = rep(c("file A", "file B", "file C"), each=81),
#'                    RT = c(20:100),
#'                    peakWidth = c(rnorm(81, mean=20), rnorm(81, mean=10), rnorm(81, mean=30)))
#'  plot_RTPeakWidth(data, c(10, 100), c(0, 40))
#'
plot_RTPeakWidth = function(data, x_lim, y_lim)
{
  p = ggplot(data) +
    geom_line(aes(x = .data$RT, y = .data$peakWidth, colour = .data$fc.raw.file), linewidth = 1, alpha = 0.7) +
    scale_color_manual(values = brewer.pal.Safe(length(unique(data$fc.raw.file)), "Set1")) +
    guides(color = guide_legend(title = "Raw file\n(avg. peak width)")) +
    xlab("retention time [min]") +
    ylab("peak width [min]") +
    coord_cartesian(xlim = x_lim, ylim = y_lim) + ## zoom in y -- do not cut data (preserve lines)
    ggtitle("EVD: Peak width over RT") +
    theme(axis.text.x = element_text(angle = 90, vjust = 0.5, hjust = 1))
  #print(p)
  return(p)
}


#'
#' Plot MaxQuant Match-between-runs alignment performance.
#'
#' The plots shows the correction function applied by MaxQuant, and the
#' residual RT (ideally 0) of each peptide to its reference. Uncalibrated peptides
#' are shown in red, calibrated ones in green.
#' The MaxQuant RT correction which was applied prior is shown in blue. The range of this function
#' can give hints if the allowed RT search window (20min by default) is sufficient or if
#' MaxQuant should be re-run with more tolerant settings.
#'
#' The input is a data.frame with columns
#'   'calibrated.retention.time' - resulting (hopefully) calibratated RT after MQ-recal (the X-axis of the plot)
#'   'retention.time.calibration' - delta applied by MaxQuant
#'   'rtdiff' - remaining RT diff to reference peptide of the same sequence
#'   'RTdiff_in' - is the feature aligned (within 'match_tol')?
#'   'fc.raw.file_ext' - raw file
#' where each row represents one peptide whose RT was corrected by MaxQuant.
#'
#' @param data A data.frame with columns as described above
#' @param y_lim Plot range of y-axis
#' @param title_sub Subtitle
#' @param match_tol Maximal residual RT delta to reference (usually ~1 min)
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'
#'  data = data.frame(fc.raw.file_ext = "file A", ## more than one would be possible
#'                    calibrated.retention.time = c(20:100),
#'                    retention.time.calibration = 6 + sin((20:100)/10))
#'  data$rtdiff = rnorm(nrow(data))
#'  data$RTdiff_in = c("green", "red")[1 + (abs(data$rtdiff) > 0.7)]
#'
#'  plot_MBRAlign(data, c(-10, 10), "fancy subtitle", 0.7)
#'
plot_MBRAlign = function(data, y_lim, title_sub, match_tol)
{
  #data = evd_RT_t[ evd_RT_t$fc.raw.file == "file 13",]
  p = ggplot(data, aes(x = .data$calibrated.retention.time, y = .data$retention.time.calibration)) +
        ## the MaxQuant correction (plot real data, no spline, since it can be very irregular)
        geom_line(aes(alpha = 0.7), color = "blue") +
        ## PTXQC correction
        geom_point(aes(x = .data$calibrated.retention.time, y = .data$rtdiff, color = .data$RTdiff_in), alpha = 0.5) +
        scale_alpha(name = 'Alignment function',
                    labels = list(expression("MaxQuant" ~ Delta*"RT")),
                    range = c(1,1)) +
        scale_colour_manual(name = expression(bold("ID pairs ("*Delta*"RT to Ref)")),
                            values = c("green" = "green", "red" = "red"),
                            labels=c("green" = paste0("good (<", match_tol, "min)"),
                                     "red" = paste0("bad (>", match_tol, "min)"))) +
        guides(colour = guide_legend(order = 2),
               alpha = guide_legend(order = 1)) +   ## alpha-legend on top, color below
        ylim(y_lim) +
        xlab("corrected RT [min]") +
        ylab(expression(Delta*"RT [min]")) +
        facet_wrap(~ fc.raw.file_ext) +
        ggtitle("EVD: MBR - alignment", title_sub)
  #print(p)
  return(p)
}


#'
#' Plot MaxQuant Match-between-runs id transfer performance.
#'
#' The plots shows the different categories of peak classes
#'
#' The input is a data.frame with columns
#'   'fc.raw.file' - raw file name
#'   'single' - fraction of peptides with are represent only once
#'   'multi.inRT' - fraction of peptides with are represent multiple times,
#'                  but within a certain RT peak width
#'   'multi.outRT' - fraction of peptides with are represent multiple times,
#'                   with large RT distance
#'   'sample' - raw file
#' where each row represents one peptide sequence.
#'
#' @param data A data.frame with columns as described above
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'  data = data.frame(fc.raw.file = rep(c("file A", "file B"), each = 3),
#'                    single = c(0.9853628, 0.8323160, 0.9438375,
#'                               0.9825538, 0.8003763, 0.9329961),
#'                    multi.inRT = c(0.002927445, 0.055101018, 0.017593087,
#'                                   0.005636457, 0.099640044, 0.031870056),
#'                    multi.outRT = c(0.01170978, 0.11258294, 0.03856946,
#'                                    0.01180972, 0.09998363, 0.03513386),
#'                    sample = rep(c("genuine", "transferred", "all"), 2))
#'  plot_MBRIDtransfer(data)
#'
plot_MBRIDtransfer = function(data)
{
  data.m = reshape2::melt(data, id.vars=c("fc.raw.file", "sample"))
  data.m$value = data.m$value * 100 ## used to be scores in [0-1]
  if (all(is.na(data.m$value)))
  {# the slice of Raw file we are looking at could have no MBR data -- and ggplot needs something to plot...
    data.m$value = 0
  }
  p = ggplot(data.m) +
        geom_col(aes(x = .data$fc.raw.file, y = .data$value, fill = .data$variable), position = position_stack(reverse = TRUE)) +
        scale_fill_manual("peak class",
                          values = c("single"="green", "multi.inRT"="lightgreen", "multi.outRT"="red"),
                          labels=c("single", "group (in width)", "group (out width)")) +
        ylim(0, 100.1) + ## ggplot might not show the last (red) group upon 100.0
        xlab("") +
        ylab("fraction of 3D-peaks [%]") +
        coord_flip() +
        scale_x_discrete_reverse(factor(data$fc.raw.file)) +
        ggtitle("EVD: MBR - ID Transfer") +
        facet_wrap(~sample)
  #print(p)
  return(p)
}


#'
#' Plot MaxQuant Match-between-runs id transfer performance as a scatterplot.
#'
#' Per Raw file, the absolute number of transferred IDs as well as the relative gain in percent.
#'
#' The input is a data.frame with columns
#'   'fc.raw.file' - raw file name
#'   'abs' - absolute number of transferred ID's
#'   'pc' - gain on top of genuine IDs [%]
#' where each row represents one rawfile.
#'
#' @param data A data.frame with columns as described above
#' @param title_sub Subtitle text
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'  data = data.frame(fc.raw.file = paste("file", letters[1:4]),
#'                    abs = c(5461, 5312, 3618, 502),
#'                    pc = c(34, 32, 22, 2))
#'  plot_MBRgain(data, "MBR gain: 18%")
#'
plot_MBRgain = function(data, title_sub = "")
{
  p = ggplot(data = data, aes(x = .data$abs, y = .data$pc, col = .data$fc.raw.file)) +
        geom_point(size = 2) +
        ggtitle("EVD: Peptides inferred by MBR", title_sub) +
        xlab("number of transferred ID's") +
        ylab("gain on top of genuine IDs [%]") +
        xlim(0, max(data$abs, na.rm = TRUE)*1.1) + ## accounting for labels near the border
        ylim(0, max(data$pc, na.rm = TRUE)*1.1) +
        guides(color = "none") +
        geom_text(aes(label = .data$fc.raw.file), hjust = -0.2, show.legend = FALSE, check_overlap = TRUE)
  #print(p)
  return(p)
}


#'
#' The plots shows the charge distribution per Raw file.
#' The output of 'mosaicize()' can be used directly.
#'
#' The input is a data.frame with columns
#'   'Var1' - name of the Raw file
#'   'Var2' - charge (used as fill color)
#'   'Var1_center' - contains X-position of the Raw file
#'   'Var2_height' - relative frequency of the charge
#'   'Margin_var1' -
#' where each row represents one peptide sequence.
#'
#' @param d_charge A data.frame with columns as described above
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'  data = data.frame(raw.file = c(rep('file A', 100), rep('file B', 40)),
#'                        data = c(rep(2, 60), rep(3, 30), rep(4, 10),
#'                                 rep(2, 30), rep(3, 7), rep(4, 3)))
#'  plot_Charge(mosaicize(data))
#'
plot_Charge = function(d_charge)
{
  p = ggplot(d_charge, aes(x = .data$Var1_center, y = .data$Var2_height, width = .data$Margin_var1)) +
        geom_col(aes(fill = .data$Var2), color = "black", position = position_stack(reverse = TRUE))  +
        geom_text(aes(label = .data$Var1, x = .data$Var1_center, y = 1.05)) +
        xlab("Raw file") +
        ylab("fraction [%]") +
        guides(fill = guide_legend(title = "charge"),
                                   color = "none") + # avoid black line in legend
        scale_x_reverse() +
        coord_flip() +
        theme(axis.text.y = element_blank(), axis.ticks = element_blank()) +
        ggtitle("EVD: charge distribution")
  #print(p)
  return(p)
}


#'
#' Plot IDs over time for each Raw file.
#'
#' Uses plot_DataOverRT() internally.
#'
#' @param data A data.frame with columns as described above
#' @param x_lim Limits of the x-axis (2-tuple)
#' @param y_max Maximum of the y-axis (single value)
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'  data = data.frame(fc.raw.file = rep(paste('file', letters[1:3]), each=30),
#'                             RT = seq(20, 120, length.out = 30),
#'                         counts = c(rnorm(30, 400, 20), rnorm(30, 250, 15), rnorm(30, 50, 15)))
#'  plot_IDsOverRT(data)
#'
plot_IDsOverRT = function(data, x_lim = range(data$RT), y_max = max(data$counts))
{
  return(plot_DataOverRT(data, "EVD: IDs over RT", "ID count", x_lim, y_max))
}


#'
#' Plot some count data over time for each Raw file.
#'
#' The input is a data.frame with columns
#'   'RT' - RT in seconds, representing one bin
#'   'counts' - number of counts at this bin
#'   'fc.raw.file' - name of the Raw file
#' where each row represents one bin in RT.
#'
#' At most nine(!) Raw files can be plotted. If more are given,
#' an error is thrown.
#'
#'
#' @param data A data.frame with columns as described above
#' @param title The plot title
#' @param y_lab Label of y-axis
#' @param x_lim Limits of the x-axis (2-tuple)
#' @param y_max Maximum of the y-axis (single value)
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'  data = data.frame(fc.raw.file = rep(paste('file', letters[1:3]), each=30),
#'                             RT = seq(20, 120, length.out = 30),
#'                         counts = c(rnorm(30, 400, 20), rnorm(30, 250, 15), rnorm(30, 50, 15)))
#'  plot_DataOverRT(data, "some title", "count data")
#'
plot_DataOverRT = function(data, title, y_lab, x_lim = range(data$RT), y_max = max(data$counts))
{
  nrOfRaws = length(unique(data$fc.raw.file))
  p = ggplot(data = data) +
    geom_line(aes(x = .data$RT, y = .data$counts, colour = .data$fc.raw.file, linetype = .data$fc.raw.file)) +
    xlim(x_lim) +
    xlab("RT [min]") +
    ylim(from = 0, to = y_max) +
    ylab(y_lab) +
    ggtitle(title) +
    guides(colour = guide_legend(title="Raw file"), linetype = "none") +
    scale_linetype_manual(values = rep_len(c("solid", "dashed"), nrOfRaws)) +
    scale_color_manual(values = brewer.pal.Safe(nrOfRaws, "Set1"))
  #print(p)
  return(p)
}

#'
#' Plot percent of identified MS/MS for each Raw file.
#'
#' Useful for a first overall impression of the data.
#'
#' The input is a data.frame with columns
#'   'fc.raw.file' - name of the Raw file
#'   'ms.ms.identified....' - fraction of identified MS/MS spectra in percent
#'   'cat' - identification category as arbitrary string
#' where each row represents one Raw file.
#'
#' @param data A data.frame with columns as described above
#' @param id_rate_bad Number below which the ID rate is considered bad
#' @param id_rate_great Number above which the ID rate is considered great
#' @param label_ID Named vector with colors for the categories given in data$cat
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'  id_rate_bad = 20; id_rate_great = 35;
#'  label_ID = c("bad (<20%)" = "red", "ok (...)" = "blue", "great (>35%)" = "green")
#'  data = data.frame(fc.raw.file = paste('file', letters[1:3]),
#'                    ms.ms.identified.... = rnorm(3, 25, 15))
#'  data$cat = factor(cut(data$ms.ms.identified....,
#'                        breaks=c(-1, id_rate_bad, id_rate_great, 100),
#'                        labels=names(label_ID)))
#'  plot_IDRate(data, id_rate_bad, id_rate_great, label_ID)
#'
plot_IDRate = function(data, id_rate_bad, id_rate_great, label_ID)
{
    p = ggplot(data, aes(y = .data$fc.raw.file, x = .data$ms.ms.identified....)) +
        geom_point(aes(colour = .data$cat)) +
        geom_vline(xintercept = id_rate_bad, color=(label_ID)[1]) +
        geom_vline(xintercept = id_rate_great, color=(label_ID)[3]) +
        ylab("") +
        xlab("MS/MS identified [%]") +
        scale_colour_manual(values=label_ID) +
        ggtitle("SM: MS/MS identified per Raw file") +
        xlim(0, max(data$ms.ms.identified...., id_rate_great)*1.1) +
        guides(color = guide_legend(title="ID class")) +
        scale_y_discrete_reverse(data$fc.raw.file, breaks = ggAxisLabels)
  #print(p)
  return(p)
}



#'
#' Colored table plot.
#'
#' Code taken from http://stackoverflow.com/questions/23819209/change-text-color-for-cells-using-tablegrob-in-r
#'
#' @param data Table as Data.frame
#' @param colours Single or set of colours (col-wise)
#' @param fill Cell fill (row-wise)
#' @param just (ignored)
#' @return gTable
#'
plotTableRaw = function(data, colours="black", fill=NA, just="centre")
{

  label_matrix = as.matrix(data)

  nc = ncol(label_matrix)
  nr = nrow(label_matrix)
  n = nc*nr

  colours <- rep(colours, length.out = n)
  fill <- rep(fill, length.out = n)
  justs <- rep(just, length.out = n)

  ## text for each cell
  labels <- lapply(seq_len(n), function(ii)
    grid::textGrob(as.character(label_matrix[ii]), gp = grid::gpar(fontsize=8, col=colours[ii]), just="left", x = grid::unit(0.05, "npc")))
  label_grobs <- matrix(labels, ncol=nc)

  ## define the fill background of cells
  fill <- lapply(seq_len(n), function(ii)
    grid::rectGrob(gp = grid::gpar(fill=fill[ii])))

  ## some calculations of cell sizes
  row_heights <- function(m){
    do.call(grid::unit.c, apply(m, 1, function(l)
      max(do.call(grid::unit.c, lapply(l, grid::grobHeight)))))
  }
  col_widths <- function(m){
    do.call(grid::unit.c, apply(m, 2, function(l)
      max(do.call(grid::unit.c, lapply(l, grid::grobWidth)))))
  }

  ## place labels in a gtable
  g <- gtable::gtable_matrix("table", grobs = label_grobs,
                             widths = col_widths(label_grobs) + grid::unit(2,"mm"),
                             heights = row_heights(label_grobs) + grid::unit(2,"mm"))

  ## add the background
  xt <- rep(seq_len(nr), each=nc)
  xl <- rep(seq_len(nc), times=nr)
  g <- gtable::gtable_add_grob(g, fill, t=xt, l=xl, z=0, name="fill")

  return(g)
}

#'
#' Create an HTML table with an extra header row
#'
#'
#' @param data A data.frame which serves as table
#' @param caption A set of headlines, e.g. c("top line", "bottom line")
#' @return table as html character string for cat()'ing into an html document
#'
#' @import htmlTable
#' @import magrittr
#'
#' @export
#'
#' @examples
#'   data = data.frame(raw.file = letters[1:4],
#'                     id.rate = 3:6)
#'   getHTMLTable(data,
#'                caption = "some header line")
#'
getHTMLTable = function(data, caption = NA)
{

  tbl = htmlTable::addHtmlTableStyle(data,
                                     align = 'l',  ## align columns left
                                     col.rgroup = c("none", "#F7F7F7"))
  tbl = htmlTable::htmlTable(tbl, rnames = FALSE,    ## no row names
                             caption = caption)

  return(tbl)
}

#'
#' Plot a table with row names and title
#'
#' Restriction: currently, the footer will be cropped at the table width.
#'
#' @param data A data.frame with columns as described above
#' @param title Table title
#' @param footer Footer text
#' @param col_names Column names for Table
#' @param fill Fill pattern (by row)
#' @param col Text color (by column)
#' @param just (ignored)
#' @return gTree object with class 'PTXQC_table'
#'
#' @export
#'
#' @examples
#'   data = data.frame(raw.file = letters[1:4],
#'                     id.rate = 3:6)
#'   plotTable(data,
#'             title = "Bad files",
#'             footer = "bottom",
#'             col_names = c("first col", "second col"),
#'             col=c("red", "green"))
#'
plotTable = function(data, title = "", footer = "", col_names = colnames(data), fill = c("grey90", "grey70"), col = "black", just="centre")
{
  ## add column names
  data2 = rbind(col_names, apply(data, 2, function(x) as.character(x)))
  ## create table
  n = nrow(data2)*ncol(data2)
  nd = nrow(data)*ncol(data)
  table = plotTableRaw(data2,
                       fill = c(rep("grey50", ncol(data)), rep(fill, each=ncol(data), length.out=nd)), ## row-wise
                       colours = unlist(lapply(col, function(cc) c("black", rep(cc, nrow(data))))), ## col-wise
                       just = c(rep("centre", ncol(data)), rep(just, each=nrow(data), length.out=nd)))

  colhead = lapply(col_names, function(ii) grid::textGrob(ii, gp = grid::gpar(fontsize=12, col="black", fontface="bold", fill="grey")))
  ## replace column names
  table = gtable::gtable_add_grob(table, colhead, t = 1, l = 1:ncol(data))

  if (nchar(title[1]) > 0)
  {
    gtitle = grid::textGrob(title, gp = grid::gpar(fontsize = 14))
    padding = grid::unit(1.5, "line")
    ## add heading (white space)
    table = gtable::gtable_add_rows(table, heights = grid::grobHeight(gtitle) + padding, pos = 0)
    ## add heading (text as overlay)
    table = gtable::gtable_add_grob(table, list(gtitle), t = 1, l = 1, r = ncol(table), clip = "off")
  }
  if (nchar(footer[1]) > 0)
  {
    gfooter = grid::textGrob(footer, gp = grid::gpar(fontsize = 10))
    padding = grid::unit(1.5, "line")
    ## add heading (white space)
    table = gtable::gtable_add_rows(table, heights = grid::grobHeight(gfooter) + padding, pos = -1) ## bottom
    ## add heading (text as overlay)
    table = gtable::gtable_add_grob(table, list(gfooter), t = nrow(table), l = 1, r = ncol(table), clip = "off")
  }


  ## neat trick to enable calling print(g), to mimic ggplot-behaviour on this object
  ## in combination with print.PTXQC_table() -- see below
  p = grid::gTree(children = grid::gList(table), cl = c("PTXQC_table"))

  ## hide the table name inside (for qcMetric::getTitles())
  p$labels$title = title

  #print(p)
  return(p)
}

#' helper S3 class, enabling print(some-plot_Table-object)
#' @param x Some Grid object to plot
#' @param ... Further arguments (not used, but required for consistency with other print methods)
#' @return NULL
#'
#' @export
#'
print.PTXQC_table = function(x, ...) {
  grid::grid.newpage();
  grid::grid.draw(x)
  return(NULL)
}

#'
#' A boxplot of uncalibrated mass errors for each Raw file.
#'
#' Boxes are optionally colored to indicate that a MQ bug was detected or
#' if PTXQC detected a too narrow search window.
#'
#' @param data A data.frame with columns 'fc.raw.file', 'uncalibrated.mass.error..ppm.'
#' @param MQBug_raw_files List of Raw files with invalid calibration values
#' @param stats A data.frame with columns 'fc.raw.file', 'sd', 'outOfCal'
#' @param y_lim Range of y-axis
#' @param extra_limit Position where a v-line is plotted (for visual guidance)
#' @param title_sub Subtitle
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'   n = c(150, 1000, 1000, 1000)
#'   data = data.frame(fc.raw.file = repEach(letters[4:1], n),
#'                     uncalibrated.mass.error..ppm. = c(rnorm(n[1], 13, 2.4),
#'                                                       rnorm(n[2], 1, 0.5),
#'                                                       rnorm(n[3], 3, 0.7),
#'                                                       rnorm(n[4], 4.5, 0.8)))
#'   stats = data.frame(fc.raw.file = letters[4:1],
#'                      sd_uncal = c(2.4, 0.5, 0.7, 0.8),
#'                      outOfCal = c(TRUE, FALSE, FALSE, FALSE))
#'   plot_UncalibratedMSErr(data, MQBug_raw_files = letters[1],
#'                          stats, y_lim = c(-20,20), 15, "subtitle")
#'
plot_UncalibratedMSErr = function(data, MQBug_raw_files, stats, y_lim, extra_limit, title_sub)
{

  data$col = "default"
  if (length(MQBug_raw_files) > 0)
  {
    data$col[data$fc.raw.file %in% MQBug_raw_files] = "MQ bug"
  }
  ## add 'out-of-calibration' Raw files:
  data$col[data$fc.raw.file %in% stats$fc.raw.file[stats$outOfCal]] = "out-of-search-tol"
  ## only show legend if special things happen
  showColLegend = ifelse(length(setdiff(data$col, "default")) > 0, "legend", "none")
  ## amend SD to fc.raw.file
  stats$fcr_new_lvl = paste0(stats$fc.raw.file, " (sd = ", stats$sd_uncal, "ppm)")

  ## use augmented name
  data$fc.raw.file_ext = stats$fcr_new_lvl[ match(data$fc.raw.file, stats$fc.raw.file) ]

  cols_sub = c("default"="black", "MQ bug"="red", "out-of-search-tol"="orange")
  cols_sub = cols_sub[names(cols_sub) %in% data$col]

  p = ggplot(data) +
        geom_boxplot(aes(x = .data$fc.raw.file_ext, y = .data$uncalibrated.mass.error..ppm., colour = .data$col), varwidth = TRUE, outlier.shape = NA) +
        scale_colour_manual("", values = cols_sub, guide = showColLegend) +
        ylab(expression(Delta~"mass [ppm]")) +
        xlab("") +
        ylim(y_lim) +
        #scale_x_discrete_reverse(data$fc.raw.file_ext) +
        scale_x_discrete(limits=rev) +
        geom_hline(yintercept = c(-extra_limit, extra_limit),
                   colour="red",
                   linetype = "longdash") +  ## == vline for coord_flip
        coord_flip() +
        ggtitle("EVD: Uncalibrated mass error", title_sub)

  #print(p)
  return(p)
}

#'
#' Plot bargraph of uncalibrated mass errors for each Raw file.
#'
#' Boxes are optionally colored to indicate that a MQ bug was detected or
#' if PTXQC detected a too narrow search window.
#'
#' @param data A data.frame with columns 'fc.raw.file', 'mass.error..ppm.'
#' @param MQBug_raw_files List of Raw files with invalid calibration values
#' @param stats A data.frame with columns 'fc.raw.file', 'outOfCal'
#' @param y_lim Range of y-axis
#' @param extra_limit Position where a v-line is plotted (for visual guidance)
#' @param title_sub Subtitle
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'   n = c(150, 1000, 1000, 1000)
#'   data = data.frame(fc.raw.file = repEach(letters[4:1], n),
#'                     mass.error..ppm. = c(rnorm(n[1], 1, 2.4),
#'                                          rnorm(n[2], 0.5, 0.5),
#'                                          rnorm(n[3], 0.1, 0.7),
#'                                          rnorm(n[4], 0.3, 0.8)))
#'   stats = data.frame(fc.raw.file = letters[4:1],
#'                      sd = c(2.4, 0.5, 0.7, 0.8),
#'                      outOfCal = c(TRUE, FALSE, FALSE, FALSE))
#'   plot_CalibratedMSErr(data, MQBug_raw_files = letters[1], stats, y_lim = c(-20,20), 15, "subtitle")
#'
plot_CalibratedMSErr = function(data, MQBug_raw_files, stats, y_lim, extra_limit = NA, title_sub = "")
{
  data$col = "default"
  if (length(MQBug_raw_files) > 0) {
    data$col = c("default", "MQ bug")[(data$fc.raw.file %in% MQBug_raw_files) + 1]
    data$mass.error..ppm.[data$fc.raw.file %in% MQBug_raw_files] = 0
    if (all(data$mass.error..ppm.==0)) data$mass.error..ppm. = rnorm(nrow(data), sd=0.0001, mean=mean(y_lim))
  }
  ## add 'out-of-calibration' Raw files:
  data$col[data$fc.raw.file %in% stats$fc.raw.file[stats$outOfCal]] = "out-of-search-tol"
  ## only show legend if special things happen
  showColLegend = ifelse(length(setdiff(data$col, "default")) > 0, "legend", "none")

  cols_sub = c("default"="black", "MQ bug"="red", "out-of-search-tol"="orange")
  cols_sub = cols_sub[names(cols_sub) %in% data$col]

  ## plot
  p = ggplot(data) +
    geom_boxplot(aes(x = .data$fc.raw.file, y = .data$mass.error..ppm., colour = .data$col), varwidth = TRUE, outlier.shape = NA) +
    scale_colour_manual("", values = cols_sub, guide = showColLegend) +
    ylab(expression(Delta~"mass [ppm]")) +
    xlab("") +
    ylim(y_lim) +
    #scale_x_discrete_reverse(data$fc.raw.file) +
    scale_x_discrete(limits=rev) +
    coord_flip() +
    ggtitle("EVD: Calibrated mass error", title_sub)
  if (!is.na(extra_limit)) {
    p = p + geom_hline(yintercept = c(-extra_limit, extra_limit), colour="red", linetype = "longdash")  ## == vline for coord_flip
  }

  #print(p)
  return (p)
}


#'
#' Plot bargraph of oversampled 3D-peaks.
#'
#' Per Raw file, at most three n's must be given, i.e.
#' the fraction of 3D-peaks for n=1, n=2 and n=3(or more).
#' The fractions must sum to 1 (=100%).
#'
#'
#' @param data A data.frame with columns 'fc.raw.file', 'n', 'fraction'
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'   data = data.frame(fc.raw.file = rep(letters[1:3], each=3),
#'                     n = 1:3,
#'                     fraction = c(0.8, 0.1, 0.1, 0.6, 0.3, 0.1, 0.7, 0.25, 0.05))
#'   plot_MS2Oversampling(data)
#'
plot_MS2Oversampling = function(data)
{
  stopifnot(length(unique(data$n)) <= 3) ## at most three -- to match color vector below
  #data = d_dups
  ## reorder factor, such that '10+' is last
  data$n = as.character(data$n)
  n_unique = sort(unique(data$n)) ## sort as character vector!
  data$n = factor(data$n, levels=n_unique[order(nchar(n_unique))], ordered = TRUE)

  p = ggplot(data) +
        geom_col(position = position_stack(reverse = TRUE), aes(x = .data$fc.raw.file, y = .data$fraction, fill = .data$n)) +
        scale_fill_manual("MS/MS\ncounts", values =c("green", "blue", "red")) +
        scale_x_discrete_reverse(data$fc.raw.file) +
        xlab("") +
        ylab("MS/MS counts per 3D-peak [%]") +
        ggtitle(paste0("EVD: Oversampling (MS/MS counts per 3D-peak)")) +
        theme(axis.text.x = element_text(angle = 90, vjust = 0.5, hjust = 1)) +
        coord_flip()

  #print(p)
  return(p)
}


#'
#' Plot bargraph of oversampled 3D-peaks.
#'
#' Per Raw file, at most three n's must be given, i.e.
#' the fraction of 3D-peaks for n=1, n=2 and n=3(or more).
#' The fractions must sum to 1 (=100%).
#'
#'
#' @param data A data.frame with columns 'file', 'msErr', 'type'
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'   n = c(100, 130, 50)
#'   data = data.frame(file = repEach(paste(letters[1:3],"\nLTQ [Da]"), n),
#'                     msErr = c(rnorm(n[1], 0.5), rnorm(n[2], 0.0), rnorm(n[3], -0.5)),
#'                     type = c("forward", "decoy")[1+(runif(sum(n))>0.95)])
#'   plot_MS2Decal(data)
#'
plot_MS2Decal = function(data)
{
  ## trim down the data to 2-98 percentiles (to avoid outliers far off)
  data2 = plyr::ddply(data, "file", function(x) {
    qnt = quantile(x$msErr, probs = c(0.02, 0.98), na.rm = TRUE)
    return (x[qnt[1] < x$msErr & x$msErr < qnt[2], ])
  })
  p = ggplot(data2, aes(x = .data$msErr, fill = .data$type)) +
    geom_histogram(bins = 30) +
    xlab("fragment mass delta") +
    ylab("count") +
    scale_fill_manual(values = c(forward = "#99d594", decoy = "#ff0000")) +
    ggtitle("MSMS: Fragment mass errors per Raw file") +
    facet_wrap(~file, scales = "fixed")

  #print(p)
  return(p)
}

#'
#' Plot bargraph of missed cleavages.
#'
#' Per Raw file, an arbitrary number of missed cleavage classes (one per column) can be given.
#' The total fraction of 3D-peaks must sum to 1 (=100%).
#' Columns are ordered by name.
#'
#' A visual threshold line is drawn at 75% (expected MC0 count).
#'
#' @param data A data.frame with columns 'fc.raw.file', '...' (missed cleavage classes)
#' @param title_sub Plot's subtitle
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'   data = data.frame(fc.raw.file = letters[1:5],
#'                     MC0 = c(0.8, 0.5, 0.85, 0.2, 0.9),
#'                     MC1 = c(0.1, 0.4, 0.05, 0.7, 0.0),
#'                     "MS2+" =  c(0.1, 0.1, 0.1, 0.1, 0.1),
#'                     check.names = FALSE)
#'   plot_MissedCleavages(data, "contaminant inclusion unknown")
#'
plot_MissedCleavages = function(data, title_sub = "")
{
  st_bin.m = reshape2::melt(data, id.vars = c("fc.raw.file"))
  p =
    ggplot(data = st_bin.m, aes(x = factor(.data$fc.raw.file), y = .data$value, fill = .data$variable)) +
        geom_col(position = position_stack(reverse = TRUE)) +
        xlab("Raw file") +
        ylab("missed cleavages [%]") +
        theme(legend.title = element_blank()) +
        scale_fill_manual(values = rep(c("#99d594", "#ffffbf", "#fc8d59", "#ff0000", "#800080", "#000000"), 10)) +
        geom_abline(alpha = 0.5, intercept = 0.75, slope = 0, colour = "black", linetype = "dashed", linewidth = 1.5) +
        coord_flip() +
        scale_x_discrete_reverse(st_bin.m$fc.raw.file) +
        ggtitle("MSMS: Missed cleavages per Raw file", title_sub)

  #print(p)
  return(p)
}

#'
#' Plot line graph of TopN over Retention time.
#'
#' Number of Raw files must be 6 at most. Function will stop otherwise.
#'
#' @param data A data.frame with columns 'fc.raw.file', 'rRT', 'topN'
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'   data = data.frame(fc.raw.file = rep(letters[1:3], each=100),
#'                     rRT = seq(20, 120, length.out = 100),
#'                     topN = c(round(runif(100, min=3, max=5)),
#'                              round(runif(100, min=5, max=8)),
#'                              round(runif(100, min=1, max=3)))
#'                     )
#'   plot_TopNoverRT(data)
#'
plot_TopNoverRT = function(data)
{
  nrOfRaws = length(unique(data$fc.raw.file))
  p = ggplot(data, aes(x = .data$rRT, y = .data$topN, col = .data$fc.raw.file)) +
        geom_line() +
        scale_color_manual(values = brewer.pal.Safe(nrOfRaws, "Set1")) +
        xlab("retention time [min]") +
        ylab("highest N [median per RT bin]") +
        #stat_smooth(method = "loess", formula = y ~ x, se = FALSE, span = 0.1) +
        guides(color=guide_legend(title="")) +
        ggtitle("MSMSscans: TopN over RT")

  #print(p)
  return (p)
}

#'
#' Plot line graph of TopN over Retention time.
#'
#' Number of Raw files must be 6 at most. Function will stop otherwise.
#'
#' @param data A data.frame with columns 'fc.raw.file', 'rRT', 'medIIT'
#' @param stats A data.frame with columns 'fc.raw.file', 'mean'
#' @param extra_limit Visual guidance line (maximum acceptable IIT)
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'   data = data.frame(fc.raw.file = rep(c("d","a","x"), each=100),
#'                     rRT = seq(20, 120, length.out = 100),
#'                     medIIT = c(round(runif(100, min=3, max=5)),
#'                                round(runif(100, min=5, max=8)),
#'                                round(runif(100, min=1, max=3)))
#'                     )
#'   stats = data.frame(fc.raw.file = c("d","a","x"),
#'                      mean = c(4, 6.5, 2))
#'   plot_IonInjectionTimeOverRT(data, stats, 10)
#'
plot_IonInjectionTimeOverRT = function(data, stats, extra_limit)
{
  data$fc.raw.file = data$fc.raw.file[,drop = TRUE] ## drop unused factor levels
  nrOfRaws = length(unique(data$fc.raw.file))
  stats_sub = stats[stats$fc.raw.file %in% data$fc.raw.file, , drop = FALSE]
  ## augment legend with average II-time[ms]
  data$fc.raw.file = paste0(data$fc.raw.file, " (~",
                            round(stats_sub$mean[match(data$fc.raw.file, stats_sub$fc.raw.file)]),
                            " ms)")
  ## manually convert to factor to keep old ordering (otherwise ggplot will sort it, since its a string)
  data$fc.raw.file = factor(data$fc.raw.file, levels = unique(data$fc.raw.file), ordered = TRUE)
  stats_sub$fc.raw.file = paste0(stats_sub$fc.raw.file, " (~",
                                 round(stats_sub$mean[match(stats_sub$fc.raw.file, stats_sub$fc.raw.file)]),
                                 " ms)")
  p = ggplot(data) +
        geom_line(aes(x = .data$rRT, y = .data$medIIT, col = .data$fc.raw.file)) +
        scale_color_manual(values = brewer.pal.Safe(nrOfRaws, "Set1")) +
        xlab("retention time [min]") +
        ylab("ion injection time [ms]") +
        geom_hline(yintercept = extra_limit, linetype = 'dashed') +
        guides(color=guide_legend(title="Raw file with\naverage inj. time")) +
        ggtitle("MSMSscans: Ion Injection Time over RT") +
        pointsPutX(x_range = range(data$rRT), x_section = c(0.03, 0.08), y = stats_sub$mean, col = stats_sub$fc.raw.file[,drop = TRUE])

  #print(p)
  return(p)
}

#'
#' Plot line graph of TopN over Retention time.
#'
#' Number of Raw files must be 6 at most. Function will stop otherwise.
#'
#' @param data A data.frame with columns 'fc.raw.file', 'scan.event.number', 'n'
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'   data = data.frame(fc.raw.file = rep(c("d","a","x"), each=10),
#'                     scan.event.number = 1:10,
#'                     n = 11:20)
#'   plot_TopN(data)
#'
plot_TopN = function(data)
{

  p = ggplot(data, aes(x = .data$scan.event.number, y = .data$n)) +
        geom_col() +
        xlab("highest scan event") +
        ylab("count") +
        facet_wrap(~ fc.raw.file, scales = "free_y") +
        ggtitle(paste0("MSMSscans: TopN"))

  #print(p)
  return(p)
}

#'
#' Plot line graph of TopN over Retention time.
#'
#' Number of Raw files must be 6 at most. Function will stop otherwise.
#'
#' @param data A data.frame with columns 'fc.raw.file', 'scan.event.number', 'ratio', 'count'
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'   data = data.frame(fc.raw.file = factor(rep(c("d","a","x"), each=10), levels = c("d","a","x")),
#'                     scan.event.number = 1:10,
#'                     ratio = seq(40, 20, length.out=10),
#'                     count = seq(400, 200, length.out=10))
#'   plot_ScanIDRate(data)
#'
plot_ScanIDRate = function(data)
{

  p = ggplot(data, aes(x = .data$scan.event.number, y = .data$ratio, alpha = .data$count)) +
        geom_col() +
        xlab("scan event") +
        ylab("percent identified") +
        facet_wrap(~ fc.raw.file) +
        ggtitle(paste0("MSMSscans: TopN % identified over N"))
  return (p)
}


#'
#' Plot Total Ion Count over time
#'
#' The input is a data.frame with already averaged counts over binned RT-slices.
#'
#' @param data A data.frame with columns 'fc.raw.file', 'RT', 'intensity'
#' @param x_lim Plot range of x-axis
#' @param y_lim Plot range of y-axis
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#'
#' @examples
#'
#'  data = data.frame(fc.raw.file = rep(c("file A", "file B", "file C"), each=81),
#'                    RT = c(20:100),
#'                    intensity = c(rnorm(81, mean=20), rnorm(81, mean=10), rnorm(81, mean=30)))
#'  plot_TIC(data, c(10, 100), c(0, 40))
#'
plot_TIC = function(data, x_lim, y_lim)
{
  p = ggplot(data) +
    geom_line(aes(x = .data$RT, y = .data$intensity, colour = .data$fc.raw.file), linewidth = 1, alpha = 0.7) +
    scale_color_manual(values = brewer.pal.Safe(length(unique(data$fc.raw.file)), "Set1")) +
    guides(color = guide_legend(title = "Raw file\n(avg. peak width)")) +
    xlab("retention time [min]") +
    ylab("intensity") +
    coord_cartesian(xlim = x_lim, ylim = y_lim) + ## zoom in y -- do not cut data (preserve lines)
    ggtitle("SM: Total Ion Count") +
    theme(axis.text.x = element_text(angle = 90, vjust = 0.5, hjust = 1))
  #print(p)
  return(p)
}

#'
#' Plot peptide modification frequencies
#' 
#' The input is a data.frame, as obtained from modsToTableByRaw().
#' 
#' 
#' 
#' @param tbl A data.frame with 'fc.raw.file', 'modification_names' (can be a factor), and 'Freq' (0-100)
#' @param y_max The upper limit of the y-axis's (==Freq); useful for multiple plots with identical limits; if 'NA' the limit is computed from the given 'tbl'
#' @param show_missing_modification_levels If 'tbl$modification_names' is a factor and has more (but missing) levels than actually used, should missing values be dropped or assumed as '0' frequency?
#' @return GGplot object
#'
#' @import ggplot2
#' @export
#' 
#' @examples 
#' 
#'  data = data.frame(fc.raw.file = rep(c("file A", "file B"), each=3),
#'                    modifications = c("Oxidation (M)",
#'                    "Unmodified",
#'                    "Oxidation (M), Acetyl (Protein N-term)",
#'                    "2 Oxidation (M)", 
#'                    "Unmodified", 
#'                    "Unmodified"))
#'  tbl = modsToTableByRaw(data)
#'  plot_peptideMods(tbl,show_missing_modification_levels = TRUE)
#' 
plot_peptideMods = function(tbl, y_max = NA, show_missing_modification_levels = TRUE)
{
  if (is.na(y_max)){
    y_max = max(tbl$Freq, na.rm = TRUE)
  } 

  ## augment '0'-frequency for missing factors
  if (show_missing_modification_levels) {
    tbl$fc.raw.file = droplevels(as.factor(tbl$fc.raw.file))  ## we need fc.raw.file to be a factor, but we do not want its unused levels
    all_combinations = expand.grid(lapply(tbl[c("fc.raw.file","modification_names")], levels))
    new_tbl = merge(tbl, all_combinations , all.y=TRUE)
    new_tbl$Freq[is.na(new_tbl$Freq)] = 0  ## replace new 'NA's with 0
    tbl = new_tbl
  }
    
  p = ggplot(tbl, aes(x = .data$fc.raw.file, y = .data$Freq, fill = .data$modification_names, colour = .data$modification_names)) + ## use 'colour' for outline of '0'-frequency features (invisible otherwise)
    geom_col(position = "dodge") +
    xlab("") +
    ylab("Occurence [%]") +
    scale_x_discrete_reverse(tbl$fc.raw.file) +
    guides(fill = guide_legend(reverse = TRUE), color = guide_none()) + ## reverse order of mods in legend, to match order in plot
    ylim(0, y_max) +
    ggtitle("EVD: variable modifications per Raw file") + 
    coord_flip()
  return(p)
}

#' Create a quality control report (in PDF format).
#'
#' This is the main function of the package and the only thing you need to call directly if you are
#' just interested in getting a QC report.
#'
#' You need to provide either
#' a) the folder name of the 'txt' output, as generated by MaxQuant or an mzTab file
#' or b) an mzTab file as generated by the OpenMS QualityControl TOPP tool (other mzTab files will probably not work)
#'
#' Optionally, provide a YAML configuration object, which allows to (de)activate certain plots and holds other parameters.
#' The yaml_obj is complex and best obtained by running this function once using the default (empty list).
#' A full YAML configuration object will be written in the 'txt' folder you provide and can be loaded using
#' \code{\link[yaml]{yaml.load}}.
#'
#' The PDF and the config file will be stored in the given txt folder.
#'
#' @note You need write access to the txt/mzTab folder!
#'
#' For updates, bug fixes and feedback please visit \url{https://github.com/cbielow/PTXQC}.
#'
#' @param txt_folder Path to txt output folder of MaxQuant (e.g. "c:/data/Hek293/txt")
#' @param mztab_file Alternative to **txt_folder**, you can provide a single mzTab file which contains PSM, PEP and PRT tables
#' @param yaml_obj   A nested list object with configuration parameters for the report.
#'                   Useful to switch off certain plots or skip entire sections.
#' @param report_filenames Optional list with names (as generated by \code{\link{getReportFilenames}}).
#'                         If not provided, will be created internally by calling \code{\link{getReportFilenames}}.
#' @param enable_log If TRUE all console output (including warnings and errors) is logged to the file given in **report_filenames$log_file**.
#'                   Note: warnings/errors can only be shown in either the log **or** the console, not both!
#' @return List with named filename strings, e.g. $yaml_file, $report_file etc..
#'
#' @export
#'
createReport = function(txt_folder = NULL,
                        mztab_file = NULL,
                        yaml_obj = list(),
                        report_filenames = NULL,
                        enable_log = FALSE)
{
  if (!exists("DEBUG_PTXQC")) DEBUG_PTXQC = FALSE ## debug only when defined externally

  ## the following code makes sure that a plotting device is open.
  ## In some scenarios, the a plotting device is not available by default, e.g. in non-interactive sessions (e.g. shinyApps).
  ## Some versions of R then open a default pdf device to rplots.pdf, but permissions might not allow to write there - so our app crashes.
  ## In any case, a plotting device is needed for some functions calls, e.g. in grid, to determine some meta-parameters, such as font size...
  ptxqc_dummy_pdffile = tempfile(fileext=".pdf")
  pdf(ptxqc_dummy_pdffile)      ## create a real file, since the pdf(NULL) might not be able to fulfill all requests-
  ptxqc_dummy_dev = dev.cur()   ## remember the device number
  on.exit({                     # when leaving the function, close our device (avoid leaving zombies)
    if (ptxqc_dummy_dev %in% dev.list()) {
      dev.off(ptxqc_dummy_dev)
      unlink(ptxqc_dummy_pdffile) ## delete the file; it might stay there for the whole session otherwise
    } else warning("Dummy graphics device was closed by someone else. This should not have happened...")
  }, add = TRUE)

  time_start = Sys.time()
  #mztab_file = "c:\\temp\\test.mzTab"
  ##mztab_file = NULL

  in_count =  (!is.null(mztab_file)) + (!is.null(txt_folder))
  if ( in_count == 2 )
  {
    stop("Please provide EITHER txt_folder or mztab_file, not both")
  }
  if ( in_count == 0 )
  {
    stop("Please provide EITHER mz_folder or mztab_file. Both are currently missing!")
  }

  ###
  ###  prepare the YAML config
  ###
  if (!inherits(yaml_obj, "list"))
  {
    stop(paste0("Argument 'yaml_obj' is not of type list\n"));
  }
  yc = YAMLClass$new(yaml_obj)


  MZTAB_MODE = !is.null(mztab_file)  ## will be TRUE if mzTab is detected

  if (MZTAB_MODE)
  {
    base_folder = dirname(mztab_file)
    mzt = MzTabReader$new()
    mzt$readMzTab(mztab_file) ## creates an inital fc.raw.file mapping from MTD
    expr_fn_map = quote(mzt$fn_map)
    txt_files = NULL
  } else
  {
    if (!any(file.info(txt_folder)$isdir, na.rm = TRUE))
    {
      stop(paste0("Argument txt_folder with value '", txt_folder, "' is not a valid directory\n"));
    }
    base_folder = txt_folder
    txt_files = list()
    txt_files$param = "parameters.txt"
    txt_files$summary = "summary.txt"
    txt_files$groups = "proteinGroups.txt"
    txt_files$evd = "evidence.txt"
    txt_files$msms = "msms.txt"
    txt_files$msmsScan = "msmsScans.txt"
    txt_files$mqpar = "mqpar.xml"
    txt_files = lapply(txt_files, function(file) file.path(txt_folder, file))

    ## prepare for readMQ()
    mq = MQDataReader$new()
    expr_fn_map = quote(mq$fn_map)

  }
  ## create names of output files (report PDF, YAML, stats, etc...)
  if (is.null(report_filenames)) {
    use_extended_reportname = yc$getYAML("PTXQC$ReportFilename$extended", TRUE)
    rprt_fns = getReportFilenames(base_folder, use_extended_reportname, mzTab_filename = mztab_file)
  } else {
    rprt_fns = report_filenames
  }
  ## read manual filename shortening & sorting (if available)
  eval(expr_fn_map)$readMappingFile(rprt_fns$filename_sorting)

  ##
  ## start logging the output
  if (enable_log)
  {
    ## establish our own error handler, so we can see the traceback etc
    options(error=function(x) {
            cat("\nTraceback:\n")
            traceback()
            cat(paste0("\nAn error occurred: '", trimws(geterrmessage()), "'. See '", rprt_fns$log_file, "' for details!\n\n\n"))})

    my_log = file(rprt_fns$log_file, open = "wt") # File name of output log
    sink(my_log, type = "output", split = TRUE) # Writing console output to log file
    sink(my_log, type = "message")  ## cannot be split ... so we need to decide where it should go

    on.exit({
      ## show warnings, before we leave
      print(warnings());

      ## Restore output to console
      sink(type="message")
      sink()
    }, add = TRUE)
  }


  cat(paste0(date(), ": Starting QC computation on report '", rprt_fns$report_file_prefix, "'\n"))

  ##
  ## YAML config (with default values if no yaml file was given)
  ##

  yc_param_lstqcMetrics_list = createYaml(yc = yc, DEBUG_PTXQC = DEBUG_PTXQC, txt_files = txt_files)

  yc = yc_param_lstqcMetrics_list$yc
  yaml_param = yc_param_lstqcMetrics_list$param
  lst_qcMetrics = yc_param_lstqcMetrics_list$lst_qcMetrics

  ## write out the final YAML file (so users can disable metrics, if they fail)
  yc$writeYAML(rprt_fns$yaml_file)

  out_formats_supported <- yaml_param$param_OutputFormats


  ## write shortnames and sorting of filenames
  eval(expr_fn_map)$writeMappingFile(rprt_fns$filename_sorting)

  ## get full filenames (and their suffix -- for mzQC metadata)
  file_meta = QCMetaFilenames$new()
  ## does only work if mqpar.xml is present (for now)
  if (!MZTAB_MODE) file_meta$data = getMetaFilenames(txt_files$mqpar, base_folder)
  ## --> wherever you need this data, simply re-grab the singleton using 'QCMetaFilenames$new()$data'

  ######
  ######  parameters.txt ...
  ######

  if (MZTAB_MODE) d_parAll = mzt$getParameters()
  else d_parAll = mq$readMQ(txt_files$param, type="par")

  lst_qcMetrics[["qcMetric_PAR"]]$setData(d_parAll)

  ######
  ######  summary.txt ...
  ######

  if (MZTAB_MODE) d_smy = mzt$getSummary()
  else d_smy = mq$readMQ(txt_files$summary, type="sm", add_fs_col = yaml_param$add_fs_col)
  #colnames(d_smy)
  #colnames(d_smy[[1]])

  ### MS/MS identified [%]
  lst_qcMetrics[["qcMetric_SM_MSMSIdRate"]]$setData(d_smy, yaml_param$id_rate_bad, yaml_param$id_rate_great)

  ### TIC
  if (MZTAB_MODE) lst_qcMetrics[["qcMetric_SM_TIC"]]$setData(d_smy)

  ######
  ######  proteinGroups.txt ...
  ######
  if (MZTAB_MODE) df_pg = mzt$getProteins()
  else df_pg = mq$readMQ(txt_files$groups, type="pg", col_subset=NA, filter="R")

  ## just a scope
  {
    ##
    ## Raw/LFQ/Reporter intensity boxplots
    ##
    clusterCols = list()

    colsSIL = grepv("^intensity\\.[hlm](\\.|$)", colnames(df_pg))
    colsLF = grepv("^intensity\\..", colnames(df_pg))
    colsOneCond = "intensity" ## just one group -- we still want to know what the overall intensity is
    if (length(colsSIL)) {
      ## ignore intensity.l and alike if real groups are present
      plain_channel = grepv("^intensity\\.[hlm]$", colnames(df_pg))
      if (all(plain_channel == colsSIL)) colsW = colsSIL else colsW = setdiff(colsSIL, plain_channel)
    } else if (length(colsLF)) {
      colsW = colsLF
    }  else {
      colsW = colsOneCond
    }

    ## a global PG name mapping
    MAP_pg_groups = data.frame(long = colsW)
    MAP_pg_groups$short = shortenStrings(simplifyNames(delLCP(MAP_pg_groups$long,
                                                              min_out_length = yaml_param$GL_name_min_length,
                                                              add_dots = TRUE),
                                                       min_out_length = yaml_param$GL_name_min_length))
    ##
    ## Contaminants plots on Raw intensity
    ##
    lst_qcMetrics[["qcMetric_PG_Cont"]]$setData(df_pg, colsW, MAP_pg_groups)


    ###
    ### Raw intensity boxplot
    ###

    clusterCols$raw.intensity = colsW ## cluster using intensity

    lst_qcMetrics[["qcMetric_PG_RawInt"]]$setData(df_pg, int_cols = colsW, MAP_pg_groups = MAP_pg_groups, thresh_intensity = yaml_param$param_PG_intThresh)

    ##
    ## LFQ boxplots
    ##
    colsSIL = grepv("^lfq.intensity\\.[hlm](\\.|$)", colnames(df_pg))
    colsLF = grepv("^lfq.intensity\\..", colnames(df_pg))

    ## a global PG name mapping
    MAP_pg_groups_LFQ = NA
    if (length(c(colsSIL, colsLF)) > 0)
    {
      if (length(colsSIL)) {
        ## unlike intensity.l, there is no lfq.intensity.l which we could remove
        colsW = colsSIL
      } else colsW = colsLF
      MAP_pg_groups_LFQ = data.frame(long = colsW)
      MAP_pg_groups_LFQ$short = shortenStrings(simplifyNames(delLCP(MAP_pg_groups_LFQ$long,
                                                                    min_out_length = yaml_param$GL_name_min_length,
                                                                    add_dots = TRUE),
                                                             min_out_length = yaml_param$GL_name_min_length))

      clusterCols$lfq.intensity = colsW ## cluster using LFQ

      lst_qcMetrics[["qcMetric_PG_LFQInt"]]$setData(df_pg, colsW, MAP_pg_groups_LFQ, yaml_param$param_PG_intThresh)
    }

    ##
    ## iTRAQ/TMT, reporter ion intensity boxplot
    ##
    ## either "reporter.intensity.0.groupname" or "reporter.intensity.0" (no groups)
    colsITRAQ = grepv("^reporter.intensity.[0-9]", colnames(df_pg)) ## we require at least one number!
    ## a global PG name mapping
    MAP_pg_groups_ITRAQ = NA
    if (length(colsITRAQ) > 0)
    {
      MAP_pg_groups_ITRAQ = data.frame(long = c(colsITRAQ))
      MAP_pg_groups_ITRAQ$short = shortenStrings(simplifyNames(delLCP(MAP_pg_groups_ITRAQ$long,
                                                                      min_out_length = yaml_param$GL_name_min_length,
                                                                      add_dots = TRUE),
                                                               min_out_length = yaml_param$GL_name_min_length))

      clusterCols$reporter.intensity = colsITRAQ ## cluster using reporters

      lst_qcMetrics[["qcMetric_PG_ReporterInt"]]$setData(df_pg, colsITRAQ, MAP_pg_groups_ITRAQ, yaml_param$param_PG_intThresh)
    }


    ##
    ## PCA
    ##
    ## some clustering (its based on intensity / lfq.intensity columns..)
    ## todo: maybe add ratios -- requires loading from txt though..
    MAP_pg_groups_ALL = rbind(MAP_pg_groups, MAP_pg_groups_LFQ, MAP_pg_groups_ITRAQ)

    lst_qcMetrics[["qcMetric_PG_PCA"]]$setData(df_pg, clusterCols, MAP_pg_groups_ALL)


    ##################################
    ## ratio plots
    ##################################
    ## get ratio column
    ratio_cols = grepv("^ratio\\.[hm]\\.l", colnames(df_pg))  ## e.g. "ratio.m.l.ARK5exp" or "ratio.m.l.variability.ARK5exp"
    ## remove everything else
    ## e.g. we do not want ratio.h.l.variability.ARK5exp, i.e. the 'variability' property
    ratio_cols = grepv("^ratio.[hm].l.normalized", ratio_cols, invert = TRUE)
    ratio_cols = grepv("^ratio.[hm].l.count", ratio_cols, invert = TRUE)
    ratio_cols = grepv("^ratio.[hm].l.variability", ratio_cols, invert = TRUE)
    ratio_cols = grepv("^ratio.[hm].l.significance.a", ratio_cols, invert = TRUE) ## from MQ 1.0.1x
    ratio_cols = grepv("^ratio.[hm].l.significance.b", ratio_cols, invert = TRUE)
    ratio_cols = grepv("^ratio.[hm].l.iso.count", ratio_cols, invert = TRUE) ## from MQ 1.5.1.2
    ratio_cols = grepv("^ratio.[hm].l.type", ratio_cols, invert = TRUE)
    ratio_cols

    if (length(ratio_cols) > 0)
    {
      lst_qcMetrics[["qcMetric_PG_Ratio"]]$setData(df_pg, ratio_cols = ratio_cols, thresh_LabelIncorp = yaml_param$pg_ratioLabIncThresh, GL_name_min_length = yaml_param$GL_name_min_length)
    }
  }

  ######
  ######  evidence.txt ...
  ######

  ## protein.names is only available from MQ 1.4 onwards
  if (MZTAB_MODE) {
    all_evd = mzt$getEvidence()
    df_evd = all_evd$genuine
    df_evd_tf = all_evd$transferred

  }
  else {
    all_evd = mq$readMQ(txt_files$evd, type="ev", filter="R",
                        col_subset=c("proteins",
                                     numeric = "Retention.Length",
                                     numeric = "retention.time.calibration",
                                     numeric = "Retention.time$",
                                     numeric = "Match.Time.Difference",
                                     numeric = "^intensity$",
                                     "^Type$",
                                     numeric = "Mass\\.Error",
                                     numeric = "^uncalibrated...calibrated." ,
                                     numeric = "^m.z$",
                                     numeric = "^score$",
                                     numeric = "^fraction$",  ## only available when fractions were given
                                     "Raw.file",
                                     "^Protein.Group.IDs$",
                                     "Contaminant",
                                     "^modifications$",
                                     numeric = "^Charge$",
                                     "modified.sequence",
                                     numeric = "^Mass$",
                                     "^protein.names$",
                                     numeric = "^ms.ms.count$",
                                     numeric = "^reporter.intensity.", ## we want .corrected and .not.corrected
                                     numeric = "Missed\\.cleavages",
                                     "^sequence$"))
    ## contains NA if 'genuine' ID
    ## ms.ms.count is always 0 when mtd has a number; 'type' is always "MULTI-MATCH" and ms.ms.ids is empty!
    #dsub = d_evd[,c("ms.ms.count", "match.time.difference")]
    #head(dsub[is.na(dsub[,2]),])
    #sum(0==(dsub[,1]) & is.na(dsub[,2]))
    ##
    ## MQ1.4 MTD is either: NA or a number
    ##
    if (!is.null(all_evd)) all_evd$is.transferred = (all_evd$type == "MULTI-MATCH")

    df_evd = all_evd[all_evd$type != "MULTI-MATCH", ]
    df_evd_tf = all_evd[all_evd$type == "MULTI-MATCH", , drop=FALSE] ## keep columns, if empty

  }
  ## just a local scope to fold evidence metrics in the editor...
  {
    if (!checkEnglishLocale(df_evd)){
      stop ("\n\nThe data in evidence.txt looks weird! MaxQuant was run under a wrong locale/region settings (i.e. make sure to use an english locale, specifically, the decimal separator should be '.'!). Please fix the locale on the PC where MaxQuant was used, and redo the computation.\n\n")
    }

    ### warn of special contaminants!
    if (inherits(yaml_param$yaml_contaminants, "list"))  ## SC are requested
    {
      if (!is.null(df_pg))
      {
        lst_qcMetrics[["qcMetric_EVD_UserContaminant"]]$setData(df_evd, df_pg, yaml_param$yaml_contaminants)
      } else {
        lst_qcMetrics[["qcMetric_EVD_UserContaminant"]]$setData(df_evd, NULL, yaml_param$yaml_contaminants)
      }
    }

    ##
    ## intensity of peptides
    ##
    lst_qcMetrics[["qcMetric_EVD_PeptideInt"]]$setData(df_evd, yaml_param$param_EV_intThresh)

    ##
    ## MS2/MS3 labeled (TMT/ITRAQ) only: reporter intensity of peptides
    ##
    lst_qcMetrics[["qcMetric_EVD_ReporterInt"]]$setData(df_evd)


    ##
    ## Variable modification frequencies
    ##
    lst_qcMetrics[["qcMetric_EVD_modTable"]]$setData(df_evd)



    ##
    ## peptide & protein counts
    ##
    lst_qcMetrics[["qcMetric_EVD_ProteinCount"]]$setData(df_evd, df_evd_tf, yaml_param$param_EV_protThresh)

    lst_qcMetrics[["qcMetric_EVD_PeptideCount"]]$setData(df_evd, df_evd_tf, yaml_param$param_EV_pepThresh)

    ####
    #### peak length (not supported in MQ 1.0.13)
    ####
    if ("retention.length" %in% colnames(df_evd))
    {
      lst_qcMetrics[["qcMetric_EVD_RTPeakWidth"]]$setData(df_evd)
      #lst_qcMetrics[["qcMetric_EVD_CarryOver"]]$setData(df_evd)
    } ## end retention length (aka peak width)

    ##
    ## retention time calibration (to see if window was sufficiently large)
    ## (not supported in MQ 1.0.13)
    ## Even if MBR=off, this column always contains numbers (usually 0, or very small)
    ##

    if ("retention.time.calibration" %in% colnames(df_evd))
    {
      ## this should enable us to decide if MBR was used (we could also look up parameters.txt -- if present)
      if (!(yaml_param$param_evd_mbr == FALSE) & nrow(df_evd_tf)>0)
      {
        lst_qcMetrics[["qcMetric_EVD_MBRAlign"]]$setData(df_evd,
                                                         tolerance_matching = yaml_param$param_EV_MatchingTolerance,
                                                         raw_file_mapping = eval(expr_fn_map)$raw_file_mapping)

        ###
        ###     MBR: ID transfer
        ###
        #debug (restore data): lst_qcMetrics[["qcMetric_EVD_RTPeakWidth"]]$setData(df_evd)
        avg_peak_width = lst_qcMetrics[["qcMetric_EVD_RTPeakWidth"]]$outData[["avg_peak_width"]]
        if (is.null(avg_peak_width)) {
          warning("RT peak width module did not run, but is required for MBR metrics. Enable it and try again or switch off MBR metrics!")
        } else lst_qcMetrics[["qcMetric_EVD_MBRIdTransfer"]]$setData(df_evd, df_evd_tf, avg_peak_width)


        ##
        ## MBR: Tree Clustering (experimental)
        ##  and
        ## MBR: additional evidence by matching MS1 by AMT across files
        ##
        lst_qcMetrics[["qcMetric_EVD_MBRaux"]]$setData(all_evd)

      } ## MBR has data
    } ## retention.time.calibration column exists


    ##
    ## charge distribution
    ##
    ##  (this uses genuine peptides only -- no MBR!)
    ##
    lst_qcMetrics[["qcMetric_EVD_Charge"]]$setData(df_evd)

    ##
    ## peptides per RT
    ##
    lst_qcMetrics[["qcMetric_EVD_IDoverRT"]]$setData(df_evd)


    ##
    ## upSet plot
    ##
    lst_qcMetrics[["qcMetric_EVD_UpSet"]]$setData(df_evd)

    ##
    ## barplots of mass error
    ##
    ## MQ seems to mess up mass recal on some (iTRAQ/TMT) samples, by reporting ppm errors which include modifications
    ## , thus one sees >1e5 ppm, e.g. 144.10 Da
    ##  this affects both 'uncalibrated.mass.error..ppm.'   and
    ##                    'mass.error..ppm.'
    ## HOWEVER, 'uncalibrated...calibrated.m.z..ppm.' seems unaffected, but is not available in all MQ versions :(
    ##    also, 'mass' and 'm/z' columns seem unaffected.
    ## We cannot always reconstruct mass_error[ppm] from 'm/z' and mass columns
    ## since 'm/z' is just too close to the theoretical value or islacking precision of the stored numbers.
    ##
    ## The MQ list reports one case with high ppm error (8000), where the KR.count was at fault. We cannot
    ## reconstruct this.
    ##
    ## Also, MaxQuant will not report uncalibrated mass errors if the data are too sparse for a given Raw file.
    ## Then, 'uncalibrated.mass.error..ppm.' will be 'NaN' throughout -- but weirdly, calibrated masses will be reported.
    ##

    ##
    ## MS1-out-of-calibration (i.e. the tol-window being too small)
    ##
    ## additionally use MS2-ID rate (should be below 1%)
    df_idrate = d_smy[, c("fc.raw.file", "ms.ms.identified....")] ## returns NULL if d_smy == NULL

    lst_qcMetrics[["qcMetric_EVD_PreCal"]]$setData(df_evd, df_idrate, yaml_param$param_EV_PrecursorTolPPM, yaml_param$param_EV_PrecursorOutOfCalSD)


    ##
    ## MS1 post calibration
    ##
    lst_qcMetrics[["qcMetric_EVD_PostCal"]]$setData(df_evd, df_idrate, yaml_param$param_EV_PrecursorTolPPM, yaml_param$param_EV_PrecursorOutOfCalSD, yaml_param$param_EV_PrecursorTolPPMmainSearch)


    ##
    ## Top5 contaminants
    ##
    lst_qcMetrics[["qcMetric_EVD_Top5Cont"]]$setData(df_evd)

    ##
    ## Oversampling: determine peaks repeatedly sequenced
    ##
    lst_qcMetrics[["qcMetric_EVD_MS2OverSampling"]]$setData(df_evd)

    ##
    ## missing values
    ##
    lst_qcMetrics[["qcMetric_EVD_MissingValues"]]$setData(df_evd)

    ## trim down to the absolute required (we need to identify contaminants in MSMS.txt later on)
    ## --> use %in% because some columns, e.g. 'missed.cleavages' are optional
    if (!DEBUG_PTXQC) df_evd = df_evd[, names(df_evd) %in% c("id", "contaminant", "fc.raw.file", "sequence", "missed.cleavages")]
  }




  ######
  ######  msms.txt ...
  ######

  if (MZTAB_MODE) df_msms = mzt$getMSMSScans(identified_only = TRUE) else
    df_msms = mq$readMQ(txt_files$msms, type="msms", filter = "", col_subset=c(numeric = "Missed\\.cleavages",
                                                                                  "^Raw.file$",
                                                                                  "^mass.deviations",
                                                                                  "^masses$",
                                                                                  "^mass.analyzer$",
                                                                                  "^sequence$",
                                                                                  "fragmentation",
                                                                                  "reverse",
                                                                                  numeric = "^evidence.id$"
  ), check_invalid_lines = FALSE)
  ## just a scope
  {
    ### missed cleavages (again)
    ### this is the real missed cleavages estimate ... but slow
    #df_msms_s = mq$readMQ(txt_files$msms, type="msms", filter = "", nrows=10)
    #colnames(df_msms_s)
    #head(df_msms)

    ##
    ##  MS2 fragment decalibration
    ##
    lst_qcMetrics[["qcMetric_MSMS_MSMSDecal"]]$setData(df_msms, eval(expr_fn_map)$raw_file_mapping$to)

    ##
    ## missed cleavages per Raw file
    ##
    # df_evd can be NULL; that's no problem
    lst_qcMetrics[["qcMetric_MSMS_MissedCleavages"]]$setData(df_msms, df_evd)

    # In case missed.cleavages is not in msms but in evd
    # metric checks if it was already done
    lst_qcMetrics[["qcMetric_MSMS_MissedCleavages"]]$setData(df_evd)



  }
  ## save RAM: msms.txt is not required any longer
  if (!DEBUG_PTXQC) rm(df_msms)
  if (!DEBUG_PTXQC) rm(df_evd)



  ######
  ######  msmsScans.txt ...
  ######
  if (MZTAB_MODE) df_msmsScans = mzt$getMSMSScans(identified_only = FALSE) else
     df_msmsScans = mq$readMQ(txt_files$msmsScan, type = "msms_scans", filter = "",
                                col_subset = c(numeric = "^ion.injection.time",
                                               numeric = "^retention.time$",
                                               "^Identified",
                                               numeric = "^Scan.event.number",
                                               numeric = 'Scan.index',    ## required for fixing scan.event.number, in case its broken
                                               numeric = 'MS.scan.index', ## required for fixing scan.event.number, in case its broken
                                               "^total.ion.current",
                                               "^base.?peak.intensity", ## basepeak.intensity (MQ1.2) and base.peak.intensity (MQ1.3+)
                                               "^Raw.file",
                                               "^dp.aa$",
                                               "^dp.modification$"),
                                check_invalid_lines = FALSE)

  # just a scope
  {
    ##
    ## MQ version 1.0.13 has very rudimentary MSMSscans.txt, with no header, so we need to skip the metrics of this file
    ##
    if (!is.null(df_msmsScans) && ncol(df_msmsScans) > 3)
    {
      # round RT to 2 min intervals
      df_msmsScans$rRT = round(df_msmsScans$retention.time/2)*2

      ##
      ## TopN over RT
      ##
      lst_qcMetrics[["qcMetric_MSMSScans_TopNoverRT"]]$setData(df_msmsScans)

      ##
      ## Injection time over RT
      ##
      lst_qcMetrics[["qcMetric_MSMSScans_IonInjTime"]]$setData(df_msmsScans, yaml_param$param_MSMSScans_ionInjThresh)

      ##
      ## MS/MS intensity (TIC and base peak)
      ##
      lst_qcMetrics[["qcMetric_MSMSScans_MSMSIntensity"]]$setData(df_msmsScans)

      ##
      ## TopN counts
      ##
      lst_qcMetrics[["qcMetric_MSMSScans_TopN"]]$setData(df_msmsScans)

      ##
      ## Scan event: % identified
      ##
      lst_qcMetrics[["qcMetric_MSMSScans_TopNID"]]$setData(df_msmsScans)

      ##
      ## Dependent peptides (no score)
      ##
      if ("dp.modification" %in% colnames(df_msmsScans)) {
        lst_qcMetrics[["qcMetric_MSMSScans_DepPep"]]$setData(df_msmsScans)
      }

    } ## end MSMSscan from MQ > 1.0.13


  }
  ## save RAM: msmsScans.txt is not required any longer
  if (!DEBUG_PTXQC) rm(df_msmsScans)

  #####################################################################
  #####################################################################
  ## write mzQC file
  try( ## if not enough metrics are produced, then writing will fail (e.g. one run or setQuality needs to be present)
    rmzqc::writeMZQC(
      rprt_fns$mzQC_file,
      assembleMZQC(lst_qcMetrics, raw_file_mapping = eval(expr_fn_map)$raw_file_mapping)
    )
  )


  #####################################################################
  ## list of qcMetric objects
  print("#Metrics: ")
  print(length(lst_qcMetrics))

  hm = getQCHeatMap(lst_qcMetrics, raw_file_mapping = eval(expr_fn_map)$raw_file_mapping)
  #print(hm[["plot"]])
  write.table(hm[["table"]], file = rprt_fns$heatmap_values_file, quote = TRUE, sep = "\t", row.names = FALSE)

  ## get MQ short name mapping plot (might be NULL if no mapping was required)
  pl_nameMapping = eval(expr_fn_map)$plotNameMapping()

  ##
  ## plot it!!!
  ##
  cat("Creating Report file ...")

  #
  #param_OutputFormats = "html pdf"
  #
  out_formats = unlist(strsplit(yaml_param$param_OutputFormats, "[ ,]+"))
  out_formats
  out_format_requested = out_formats_supported[match(out_formats, out_formats_supported)]
  if (any(is.na(out_format_requested)))
  {
    stop("Output format(s) not supported: '", paste(out_formats[is.na(out_format_requested)], collapse="', '"), "'")
  }

  ## a bit hacky, but we want gridExtra plots to plot when we call print() -- similar to ggplot's print
  print.gtable = function(t) { plot(t)}

  if ("html" %in% out_format_requested)
  {
    if (rmarkdown::pandoc_available()) {
      ## HTML reports require Pandoc for converting Markdown to Html via the rmarkdown package
      if (DEBUG_PTXQC) {
        html_template = paste0(getwd(), "/inst/reportTemplate/PTXQC_report_template.Rmd")
        if (!file.exists(html_template)) stop("Wrong working directory. Please set your working directory to the PTXQC main dir such that 'paste0(getwd(), '/inst/reportTemplate/PTXQC_report_template.Rmd')' is a valid file.")
      } else {
        html_template = system.file("./reportTemplate/PTXQC_report_template.Rmd", package="PTXQC")
      }
      cat(paste0("HTML TEMPLATE: ", html_template, "\n"))
      out_dir = dirname(rprt_fns$report_file_HTML)
      file.copy(html_template, out_dir, overwrite = FALSE)
      out_template = file.path(out_dir, basename(html_template))
      ## Rmarkdown: convert to Markdown, and then to HTML (or PDF) ...
      ## Intermediates_dir is required if inputdir!=outputdir, since Shiny server might not allow write-access to input file directory
      res_html = try(
        rmarkdown::render(out_template, output_file = rprt_fns$report_file_HTML) #, intermediates_dir = dirname(rprt_fns$report_file_HTML))
      )
      if (inherits(res_html, "try-error")) {
        warning("Creating the HTML template did not succeed, probably due to an outdated markdown template the in
                txt folder. PTXQC will use the default template now instead. Fix or remove the broken/old .Rmd file from the ", txt_folder,
                " to avoid this warning.", immediate. = TRUE)
        rmarkdown::render(html_template, output_file = rprt_fns$report_file_HTML) #, intermediates_dir = dirname(rprt_fns$report_file_HTML))
      }
   } else {
      warning("The 'Pandoc' converter is not installed on your system or you do not have read-access to it!\n",
              "Pandoc is required for HTML reports.\n",
              "Please install Pandoc <http://pandoc.org/installing.html> or make sure you have access to pandoc(.exe).\n",
              "Restart your R-session afterwards.",
              immediate. = TRUE)
    }
  }

  if ("plainPDF" %in% out_format_requested)
  {
    report_file_PDF = rprt_fns$report_file_PDF
    ## give the user a chance to close open reports which are currently blocked for writing
    if (!wait_for_writable(report_file_PDF))
    {
      stop("Target file not writable")
    }

    if (yaml_param$param_PageNumbers == "on")
    {
      printWithPage = function(gg_obj, page_nr, filename = report_file_PDF)
      {
        filename = basename(filename)
        printWithFooter(gg_obj, bottom_left = filename, bottom_right = page_nr)
      }
    } else {
      ## no page number and filename at bottom of each page
      printWithPage = function(gg_obj, page_nr, filename = report_file_PDF)
      {
        print(gg_obj)
      }
    }
    cat("\nCreating PDF ...\n")
    grDevices::pdf(report_file_PDF)
    printWithPage(hm[["plot"]], "p. 1")      # summary heatmap
    printWithPage(pl_nameMapping$plots, "p. 2")    # short file mapping
    pc = 3; ## subsequent pages start at #4
    for (qcm in lst_qcMetrics)
    {
      for (p in qcm$plots)
      {
        printWithPage(p, paste("p.", pc))
        pc = pc + 1
      }
    }
    grDevices::dev.off();
    cat(" done\n")
  }

  ## save plot object (for easier access, in case someone wants high-res plots)
  ## (...disabled for now until concrete use case pops up)
  #cat("Dumping plot objects as Rdata file ...")
  #save(file = rprt_fns$R_plots_file, list = "GPL")
  #cat(" done\n")

  ## write shortnames and sorting of filenames (again)
  eval(expr_fn_map)$writeMappingFile(rprt_fns$filename_sorting)

  cat(paste("Report file created at\n\n    ", rprt_fns$report_file_prefix, ".*\n\n", sep=""))
  cat(paste0("\n\nTime elapsed: ", round(as.double(Sys.time() - time_start, units="mins"), 1), " min\n\n"))


  ## return path to PDF report and YAML config, etc
  return(rprt_fns)
}

library(PTXQC)

## the next require() is needed to prevent a spurious error in certain R versions (might be a bug in R or a package)
## error message is:
##    Error in Scales$new : could not find function "loadMethod"
require(methods)

context("createReport.R")

test_that("createReport", {
  ## this is a rather lengthy function, and its hard to test in all its granularity (hence we test
  ## the functions within in other tests), but at least we want to see that it produces
  ## results -- even though its hard to test if the resulting PDF report is exactly what we expect
  ## without going into mindboggling image-comparison tests...
  
  
  ## get some data
  local_zip = tempfile(fileext=".zip")
  ## getting the correct URL is tricky, since download methods like 'curl' will not follow
  ## redirects, but download the server message instead. Try 'curl' etc on the command line, to see
  ## what's going on and find the right URL (for GitHub: we need https and raw.github....)
  target_url = "https://raw.githubusercontent.com/cbielow/PTXQC_data/master/txt_Ecoli.zip"
  dl = NULL
  tryCatch({
    dl = download.file(target_url, destfile = local_zip, quiet = TRUE)
  }, silent = TRUE, warning = function(w) { }, error = function(err) {
    ## in case of error, try with Curl
    tryCatch({
      dl = download.file(target_url, destfile = local_zip, method='curl', quiet = TRUE) ## for Linux/MacOSX
    }, silent = TRUE, warning = function(w) { }, error = function(err2) {
      print("Internet down. Aborting test gracefully")
      return(); ## fail gracefully
    })
  })
  if (is.null(dl)) return()
  
  unzip(local_zip, exdir = tempdir()) ## extracts content
  txt_folder = file.path(tempdir(), "txt")
  
  ##
  ## test mqPar parser while we have the data here..
  ## 
  expect_equal(getMQPARValue(paste0(txt_folder, "/mqpar.xml"), "//firstSearchTol"), "20")
  expect_equal(getMQPARValue(paste0(txt_folder, "/mqpar.xml"), "//matchingTimeWindow"), "0.7")
  expect_equal(getMQPARValue(paste0(txt_folder, "/mqpar.xml"), "//string[parent::filePaths|parent::Filenames]", allow_multiple = TRUE),
               c("M:\\projects\\QC\\ecoli_small\\Toni_20120502_GM_Ecoli_01.raw", "M:\\projects\\QC\\ecoli_small\\Toni_20120502_GM_Ecoli_02.raw"))
  
  
  
  yaml_obj = list() ## no special config...
  r = PTXQC::createReport(txt_folder, NULL, yaml_obj)
  expect_equal(c("yaml_file", "heatmap_values_file", "R_plots_file", "filename_sorting", "mzQC_file",         
                 "log_file", "report_file_prefix", "report_file_PDF", "report_file_HTML"), names(r))
  rep_files = c(r[["report_file_PDF"]], r[["report_file_HTML"]])
  
  print(list.files(path = txt_folder))
  
  for (f in rep_files)
  {
    cat("Checking file ", f, "\n")
    # HTML file might not exist if PANDOC is not installed
    if (file.exists(f)) expect_equal(file.info(f)$size > 50*1024, TRUE) ## PDF 70kb & HTML 161kb
  }
  
  expect_equal(file.exists(r[["heatmap_values_file"]]), TRUE)
  d_heatmap = read.delim(r[["heatmap_values_file"]])
  expect_equal(dim(d_heatmap)[1], 2) ## two files
  expect_equal(dim(d_heatmap)[2] >= 22, TRUE) ## 22 (or more) metrics
  expect_equal(as.character(d_heatmap$fc.raw.file), c("..Ecoli_01", "..Ecoli_02"))
  
  expect_equal(file.exists(r[["filename_sorting"]]), TRUE)
  d_filenamesort = read.delim(r[["filename_sorting"]], comment.char="#")
  expect_equal(dim(d_filenamesort), c(2, 3)) ## two files, three columns
  expect_equal(as.character(d_filenamesort$new.Name), c("..Ecoli_01", "..Ecoli_02"))
  
  ## remove creationDate, contactName, and URI from mzQC file, since its changing dynamically
  no_volatile = function(file)
  {
    lines = readLines(file, warn = FALSE)
    grep("\"creationDate\"|\"contactName\"|\"location\"|\"version\"|\"uri\"", lines, invert = TRUE, value = TRUE)
  }
  
  expect_equal(no_volatile(r[["mzQC_file"]]), no_volatile(system.file("./examples/report_ecoli_small.mzQC", package="PTXQC")))
  
  unlink(local_zip) ## delete zip
  unlink(txt_folder, recursive = TRUE) ## delete txt-folder
  
})

#' Create a principal component analysis (PCA) plot for the first two dimensions.
#' 
#' @param data               Matrix(!) where each row is one high-dimensional point, with ncol dimensions, 
#'                           e.g. a mouse as an array of proteinexpressions
#'                           rownames(data) give classes for colouring (can be duplicates in matrices, as opposed to data.frames)
#' @param do_plot            Show PCA plot? if ==2, then shows correlations plot as well
#' @param connect_line_order Connect points by lines, the order is given by this vector.
#'                           Default: NA (no lines)
#' @param gg_layer           More parameters added to a ggplot object (ggplot(x) + gg_layer)

#' @return [invisible] Named list with "PCA": The PCA object as returned by \code{\link[stats]{prcomp}}, access $x for PC values
#'                                 and "plots": list of plot objects (one or two)
#' 
#' @import ggplot2
#' 
#' @examples
#' n = 5
#' m = 10
#' data = matrix(runif(n * m), nrow = n, ncol = m)
#' rownames(data) = 1:n
#' getPCA(data, connect_line_order = 1:n, gg_layer = ggplot2::ggtitle("test"))
#' 
#' @export
#' 
getPCA = function(data, do_plot = TRUE, connect_line_order = NA, gg_layer)
{
  pc = stats::prcomp(data, scale. = TRUE)
  useOrd = !is.na(connect_line_order[1])
  # create data frame with scores
  scores = as.data.frame(pc$x)
  scores$class = rownames(data)
  rownames(scores) = paste0(rownames(scores),1:nrow(scores))
  show.Line = !is.na(connect_line_order[1])
  if (show.Line)
  {
    scores$pathX = scores$PC1[connect_line_order]
    scores$pathY = scores$PC2[connect_line_order]
  }
  if (do_plot & show.Line)
  {
    scores$ord = connect_line_order / max(connect_line_order) * 10
  } else
  {
    scores$ord = as.numeric(factor(rownames(data)))
  }
  
  lpl = list();
  
  # plot of observations
  if (do_plot)
  {
    
    var_PC1 = round(pc$sdev[1]^2 / sum(pc$sdev^2) * 100)
    var_PC2 = round(pc$sdev[2]^2 / sum(pc$sdev^2) * 100)
    
    pl = ggplot(data=scores, aes(x = .data$PC1, y = .data$PC2));
    if (show.Line) pl = pl + geom_path(aes(x = .data$pathX, y = .data$pathY, color = .data$ord), alpha = 0.5)
    pl = pl + scale_colour_gradient(low="red", high="darkgreen")
    pl = pl + 
              gg_layer + 
              geom_point(aes(colour = .data$ord), size = 1) +
              geom_text(aes(label = .data$class, colour = .data$ord, vjust = 1), size = 4, angle=0) +
              geom_hline(yintercept=0, colour="gray65") +
              #theme(panel.background=element_rect("black")) +
              geom_vline(xintercept=0, colour="gray65") +
              theme_bw() + 
              xlab(paste0("PC #1 (", var_PC1, "% var)")) +
              ylab(paste0("PC #2 (", var_PC2, "% var)"))
    lpl[[1]] = pl
  }
  
  correlations = NA
  
  if (do_plot==2)
  {
    my_circle = function(center=c(0,0), npoints=100)
    {
      r = 1
      tt = seq(0, 2*pi, length=npoints)
      xx = center[1] + r * cos(tt)
      yy = center[1] + r * sin(tt)
      return(data.frame(x = xx, y = yy))
    }
    corcir = my_circle(c(0,0), npoints = 100)
    # create data frame with correlations between variables and PCs
    correlations = as.data.frame(cor(data, pc$x))
    # data frame with arrows coordinates
    arrows = data.frame(x1=rep(0, ncol(data)), y1=rep(0, ncol(data)),
                        x2=correlations$PC1, y2=correlations$PC2)
    
    
    lpl[[2]] = 
      ggplot() +
      geom_path(data = corcir, aes(x = .data$x, y = .data$y), colour="gray65") +  ## open circles
      geom_segment(data = arrows, aes(x = .data$x1, y = .data$y1, xend = .data$x2, yend = .data$y2), colour="gray65") +
      geom_text(data = correlations, aes(x  = .data$PC1, y = .data$PC2, label = rownames(correlations))) +
      geom_hline(yintercept = 0, colour = "gray65") +
      geom_vline(xintercept = 0, colour = "gray65") +
      xlim(-1.1,1.1) + ylim(-1.1,1.1) +
      labs(x="pc1 axis", y="pc2 axis") +
      ggtitle("Circle of correlations")

  }
  return (list("PCA" = invisible(pc), "plots" = lpl, "correlations" = correlations))
}

qcMetric_SM_MSMSIdRate =  setRefClass(
  "qcMetric_SM_MSMSIdRate",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(  
    helpTextTemplate = 
      "MS/MS identification rate per Raw file from summary.txt (SM). Each Raw file is colored
according to its ID rate and categorized into performance bins as 'bad', 'ok' and 'great'. 
Raw files below 'ok', are listed separately on the next page of the report for convenient follow-up.

The thresholds for the bins are

%s


Heatmap score [SM: MS<sup>2</sup> IDrate (>%1.0f)]: reaches 1 (=100%%) if the threshold for 'great' is reached or exceeded. 
",

    workerFcn = function(.self, df_summary, id_rate_bad, id_rate_great)
    {
      if (!checkInput(c("fc.raw.file", "ms.ms.identified...."), df_summary)) return()
      
      if (nrow(df_summary) == 0) return() ## empty for DIA data
      
      dms = df_summary$"ms.ms.identified...."
      dms[is.na(dms)] = 0  ## ID rate can be NaN for some raw files if NOTHING was acquired
      lab_IDd = c("red", 
                  "blue", 
                  "green")
      names(lab_IDd) = c("bad (<" %+% id_rate_bad %+% "%)", 
                         "ok (" %+% id_rate_bad %+% "-" %+% id_rate_great %+% "%)",
                         "great (>" %+% id_rate_great %+% "%)")
      df_summary$cat = factor(cut(dms, breaks=c(-1, id_rate_bad, id_rate_great, 100), labels=names(lab_IDd)))
      
      pl = plot_IDRate(df_summary, id_rate_bad, id_rate_great, lab_IDd)
      lpl = list(pl)
      
      ## table of files with 'bad' MS/MS id rate
      bad_id_count = sum(df_summary$cat==names(lab_IDd[1]))
      if (bad_id_count>0)
      {
        sm_badID = df_summary[df_summary$cat==names(lab_IDd[1]), c("raw.file","ms.ms.identified....")]
        if (nrow(sm_badID) > 40)
        {
          sm_badID[40, "raw.file"] = paste(nrow(sm_badID) - 39, "more ...");
          sm_badID[40, "ms.ms.identified...."] = ""
          sm_badID = sm_badID[1:40, ]
        }
        plot_title = paste0("SM: Files with '", lab_IDd[1], "' ID rate ")
        p_tbl = plotTable(sm_badID, 
                          title = plot_title,
                          footer = paste0(round(bad_id_count*100/nrow(df_summary)), "% of samples)"),
                          col_names = c("Raw file", "% identified"))
        lpl[["bad_id_table"]] =  p_tbl
      }
      
      ## update help text according to actual limits
      inText = paste(paste0("  - ", names(lab_IDd)), collapse="\n", sep="")
      
      .self$helpText = sprintf(.self$helpTextTemplate, inText, id_rate_great)
      
      ## QC measure for ID rate (threshold reached?)
      ## update QC name based on parameter values
      ## QC measure for ID rate (threshold reached?)
      qc_sm_id = df_summary[, c("fc.raw.file", "ms.ms.identified....")]
      cname = sprintf(.self$qcName, id_rate_great)
      qc_sm_id[, cname] = qualLinThresh(qc_sm_id$ms.ms.identified.... , id_rate_great)
      qcScore = qc_sm_id[,c("fc.raw.file", cname)]
      
      return(list(plots = lpl, qcScores = qcScore))
    }, 
    qcCat = "MS", 
    qcName = "SM:~MS^2~ID~rate (\">%1.0f\")", 
    orderNr = 300
  )
    return(.self)
  })
)


qcMetric_SM_TIC =  setRefClass(
  "qcMetric_SM_TIC",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate = 
      "Total Ion Count: Returns the summed intensity of all MS1 signals (regardless of identification state).

Heatmap score [SM: TIC]: reaches 1 (=100%%) if the TIC is uniform (i.e. a flat line)
",
    workerFcn = function(.self, d_smy)
    {
      ## completeness check
      if (!checkInput(c("fc.raw.file", "TIC"), d_smy)) return()
      
      df_long = plyr::ddply(d_smy, "fc.raw.file", function(x) {
        n = length(x$TIC[[1]])
        df = data.frame(RT = x$TIC[[1]][seq(1,n,2)], intensity = x$TIC[[1]][seq(2,n,2)])
        df$RT = round(df$RT / 60) ## seconds to minutes
        df2 = data.frame(RT = df$RT[!duplicated(df$RT)], intensity = tapply(df$intensity, df$RT, FUN = mean))
        return(df2)
      })
      
      head(df_long)
      
      lpl =
        byXflex(df_long, df_long$fc.raw.file, 6, plot_TIC, x_lim = range(df_long$RT), y_lim = range(df_long$intensity), sort_indices = FALSE)
      
      ## QC measure for smoothness of TopN over RT
      qc_TIC = plyr::ddply(df_long, "fc.raw.file", function(x) data.frame(val = qualUniform(x$intensity)))
      colnames(qc_TIC)[colnames(qc_TIC) == "val"] = .self$qcName
      
      return(list(plots = lpl, qcScores = qc_TIC))
    }, 
    qcCat = "LC", 
    qcName = "SM:~TIC", 
    orderNr = 0025
  )
    return(.self)
  })
)

#' 
#' Define a Singleton class which holds the full raw filenames (+path) and their PSI-MS CV terms for usage in the mzQC metadata
#' 
#' The internal data is filled using, e.g. 'getMetaFilenames()'
#' 
#' @export
#' 
QCMetaFilenames <- R6::R6Class("QCMetaFilenames", inherit = R6P::Singleton, public = list(
  #' @field data Stores the data of the singleton. Set the data once before using the singleton all over the place
  data = NULL
))


#'
#' Parses the given mqpar.xml file (or, if not found, tries the 'txt_folder' + '/../../' folder (i.e. where the raw data should be)) to extract the full filepaths for all Raw files
#' 
#' @param mqpar_file Location of the mqpar.xml (can be empty, if unknown)
#' @param txt_folder Fallback option: path to the txt folder (which contains evidence.txt, etc)
#'
#' @return May return 'NULL' if no mqpar.xml could be found.
#'         Otherwise: data.frame with columns:
#'         \itemize{
#'           \item 'file' (no path), 'path' (full path incl. names)
#'           \item 'file_no_suffix' (as 'file' but without suffix) 
#'           \item 'CV' (CV term for filetype, e.g. for Thermo Raw)
#'         }
#'
getMetaFilenames = function(mqpar_file, txt_folder)
{
  out = NULL
  ## mqpar_file = "Z:\\projects\\QC\\PTXQC\\data\\ecoli_small\\mqpar.xml"
  ## txt_folder = "Z:\\projects\\QC\\PTXQC\\data\\ecoli_small\\combined\\txt\\"
  xml_rawfiles = getMQPARValue(mqpar_file, "//string[parent::filePaths|parent::Filenames]", allow_multiple = TRUE)
  if (is.null(xml_rawfiles)) {
    ## try again using parent directory
    warning("No mqpar.xml found in '", mqpar_file, "'. Trying two folders up.")
    up_dir = paste0(txt_folder, "/../../mqpar.xml")
    xml_rawfiles = getMQPARValue(up_dir, "//string[parent::filePaths|parent::Filenames]", allow_multiple = TRUE)
  }
  ## second try...
  if (is.null(xml_rawfiles)) {
    warning("No mqpar.xml found in '", up_dir, "'. Giving up to read full file paths The mqQC file will contain incomplete information and may not validate.")
  } else {
    ## cannot use 'eval(expr_fn_map)$raw_file_mapping' yet, since we do not have read any .txt files which fills the mapping
    out = data.frame(file = basename(xml_rawfiles), path = xml_rawfiles, file_no_suffix = rmzqc::removeFileSuffix(basename(xml_rawfiles)), CV = rmzqc::filenameToCV(xml_rawfiles))
  }
  return (out)
}



#'
#' Get an mzQC runQuality without actual metrics, but with full metadata
#' 
#' @param fc.raw.file For which run
#' @param raw_file_mapping A data.frame with cols 'from', 'to' and maybe 'best.effort' (if shorting was unsuccessful), as e.g. obtained by a FilenameMapper$raw_file_mapping
#' @return An MzQCrunQuality object
#' 
#' @import rmzqc
#'
getRunQualityTemplate = function(fc.raw.file, raw_file_mapping)
{
  
  idx = which(raw_file_mapping$to == fc.raw.file)
  if (length(idx) != 1) stop("fc.raw.file '", fc.raw.file, "' is not (or not unique) in mapping table.")
  
  raw_file = as.character(raw_file_mapping$from[idx])
  meta = QCMetaFilenames$new()$data
  if (is.null(meta) || sum(meta$file_no_suffix == raw_file) == 0) {
    ## no mqpar.xml found
    ## we're just guessing here...
    warning("Cannot properly fill metadata of mzQC file, since full filenames are unknown due to missing mqpar.xml. Using placeholders.")
    filename = paste0(raw_file, ".raw"); 
    location = paste0("???/", filename);
    accession = rmzqc::filenameToCV(filename)
  } else {
    idx_meta = which(meta$file_no_suffix == raw_file)
    filename = as.character(meta$file[idx_meta])
    location = as.character(meta$path[idx_meta])
    accession = as.character(meta$CV[idx_meta])
  }
  ## make location a proper URI
  if (!startsWith(location, "file:///")) location = paste0("file:///", location);
  location = gsub("\\", "/", location, fixed = TRUE)
  file_format = rmzqc::getCVTemplate(accession = accession)
  ptxqc_software = rmzqc::toAnalysisSoftware(id = "MS:1003162", version = as.character(utils::packageVersion("PTXQC")))
  
  out = rmzqc::MzQCrunQuality$new(rmzqc::MzQCmetadata$new(raw_file,  ## label
                                                          list(rmzqc::MzQCinputFile$new(filename, location, file_format)),
                                                          list(ptxqc_software)),
                                  list())
  
  return(out)
}

#'
#' Collects all 'mzQC' members from each entry in lst_qcMetrics and stores them in an overall mzQC object, which can be written to disk (see writeMZQC()) or augmented otherwise
#'
#' @param lst_qcMetrics A list of qcMetric objects which have their mzQC member populated with "MzQCrunQuality" and/or "MzQCsetQuality" objects
#' @param raw_file_mapping A data.frame with cols 'from', to' and maybe 'best.effort' (if shorting was unsuccessful), as e.g. obtained by a FilenameMapper$raw_file_mapping
#' @return An MzQCmzQC object (root object of an mzQC document)
#' 
#' @export
#' 
assembleMZQC = function(lst_qcMetrics, raw_file_mapping)
{
  out = rmzqc::MzQCmzQC$new(version = "1.0.0", 
                            creationDate = MzQCDateTime$new(), 
                            contactName = Sys.info()["user"], 
                            contactAddress = NA_character_, 
                            description = NA_character_,
                            runQualities = list(),
                            setQualities = list(), 
                            controlledVocabularies = list(rmzqc::getCVInfo()))

  run_qualities = list()
  set_qualities = list()
  
  for (metric in lst_qcMetrics)
  {
    mzqc_data = metric$mzQC
    if (is.null(mzqc_data)) next
    if (!inherits(mzqc_data, "list")) stop("mzQC member of metric must be of class 'list()'")
    if (length(mzqc_data) == 0) next
    if (any(is.null(names(mzqc_data)))) stop("mzQC member of metric '", metric$qcName, "' is a list, but has no names (must be fc.raw.file names or a concatenation of those)")
    
    ## either an fc.raw.file or a concatenation of them
    for (name in names(mzqc_data))
    {
      if (name %in% raw_file_mapping$to)
      { ##
        ## runQuality
        ##
        rc = run_qualities[[name]];
        if (is.null(rc)) rc = getRunQualityTemplate(name, raw_file_mapping)
        ## append ...
        l = 1 + length(rc$qualityMetrics)
        rc$qualityMetrics[[l]] = mzqc_data[[name]]
        ## write back
        run_qualities[[name]] = rc
      } else {
        if (!grepl(";", name, fixed=TRUE)) stop("mzQC metric data must be an fc.raw.file or a contatenation of those using ';'. No ';' found!")
        ##
        ## setQuality
        ## 
        stop("setQuality not implemented yet")
      }
    }
    
  }
  ## remove the names from lists (to make it a JSON array; otherwise it would be an object)
  names(run_qualities) = NULL
  names(set_qualities) = NULL
  
  out$runQualities = run_qualities
  out$setQualities = set_qualities
  
  return(out)
}

#####################################################################

qcMetric_MSMS_MSMSDecal =  setRefClass(
  "qcMetric_MSMS_MSMSDecal",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate =
      "MS/MS decalibration metric. If most of the fragments are within tighter bounds,
you can reduce the fragment mass tolerance to obtain more
identifications under the same FDR. On the other hand, if the fragment mass errors are not centered on
zero, a recalibration of the instrument should be performed.
If the (Gaussian-like) distribution is cut too severely on either side by the search tolerance window in MaxQuant,
you might be able to increase the number of identifications by allowing for a wider MS/MS search window when re-running MaxQuant.
However, the number of decoy identifications will increase as well, potentially offsetting any gain when FDR is applied.

Heatmap score [MSMS: MS<sup>2</sup> Cal (Analyzer)]: rewards centeredness around 0 ppm/Da (function Centered).
",
    workerFcn = function(.self, df_msms, fc_raw_files)
    {
      ## completeness check
      if (!checkInput(c("fc.raw.file", "fragmentation", "reverse", "mass.deviations..da."), df_msms)) return ()
      ## older MQ versions do not have 'mass.analyzer' or 'mass.deviations..ppm.'
      ## , so we use fragmentation instead (this is a little risky, since you could do CID fragmentation and forward to Orbi, but hey...)
      if (!("mass.analyzer" %in% colnames(df_msms)) || all(is.na(df_msms$mass.analyzer))) df_msms$mass.analyzer = df_msms$fragmentation

      sampleMax = function(x, max = 300) {
        ## sample at most 'max' items from 1:x
        ## if 'x' < max, return 1:x
        if (x < max) return(1:x)
        return (sort(sample.int(x, size = max)))
      }

      ms2_decal = plyr::ddply(df_msms, c("fc.raw.file", "mass.analyzer"), .fun = function(x) {
        df.ms = NULL
        ##
        ##  Forwards
        ##
        if (any(!x$reverse))
        {
          idx_nr = which(!x$reverse)
          ## select a representative subset, otherwise the number of datapoints is just too large
          idx_nr_subset = idx_nr[sampleMax(length(idx_nr))]
          df.ms = getFragmentErrors(x[idx_nr_subset, , drop=FALSE])
          if (!is.null(df.ms)) df.ms$type="forward"
        }

        if (any(x$reverse))
        {
          idx_nr = which(x$reverse)
          ## select a representative subset, otherwise the number of datapoints is just too large
          idx_nr_subset = idx_nr[sampleMax(length(idx_nr))]
          df.ms_r = getFragmentErrors(x[idx_nr_subset, , drop=FALSE])
          if (!is.null(df.ms) & !is.null(df.ms_r)) {
            df.ms_r$type="decoy"
            ## only merge if we have hits
            df.ms = rbind(df.ms, df.ms_r)
          }
        }
        return (df.ms)
      })

      #ms2_range = diff(range(ms2_decal$msErr, na.rm = TRUE))
      #ms2_binwidth = ms2_range/20
      ## precision (plotting is just so much quicker, despite using a fixed binwidth)
      #ms2_decal$msErr = round(ms2_decal$msErr, digits=ceiling(-log10(ms2_binwidth)+1))

      ## we may not have that information (MaxQuant 2.7.5 is known to produce all "NaN;NaN..." values)
      if (nrow(ms2_decal) == 0) {
        warning("Metric qcMetric_MSMS_MSMSDecal: all data is NaN. Try a different MaxQuant version if you want to have fragment error information.", immediate. =  TRUE)
        return();
      }

      ## separate plots for each mass analyzer, since we want to keep 'fixed' scales for all raw.files (comparability)
      lpl = plyr::dlply(ms2_decal, "mass.analyzer", function(ms2_decal) {
        ## create filename inside, since we need to retain the factor levels (i.e. ordering)
        ## and this only works if raw file + massanalyzer is unique
        ms2_decal$new_filename = paste(ms2_decal$fc.raw.file, paste(ms2_decal$mass.analyzer, ms2_decal$unit), sep="\n")
        ## i.e. change name without underlying value
        ms2_decal$file = ms2_decal$fc.raw.file
        levels(ms2_decal$file) = ms2_decal$new_filename[ match(levels(ms2_decal$file), ms2_decal$file) ]
        byXflex(ms2_decal, ms2_decal$fc.raw.file, 9, plot_MS2Decal, sort_indices = TRUE)
      })
      ## currently lpl is a list of lists() -- flatten
      lpl = Reduce(append, lpl)

      ##
      ## QC measure for centered-ness of MS2-calibration
      ##
      head(ms2_decal)
      qcScore = list()
      for (analyzer in unique(ms2_decal$mass.analyzer)) {
        qc_name = sprintf(.self$qcName, analyzer)
        qc_MS2_decal = plyr::ddply(ms2_decal[ms2_decal$mass.analyzer==analyzer, ], "fc.raw.file",
                             function(x)
                             {
                               xx = na.omit(x$msErr);
                               return(data.frame(X1 = qualCentered(xx)))
                             })
        ## augment fragmentation methods with -Inf for missing raw files (otherwise they would become 'red'=fail)
        if (length( setdiff(fc_raw_files, qc_MS2_decal$fc.raw.file) )) {
          frag_missing = data.frame(fc.raw.file = setdiff(fc_raw_files, qc_MS2_decal$fc.raw.file), X1=-Inf)
          qc_MS2_decal = rbind(qc_MS2_decal, frag_missing)
        }
        colnames(qc_MS2_decal)[colnames(qc_MS2_decal)=="X1"] = qc_name
        qcScore[[analyzer]] = qc_MS2_decal[, c("fc.raw.file", qc_name)]
      }
      qcScore_all = Reduce(function(x,y) merge(x,y,all=TRUE), qcScore)

      return(list(plots = lpl, qcScores = qcScore_all))
    },
    qcCat = "MS",
    qcName = "MSMS:~MS^2~Cal~(%s)",
    orderNr = 0280
  )
    return(.self)
  })
)

#####################################################################

qcMetric_MSMS_MissedCleavages =  setRefClass(
  "qcMetric_MSMS_MissedCleavages",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate =
      "Under optimal digestion conditions (high enzyme grade etc.), only few missed cleavages (MC) are expected. In
general, increased MC counts also increase the number of peptide signals, thus cluttering the available
space and potentially provoking overlapping peptide signals, biasing peptide quantification.
Thus, low MC counts should be favored. Interestingly, it has been shown recently that
incorporation of peptides with missed cleavages does not negatively influence protein quantification (see
Chiva, C., Ortega, M., and Sabido, E. Influence of the Digestion Technique, Protease, and Missed
Cleavage Peptides in Protein Quantitation. J. Proteome Res. 2014, 13, 3979-86, DOI: 10.1021/pr500294d).
However this is true only if all samples show the same degree of digestion. High missed cleavage values
can indicate for example, either a) failed digestion, b) a high (post-digestion) protein contamination, or
c) a sample with high amounts of unspecifically degraded peptides which are not digested by trypsin.

If MC>=1 is high (>20%) you should increase the missed cleavages settings in MaxQuant and compare the number of peptides.
Usually high MC correlates with bad identification rates, since many spectra cannot be matched to the forward database.

In the rare case that 'no enzyme' was specified in MaxQuant, neither scores nor plots are shown.

Heatmap score [MSMS: MC]: the fraction (0% - 100%) of fully cleaved peptides per Raw file

Heatmap score [MSMS: MC Var]: each Raw file is scored for its deviation (score: MedianDist) from the 'average' digestion state of the
current study. ",
    workerFcn = function(.self, df_any, df_evd = NULL)
    {
      ## metric already ran... return result we have
      if (length(.self$plots) != 0 ) return(list(plots = .self$plots, qcScores = .self$qcScores))
      ## completeness check
      if (!checkInput(c("fc.raw.file", "sequence", "missed.cleavages"), df_any)) return()
      if (!is.null(df_evd) && !checkInput(c("contaminant", "id"), df_evd)) {df_evd = NULL}

      max_mc = max(-Inf, df_any$missed.cleavages, na.rm = TRUE) ## will be -Inf iff enzyme was not specified and columns is 100% NA
      if (!is.infinite(max_mc))
      { ## MC's require an enzyme to be set
        ## remove contaminants
        msg_cont_removed = "(excludes contaminants)"
        if ("contaminant" %in% colnames(df_any)) { # for MzTab
          df_any_filt = df_any[!df_any$contaminant,]
        } else if (!is.null(df_evd)) {
          if (!checkInput(c("evidence.id"), df_any)) return()
          df_any_filt = df_any[!df_evd$contaminant[match(df_any$evidence.id, df_evd$id)], ]
        } else {
          msg_cont_removed = "(includes contaminants -- no evidence.txt read)"
          df_any_filt = df_any
        }

        if (nrow(df_any_filt) == 0)
        { ## rare case when all PSMs are contaminants...
          df_any_filt = df_any ## roll back to at least get a statistic...
          msg_cont_removed = "(100% contaminants!)"
        }

        st_bin = plyr::ddply(df_any_filt[, c("missed.cleavages", "fc.raw.file")], "fc.raw.file", .fun = function(x) {
          t = table(x$missed.cleavages)/nrow(x)
          r = rep(0, max_mc + 1)
          names(r) = as.character(0:max_mc)
          r[names(t)] = t
          return (r)
        })

        lpl =
          byXflex(st_bin, st_bin$fc.raw.file, 25, plot_MissedCleavages, sort_indices = TRUE, title_sub = paste(msg_cont_removed))

        ## QC measure for missed-cleavages variation
        qc_score = data.frame(fc.raw.file = st_bin$fc.raw.file, valMC = st_bin[, "0"])
        qc_score$valMCVar = qualMedianDist(qc_score$valMC)

      } else {
        lpl = list(ggText("MSMS: Missed cleavages per Raw file",
                          "No enzyme was specified.\nDigestion efficiency cannot be scored."))
        qc_score = data.frame(fc.raw.file = unique(df_any$fc.raw.file),
                              valMC = HEATMAP_NA_VALUE,
                              valMCVar = HEATMAP_NA_VALUE)
      }## end enzyme check

      colnames(qc_score)[colnames(qc_score) == "valMC"] = sprintf(.self$qcName, "")
      colnames(qc_score)[colnames(qc_score) == "valMCVar"] = sprintf(.self$qcName, "~Var")

      return(list(plots = lpl, qcScores = qc_score))
    },
    qcCat = "Prep",
    qcName = "MSMS:~Missed~Cleavages%s",
    orderNr = 0040
  )
    return(.self)
  })
)

#'
#' Class to read an mzTab file and store the tables internally.
#' 
#' The 'sections' field is initialized after $readMzTab was called.
#' The 'fn_map' fields should be initialized via ...$fn_map$readMappingFile(...) manually if user-defined filename mappings are desired
#' and is automatically updated/queried when $readMzTab is called.
#' 
#' @field sections MzTab sections as list. Valid list entries are: "MTD", "PRT", "PEP", "PSM", "SML", "filename" and "comments"
#' @field fn_map FilenameMapper which can translate raw filenames into something shorter
#'
#'
#'
MzTabReader = setRefClass("MzTabReader",
                       fields = list(sections = "list",
                                     fn_map = "FilenameMapper"
                       ),
                       methods = list(
                         initialize=function() {
                           .self$sections = list();
                           .self$fn_map = FilenameMapper$new();
                           
                           return(.self)
                         },
                         #'
readMzTab = function(.self, mztab_file) {
  "Read an mzTab file into a list of 5 data.frames (one df per mzTab section).
   Data.frames in the resulting list are named as follows:
     'MTD', 'PRT', 'PEP', 'PSM', 'SML',.
   Additionally, 'filename' and 'comments' are valid list elements.
  "

  cat("Reading mzTab '", mztab_file, "' ...", sep = "")
  
  ## this implementation is derived from with minor modifications
  ## https://github.com/lgatto/MSnbase/blob/master/R/MzTab.R
  
  f_con = file(mztab_file, open = "r") ## for better error messages
  lines = readLines(f_con)
  close(f_con)
  # remove empty lines
  lines = lines[ nzchar(lines) ]
  
  ## Split on the first two characters (so headers stay in
  ## the same group as table content rows)
  lineType = substring(lines, 1, 2)
  
  ## Could be stricter in the type checking to check that all
  ## three of the first characters match the 10 allowed types
  ## but since it doesn't affect parsing, I don't think it's
  ## worth bothering.
  allowed_types = c("CO", "MT", "PR", "PE", "PS", "SM")
  stopifnot(all(lineType %in% allowed_types))
  linesByType = split(lines, lineType)
  
  ## Comments are easy: just strip the first four characters
  ## from each line.
  comments = substring(linesByType[["CO"]], 5)
  
  ## Parse the other five blocks in a loop, then fix up
  ## metadata afterwards
  res = setNames(
    lapply(
      linesByType[c("MT", "PR", "PE", "PS", "SM")],
      function(x) {
        if (length(x) == 0) return(data.frame())
        ## MTD section has no header...
        if (startsWith(x[1], "MTD")) {
          d = read.delim(text = x,
                         header = FALSE,
                         col.names = c("MTD", "key", "value"),
                         na.strings = c("", "null"),
                         stringsAsFactors = FALSE,
                         fill = FALSE)
        } else {
          d = read.delim(text = x,
                         header = TRUE,
                         na.strings = c("", "null", "not mapped"),
                         stringsAsFactors = FALSE,
                         fill = FALSE)
          colnames(d) = make.names(colnames(d), allow_ = FALSE)
        }
        return(d[,-1])
      }),
    c("MTD", "PRT", "PEP", "PSM", "SML"))
  
  ## rewrite MetaData as named vector
  #res[["MTD"]] = setNames(res[["MTD"]][,2], res[["MTD"]][, 1])
  
  ## create Raw filename mapping internally
  mtd = res[["MTD"]]
  ## in addition to the correct 'ms_run[xx]-location' we also accept 'ms_run[xx]_location' (ProteomeDiscoverer2.2)
  idx_run = grep("^ms_run\\[\\d*\\][_-]location", mtd$key, value = FALSE)
  raw_filenames = mtd$value[idx_run]
  ms_runs = gsub("([.]*)[_-]location", "\\1", mtd$key[idx_run])
  .self$fn_map$getShortNames(raw_filenames, ms_runs = ms_runs)
  
  
  res[["filename"]] = mztab_file
  res[["comments"]] = comments
  .self$sections = res
  return (NULL)
},

getParameters = function()
{
  "Converts internal mzTab metadata section to a two column key-value data.frame similar to MaxQuants parameters.txt."
  
  # copy the whole MTD for now, since R likes shallow copies and we are about to rename columns by reference (see ?setnames)
  res = data.table::copy(.self$sections[["MTD"]])
  if (!is.na(unique(.self$sections$PSM$database))) {
    res = rbind(res, data.frame(key= "fasta file", value = paste(basename(unique(.self$sections$PSM$database)), collapse=";")))
  }
  else {
    res = rbind(res, data.frame(key= "fasta file", value = "NULL"))
  }
  
  res = res[grep("^custom", res$key, invert = TRUE),]
  
  res[is.na(res)] = "NULL" # temp workaround
  
  ## todo: remove at some point, since it forces us to use `::copy`
  renameColumns(res, list(key = "parameter"))

  return (res)
},

getSummary = function()
{
  "Converts internal mzTab metadata section to a two data.frame with columns 'fc.raw.file', 'ms.ms.identified....'
   similar to MaxQuants summary.txt."
  res = .self$fn_map$getRawfm()[ , c("from", "to")]
  colnames(res) = c("raw.file", "fc.raw.file")

  ## read all custom entries
  mtd_custom_df = .self$sections$MTD[grep("^custom", .self$sections$MTD$key), ]
  
  if (nrow(mtd_custom_df) == 0) return(NULL)

  ## ... and subselect the ms2-ID-Rate
  ms2_df = mtd_custom_df[grep("^\\[MS2 identification rate", mtd_custom_df$value), ] 
  res$ms.ms.identified.... = unlist(lapply(gsub(".*, ?(\\d*\\.\\d*)\\]", "\\1", ms2_df$value), as.numeric))

  ## read TIC
  tic_df = mtd_custom_df[grep("total ion current", mtd_custom_df$value),]
  if (nrow(tic_df) > 0) res$TIC = lapply(strsplit(sub(".* \\[(.*)\\]]", "\\1", tic_df$value), ","), as.numeric)

  return (res)
},

## MaxQuant-like representation of PRT table, i.e. augmented this with more columns (or renamed) if a metric requires it
getProteins = function()
{
  "Basically the PRT table ..."
  
  res = .self$sections$PRT

  return ( res )
},

## MaxQuant-like representation of PEP table, i.e. augmented this with more columns (or renamed) if a metric requires it
getEvidence = function()
{
  "Basically the PSM table and additionally columns named 'raw.file' and 'fc.raw.file'."
  
  res = data.table::as.data.table(.self$sections$PSM)

  ## remove empty PepIDs 
  ## ... unidentfied MS2 scans (NA)      or with no ConsensusFeature group (-1), i.e. unassigned PepIDs
  
  if (all(c("opt.global.cf.id") %in% colnames(res))) {
    res = res[!(is.na(res$opt.global.cf.id) | (res$opt.global.cf.id == -1)),]
    stopifnot(min(res$opt.global.cf.id) >= 0) ## would stop on NA as well
  }
  
  ## augment with fc.raw.file
  ## The 'spectra_ref' looks like 'ms_run[x]:index=y|ms_run'
  res = cbind(res, .self$fn_map$specrefToRawfile(res$spectra.ref))
  stopifnot(all(!is.na(res$fc.raw.file))) # Spectra-Ref in PSM table not set for all entries
  
  
  res$retention.time.calibration = NA
  if (all(c("opt.global.rt.align", "opt.global.rt.raw") %in% colnames(res))) 
  {
    renameColumns(res, list(               retention.time = "retention.time.pep",
                                        opt.global.rt.raw = "retention.time",
                                      opt.global.rt.align = "calibrated.retention.time"
                            ))
    res$retention.time.calibration = res$calibrated.retention.time - res$retention.time 
  }

  renameColumns(res, list(         opt.global.calibrated.mz.error.ppm = "mass.error..ppm.",
                                 opt.global.uncalibrated.mz.error.ppm = "uncalibrated.mass.error..ppm.", 
                                                   exp.mass.to.charge = "m.z", 
                                                      opt.global.mass = "mass", 
                                                opt.global.identified = "identified",
                                           opt.global.ScanEventNumber = "scan.event.number",
                                                               PSM.ID = "id", 
                                         opt.global.modified.sequence = "modified.sequence",
                        opt.global.cv.MS.1000889.peptidoform.sequence = "modified.sequence",
                                            opt.global.is.contaminant = "contaminant",
                                    opt.global.fragment.mass.error.da = "mass.deviations..da.",
                               opt.global.cv.MS.1002217.decoy.peptide = "reverse"
          ))

  if (!"modified.sequence" %in% colnames(res)){
    res$modified.sequence = res$sequence
    warning("modified.sequence is not present in input data, metrics use sequence instead", immediate. = TRUE)
  }
  
  if(!"contaminant" %in% colnames(res)){
    res$contaminant = 0
  }
  # set contaminant to TRUE/FALSE
  res$contaminant = (res$contaminant > 0)

  ## optional in MzTab (depending on which FeatureFinder was used)
  if ("opt.global.FWHM" %in% colnames(res)){
    renameColumns(res, list(opt.global.FWHM = "retention.length"))
  }
  
  ## de-duplicate protein-accession entries
  accessions = res[, .(l_accession = list(accession)), by = id]$l_accession
  res = unique(res, by = "id")
  res$proteins = unlist(lapply(accessions, paste, collapse = ";"))
  
  ## todo: these are protein names not IDs, but the code does not care (its not very clean though)
  res$protein.group.ids = res$proteins 
  
  ## annotate ms.ms.count for identical sequences per rawfile, but only the first member of the group; 
  ## all others get NA to prevent double counting
  res[, ms.ms.count := c(.N, rep(NA, .N-1)), by = list(raw.file, modified.sequence, charge)]

  ## convert values from seconds to minutes for all RT columns
  RTUnitCorrection(res)

  ##
  ## intensity from PEP to PSM: only labelfree ('opt.global.cf.id' links all PSMs belonging to a ConsensusFeature)
  ##
  df_pep = data.frame()
  if ("opt.global.cf.id" %in% colnames(res)){
    df_pep = data.table::as.data.table(.self$sections$PEP)[!is.na(sequence), ]
    renameColumns(df_pep, list(opt.global.modified.sequence = "modified.sequence", 
                               opt.global.cv.MS.1000889.peptidoform.sequence = "modified.sequence"))
    ## add raw.file...
    df_pep = cbind(df_pep, .self$fn_map$specrefToRawfile(df_pep$spectra.ref))
    ## .. a unique index
    df_pep$idx = 1:nrow(df_pep)
    ## map from PSM -> PEP row
    ## ... do NOT use spectra.ref since this is ambiguous (IDMapper duplicates MS2 PepIDs to multiple features)
    if (!("opt.global.cf.id" %in% colnames(df_pep)))
    {
      stop("Please re-run the TOPP QualityControl tool to generate an updated version of OpenMS mzTab. Your mzTab is missing some information.")
    } 
    #old: res$pep_idx = match(res$opt.global.feature.id, df_pep$opt.global.feature.id, nomatch = NA_integer_)
    res$pep_idx = match(res$opt.global.cf.id, df_pep$opt.global.cf.id, nomatch = NA_integer_)
    
    res$ms_run_number = as.numeric(gsub("^ms_run\\[(\\d*)\\].*", "\\1", res$ms_run))
    col_abd_df_pep = grepv( "^peptide.abundance.study.variable.", names(df_pep))
    col_RT_df_pep = grepv( "^opt.global.retention.time.study.variable", names(df_pep))
    ## transposed matrix for all abundances (rows = study; cols = pep_idx)
    ## .. this is a significant speedup compared to indexing into .SD[,] in subqueries, since that requires unlist()
    m_pep_abd = t(df_pep[, ..col_abd_df_pep])
    m_pep_rt = t(df_pep[, ..col_RT_df_pep])
    N.studies = length(col_RT_df_pep)
    stopifnot(N.studies == length(col_abd_df_pep))
  }
 

  NA_duplicates = function(vec_abd, idx) {
    ## replaces duplicate indices into the same ms_run intensities with NA
    ## (to avoid counting a feature more than once due to oversampled PSMs from one run assigned to a CF)
    r = vec_abd[idx]
    r[duplicated(idx)] = NA
    return(r)
  }
  
  if (all(c("opt.global.cf.id") %in% colnames(res))) {
    ## assign intensity to genuine PSMs
    res$intensity = NA_real_ ## unassigned PSMs have no MS1 intensity
    res[,
      intensity := NA_duplicates(m_pep_abd[, .SD$pep_idx[1]],
                                 .SD$ms_run_number),
      by = "opt.global.cf.id"]
    summary(res$intensity)
  }
  

  res$is.transferred = FALSE
  res$type = "MULTI-MSMS"

  #set reverse to needed values
  if ("reverse" %in% colnames(res)){
    res$reverse=(res$reverse=="decoy")
  }
  
  ## remove the data.table info, since metrics will break due to different syntax
  class(res) = "data.frame"


  ##
  ## Infer MBR 
  ## --> find all subfeatures in a CF with abundance but missing PSM --> create as MBR-dummy-PSMs
  ##
  ## : df_pep: the PEP table (with some extras)
  ## : res:    the PSM table (with some extras)
  ##     res$pep_idx: row in df_pep (consensusFeature) to which this PSM was assigned
  res_tf = res[NULL,]
  stopifnot(ncol(res_tf) > 0)
  ## skip if MS2 isotope labeling detected
  quant_methods = grep("quantification_method", .self$sections$MTD$key)
  if (!any(grepl("PRIDE_0000317", .self$sections$MTD$value[quant_methods])) & (!plyr::empty(df_pep)))
  {
    ## iterate though all consensusFeatures .. find subfeatures with intensity but missing PSM
    res_tf_tmp = df_pep[, {#print(idx)
      #idx = 44
      idx_PSM = which(res$pep_idx == idx)
      runs_with_MS2 = unique(res$ms_run_number[idx_PSM]) ## all existing PSMs for this PEP (=consensusFeature)
      runs_wo_MS2 = (1:N.studies)[-runs_with_MS2]
      df = .SD[rep(1, length(runs_wo_MS2)), c("charge", "modified.sequence", "sequence")]
      df$pep_idx = idx
      df$ms_run_number = runs_wo_MS2 ## vector
      df$calibrated.retention.time = m_pep_rt[runs_wo_MS2, idx]
      df$intensity = m_pep_abd[runs_wo_MS2, idx]
      df$protein.group.ids = res$protein.group.ids[idx_PSM[1]] ## use PGI from genuine PSMs
      ## remove all inferred PSMs which do not have an intensity(= MS1 feature)
      df = df[!is.na(df$intensity),]
      df ## return
    }, by = "idx"] ## one PEP row at a time
   
    if(nrow(res_tf_tmp) > 0){
      res_tf = res_tf_tmp
      ## convert from "ms_run_number" to fc.raw.file
      res_tf = cbind(res_tf, .self$fn_map$msrunToRawfile(paste0("ms_run[", res_tf$ms_run_number, "]")))
      res_tf$is.transferred = TRUE
      res_tf$type = "MULTI-MATCH"
      
      ## check: summed intensities should be equal
      if("intensity" %in% colnames(res) && "intensity" %in% colnames(res_tf)){
        stopifnot(sum(df_pep[, ..col_abd_df_pep], na.rm = TRUE) == sum(res$intensity, na.rm = TRUE) + sum(res_tf$intensity, na.rm = TRUE))
      }
    }
  }
  
  ## remove the data.table info, since metrics will break due to different syntax
  class(res_tf) = "data.frame"

  message("Evidence table generated: ", nrow(res), "x", ncol(res), "(genuine); ", nrow(res_tf), "x", ncol(res_tf), "(transferred)")
  
  ## must at least have column names, but can have 0 rows
  stopifnot(ncol(res_tf) > 0)
  
  return (list("genuine" = res, "transferred" = res_tf))
},


getMSMSScans = function(identified_only = FALSE)
{
  "Basically the PSM table (partially renamed columns) and additionally two columns 'raw.file' and 'fc.raw.file'. 
   If identified_only is TRUE, only MS2 scans which were identified (i.e. a PSM) are returned -- this is equivalent to msms.txt in MaxQuant."
  
  res = data.table::as.data.table(.self$sections$PSM)

  stopifnot(all((res$opt.global.identified == 1) == (!is.na(res$sequence))))
  if (identified_only) {
    res = res[!is.na(res$sequence), ] # == NA sequence
  }
  
  ## de-duplicate PSM.ID column: take first row for each PSM.ID 
  res_temp = res[!duplicated(res$PSM.ID), ]
  ## ... and append accessions of the complete subset (.SD)
  res_temp$accessions = res[, paste0(.SD$accession, collapse=";"), by = "PSM.ID"]$V1
  res = res_temp
  
  ## Augment fc.raw.file column
  ## ... the `spectra_ref` looks like "ms_run[12]:controllerType=0 controllerNumber=1 scan=25337"
  res = cbind(res, .self$fn_map$specrefToRawfile(res$spectra.ref))

  ## IDMapper might duplicate PepIDs if two or more features are a good match
  ## ... but we only want each MS2 scan represented once here
  res = unique(res, by = c("fc.raw.file", "spectra.ref"))
  
  if (all(c("opt.global.rt.align", "opt.global.rt.raw") %in% colnames(res))) 
  {
    renameColumns(res, list(     retention.time = "retention.time.pep",
                              opt.global.rt.raw = "retention.time",
                            opt.global.rt.align = "calibrated.retention.time"))
    res$retention.time.calibration = res$calibrated.retention.time - res$retention.time 
  }
  else res$retention.time.calibration = NA
  
  
  name = list(         opt.global.calibrated.mz.error.ppm = "mass.error..ppm",
                     opt.global.uncalibrated.mz.error.ppm = "uncalibrated.mass.error..ppm.", 
                                       exp.mass.to.charge = "m.z", 
                                          opt.global.mass = "mass", 
                        opt.global.fragment.mass.error.da = "mass.deviations..da.", 
                       opt.global.fragment.mass.error.ppm = "mass.deviations..ppm.",
                                    opt.global.identified = "identified",
                               opt.global.ScanEventNumber = "scan.event.number",
                                                   PSM.ID = "id", 
                             opt.global.modified.sequence = "modified.sequence",
            opt.global.cv.MS.1000889.peptidoform.sequence = "modified.sequence",
                                opt.global.is.contaminant = "contaminant",
                              opt.global.missed.cleavages = "missed.cleavages",
                   opt.global.cv.MS.1002217.decoy.peptide = "reverse",
                             opt.global.activation.method = "fragmentation",
                               opt.global.total.ion.count = "total.ion.current",
                           opt.global.base.peak.intensity = "base.peak.intensity",
                            opt.global.ion.injection.time = "ion.injection.time")
 
  renameColumns(res, name)
 
  if ("mass.deviations..ppm." %in% colnames(res)) {
    res$mass.deviations..ppm. = substr(res$mass.deviations..ppm., 2, nchar(res$mass.deviations..ppm.) - 1)
    res$mass.deviations..ppm. = gsub(",", ";", res$mass.deviations..ppm., fixed = TRUE)
  }
  if ("mass.deviations..da." %in% colnames(res)) {   
    res$mass.deviations..da. = substr(res$mass.deviations..da., 2, nchar(res$mass.deviations..da.) - 1)
    res$mass.deviations..da. =  gsub(",", ";", res$mass.deviations..da., fixed = TRUE)
  }
  
  #set reverse to needed values
  if ("reverse" %in% colnames(res)){
    res$reverse=(res$reverse=="decoy")
  }
  
  if (!"contaminant" %in% colnames(res)){
    res$contaminant = 0
  }
  #set contaminant to TRUE/FALSE
  res$contaminant = (res$contaminant > 0)
  
  #set identified to needed values
  if ("identified" %in% colnames(res)){
    stopifnot(unique(res$identified) %in% c(0,1)) ## make sure the column has the expected values (0/1)
    # set $identified to MaxQuant values (+/-)
    res$identified = c("-", "+")[(res$identified==1) + 1]
  }

  RTUnitCorrection(res)


  ## remove the data.table info, since metrics will break due to different syntax
  class(res) = "data.frame"
  
  ## order by file and specRef as RT proxy (do NOT use RT directly, since it might be NA or non-linearly transformed)
  ## e.g. spectra.ref might be 'ms_run[1]:controllerType=0 controllerNumber=1 scan=13999'
  ##                        or 'ms_run[2]:spectrum=33'
  ##      --> extract scan as numeric, since string compare is insufficient for numbers ("13999" > "140")
  res$scan = as.numeric(gsub(".*index=(\\d*)|.*scan=(\\d*)|.*spectrum=(\\d*)", "\\1\\2\\3", res$spectra.ref))
  stopifnot(all(!is.na(res$scan))) 
  res = res[order(res$fc.raw.file, res$scan), ]
  
  return ( res )
},

RTUnitCorrection = function(dt)
{
  "Convert all RT columns from seconds (OpenMS default) to minutes (MaxQuant default)"
  
  # heuristic to detect presence of unit:seconds; if retention.time has is, we assume that all rt-columns are in seconds
  # retention.time is mandatory for mzTab
  if (max(dt[, "retention.time"], na.rm = TRUE) > 300)
  {
    cn_rt = grepv("retention.time|retention.length", names(dt))
    dt[, c(cn_rt) := lapply(.SD, function(x) x / 60 ), .SDcols = cn_rt]
  }
  
  #dt[, ..cn_rt]
  return(NULL)
},

renameColumns = function(dt, namelist)
{
  "Renames all columns and throws a warning if a column does not exist in the data"
  
  from = names(namelist)
  to = unlist(namelist)
  data.table::setnames(dt, old = from, new = to, skip_absent = TRUE)
  
  existName = to %in% colnames(dt)
  if (!all(existName))
  {
    warning(paste0("Columns\n '", 
                   paste(from[!existName], "' (mzTab name) --> '", to[!existName], collapse="' (internal name),\n '", sep=""),
                   "'\n are not present in input data!"),
            immediate. = TRUE)
  }
}

) # methods
) # class

#'
#' Convert list of (mixed)modifications to a frequency table
#'
#' @param df_evd data.frame with 'fc.raw.file' and a 'modifications' column, which contains the modifications for each peptide.
#' @param name_unmod String in 'modifications' which represents an unmodified peptide
#' @param name_unmod_inverse If non-empty, then inverse the frequencies of the 'name_unmod' modifications (i.e. 100-x) IFF they are >=50\% on average (across Raw files) and rename them to this string
#' @return A data.table with 'fc.raw.file', 'modification_names' (factor), and 'Freq' (0-100)
#'
#' @export
#'
#' @examples
#' data = data.frame(fc.raw.file = rep(c("file A", "file B"),
#'                                 each = 3),
#'        modifications = c("Oxidation (M)",
#'                         "Unmodified",
#'                         "Oxidation (M),Acetyl (Protein N-term)",
#'                         "2 Oxidation (M)",
#'                         "Unmodified", "Unmodified"))
#' modsToTableByRaw(data)
#'
#'
modsToTableByRaw = function(df_evd, name_unmod = "Unmodified", name_unmod_inverse = "Modified (total)")
{
  fc.raw.file <- modification_names <- Freq <- NULL  ## make R check happy and avoid 'no visible binding for global variable'
  dt_evd = data.table::data.table(df_evd)
  mods_tbl = dt_evd[, modsToTable(.SD$modifications), by = fc.raw.file]

  mods_tbl$modification_names = as.factor(mods_tbl$modification_names) ## ensure all mods are known when subsetting the data later

  ## inverse frequencies for 'modified' if they are >=50% (to make the plot more compact)
  if (nchar(name_unmod_inverse)){
    if (mean(mods_tbl[modification_names==name_unmod, Freq]) > 50)
    {
      mods_tbl[modification_names==name_unmod, Freq := 100 - Freq]
      mods_tbl[modification_names==name_unmod, modification_names := name_unmod_inverse]
    }
    mods_tbl = droplevels(mods_tbl) ## drop 'Unmodified' level
  }

  return(as.data.frame(mods_tbl))
}

context("modstoTableByRaw.R")

test_that("modstoTableByRaw", {
  dt_in = data.frame(  fc.raw.file = rep(c("file A", "file B"), each = 3),
                     modifications = c("A", "Unmodified", "A,B", "C", "Unmodified", "Unmodified"))
  r = modsToTableByRaw(dt_in)
 
  testthat::expect_identical(colnames(r), c('fc.raw.file', 'modification_names', 'Freq'))
  
  exp = data.frame(fc.raw.file = rep(c("file A",                     "file B"), times = c(3, 2)),
                   modification_names = factor(c("A", "B", "Unmodified",    "C", "Unmodified"), levels = levels(r$modification_names)),
                   Freq = c(200/3, 100/3, 100/3,                     100/3, 200/3))
  testthat::expect_equal(r, exp)
  
  dt_in = data.frame(  fc.raw.file = rep(c("file A", "file B"), each = 3),
                       modifications = c("C", "Unmodified", "Unmodified",    "Unmodified", "A", "Unmodified"))
  r_rename = modsToTableByRaw(dt_in, name_unmod = "Unmodified", name_unmod_inverse = "MOD_TOTAL")
  exp_rename = data.frame(fc.raw.file = rep(c("file A",                     "file B"), times = c(2, 2)),
                          modification_names = factor(c("C", "MOD_TOTAL",    "A", "MOD_TOTAL"), levels = levels(r_rename$modification_names)),
                          Freq = c(100/3, 100/3,                             100/3, 100/3))
  testthat::expect_equal(r_rename, exp_rename)
  testthat::expect_equal("Unmodified" %in% levels(r_rename$modification_names), FALSE)
  
})

#'
#' Convenience wrapper for MQDataReader when only a single MQ file should be read
#' and file mapping need not be stored.
#'
#' For params, see \code{MQDataReader::readMQ()}.
#'
#' @param file   see \code{MQDataReader::readMQ()}
#' @param filter see \code{MQDataReader::readMQ()}
#' @param type   see \code{MQDataReader::readMQ()}
#' @param col_subset see \code{MQDataReader::readMQ()}
#' @param add_fs_col see \code{MQDataReader::readMQ()}
#' @param LFQ_action see \code{MQDataReader::readMQ()}
#' @param ... see \code{MQDataReader::readMQ()}
#' @return see \code{MQDataReader::readMQ()}
#'
#' @export
#'
read.MQ = function(file, filter = "", type = "pg", col_subset = NA, add_fs_col = 10, LFQ_action = FALSE, ...)
{
  mq = MQDataReader$new()
  mq$readMQ(file, filter, type, col_subset, add_fs_col, LFQ_action, ...)
}

#' Determine if a file is 'UTF-8' or 'UTF-8-BOM' (as of MQ2.4) or 'UTF-16BE' or 'UTF-16LE'
#' @param filename Relative or absolute path to a file
#' @return '' if the file does not exist or is not readable
getFileEncoding = function(filename)
{
  file_handle = try(file(filename, "rb"))

  if (inherits(file_handle, 'try-error')) return("")

  data = readBin(file_handle, "raw", n = 4)
  close(file_handle)

  if (data[1]==as.raw(0xef) & data[2]==as.raw(0xbb) & data[3]==as.raw(0xbf))
    return("UTF-8-BOM")
  if (data[1]==as.raw(0xfe) & data[2]==as.raw(0xff))
    return("UTF-16BE")  ##UTF16 big endian
  if (data[1]==as.raw(0xff) & data[2]==as.raw(0xfe))
    return("UTF-16LE")  ##UTF16 little endian
  return("UTF-8")
}


#'
#' S5-RefClass to read MaxQuant .txt files
#'
#' This class is used to read MQ data tables using \code{MQDataReader::readMQ()} while holding
#' the internal raw file --> short raw file name mapping (stored in a member called
#' 'fn_map') and updating/using it every time \code{MQDataReader::readMQ()} is called.
#'
#' Since MaxQuant changes capitalization and sometimes even column names, it seemed convenient
#' to have a function which just reads a txt file and returns unified column names, irrespective of the MQ version.
#' So, it unifies access to columns (e.g. by using lower case for ALL columns) and ensures columns are
#' identically named across MQ versions:
#' \preformatted{
#'  alternative term          new term
#'  -----------------------------------------
#'  protease                  enzyme
#'  protein.descriptions      fasta.headers
#'  potential.contaminant     contaminant
#'  mass.deviations           mass.deviations..da.
#'  basepeak.intensity        base.peak.intensity
#' }
#'
#' We also correct 'reporter.intensity.*' naming issues to MQ 1.6 convention, when 'reporter.intensity.not.corrected' is present.
#' MQ 1.5 uses: reporter.intensity.X and reporter.intensity.not.corrected.X
#' MQ 1.6 uses: reporter.intensity.X and reporter.intensity.corrected.X
#'
#' Note: you must find a regex which matches both versions, or explicitly add both terms if you are requesting only a subset
#'       of columns!
#'
#' Fixes for msmsScans.txt:
#'  negative Scan Event Numbers in msmsScans.txt are reconstructed by using other columns
#'
#' Automatically detects UTF8-BOM encoding and deals with it (since MQ2.4).
#'
#' Example of usage:
#' \preformatted{
#'   mq = MQDataReader$new()
#'   d_evd = mq$readMQ("evidence.txt", type="ev", filter="R", col_subset=c("proteins", "Retention.Length", "retention.time.calibration"))
#' }
#'
#' If the file is empty, this function shows a warning and returns NULL.
#' If the file is present but cannot be read, the program will stop.
#'
MQDataReader = setRefClass("MQDataReader",
 fields = list(mq.data = "data.frame",
               other = "list",
               fn_map = "FilenameMapper"
 ),
 methods = list(
   initialize = function() {
     .self$mq.data = data.frame();
     .self$other = list();
     .self$fn_map = FilenameMapper$new();
     return(.self)
   },

readMQ = function(file, filter = "", type = "pg", col_subset = NA, add_fs_col = 10, check_invalid_lines = TRUE, LFQ_action = FALSE, ...)
{
  "Wrapper to read a MQ txt file (e.g. proteinGroups.txt).

   Arguments:
   \\describe{
     \\item{\\code{file}}{(Relative) path to a MQ txt file.}
     \\item{\\code{filter}}{Searched for \"C\" and \"R\". If present, [c]ontaminants and [r]everse hits are removed if the respective columns are present.
       E.g. to filter both, \\code{filter = \"C+R\"}}
     \\item{\\code{type}}{Allowed values are:
       \\itemize{
         \\item \"pg\" (proteinGroups) [default], adds abundance index columns (*AbInd*, replacing 'intensity')
         \\item \"sm\" (summary), splits into three row subsets (raw.file, condition, total)
         \\item \"ev\" (evidence), will fix empty modified.sequence cells for older MQ versions (when MBR is active)
         \\item \"msms_scans\", will fix invalid (negative) scan event numbers
       }
       Any other value will not add/modify any columns.}
     \\item{\\code{col_subset}}{A vector of column names as read by read.delim(), e.g., spaces are replaced by dot already.
       If given, only columns with these names (ignoring lower/uppercase) will be returned (regex allowed).
       E.g. \\code{col_subset = c(\"^lfq.intensity.\", \"protein.name\")}}
     \\item{\\code{add_fs_col}}{If TRUE and a column 'raw.file' is present, an additional column 'fc.raw.file' will be added with
       common prefix AND common substrings removed (\\code{\\link{simplifyNames}}).
       E.g. two rawfiles named 'OrbiXL_2014_Hek293_Control', 'OrbiXL_2014_Hek293_Treated' will give 'Control', 'Treated'.
       If \\code{add_fs_col} is a number AND the longest short-name is still longer, the names are discarded and replaced by
       a running ID of the form 'file <x>', where <x> is a number from 1 to N.
       If the function is called again and a mapping already exists, this mapping is used.
       Should some raw.files be unknown (ie the mapping from the previous file is incomplete), they will be augmented.}
     \\item{\\code{check_invalid_lines}}{After reading the data, check for unusual number of NA's to detect if file was corrupted by Excel or alike.}
     \\item{\\code{LFQ_action}}{[For type=='pg' only] An additional custom LFQ column ('cLFQ...') is created where
       zero values in LFQ columns are replaced by the following method IFF(!) the corresponding raw intensity is >0 (indicating that LFQ is erroneusly 0):
       \\itemize{
         \\item \"toNA\": replace by NA
         \\item \"impute\": replace by lowest LFQ value >0 (simulating 'noise')
       }}
     \\item{\\code{...}}{Additional parameters passed on to read.delim().}
   }

   Returns a data.frame of the respective file.
  "

  if (!file.exists(file)) {
    cat(paste0("MaxQuant file ", file, " was not found. Reading skipped.\n"))
    return (NULL);
  }
  cat(paste("Reading file", file,"...\n"))
  ## error message if failure should occur below
  msg_parse_error = paste0("\n\nParsing the file '", file, "' failed. See message above why. If the file is not usable but other files are ok, disable the corresponding section in the YAML config. You might also be running a foreign locale (e.g. Chinese) - switch to an English locale and make sure that txt files are encoded in ASCII (Latin-1)!")

  ## get encoding, to pass on to read.delim
  file_encoding = getFileEncoding(file)


  ## resolve set of columns which we want to keep
  #example: col_subset = c("Multi.*", "^Peaks$")
  colClasses = NA ## read.table: keep all columns by default
  if (sum(!is.na(col_subset)) > 0)
  { ## just read a tiny bit to get column names
    ## do not use data.table::fread for this, since it will read the WHOLE file and takes ages...
    data_header = try(read.delim(file, comment.char="", nrows=2, fileEncoding = file_encoding))
    if (inherits(data_header, 'try-error')) stop(msg_parse_error, call. = FALSE);

    colnames(data_header) = tolower(colnames(data_header))
    colClasses = rep("NULL", ncol(data_header)) ## either "NULL" (discard), NA (auto-detect), or specific, i.e. "numeric", ...
    for (idx in 1:length(col_subset))
    {
      idx_new = grepl(col_subset[idx], colnames(data_header), ignore.case = TRUE)
      if (sum(idx_new) == 0)
      {
        cat(paste0("WARNING: Could not find column regex '", col_subset[idx], "' using case-INsensitive matching.\n"))
      } else {
        colClasses[idx_new] = NA  ## auto
        if (nchar(names(col_subset)[idx]) > 0) {
          colClasses[idx_new] = names(col_subset)[idx]
          cat(paste0("Requiring column(s) '", paste(colnames(data_header)[idx_new], sep="", collapse="', '"), "' to be of type '", names(col_subset)[idx], "'!\n"))
        }
      }

    }
    ## keep the 'id' column if available (for checking data integrity: invalid line-breaks)
    if ("id" %in% colnames(data_header) & check_invalid_lines == TRUE) colClasses[which("id"==colnames(data_header))] = NA
    cat(paste("Keeping", sum(colClasses != "NULL", na.rm=TRUE), "of", ncol(data_header), "columns!\n"))
    if (sum(colClasses != "NULL", na.rm=TRUE) == 0) {
      ## can happen for very old MQ files without header, or if the user just gave the wrong colClasses
      .self$mq.data = data.frame()
      return (.self$mq.data)
    }
  }

  ## higher memory consumption during load (due to memory mapped files) compared to read.delim... but about 5x faster
  ## , but also different numerical results when parsing numbers!!!
  #.self$mq.data = try(
  #  fread(file, header = TRUE, sep='\t', na.strings=c("NA", "n. def."), verbose = TRUE, select = idx_keep, data.table = FALSE, ...)
  #)
  #colnames(.self$mq.data) = make.names(colnames(.self$mq.data), unique = TRUE)

  ## comment.char should be "", since lines will be TRUNCATED starting at the comment char.. and a protein identifier might contain just anything...
  ## na.strings:
  ##  - use BOTH 'n. def.' and 'n.def.' even though only the former is present in the data
  ##    However, when the colClass is 'numeric', whitespaces are stripped, and only AFTERWARDS the string
  ##    is checked against na.strings
  ##  - the '\u975E\u6570\u5B57' na-string is the chinese UTF-8 representation of "NA"
  .self$mq.data = try(read.delim(file, na.strings=c("NA", "n. def.", "n.def.", "\u975E\u6570\u5B57"),
                                 encoding="UTF-8", comment.char="", stringsAsFactors = FALSE, colClasses = colClasses,
                                 fileEncoding = file_encoding, ...))
  if (inherits(.self$mq.data, 'try-error')) stop(msg_parse_error, call. = FALSE);

  #colnames(.self$mq.data)

  cat(paste0("Read ", nrow(.self$mq.data), " entries from ", file,".\n"))

  ### checking for invalid rows
  if (check_invalid_lines == TRUE & type != "sm") ## summary.txt has irregular structure
  {
    inv_lines = .self$getInvalidLines();
    if (length(inv_lines) > 0)
    {
      stop(paste0("\n\nError: file '", file, "' seems to have been edited in Microsoft Excel and",
                                             " has artificial line-breaks which destroy the data at lines (roughly):\n",
                                             paste(inv_lines, collapse="\n"), "\nPlease fix (e.g. try LibreOffice 4.0.x or above)!",
                                             "You might also have unknown UTF-8 characters (Asian?) in your file which PTXQC does not recognize. Try running MaxQuant on an english locale!"))
    }
  }

  cat(paste0("Updating colnames\n"))
  cn = colnames(.self$mq.data)
  ### just make everything lower.case (MQ versions keep changing it and we want it to be reproducible)
  cn = tolower(cn)
  ## rename some columns since MQ 1.2 vs. 1.3 differ....
  cn[cn=="protease"] = "enzyme"
  cn[cn=="protein.descriptions"] = "fasta.headers"
  cn[cn=="basepeak.intensity"] = "base.peak.intensity"
  ## MQ 1.0.13 and 1.1.x have 'mass.deviations', later versions have 'mass.deviations..da.'
  cn[cn=="mass.deviations"] = "mass.deviations..da."
  ## MQ 1.5 uses 'potential.contaminant' instead of 'contaminant'
  cn[cn=="potential.contaminant"] = "contaminant"
  ## MQ 1.5 uses     : reporter.intensity.X and reporter.intensity.not.corrected.X
  ## whereas 1.6 has : reporter.intensity.X and reporter.intensity.corrected.X
  ## we rename to 1.6 convention:
  if (sum(grepl("reporter.intensity.not.corrected", cn)) > 0)
  {
    idx_nc = grep("reporter.intensity.not.corrected", cn);
    idx_c = grep("reporter.intensity.[0-9]", cn);
    cn[idx_nc] = gsub(".not.corrected", "", cn[idx_nc]);
    cn[idx_c] = gsub(".intensity", ".intensity.corrected", cn[idx_c]);
  }

  colnames(.self$mq.data) = cn



  ## work in-place on 'contaminant' column
  cat(paste0("Simplifying contaminants\n"))
  .self$substitute("contaminant");
  cat(paste0("Simplifying reverse\n"))
  .self$substitute("reverse");
  if (grepl("C", filter) & ("contaminant" %in% colnames(.self$mq.data))) .self$mq.data = .self$mq.data[!(.self$mq.data$contaminant),]
  if (grepl("R", filter) & ("reverse" %in% colnames(.self$mq.data))) .self$mq.data = .self$mq.data[!(.self$mq.data$reverse),]

  ## proteingroups.txt special treatment
  if (type=="pg") {
    stats = data.frame(n=NA, v=NA)
    if (LFQ_action!=FALSE)
    { ## replace erroneous zero LFQ values with something else
      cat("Starting LFQ action.\nReplacing ...\n")
      lfq_cols = grepv("^lfq", colnames(.self$mq.data))
      for (cc in lfq_cols)
      {
        ## get corresponding raw intensity column
        rawint_col = sub("^lfq\\.", "", cc)
        if (!(rawint_col %in% colnames(.self$mq.data))) {stop(paste0("Could not find column '", rawint_col, "' in dataframe with columns: ", paste(colnames(.self$mq.data), collapse=",")), "\n")}
        vals = .self$mq.data[, cc]
        ## affected rows
        bad_rows = (.self$mq.data[, rawint_col]>0 & .self$mq.data[, cc]==0)
        if (sum(bad_rows, na.rm = TRUE)==0) {next;}
        ## take action
        if (LFQ_action=="toNA" | LFQ_action=="impute") {
          ## set to NA
          impVal = NA;
          vals[bad_rows] = impVal;
          if (LFQ_action=="impute") {
            impVal = min(vals[vals>0], na.rm = TRUE)
            ## replace with minimum noise value (>0!)
            vals[bad_rows] = impVal;
          }
          cat(paste0("   '", cc, "' ", sum(bad_rows, na.rm = TRUE), ' entries (', sum(bad_rows, na.rm = TRUE)/nrow(.self$mq.data)*100,'%) with ', impVal, '\n'))
          ## add column
          .self$mq.data[, paste0("c", cc)] = vals;
          ##
          stats = rbind(stats, c(sum(bad_rows, na.rm = TRUE), impVal))

        }
        else {
          stop(paste0("Unknown action '", LFQ_action, "' for LFQ_action parameter! Aborting!"));
        }
      }
      if (LFQ_action=="impute") {
        plot(stats, log="y", xlab="number of replaced zeros", ylab='imputed intensity', main='Imputation statistics')
      }
    }

    int_cols = grepv("intensity", colnames(.self$mq.data))

    ##
    ## apply potential fix (MQ 1.5 writes numbers in scientific notation with ',' -- which parses as String :( )
    ##
    int_cols_nn = (apply(.self$mq.data[,int_cols, drop = FALSE], 2, class) != "numeric")
    if (any(int_cols_nn))
    {
      .self$mq.data[, int_cols[int_cols_nn]] = sapply(int_cols[int_cols_nn], function(x_name)
      {
        x = .self$mq.data[, x_name]
        if (inherits(x, "integer"))
        {
          x = as.numeric(x)
        }
        if (!inherits(x, "numeric"))
        {
          cat(paste(x[grep(",", x)], collapse = ",", sep=", "))
          cat(paste(unique(x), collapse = ",", sep=", "))
          warning(paste0("Column '", x_name, "' in file '", file, "' is expected to contain numbers, but was classified as '",
                         class(x), "'.",
                         " A bug in MaxQuant 1.5.2.8 (and related?) versions causes usage of comma instead of dot for large numbers in scientific notation,",
                         " e.g. '1,73E+011' instead of '1.73E+011'. We will try to fix this now ..."),
                  call. = FALSE, immediate. = TRUE)
          x = as.numeric(gsub(",",".", x))
        }
        return(x)
      })
    }

    ##
    ## add Abundance index
    ##
    if ("mol..weight..kda." %in% colnames(.self$mq.data)){
      ### add abundance index columns (for both, intensity and lfq.intensity)
      .self$mq.data[, sub("intensity", "AbInd", int_cols)] = apply(.self$mq.data[,int_cols, drop = FALSE], 2, function(x)
      {
        return (x / .self$mq.data[,"mol..weight..kda."])
      })
    } else {
      stop("MQDataReader::readMQ(): Cannot add abundance index since 'mol..weight..kda.' was not loaded from file. Did you use the correct 'type' or forgot to add the column in 'col_subset'?")
    }

  } else if (type=="sm") {
    ## for MQ-DIA, summary.txt contains only a 'total' column for some reason (at least using MQ 2.4.2.0)
    ## --> this will mess with raw.file mapping, thus abort at this point
    if (nrow(.self$mq.data) == 1 && .self$mq.data$raw.file[1] == "Total") {
        .self$mq.data = data.frame()  ## make the data empty (cannot assign NULL, due to member restriction via RefClass)
        return(NULL)
    }


    ## summary.txt special treatment
    ## find the first row, which lists Groups (after Raw files): it has two non-zero entries only
    ##                                                           (or even less if the group name is empty)
    ##dx <<- .self$mq.data;
    idx_group = which(apply(.self$mq.data, 1, function(x) sum(x!="", na.rm = TRUE))<=2)[1]
    ## summary.txt will not contain groups, if none where specified during MQ-configuration
    ## index of last row (Total)
    idx_Total = which(.self$mq.data[, 1] == "Total")
    idx_beforeTotal = idx_Total - 1
    if (is.na(idx_group)) {
      ## now its either DIA data with no raw files or DDA data with ONLY raw files (no groups)
      raw.files = .self$mq.data[1:idx_beforeTotal, ]
      groups= NA
    } else {  ## with raw files AND groups
      raw.files = .self$mq.data[1:(idx_group-1), ]
      groups = .self$mq.data[idx_group:idx_beforeTotal, ]
    }
    total = .self$mq.data
    .self$mq.data = raw.files ## temporary, until we have assigned the fc.raw.files
  } else if (type == "ev") {
    ## check if data is missing
    if (all(c("type", "modified.sequence") %in% colnames(.self$mq.data)) &
        any("MULTI-MATCH" %in% .self$mq.data$type) &
        all(.self$mq.data$modified.sequence[.self$mq.data$type=="MULTI-MATCH"]==""))
    {
      warning(immediate. = TRUE, "readMQ(): Input data has empty cells for column 'modified.sequence' of type 'MULTI-MATCH'. Early MaxQuant versions (e.g. 1.2.2) have this problem. We will try to reconstruct the data.")
      ## use the preceeding sequence (and hope that there are no missing rows in between)
      .self$mq.data = .self$mq.data[order(.self$mq.data$id), ]
      ## find blocks of MATCHed rows ...
      idx_mm = which(.self$mq.data$type=="MULTI-MATCH") ## row index
      head(idx_mm)
      idx_block_start = idx_mm[ c(1, which(diff(idx_mm)>1) + 1) ] ## index to block of MATCHES
      head(idx_block_start)
      idx_block_end = c(idx_mm[match(idx_block_start, idx_mm)[-1]-1], idx_mm[length(idx_mm)])
      head(idx_block_end)
      .self$mq.data$modified.sequence[idx_mm] = rep(.self$mq.data$modified.sequence[idx_block_start-1],
                                                    idx_block_end-idx_block_start+1)
    }
  } else if (type == "msms_scans") {

    ## fix scan.event.number (some MQ 1.6.17.0 results have negative values...)
    if (min(.self$mq.data$scan.event.number, na.rm = TRUE) < 1)
    { ## fix by manually computing it from 'Scan index' and 'MS scan index' (the precursor MS1)
      warning("Found MaxQuant bug in msmsScans.txt (Scan Event Numbers are negative)")
      req_cols = c("raw.file", "ms.scan.index", "scan.index")
      if (!checkInput(req_cols, .self$mq.data)) stop("Could not find all of '", paste0(req_cols, sep="', '"), "' in msmsScans.txt load() request. Please request loading these columns in order to fix the scan.event.number values.")
      dtemp = as.data.table(.self$mq.data)
      ## sort by precursor index + MS2 index, such that subsetting later already has the right order
      setorder(dtemp, raw.file, ms.scan.index, scan.index)
      dtemp[, scan.event.number := 1:.N, by = .(raw.file, ms.scan.index)]
      .self$mq.data = as.data.frame(dtemp)
    }
  }



  if (add_fs_col & "raw.file" %in% colnames(.self$mq.data))
  {
    .self$mq.data$fc.raw.file = .self$fn_map$getShortNames(.self$mq.data$raw.file, add_fs_col)
  }

  if (type=="sm") { ## post processing for summary
    ## .self$mq.data is basically "raw.files#, but with fc.raw.files augmented
    .self$other = list(groups = groups, total = total)
  }

  return (.self$mq.data);
}, ## end readMQ()



substitute = function(colname, valid_entries = c(NA, "", "+"), replacements = c(FALSE, FALSE, TRUE))
{
  "Replaces values in the mq.data member with (binary) values.

   Most MQ tables contain columns like 'contaminants' or 'reverse', whose values are either empty strings
   or \"+\", which is inconvenient and can be much better represented as TRUE/FALSE.
   The params \\code{valid_entries} and \\code{replacements} contain the matched pairs, which determine what is replaced with what.

   Arguments:
   \\describe{
     \\item{\\code{colname}}{Name of the column (e.g. 'contaminants') in the mq.data table.}
     \\item{\\code{valid_entries}}{Vector of values to be replaced (must contain all values expected in the column -- fails otherwise).}
     \\item{\\code{replacements}}{Vector of values inserted with the same length as \\code{valid_entries}.}
   }

   Returns \\code{TRUE} if successful.
  "

  if (length(valid_entries) == 0)
  {
    stop("Entries given to $substitute() must not be empty.")
  }
  if (length(valid_entries) != length(replacements))
  {
    stop("In function $substitute(): 'valid_entries' and 'replacements' to not have the same length!")
  }
  if (colname %in% colnames(.self$mq.data))
  {
    ## verify that there are only known entries (usually c("","+") )
    setD_c = setdiff(.self$mq.data[, colname], valid_entries)
    if (length(setD_c) > 0) stop(paste0("'", colname, "' column contains unknown entry (", paste(setD_c, collapse=",", sep="") ,")."))
    ## replace with TRUE/FALSE
    .self$mq.data[, colname] = replacements[ match(.self$mq.data[, colname], valid_entries) ];
  }
  return (TRUE)
},


getInvalidLines = function()
{
  "Detect broken lines (e.g. due to Excel import+export)

   When editing a MQ txt file in Microsoft Excel, saving the file can cause it to be corrupted,
   since Excel has a single cell content limit of 32k characters
   (see http://office.microsoft.com/en-001/excel-help/excel-specifications-and-limits-HP010342495.aspx)
   while MQ can easily reach 60k (e.g. in oxidation sites column).
   Thus, affected cells will trigger a line break, effectively splitting one line into two (or more).

   If the table has an 'id' column, we can simply check the numbers are consecutive. If no 'id' column is available,
   we detect line-breaks by counting the number of NA's per row and finding outliers.
   The line break then must be in this line (plus the preceeding or following one). Depending on where
   the break happened we can also detect both lines right away (if both have more NA's than expected).

   Currently, we have no good strategy to fix the problem since columns are not aligned any longer, which
   leads to columns not having the class (e.g. numeric) they should have.
   (thus one would need to un-do the linebreak and read the whole file again)

   [Solution to the problem: try LibreOffice 4.0.x or above -- seems not to have this limitation]

   @return Returns a vector of indices of broken (i.e. invalid) lines
  "

  if (!inherits(.self$mq.data, 'data.frame'))
  {
    stop("In 'MQDataReader$getInvalidLines': function called before data was loaded. Internal error. Exiting.", call. = FALSE);
  }
  broken_rows = c()
  if ("id" %in% colnames(.self$mq.data))
  {
    last_id = as.numeric(as.character(.self$mq.data$id[nrow(.self$mq.data)]))
    if (is.na(last_id) || (last_id+1)!=nrow(.self$mq.data))
    {
      print(paste0("While checking ID column: last ID was '", last_id, "', while table has '", nrow(.self$mq.data), "' rows."))
      broken_rows = which(!is.numeric(as.character(.self$mq.data$id)))
    }
  } else
  {
    cols = !grepl("ratio", colnames(.self$mq.data)) ## exclude ratio columns, since these can have regular NA's in unpredictable frequency
    counts = apply(.self$mq.data[, cols], 1, function(x) sum(is.na(x)));
    ## NA counts should be roughly equal across rows
    expected_count = quantile(counts, probs = 0.75)
    broken_rows = which(counts > (expected_count * 3 + 10))
    if (length(broken_rows) > 0)
    {
      print("Table:")
      print(table(counts))
      print(paste0("NAn count limit: 3*", expected_count, " + 10 = ", expected_count * 3 + 10))
    }
  }

  return (broken_rows);
}
) ## methods
) ## RefClass

qcMetric_PAR =  setRefClass(
  "qcMetric_PAR",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpText="MaxQuant parameters, extracted from parameters.txt (abbreviated as 'PAR'), summarizes the settings used for the MaxQuant analysis. 
Key parameters are MaxQuant version, Re-quantify, Match-between-runs and mass search tolerances. 
A list of protein database files is also provided, allowing to 
track database completeness and database version information (if given in the filename).", 
    workerFcn=function(.self, d_parAll)
    {
      ##todo: read in mqpar.xml to get group information and ppm tolerances for all groups (parameters.txt just gives Group1)

      line_break = "\n"; ## use space to make it work with table
      ## remove AIF stuff
      
      df_mqpar = d_parAll[!grepl("^AIF ", d_parAll$parameter),]
      df_mqpar$value = gsub(";", line_break, df_mqpar$value)
      ## seperate FASTA files (usually they destroy the layout)
      idx_fastafile = grepl("fasta file", df_mqpar$parameter, ignore.case = TRUE)
      d_par_file = df_mqpar[idx_fastafile, ]
      fasta_files = sapply(unlist(strsplit(d_par_file$value, "\n")), function(x) rev(strsplit(x,"\\", fixed = TRUE)[[1]])[1])
      d_par = df_mqpar[!idx_fastafile, ]
      ## remove duplicates
      d_par = d_par[!duplicated(d_par$parameter),]
      rownames(d_par) = d_par$parameter
      
      ## trim long param names (the user should know what they mean)
      allowed_len = nchar("Min. score for unmodified .."); 
      d_par$parameter = sapply(d_par$parameter, function (s) {
        if (nchar(s) > allowed_len) {
          s = paste(substring(s, 1, allowed_len), "..", collapse = "", sep="")
        }
        return (s)
      })
      
      
      ## sort by name
      d_par = d_par[order(d_par$parameter), ]

      ## two column layout
      if (nrow(d_par) %% 2 != 0) d_par[nrow(d_par)+1,] = ""  ## make even number of rows
      mid = nrow(d_par) / 2
      d_par$page = 1
      d_par$page[1:mid] = 0
      
      parC = c("parameter", "value")
     
      d_par2 = cbind(d_par[d_par$page==0, parC], d_par[d_par$page==1, parC])
      
      ## HTML: alternative table
      ## (do this before line breaks, since Html can handle larger strings)
      tbl_f = getHTMLTable(d_par2, caption = fasta_files)

      ## break long values into multiple lines (to preserve PDF table width)
      splitMaxLen = function(s) {
        allowed_len = nchar("Use least modified peptide"); ## this is a typical entry -- everything which is longer gets split
        r = paste(sapply(unlist(strsplit(s, line_break, fixed = TRUE)), function(s1) {
          if (!is.na(s1) && nchar(s1) > allowed_len) {
            s_beg = seq(from = 1, to = nchar(s1) - 1, by = allowed_len)
            s1 = paste(unlist(substring(s1, s_beg, s_beg + allowed_len - 1)), collapse = line_break)
          }
          return(s1)
        }), collapse = line_break)
        return (r)
      }
      d_par2[ , 2] = sapply(d_par2[ , 2], splitMaxLen)
      d_par2[ , 4] = sapply(d_par2[ , 4], splitMaxLen)
      
      plot_title = "PAR: parameters"
      ## PDF: split table onto multiple pages if necessary...
      par_pl = byXflex(d_par2, 1:nrow(d_par2), 25, plotTable, sort_indices = TRUE, title = plot_title, footer = fasta_files)
      
      return(list(plots = par_pl, htmlTable = tbl_f))
    }, 
    qcCat = NA_character_, 
    qcName = "PAR:~MQ~Parameters", 
    orderNr = 0001
  )
    return(.self)
  })
)

###
### Author: Chris Bielow
###
###

#' Boxplots - one for each condition (=column) in a data frame.
#'
#' Given a data.frame with two/three columns in long format (name, value, [contaminant]; in that order), each group (given from 1st column)
#' is plotted as a bar.
#' Contaminants (if given) are separated and plotted as yellow bars.
#'
#' Boxes are shaded: many NA or Inf lead to more transparency. Allows to easily spot sparse groups
#'
#'
#' @param data    Data frame in long format with numerical expression data
#' @param log2    Apply log2 to the data (yes/no)
#' @param ylab    Label on Y-axis
#' @param mainlab Main title
#' @param sublab  Sub title
#' @param boxes_per_page  Maximum number of boxplots per plot. Yields multiple plots if more groups are given.
#' @param abline          Draw a horizontal green line at the specified y-position (e.g. to indicate target median values)
#' @param coord_flip      Exchange Y and X-axis for better readability
#' @param names           An optional data.frame(long=.., short=..), giving a renaming scheme (long->short) for the 'name' column
#' @return List of ggplot objects
#'
#' @import ggplot2
#' @export
#'
boxplotCompare = function(data,
                          log2 = TRUE,
                          ylab = "intensity",
                          mainlab = ylab,
                          sublab = "",
                          boxes_per_page = 30,
                          abline = NA,
                          coord_flip = TRUE,
                          names = NA)
{

  if (ncol(data) == 2) {
    data$contaminant = FALSE ## add a third column, if missing
  }
  colnames(data) = c("group", "value", "contaminant")

  if (log2) {
    data$value = log2(data$value)
  }

  ## shorten group names
  if (inherits(names, 'data.frame'))
  {
    stopifnot(sort(colnames(names)) == c("long", "short"))
    data$group = names$short[match(data$group, names$long)]
    if (any(is.na(data$group)))
    {
      print(names)
      stop("Group renaming is incomplete! Aborting...")
    }
  }
  ## make it a factor
  if (!inherits(data$group, "factor")) data$group = factor(data$group)

  ## actual number of entries in each column (e.g. LFQ often has 0)
  ncol.stat = plyr::ddply(data, "group", function(x)
    {
      notNA = sum(!is.infinite(x$value) & !is.na(x$value));
      data.frame(n = nrow(x), notNA = notNA, newname = paste0(x$group[1], " (n=", notNA, ")"))
    })
  head(ncol.stat)
  ## rename (augment with '#n')
  data$group2 = ncol.stat$newname[match(data$group, ncol.stat$group)]

  ## check (ddply makes  ncol.stat$newname a factor with matching levels)
  #stopifnot(inherits(data$group2, "factor"))
  #stopifnot(all(as.numeric(data$group) == as.numeric(data$group2))) ## as.numeric() is meaningless on factors with different levels
  data$group = data$group2
  ## remove -inf and NA's
  data = data[!is.infinite(data$value) & !is.na(data$value), ]

  if (nrow(data)== 0){
    return(list(ggText(mainlab,
                   "No intensity found",
                   "red")))
  }
  groups = unique(data$group);
  ## add color for H vs L (if SILAC)
  cols = c("sample" = "black",
           "sample (light)" = "black",
           "sample (medium)" = "blue",
           "sample (heavy)" = "green",
           "contaminant" = "yellow")
  dark_cols = darken(cols)
  names(dark_cols) = names(cols)
  ## assign categories to channels
  cat_names = names(cols)
  cat = factor(cat_names, levels=cat_names)
  data$cat = cat[1]
  if (sum(grepl("^[^HLM]", groups )) == 0 || sum(grepl("^intensity\\.[hlm]\\.", groups )) > 0)
  { ## all start with either L, M or H
    data$cat = cat[2]
    data$cat[grepl("^M", data$group) | grepl("^intensity\\.m\\.", data$group)] = cat[3]
    data$cat[grepl("^H", data$group) | grepl("^intensity\\.h\\.", data$group)] = cat[4]
  }
  data$cat[data$contaminant] = cat[5]

  levels(data$cat)
  ## get rid of unused levels
  data$cat = data$cat[, drop = TRUE]

  ## subset the factor as well (otherwise ggplot will show unused factors in the legend)
  cols_sub = cols[names(cols) %in% levels(data$cat)]
  dark_cols_sub = dark_cols[names(dark_cols) %in% levels(data$cat)]

  ## compute global y-limits (so we can fix it across plots)
  ylims = grDevices::boxplot.stats(data$value)$stats[c(1, 5)]
  ## make sure to include abline (if existing)
  if (!is.na(abline))
  {
    ylims = c(ylims[1], max(ylims[2], abline))
  }

  fcn_boxplot_internal = function(data, abline = NA)
  {
    pl = ggplot(data=data, aes(x = .data$group, y = .data$value, fill = .data$cat, col = .data$cat)) +
      geom_boxplot(varwidth = TRUE) +
      xlab("") +
      ylab(ylab)
    if (coord_flip) {
      pl = pl + coord_flip(ylim = ylims)
    } else {
      pl = pl + coord_cartesian(ylim = ylims)
    }
    pl = pl +
      scale_alpha(guide = "none") +
      scale_fill_manual(values = cols_sub, name = "Category", drop = TRUE) +
      scale_color_manual(values = dark_cols_sub, name = "Category", drop = TRUE) +
      theme(axis.text.x = element_text(angle = 90, vjust = 0.5)) +
      theme(legend.position = ifelse(length(cols_sub)==1, "none", "right")) +
      ggtitle(mainlab, sublab) +
      scale_x_discrete_reverse(factor(unique(data$group)))

    if (!is.na(abline))
    {
      pl = pl + geom_abline(alpha = 0.5, intercept = abline, slope = 0, colour = "green")
    }
    #print(pl)
    return(pl)
  }
  lpl = byXflex(data = data, indices = data$group, subset_size = boxes_per_page, sort_indices = TRUE, FUN = fcn_boxplot_internal, abline)
  return (lpl)
}


#'
#' Extract fragment mass deviation errors from a data.frame from msms.txt
#'
#' Given a data.frame as obtainable from a msms.txt with
#'  - a 'mass.analyzer' column which contains only a single value for the whole column
#'  - a 'mass.deviations..da.' and (if available) 'mass.deviations..ppm.'
#'  - a 'masses' column (only required if 'mass.deviations..ppm.' is unavailable and the mass.analyzer
#'    indicates hig-res data)
#'
#' Mass deviations are extracted from the columns, e.g. each cell containing values separated by
#' semicolons is split into single values. The appropriate unit is chosen (Da or ppm, depending on
#' ITMS or FTMS data). Also the fragmentation type can be used: CID indicates ITMS, HCD to FTMS.
#' This is not 100% safe, but older MQ versions do not report the mass analyzer properly.
#'
#' Sometimes, peptides are identified purely based on MS1, i.e. have no fragments. These will be ignored.
#'
#' If ppm mass deviations are not available, errors in Da will be converted to ppm using the corresponding mass values.
#'
#' @param x Data frame in long format with numerical expression data
#' @param recurse Internal usage only. Leave at 0 when calling.
#' @return  Data frame with mass errors ('msErr') and their 'unit' (Da or ppm) or NULL (if no fragments were given)
#'
#' @export
#'
getFragmentErrors = function(x, recurse = 0)
{
  ## require only one mass analyzer type:
  stopifnot(length(unique(x$mass.analyzer))==1)
  stopifnot(all(c("mass.analyzer", "mass.deviations..da.") %in% colnames(x)))

  convert_Da2PPM = FALSE
  ## note: the "^" is important to avoid matching to " ITMS" -- see fallback below
  if (grepl("^ITMS|^TOF|^CID", x$mass.analyzer[1]) & ("mass.deviations..da." %in% colnames(x)))
  {
    ms2_unit = "[Da]"; ms2_col = "mass.deviations..da."
  } else if (grepl("^FTMS|^HCD|^HCID", x$mass.analyzer[1]) & ("mass.deviations..ppm." %in% colnames(x))) {
    ms2_unit = "[ppm]"; ms2_col = "mass.deviations..ppm."
  } else if (grepl("^FTMS|^HCD|^HCID", x$mass.analyzer[1]) & ("mass.deviations..da." %in% colnames(x))) {
    ## we know its high resolution, but this MQ version only gave us Dalton mass deviations
    ## --> convert back to ppm
    ms2_unit = "[ppm]"; ms2_col = "mass.deviations..da."
    convert_Da2PPM = TRUE
  } else {
    ## fallback if mass.analyzer is 'Unknown' (e.g. for mzXML input)
    if ("mass.deviations..ppm." %in% colnames(x)) {
      ms2_unit = "[ppm]"; ms2_col = "mass.deviations..ppm."
    } else {
      ms2_unit = "[Da]"; ms2_col = "mass.deviations..da."
    }
  }

  ## sometimes, peptides are identified purely based on MS1, i.e. have no fragments
  ## if only those peptides are present here in 'x', then 'err' below will be empty
  ## and no data.frame can be constructed, so...
  if (all(is.na(x[, ms2_col])) || all(nchar(x[, ms2_col]) == 0)) return(NULL)

  err = as.numeric(unlist(strsplit(x[, ms2_col], split=";", fixed = TRUE)))

  ## sometimes, ALL entries are NaN...
  if (all(is.na(err)))  return(NULL)

  if (convert_Da2PPM) {
    stopifnot("masses" %in% colnames(x))
    mass = unlist(strsplit(x$masses, split=";", fixed = TRUE))
    err = err / as.numeric(mass) * 1e6
  }

  abs_error95 = quantile(abs(err), probs = 0.95)
  ## if dimension (Da vs ppm) seem weird, try switching to the other -- but avoid infinite recursion
  if (recurse == 0 & (ms2_unit == "[ppm]") & (abs_error95 > 200)) {
    message(paste0("MS/MS fragment error seems rather large ", abs_error95, ". Reporting in [Da]...\n"))
    x$mass.analyzer = "ITMS"
    return (getFragmentErrors(x, recurse = 1))
  } else if (recurse == 0 & (ms2_unit == "[Da]") & (abs_error95 < 0.2)) {
    message(paste0("MS/MS fragment error seems rather small ", abs_error95, ". Reporting in [ppm]...\n"))
    x$mass.analyzer = paste0(" ", x$mass.analyzer); ## just something which the regex above does not recognize and fallback to ppm
    return (getFragmentErrors(x, recurse = 1))
  }
  return(data.frame(msErr = err, unit = ms2_unit))
}

#'
#' Detect (and fix) MaxQuant mass recalibration columns, since they
#' sometimes report wrong values.
#'
#' Returns a list of items for both diagnostics and possibly a fixed evidence data.frame.
#' Also two strings with messages are returned, which can serve as user message for
#' pre and post calibration status.
#'
#' @param df_evd Evidence data.frame with columns ()
#' @param df_idrate Data.frame from summary.txt, giving ID rates for each raw file (cols: "ms.ms.identified....", "fc.raw.file"). Can also be NULL.
#' @param tolerance_sd_PCoutOfCal Maximal standard deviation allowed before considered 'failed'
#' @param low_id_rate Minimum ID rate in Percent before a Raw file is considered 'failed'
#' @return list of data (stats, affected_raw_files, df_evd, recal_message, recal_message_post)
#'
#'
fixCalibration = function(df_evd, df_idrate = NULL, tolerance_sd_PCoutOfCal = 2, low_id_rate = 1)
{

  if (!checkInput(c("fc.raw.file", "mass.error..ppm.", "uncalibrated.mass.error..ppm."), df_evd)) return(NULL)

  ## heuristic to determine if the instrument is completely out of calibration,
  ## i.e. all ID's are false positives, since the Precursor mass is wrong
  ## -- we use the SD; if larger than 2ppm,
  ## then ID's are supposedly random
  ## -- alt: we use the 1%-to-99% quantile range: if > 10ppm
  ## -- uninformative for detection is the distribution (it's still Gaussian for a strange reason)
  MS1_decal_smr = plyr::ddply(df_evd, "fc.raw.file", function(x)
    data.frame(n = nrow(x),
               sd = round(sd(x$mass.error..ppm., na.rm = TRUE), 1),
               sd_uncal = round(sd(x$uncalibrated.mass.error..ppm., na.rm = TRUE), 1),
               range = diff(quantile(x$mass.error..ppm., c(0.01, 0.99), na.rm = TRUE)),
               decal = (median(abs(x$uncalibrated.mass.error..ppm.), na.rm = TRUE) > 1e3),
               hasMassErrorBug = FALSE,
               hasMassErrorBug_unfixable = FALSE)
  )
  ## additionally use MS2-ID rate (should be below 1%)
  if (is.null(df_idrate)) {
    MS1_decal_smr$ms.ms.identified.... = 0 ## upon no info: assume low IDrate
  } else {
    MS1_decal_smr = merge(MS1_decal_smr, df_idrate)
  }
  MS1_decal_smr$lowIDRate = MS1_decal_smr$ms.ms.identified.... < low_id_rate

  recal_message = ""
  recal_message_post = ""
  ## check each raw file individually (usually its just a few who are affected)
  if (any(MS1_decal_smr$decal, na.rm = TRUE))
  {
    if (!checkInput(c("mass", "charge", "m.z"), df_evd)) return(NULL)
    recal_message = "MQ bug: data rescued"
    recal_message_post = 'MQ bug: data cannot be rescued'

    MS1_decal_smr$hasMassErrorBug[ MS1_decal_smr$fc.raw.file %in% MS1_decal_smr$fc.raw.file[MS1_decal_smr$decal > 0] ] = TRUE

    ## re-compute 'uncalibrated.mass.error..ppm.' and 'mass.error..ppm.'
    df_evd$theomz = df_evd$mass / df_evd$charge + 1.00726
    df_evd$mass.error..ppm.2 = (df_evd$theomz - df_evd$m.z) / df_evd$theomz * 1e6
    df_evd$uncalibrated.mass.error..ppm.2 = df_evd$mass.error..ppm.2 + df_evd$uncalibrated...calibrated.m.z..ppm.

    ## check if fix worked
    de_cal2 = plyr::ddply(df_evd, "fc.raw.file", .fun = function(x)  data.frame(q = (median(abs(x$uncalibrated.mass.error..ppm.2), na.rm = TRUE) > 1e3)))
    if (any(de_cal2$q, na.rm = TRUE))
    { ## fix did not work
      MS1_decal_smr$hasMassErrorBug_unfixable[ MS1_decal_smr$fc.raw.file %in% de_cal2$fc.raw.file[de_cal2$q] ] = TRUE
      recal_message = "m/z recalibration bugfix applied but failed\n(there are still large numbers)"
    }

    idx_overwrite = (df_evd$fc.raw.file %in% MS1_decal_smr$fc.raw.file[MS1_decal_smr$decal > 0])
    ## overwrite original values
    df_evd$mass.error..ppm.[idx_overwrite] = df_evd$mass.error..ppm.2[idx_overwrite]
    df_evd$uncalibrated.mass.error..ppm.[idx_overwrite] = df_evd$uncalibrated.mass.error..ppm.2[idx_overwrite]
  }

  MS1_decal_smr$outOfCal = (MS1_decal_smr$sd > tolerance_sd_PCoutOfCal) &
    (MS1_decal_smr$lowIDRate) &
    (MS1_decal_smr$sd < 100)  ## upper bound, to distinguish from MQ bug (which has much larger SD's)
  ## NA's are possible if only a single peptide was seen (yielding MS1_decal_smr$sd == NA)
  MS1_decal_smr$outOfCal[is.na(MS1_decal_smr$outOfCal)] = FALSE

  ## report too small search tolerance
  if (any(MS1_decal_smr$outOfCal)) recal_message = paste0("search tolerance too small (since sd > ", tolerance_sd_PCoutOfCal,")", sep="")


  ## Raw files where the mass error is obviously wrong (PSM's not substracted etc...)
  affected_raw_files = MS1_decal_smr$fc.raw.file[MS1_decal_smr$hasMassErrorBug]


  return(list(stats = MS1_decal_smr,
              affected_raw_files = affected_raw_files,
              df_evd = df_evd[, c("mass.error..ppm.", "uncalibrated.mass.error..ppm.", "fc.raw.file")],
              recal_message = recal_message,
              recal_message_post = recal_message_post
              ))
}

#'
#' Return list of raw file names which were reported by MaxQuant as reference point for alignment.
#' 
#' There is only one reference point which has '0' in 'retention.time.calibration' column in evidence.txt 
#' as corrected RT. This is true for most MaxQuant versions
#' and also true for fractions. However, some evidence.txt files show 0.03 as an averaged minimum per Raw file.
#' We use the raw.file with the smallest average as reference.
#' 
#' Note that MaxQuant uses a guide tree to align the Raw files, so the order of files does not influence the 
#' alignment. But the first file will always be used as reference point when reporting delta-RTs. And this file is also
#' used by PTXQC as reference file vs all other files to find the real calibration function (see alignmentCheck()).
#'
#' This function might return multiple raw file names (if MQ decides to change its mind at some point in the future).
#' In this case the result should be treated with caution or (better) regarded as failure.
#' 
#' @param data The data.frame with columns 'retention.time.calibration' and 'raw.file'
#' @return List of reference raw files (usually just one)
#'
findAlignReference = function(data)
{
  colnames(data) = tolower(colnames(data))
  if (!("retention.time.calibration" %in% colnames(data)))
  {
    stop("findAlignReference(): Error, could not find column 'retention.time.calibration' in data. Aborting!")  
  }
  if (!("raw.file" %in% colnames(data)))
  {
    stop("findAlignReference(): Error, could not find column 'raw.file' in data. Aborting!")  
  }
  fr = plyr::ddply(data, "raw.file", function(x) data.frame(range = diff(range(x$retention.time.calibration, na.rm = TRUE))))
  ref = as.character(fr$raw.file[fr$range <= min(fr$range)])
  return (ref)  
}


#'
#' Verify an alignment by checking the retention time differences of identical peptides across Raw files
#' 
#' The input is a data frame containing feature evidence with corrected retention times,
#' e.g. a 'calibrated.retention.time' column.
#'
#' Note that this function must be given real MS/MS identifications only (type "MULTI-MSMS")
#' in order to work correctly!
#' 
#' For each peptide sequence (and charge) in the reference Raw file, this function looks up the
#' already calibrated retention time difference of the same feature in all other files. For every comparison made,
#' we report the RT difference. If alignment worked perfectly, the differences are very small (<1 min).
#' 
#' An 'id' column must be present, to enable mapping the result of this function to the original data frame.
#' 
#' A reference Raw file can be identified using 'findAlignReference()'. If Maxquants experimental design included
#' pre-fractionation, a column named 'fraction' should be given and 'referenceFile' should be empty. This function will
#' pick the one Raw file for each fraction (the first in order) to use as reference. Only the immediately neighbouring
#' fractions will be matched to this reference.
#' 
#' @param data A data.frame with columns 'calibrated.retention.time', 'id', 'modified.sequence', 'charge', 'raw.file' and 'fraction' (if present)
#' @param referenceFile A raw file name as occuring in data$raw.file, serving as alignment reference (when no fractions are used).
#' @return A data.frame containing the RT diff for each feature found in a Raw file and the reference.
#'
alignmentCheck = function(data, referenceFile) {
  colnames(data) = tolower(colnames(data))
  
  if (!all(c('calibrated.retention.time', 'id', 'modified.sequence', 'charge', 'raw.file') %in% colnames(data)))
  {
    stop("alignmentCheck(): columns missing!")  
  }
  if (is.na(referenceFile) & !('fraction' %in% colnames(data)))
  {
    stop("alignmentCheck(): Either a referenceFile or 'fraction' column must be provided!")  
  }
  if (!is.na(referenceFile) & ('fraction' %in% colnames(data)))
  {
    stop("alignmentCheck(): referenceFile and 'fraction' cannot be provided at the same time!")
  }
  
  if (!is.na(referenceFile)) {
    RefRounds = list( list("file_ref" = referenceFile, "file_clients" = unique(data$raw.file)) )
  } else {
    ## Fractions ...
    RefRounds = list()
    ## get Raw files and their fractions
    df.f = data.frame(raws = unique(data$raw.file))
    df.f$fraction = data$fraction[match(df.f$raws, data$raw.file)]
    df.f
    ## make a list of references:
    df.ref = df.f[!duplicated(df.f$fraction),]
    df.ref
    ## for each fraction: get list of references and the files that are allowed to match against them
    for (idx_ref in 1:nrow(df.ref))
    { # idx_ref = 1
      frac = df.ref$fraction[idx_ref]
      rFile = df.ref$raws[idx_ref]
      rFile
      clients = df.f$raws[df.f$fraction %in% (frac-1):(frac+1)]
      RefRounds[[idx_ref]] = list("file_ref" = rFile, "file_clients" = clients)
    }
    RefRounds
  }

  alignQ = data.frame()

  for (lRef in RefRounds)
  { ## lRef = RefRounds[[1]]
    file_ref = lRef[["file_ref"]]
    file_clients = lRef[["file_clients"]]
    
    ## prefilter data: only peptides which occur in reference
    ##            and are part of client list
    data_round = data[data$modified.sequence %in% data$modified.sequence[data$raw.file==file_ref] &
                      data$raw.file %in% file_clients,]
    
    alignQ_round = plyr::ddply(data_round, c("modified.sequence", "charge"), function(x)
    {
      ## reference must be present
      if ((nrow(x)==1) | (sum(x$raw.file == file_ref)==0) | (sum(duplicated(x$raw.file))>0)) {
        return (NULL)
      }
      
      ## we could rescue this case, but for performance reasons we don't
      ## duplicates per raw file... ignore this raw file
      #if (sum(duplicated(x$raw.file))>0) {
      #  x$calibrated.retention.time[x$raw.file %in% x$raw.file[duplicated(x$raw.file)]] = NA
      #}
      
      ## RT of reference (we use 'mean' since there could be multiple ones when using fractions)
      m = x$calibrated.retention.time[x$raw.file == file_ref]
      
      x$rtdiff = x$calibrated.retention.time - m;
      return(x[,c("raw.file", "id", "rtdiff", 'calibrated.retention.time')])
    })
    ## alignQ_round might be an empty data.frame now, if
    ## there was only one fraction with no neighbours
    if (!plyr::empty(alignQ_round))
    {
      alignQ_round$refFile = file_ref
      ## remove matches to self (i.e. with rtDiff==0)
      alignQ_round = alignQ_round[ alignQ_round$raw.file != file_ref, ]
      alignQ = rbind(alignQ, alignQ_round)
    }
  }
  
  return(alignQ)
}


#'
#' Compute the fraction of features per Raw file which have an acceptable RT difference after alignment
#' 
#' Using the result from 'alignmentCheck()', score the features of every Raw file and see if they 
#' have been properly aligned.
#' Returned value is between 0 (bad) and 1 (all aligned).
#' 
#' @param data A data.frame with columns 'rtdiff' and 'raw.file'
#' @param allowed.deltaRT The allowed matching difference (1 minute by default)
#' @return A data.frame with one row for each raw.file and columns 'raw.file' and 'withinRT' (0-1)
#' 
ScoreInAlignWindow = function(data, allowed.deltaRT = 1)
{
  colnames(data) = tolower(colnames(data))
  
  if (!all(c('rtdiff', 'raw.file') %in% colnames(data)))
  {
    stop("alignmentCheck(): columns missing!")  
  }
  alignQC = plyr::ddply(data, "raw.file", function(x) {
    withinRT = sum(abs(x$rtdiff) < allowed.deltaRT, na.rm = TRUE) / sum(!is.na(x$rtdiff))
    return(data.frame(withinRT = withinRT))
  })
  
  return(alignQC)
}
  

#'
#' Check how close transferred ID's after alignment are to their genuine IDs within one Raw file.
#' 
#' The input is a data.frame containing feature evidence with corrected retention times,
#' e.g. a 'calibrated.retention.time' column.
#'
#' Note that this function must be given MS/MS identifications of type "MULTI-MSMS" and "MSMS-MATCH".
#' It will stop() otherwise.
#'  
#' We compare for each peptide sequence (and charge) the RT difference within groups of either genuine as well as mixed pairs.
#' For every comparison made, we report the RT span If alignment worked perfectly, the span are very small (<1 min),
#' for the mixed group, i.e. the pairs are accidentally split 3D peaks. Alignment performance has no influence on the
#' genuine-only groups.
#' 
#' Note: We found early MaxQuant versions (e.g. 1.2.2.5) to have an empty 'modified.sequence' column for 'MULTI-MATCH' entries.
#' The sequence which SHOULD be present is equal to the immediate upper row. This is what we use to guess the sequence.
#' However, this relies on the data.frame not being subsetted before (we can sort using the 'id' column)!
#' 
#' @param df_evd_all A data.frame with columns 'type', 'calibrated.retention.time', 'modified.sequence', 'charge', 'raw.file'
#' @return A data.frame containing the RT diff for each ID-group found in a Raw file (bg = genuine).
#'
idTransferCheck = function(df_evd_all) {
  colnames(df_evd_all) = tolower(colnames(df_evd_all))
  
  if (!checkInput(c('id', 'type', 'calibrated.retention.time', 'modified.sequence', 'charge', 'fc.raw.file'), df_evd_all)) return()
 
  
  if (!all(c("MULTI-MSMS", "MULTI-MATCH") %in% unique(df_evd_all$type)))
  {
    stop('idTransferCheck(): scan types missing! Required: "MULTI-MSMS" and "MULTI-MATCH".')  
  }
  
  
  df_evd_all$seq_charge = paste(factor(df_evd_all$modified.sequence), df_evd_all$charge, sep="_")
  alignQ = plyr::ddply(df_evd_all[,c("fc.raw.file", "type", "calibrated.retention.time", "seq_charge")],
                       "fc.raw.file",
                       function(x) {
    # unique(df_evd_all$fc.raw.file)
    # x = df_evd_all[ df_evd_all$fc.raw.file == "file 01", ]
    
    ## genuine groups only (within this Raw file):
    x_genuine = x[x$type=="MULTI-MSMS",]
    rt_diffs_genuine = plyr::ddply(x_genuine, "seq_charge", 
                             function(x2) {
                               if (nrow(x2)==1) return(NULL) ## we do not want singlets
                               return (data.frame(rtdiff_genuine = diff(range(x2$calibrated.retention.time))))
                             })
    # rt_diffs_genuine might be empty, if no oversampling is seen
    if (nrow(rt_diffs_genuine) == 0) rt_diffs_genuine = data.frame(seq_charge = character(0), rtdiff_genuine = numeric(0)) ## remove columns from empty DF; avoid merge()ing NA columns
    
    ## mixed class:
    ## retain only IDs which have at least one transferred ID
    x_mixed = x[x$seq_charge %in% x$seq_charge[x$type=="MULTI-MATCH"], ]
    if (nrow(x_mixed)>0) {    
      rt_diffs_mixed = plyr::ddply(x_mixed, "seq_charge", 
                             function(x2) {
                               if (nrow(x2)==1) return(NULL) ## we do not want singlets
                               return (data.frame(rtdiff_mixed = diff(range(x2$calibrated.retention.time))))
                             })
      ## rtdiff_mixed might be empty, if only singlets were transferred
      ## only merge if non-empty (otherwise the whole merge is empty)
      if (nrow(rt_diffs_mixed) > 0) {
        rt_diffs_genuine = merge(rt_diffs_genuine, rt_diffs_mixed, all = TRUE)
      }
    }
    return (rt_diffs_genuine)    
  })
  #head(alignQ)
  #hist(log(1+alignQ$rtdiff), 100)
  #hist(log(1+alignQ$rtdiff_bg), 100, add=TRUE, col="red")
  
  return(alignQ)
}


#'
#' For grouped peaks: separate them into in-width vs. out-width class.
#' 
#' Looking at groups only: Compute the fraction of 3D-peak pair groups per Raw file which 
#' have an acceptable RT difference after alignment using the result from 'idTransferCheck()',
#' i.e. compute the fraction of groups which are within a certain RT tolerance.
#' 
#' Returned value is between 0 (bad) and 1 (all within tolerance).
#' 
#' @param data A data.frame with columns 'fc.raw.file', 'rtdiff_mixed', 'rtdiff_genuine'
#' @param df.allowed.deltaRT The allowed matching difference for each Raw file (as data.frame(fc.rawfile, m))
#' @return A data.frame with one row for each raw.file and columns 'raw.file' and score 'withinRT' (0-1)
#'
inMatchWindow = function(data, df.allowed.deltaRT)
{
  ## 'data' columns of interest : 
  cols_diffs = c("rtdiff_mixed", "rtdiff_genuine")
  
  colnames(data) = tolower(colnames(data))
  
  if (!all(c(cols_diffs, 'fc.raw.file') %in% colnames(data)))
  { ## be error tolerant and return an empty data frame.
    return (data.frame(fc.raw.file = NA, withinRT_genuine = NA, withinRT_mixed = NA, withinRT_all = NA)[numeric(0), ])
  }
  
  alignQC = plyr::ddply(data, "fc.raw.file", function(x) {
    # x=data[ data$fc.raw.file=="file 01",]
    allowed.deltaRT = df.allowed.deltaRT$m[match(x$fc.raw.file[1], df.allowed.deltaRT$fc.raw.file)]
    
    withinRT_genuine = sum(abs(x$rtdiff_genuine) < allowed.deltaRT, na.rm = TRUE) / sum(!is.na(x$rtdiff_genuine))
    withinRT_mixed   = sum(abs(x$rtdiff_mixed) < allowed.deltaRT, na.rm = TRUE) / sum(!is.na(x$rtdiff_mixed))
    ## compute the RT-span if all evidence is considered 
    ##  --> simply the max of mixed and genuine, since each evidence must be in either of these two
    x$rtdiff_all = pmax(x$rtdiff_mixed, x$rtdiff_genuine, na.rm = TRUE)
    withinRT_all = sum(abs(x$rtdiff_all) < allowed.deltaRT, na.rm = TRUE) / sum(!is.na(x$rtdiff_all))
    return(data.frame(withinRT_genuine, withinRT_mixed, withinRT_all))
  })
  
  return(alignQC)
}


#'
#' Determine fraction of evidence which causes segmentation, i.e. sibling peaks at different RTs
#' confirmed either by genuine or transferred MS/MS.
#'
#' Sometimes, MQ splits a feature into 2 or more if the chromatograpic conditions are not optimal and there
#' is a drop in RT intensity.
#' If both features contain successful MS/MS scans, we will find the same peptide twice (with slightly different RT)
#' in the same charge state. This constitutes a natively split peak and is rare (95% of all genuine peaks are unique).
#' 
#' If Match-between-runs is used and the RT alignment is not perfect, then a peptide might be inferred at a wrong
#' RT position, even though this Raw file already contains MS/MS evidence of this peptide.
#' Usually the number of peak duplicates rises drastically (e.g. only 75% of peaks are unique after MBR was used).
#' In most cases, the RT is too far off to be a split peak. It's rather a lucky hit with accidentally the same mass-to-charge,
#' and thus the intensity is random.
#' To find by how much these peak pairs differ in RT, use idTransferCheck() and inMatchWindow().
#' 
#' Required columns are 'is.transferred', 'fc.raw.file', 'modified.sequence', 'charge', 'type'.
#'
#' Note that this function must be given MS/MS identifications of type "MULTI-MSMS" and "MSMS-MATCH".
#' It will stop() otherwise.
#' 
#' @param df_evd_all A data.frame of evidences containing the above columns
#' @return A data.frame with one row per Raw file and 
#'         three columns: 
#'           1) % of native single peaks (ignoring transferred IDs)
#'           2) % of single peaks (group of size=1) using only groups which have one transferred evidence
#'           3) % of single peaks using all groups
#'
peakSegmentation = function(df_evd_all)
{
  if (!checkInput(c("is.transferred", "fc.raw.file", "modified.sequence", "charge", 'type'), df_evd_all)) return()

  if (!all(c("MULTI-MSMS", "MULTI-MATCH") %in% unique(df_evd_all$type)))
  {
    stop('peakSegmentation(): scan types missing! Required: "MULTI-MSMS" and "MULTI-MATCH".')  
  }
  
  fc.raw.files = unique(df_evd_all$fc.raw.file)
  
  ## just keep "MULTI-MATCH" and "MULTI-MSMS", to keep results comparable to idTransferCheck()
  df_evd_all = df_evd_all[df_evd_all$type %in% c("MULTI-MSMS", "MULTI-MATCH"), ]

  cols = c("is.transferred", "fc.raw.file", "modified.sequence", "charge")
  countSeqs = plyr::ddply(df_evd_all[, cols], cols[-1], function(x)
  {
    return(data.frame(nNative = sum(!x$is.transferred), nMatched = sum(x$is.transferred)))#, ratio = ratio))
  })
  
  mbr_score = plyr::ddply(countSeqs, "fc.raw.file", function(countSeqs_sub)
  {
    #unique(countSeqs$fc.raw.file)
    #countSeqs_sub = countSeqs[countSeqs$fc.raw.file == "file 02", ]
    
    ddt = table(countSeqs_sub[, c("nMatched", "nNative")])
    ### ddt might look like this:
    #                 nNative
    #     nMatched    0    1   2  3  4 5 6 7 8 9 10 11 13 14 15 18 20 26 27 28 31 37 39
    #            0    0 6249 230 41 16 4 1 4 2 0  0  0  1  1  0  0  0  1  1  0  0  0  0
    #            1 1352  303  34 16  6 4 2 1 1 0  1  1  0  1  0  1  0  0  0  0  0  0  0
    #            2   49   29   4  2  1 0 0 1 0 0  1  
    
    ### scale ddt matrix from counts of events to counts of peaks,
    ### e.g. event: each 1native+2matched affects three peaks
    d1 = dimnames(ddt)$nMatched
    d2 = dimnames(ddt)$nNative
    #ddt_b = ddt
    for (i in 1:length(d1)) {
      for (j in 1:length(d2)) {
        ddt[i,j] = ddt[i,j] * (as.numeric(d1[i]) + as.numeric(d2[j]))
      }
    }
    ## add nMatched=1 row -- if already present, we just insert an empty row at the end (no harm done)
    ddt = rbind(ddt, 0)

    ## segmentation genuine (ignore the nMatched, i.e. project everything onto first row)
    n.perf = sum(ddt[,colnames(ddt)==1]) ## uniquely matched
    n.all = sum(ddt[,!colnames(ddt)==0]) ## all natives (==0 is the matched-only column)
    single.nat = max(0, n.perf/n.all, na.rm=TRUE) ## e.g. 94%, but n.all could be 0, yielding NaN
    
    ## segmentation of matched-only
    i.perf = ddt[rownames(ddt)==1, colnames(ddt)==0] ## transferred singlets (nMatched=1,nNative=0)
    i.all = sum(ddt[rownames(ddt)!=0, ])             ## all groups which contain a match (==without first row)
    single.matched = max(0, i.perf) / i.all  ## e.g. 50%
    ## no ID's were transferred (div/0) -- this remains NaN
    
    ## segmentation of all evidence (weighed average of above)
    comb.perf = max(0, ddt[rownames(ddt)==1, colnames(ddt)==0]) + ## transferred singlets (nMatched=1,nNative=0) -- could be empty
                ddt[rownames(ddt)==0, colnames(ddt)==1]
    comb.all = sum(ddt)             ## all groups
    single.all = max(0, comb.perf) / comb.all       ## e.g. 87%
    
    ## final result
    data.frame(single.nat = single.nat, single.matched = single.matched, single.all = single.all)
  })
  
  ## some Raw files might be REALLY sparse and up to now, they would be missing
  empty_fc.raw.files = fc.raw.files[!(fc.raw.files %in% mbr_score$fc.raw.file)]
  if (length(empty_fc.raw.files) > 0)
  { ## ... add them with NA values
    mbr_score = merge(mbr_score, data.frame(fc.raw.file = empty_fc.raw.files), all = TRUE)   
  }
  
  return (mbr_score = mbr_score)
}

#'
#' Combine several data structs into a final picture for segmentation incurred by 'Match-between-runs'.
#' 
#' qMBRSeg_Dist_inGroup might be empty if there are only singlets (transferred and genuine), but then the scores will be pretty
#' boring as well (100%).
#'
#' @param qMBR                 A data.frame as computed by peakSegmentation()
#' @param qMBRSeg_Dist_inGroup A data.frame as computed by inMatchWindow()
#' @return A data.frame which details the distribution of singlets and pairs (inRT and outRT) for each Raw file and genuine vs. all
#' 
computeMatchRTFractions = function(qMBR, qMBRSeg_Dist_inGroup)
{
  ## data might look like this:
#     fc.raw.file single.nat  single.matched   single.all
#   1  file 01     0.9740162       0.3393057    0.9517918
#   2  file 02     0.9813251       0.7689088    0.9433782
#   3  file 03     0.9671893       0.6986601    0.9121866
# 
# qMBRSeg_Dist_inGroup
#      fc.raw.file withinRT_genuine withinRT_mixed withinRT_all
#   1      file 01        0.5000000            NaN    0.5000000
#   2      file 02        0.3589744      0.3076923    0.3750000
#   3      file 03        0.2068966      0.1764706    0.1851852
#   4      file 04        0.2500000      0.1034483    0.1818182
#   5      file 05        0.3125000      0.3333333    0.3333333
#   6      file 06        0.2500000      0.4090909    0.3529412
  
  ## compute percentage of outside dRT peaks in genuine, matched and combined(=all)
  ## then calc the drop.
  f = plyr::ddply(qMBR, "fc.raw.file", function(x) {
    #x = qMBR[3, , drop = FALSE]
    rr = qMBRSeg_Dist_inGroup$fc.raw.file==x$fc.raw.file
    
    nat.inRT = qMBRSeg_Dist_inGroup$withinRT_genuine[rr] ## e.g. 0.87
    if (length(nat.inRT) == 0) nat.inRT = NA
    x$single.nat.inRT = (1-x$single.nat)*nat.inRT
    x$single.nat.outRT = (1-x$single.nat)*(1-nat.inRT)

    matched.inRT = qMBRSeg_Dist_inGroup$withinRT_mixed[rr]
    if (length(matched.inRT) == 0) matched.inRT = NA
    x$single.matched.inRT = (1-x$single.matched)*matched.inRT
    x$single.matched.outRT = (1-x$single.matched)*(1-matched.inRT)
    
    combined.inRT = qMBRSeg_Dist_inGroup$withinRT_all[rr]
    if (length(combined.inRT) == 0) combined.inRT = NA
    x$single.all.inRT = (1-x$single.all)*combined.inRT
    x$single.all.outRT = (1-x$single.all)*(1-combined.inRT)
    
    r = data.frame(fc.raw.file=x$fc.raw.file, 
                   single=c(x$single.nat, x$single.matched, x$single.all),
                   multi.inRT = c(x$single.nat.inRT, x$single.matched.inRT, x$single.all.inRT),
                   multi.outRT = c(x$single.nat.outRT, x$single.matched.outRT, x$single.all.outRT),
                   sample = factor(c("genuine", "transferred", "all"), levels=c("genuine", "transferred", "all", ordered = TRUE)))
    return(r)    
  })
  return(f)
}

#'
#' Return a tree plot with a possible alignment tree.
#' 
#' This allows the user to judge which Raw files have similar corrected RT's (i.e. where aligned successfully).
#' If there are clear sub-clusters, it might be worth introducing artifical fractions into MaxQuant,
#' to avoid ID-transfer between these clusters (use the MBR-Align and MBR-ID-Transfer metrics to support the decision).
#' 
#' If the input contains fractions, leaf nodes will be colored accordingly.
#' Distinct sub-clusters should have their own color.
#' If not, MaxQuant's fraction settings should be optimized.
#' Note that introducing fractions in MaxQuant will naturally lead to a clustering here (it's somewhat circular).
#' 
#' @param df_evd  Evidence table containing calibrated retention times and sequence information.
#' @param col_fraction Empty vector or 1-values vector giving the name of the fraction column (if existing)
#' @return ggplot object containing the correlation tree
#' 
#' @import ggplot2
#' @export
#'
RTalignmentTree = function(df_evd, col_fraction = c())
{
  #df_evd$fc.raw.file=df_evd$raw.file
  head(df_evd)
  
  req_cols = c("calibrated.retention.time", "fc.raw.file", col_fraction, "modified.sequence", "charge")
  if (!all(req_cols %in% colnames(df_evd)))
  {
    stop("RTalignmentTree: Missing columns! Please fix the code: ", 
         setdiff(req_cols, colnames(df_evd)), "!")
  }
  
  d_cast = reshape2::dcast(df_evd, modified.sequence + charge ~ fc.raw.file, mean, value.var = "calibrated.retention.time")
  
  head(d_cast[,-(1:2)])
  d_cast.m = as.matrix(d_cast[,-(1:2)])
  head(d_cast.m)
  
  #Dissimilarity = 1 - Correlation
  #Dissimilarity = (1 - Correlation)/2
  #Dissimilarity = 1 - Abs(Correlation)
  #Dissimilarity = Sqrt(1 - Correlation2)
  dissimilarity = (1 - abs(cor(d_cast.m, use="pairwise.complete.obs")))
  ## if some samples have no overlap, their cell is NA --> set to 1 (max distance)
  dissimilarity[is.na(dissimilarity)] = 1
  #plot(hclust(as.dist(dissimilarity), method="ward.D"))
  ddata = ggdendro::dendro_data(hclust(as.dist(dissimilarity)), type = "rectangle")
  
  if (length(col_fraction))
  {
    idx_raw = match(ddata$labels$label, df_evd$fc.raw.file)
    ddata$labels$col = factor(df_evd[idx_raw, col_fraction])
  } else {
    ddata$labels$col = "black"
  }
  p = ggplot(ggdendro::segment(ddata)) + 
      geom_segment(aes(x = .data$x, y = .data$y, xend = .data$xend, yend = .data$yend)) +
      scale_x_continuous(breaks = ddata$labels$x, labels = ddata$labels$label) +
      theme_blank() +
      theme(axis.text.y = element_text(colour = ddata$labels$col), ## is a vector not officially supported...
                     axis.text.x = element_blank()) +
      coord_flip() +
      ggtitle("[experimental] EVD: Clustering Tree of Raw files", "by Correlation of Corrected Retention Times")
  #p
  return(p)
}

#####################################################################

#'
#' Metric for msmsscans.txt, showing TopN over RT.
#'
qcMetric_MSMSScans_TopNoverRT =  setRefClass(
  "qcMetric_MSMSScans_TopNoverRT",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate = 
      "TopN over retention time. Similar to ID over RT, this metric reflects the complexity of the sample
at any point in time. Ideally complexity should be made roughly equal (constant) by choosing a proper (non-linear) LC gradient.
See Moruz 2014, DOI: 10.1002/pmic.201400036 for details.
    
Heatmap score [MS<sup>2</sup> Scans: TopN over RT]: Rewards uniform (function Uniform) TopN events over time.
",
    workerFcn = function(.self, df_msmsScans)
    {
      ## completeness check
      if (!checkInput(c("fc.raw.file", "retention.time", "scan.event.number", "rRT"), df_msmsScans)) return()
      dd = data.table::as.data.table(df_msmsScans[, c("fc.raw.file", "retention.time", "scan.event.number", "rRT")])
      data.table::setkey(dd, fc.raw.file, retention.time) ## sort by RT

      ## find the highest scan event (SE) after an MS1 scan
      DF_max = dd[, {
          idx = which(getMaxima(scan.event.number, thresh_rel = 0.0))
          maxSE = scan.event.number[idx]
          RTbin = rRT[idx]
          list(maxSE = maxSE, rRT = RTbin)
        }, by = "fc.raw.file"]
      
      DFmse = DF_max[, list(topN = as.double(median(maxSE))),
                      by = c("fc.raw.file", "rRT")]
      head(DFmse)
      
      lpl =
        byXflex(DFmse, DFmse$fc.raw.file, 6, plot_TopNoverRT, sort_indices = FALSE)
      
      ## QC measure for smoothness of TopN over RT
      qc_TopNRT = plyr::ddply(DFmse, "fc.raw.file", function(x) data.frame(val = qualUniform(x$topN)))
      colnames(qc_TopNRT)[colnames(qc_TopNRT) == "val"] = .self$qcName
      
      return(list(plots = lpl, qcScores = qc_TopNRT))
    }, 
    qcCat = "LC", 
    qcName = "MS^2*Scans:~TopN~over~RT", 
    orderNr = 0120
  )
    return(.self)
  })
)

#####################################################################

qcMetric_MSMSScans_MSMSIntensity =  setRefClass(
  "qcMetric_MSMSScans_MSMSIntensity",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate = 
      "MS/MS identifications can be 'bad' for a couple of reasons. It could be computational, i.e. ID rates
are low because you specified the wrong protein database or modifications (not our concern here).
Another reason is low/missing signals for fragment ions,
e.g. due to bad (quadrupole/optics) ion transmission (charging effects), too small isolation windows, etc.

Hence, we plot the TIC and base peak intensity of all MS/MS scans (incl. unidentified ones) per Raw file.
Depending on the setup, these intensities can vary, but telling apart good from bad samples
should never be a problem. If you only have bad samples, you need to know the intensity a good sample would reach.

To automatically score this, we found that the TIC should be 10-100x larger than the base peak, i.e. there 
should be many other ions which are roughly as high (a good fragmentation ladder).
If there are only a few spurious peaks (bad MS/MS), the TIC is much lower. Thus, we score the ratio 
BP * 10 < TIC (this would be 100% score). If it's only BP * 3 > TIC, we say this MS/MS failed (0%).
Anything between 3x and 10x gets a score in between. The score for the Raw file is computed as the
median score across all its MS/MS scans.

Heatmap score [MS<sup>2</sup> Scans: Intensity]: Linear score (0-100%) between 3 < (TIC / BP) < 10. 
    ",
    workerFcn = function(.self, d_msmsScan, score_min_factor = 3, score_max_factor = 10)
    {
      ## completeness check
      if (!checkInput(c("fc.raw.file", "total.ion.current", "base.peak.intensity"), d_msmsScan)) return ()
      
      ## use data.table for aggregation, its MUCH faster than ddply() and uses almost no extra memory
      dd = data.table::as.data.table(d_msmsScan[, c("fc.raw.file", "total.ion.current", "base.peak.intensity")])
      dd$log.total.ion.current = log10(dd$total.ion.current)
      dd$log.base.peak.intensity = log10(dd$base.peak.intensity)
      log.dd.tic = dd[,list(mean=mean(log.total.ion.current),
                        min=min(log.total.ion.current),
                        lower=quantile(log.total.ion.current, .25, na.rm=TRUE),
                        middle=quantile(log.total.ion.current, .50, na.rm=TRUE),
                        upper=quantile(log.total.ion.current, .75, na.rm=TRUE),
                        max=max(log.total.ion.current)),
                    by='fc.raw.file']
      log.dd.bpi = dd[,list(mean2=mean(log.base.peak.intensity),
                        min2=min(log.base.peak.intensity),
                        lower2=quantile(log.base.peak.intensity, .25, na.rm=TRUE),
                        middle2=quantile(log.base.peak.intensity, .50, na.rm=TRUE),
                        upper2=quantile(log.base.peak.intensity, .75, na.rm=TRUE),
                        max2=max(log.base.peak.intensity)),
                  by='fc.raw.file']
      ## merge, so we have one table for byXflex()
      dd.all = merge(log.dd.tic, log.dd.bpi, by='fc.raw.file')
      
      ## for scoring...
      dd.ratio = dd[,list(ratio=median(total.ion.current/base.peak.intensity, na.rm=TRUE)), by ='fc.raw.file']
      dd.ratio
      
      
      plot_MSMSintensity = function(dd.all) {
        pl = ggplot(data = dd.all, aes(x = fc.raw.file)) + 
                geom_boxplot(stat = "identity", aes(col = "TIC", ymin = min, lower = lower, middle = middle, upper = upper, ymax = max)) +
                geom_boxplot(stat = "identity", aes(col = "Base\nPeak", ymin = min2, lower = lower2, middle = middle2, upper = upper2, ymax = max2), width = 0.3) +
                scale_color_manual("MS/MS\nintensity", values = c("TIC" = "black", "Base\nPeak" = "blue")) +
                ylim(0, NA) +
                scale_x_discrete_reverse(dd.all$fc.raw.file) +
                ggtitle("[experimental] MSMSscans: MS/MS intensity") +
                xlab("") +
                ylab(expression('intensity (' * log[10] * ')')) +
                coord_flip()
      }
      
      lpl = byXflex(dd.all, dd.all$fc.raw.file, 12, plot_MSMSintensity, sort_indices = FALSE)
      
      
      ## QC measure for intensity ratio below expected threshold (3x-10x by default)
      qc_MSMSint = plyr::ddply(dd.ratio, "fc.raw.file", 
                     function(x) data.frame(val = qualLinThresh(pmax(0, x$ratio - score_min_factor), t = score_max_factor - score_min_factor)))
      colnames(qc_MSMSint)[colnames(qc_MSMSint) == "val"] = .self$qcName
      
      return(list(plots = lpl, qcScores = qc_MSMSint))
    }, 
    qcCat = "MS", 
    qcName = "MS^2*Scans:~Intensity", 
    orderNr = 0245
  )
    return(.self)
  })
)

#####################################################################

qcMetric_MSMSScans_IonInjTime =  setRefClass(
  "qcMetric_MSMSScans_IonInjTime",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate = 
      "Ion injection time score - should be as low as possible to allow fast cycles. Correlated with peptide intensity.
Note that this threshold needs customization depending on the instrument used (e.g., ITMS vs. FTMS).

Heatmap score [MS<sup>2</sup> Scans: Ion Inj Time]: Linear score as fraction of MS/MS below the threshold. 
",
    workerFcn = function(.self, df_msmsScans, threshold_iit)
    {
      ## completeness check (mzTab might not have IIJ)
      if ( any(!("ion.injection.time" %in% colnames(df_msmsScans)), all(is.na(df_msmsScans$ion.injection.time)) ) ) {
        return(NULL)
      } 
      
      if (!checkInput(c("fc.raw.file", "ion.injection.time", "rRT"), df_msmsScans)) return ()
      
      ## use data.table for aggregation, its MUCH faster than ddply() and uses almost no extra memory
      dd = data.table::as.data.table(df_msmsScans[, c("fc.raw.file", "ion.injection.time", "rRT")])

      ## average injection time over RT
      DFmIIT = dd[, list(medIIT = median(ion.injection.time)), by=c("fc.raw.file", "rRT")]

      ## average injection time overall
      DFmIITglob = dd[, list(mean = mean(ion.injection.time)), by = "fc.raw.file"]

      lpl =
        byXflex(DFmIIT, DFmIIT$fc.raw.file, 6, plot_IonInjectionTimeOverRT, sort_indices = FALSE,
                stats = DFmIITglob,
                extra_limit = threshold_iit)
      
      
      ## QC measure for injection times below expected threshold
      DFmIIT_belowThresh = dd[,
                              list(belowThresh_IIT = sum(ion.injection.time < threshold_iit, na.rm = TRUE) / .N),
                              by = "fc.raw.file"]
      
      qc_IIT = plyr::ddply(DFmIIT_belowThresh, "fc.raw.file", 
                     function(x) data.frame(val = qualLinThresh(x$belowThresh_IIT, t = 1)))
      colnames(qc_IIT)[colnames(qc_IIT) == "val"] = .self$qcName
      
      return(list(plots = lpl, qcScores = qc_IIT))
    }, 
    qcCat = "MS", 
    qcName = "MS^2*Scans:~Ion~Inj~Time", 
    orderNr = 0240
  )
    return(.self)
  })
)

#####################################################################

qcMetric_MSMSScans_TopN =  setRefClass(
  "qcMetric_MSMSScans_TopN",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate = 
      "Reaching TopN on a regular basis indicates that all sections of the LC gradient
deliver a sufficient number of peptides to keep the instrument busy. This metric somewhat summarizes 'TopN over RT'.

Heatmap score [MS<sup>2</sup> Scans: TopN high]: rewards if TopN was reached on a regular basis (function qualHighest)
",
    workerFcn = function(.self, df_msmsScans)
    {
      ## completeness check
      if (!checkInput(c("fc.raw.file", "scan.event.number"), df_msmsScans)) return ()
      
      ## check if scan.event.number requires fixing
      ## (e.g. when MS3 events are recorded between MS2 events, there are gaps in the numbering)
      ## we close the gaps by requiring consecutive scan event numbers in MS2
      scan.events = df_msmsScans[, c("scan.event.number", "fc.raw.file")]
      while (TRUE) { ## should be at most max(scan.event.number) iterations
        se_pos = 1 + which(diff(scan.events$scan.event.number) > 1) ## position of gaps>1
        if (length(se_pos) == 0) break;
        scan.events$scan.event.number[se_pos] = scan.events$scan.event.number[se_pos] - 1
      }

      ## use data.table for aggregation, its MUCH faster than ddply() and uses almost no extra memory
      DFc = data.table::as.data.table(scan.events)[, list(n=.N), by=c("scan.event.number", "fc.raw.file")]

      dfc.ratio = plyr::ddply(DFc, "fc.raw.file", function(x, maxn)
      {
        ## sort x by scan event
        event_count = x$n
        ## verify its monotonically increasing
        if (is.unsorted(rev(event_count))) {
          #print(x)
          stop("Scan event distribution is not monotonically increasing!")
        } 
        ## verify that there are no gaps
        if (max(x$scan.event.number) != nrow(x)) {
          #print(x)
          stop("Scan event distribution has unexpected holes...!")
        }
        
        event_pre = c(event_count[-1], 0)
        event_diff = event_count - event_pre
        
        ## build new DF of fixed length
        sn = x$scan.event.number
        if (max(sn) < maxn) 
        {
          event_diff = c(event_diff, rep(0, maxn-max(sn)))
          sn = c(sn, (max(sn)+1):maxn)
        }
        DF.new = data.frame(scan.event.number = sn, n = event_diff)
        return (DF.new)
      }, maxn = max(DFc$scan.event.number))
      head(dfc.ratio)
      
      lpl =
        byXflex(dfc.ratio, dfc.ratio$fc.raw.file, 9, plot_TopN, sort_indices = FALSE)
      
      ## QC measure for always reaching the maximum TopN
      maxTopN = max(dfc.ratio$scan.event.number)
      qc_TopN = plyr::ddply(dfc.ratio, "fc.raw.file", function(x) data.frame(val = qualHighest(x$n, maxTopN)))
      colnames(qc_TopN)[colnames(qc_TopN) == "val"] = .self$qcName
      
      return(list(plots = lpl, qcScores = qc_TopN))
    }, 
    qcCat = "MS", 
    qcName = "MS^2*Scans:~TopN~high", 
    orderNr = 0350
  )
    return(.self)
  })
)

#####################################################################

qcMetric_MSMSScans_TopNID =  setRefClass(
  "qcMetric_MSMSScans_TopNID",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate = 
      "Looking at the 
identification rates per scan event (i.e. the MS/MS scans after a survey scan) can give 
hints on how well scheduled precursor peaks could be fragmented and identified.
If performance drops for the later MS/MS scans, then the LC peaks are probably not wide enough to deliver
enough eluent or the intensity threshold to trigger the MS/MS event should be lowered (if LC peak is already over),
or increased (if LC peak is still to weak to collect enough ions).

Heatmap score [MS<sup>2</sup> Scans: TopN ID over N]: Rewards uniform identification performance across all scan events.
",
    workerFcn = function(.self, df_msmsScans)
    {
      ## completeness check
      if (!checkInput(c("fc.raw.file", "scan.event.number", "identified"), df_msmsScans)) return()
      
      ## use data.table for aggregation, its MUCH faster than ddply() and uses almost no extra memory
      DF = data.table::as.data.table(df_msmsScans[, c("fc.raw.file", "scan.event.number", "identified")])[, list(n=.N), by=c("fc.raw.file", "scan.event.number", "identified")]
      
      # try KS on underlying data instead of using qualUniform()
      #   DF2= ddply(df_msmsScans, "fc.raw.file", function(rf){
      #     cat(class(rf))
      #     cat(rf$fc.raw.file[1])
      #     idx_p = rf$identified=="+"
      #     cat(length(idx_p) %+% "\n")
      #     kk = ks.test(rf$scan.event.number[idx_p], rf$scan.event.number[-idx_p])
      #     cat(kk$p.value)
      #     kk$statistic  # = 'D' ,,  p.value is much smaller (~0)
      #   })
      #   --> fail, 'D' and p-values are too low
      df.ratio = plyr::ddply(DF, c("scan.event.number", "fc.raw.file"), function(x)
      {
        xp = xm1 = xm2 = 0
        if ("+" %in% x$identified) xp = x$n[x$identified=="+"]
        if ("-" %in% x$identified) xm1 = x$n[x$identified=="-"]
        if ("" %in% x$identified) xm2 = x$n[x$identified==""] # MQ 2.x leaves unidentified empty
        
        ratio = xp * 100 / sum(xp, xm1, xm2)
        if (is.na(ratio)) 
        { # the whole 'identified' column is empty (no '+', no '-')
          ratio = 0 
        }
        return (data.frame(ratio = ratio, count = sum(x$n)))
      })
      head(df.ratio)
      
      lpl = byXflex(df.ratio, df.ratio$fc.raw.file, 9, plot_ScanIDRate, sort_indices = FALSE)
      
      ## QC measure for constantly identifiying peptides, irrespective of scan event number
      ## -- we weight scan events by their number of occurence
      qc_TopN_ID = plyr::ddply(df.ratio, "fc.raw.file", function(x) {
        data.frame(val = qualUniform(x$ratio, x$count))
      })
        
      colnames(qc_TopN_ID)[colnames(qc_TopN_ID) == "val"] = .self$qcName
      
      return(list(plots = lpl, qcScores = qc_TopN_ID))
    }, 
    qcCat = "MS", 
    qcName = "MS^2*Scans:~TopN~ID~over~N", 
    orderNr = 0380
  )
    return(.self)
  })
)


#####################################################################

qcMetric_MSMSScans_DepPep =  setRefClass(
  "qcMetric_MSMSScans_DepPep",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate = 
      "Prominent dependent peptides are shown, and how they contribute to increase identification numbers.
You can use this metric to assess if specifying another variable (or even fixed) modification makes sense to boost
your ID rate (remember that dependent peptides are only an add-on in MaxQuant and do not count towards global ID rates or
quantification!).
You can also the use the DP-modifications to compare samples (e.g. with modified sample preparation or 
biological conditions where you expect drastic changes).

Hits with 'Unmodified' are removed since they do not provide a lot of additional information.
The legend provides the top target sites (AAs) for each modification in percent
(e.g. 'Oxidation - nterm(8) L(9) H(7)' means that of all dependent peptides with Oxidation, 
8% of there are nterm, 9% are on L, 7% on H).
    
Heatmap score [MS<sup>2</sup> Scans: DepPep]: No score.
",
    workerFcn = function(.self, d_msmsScan)
    {
      ## completeness check
      if (!checkInput(c("fc.raw.file", "dp.modification", "dp.aa", "identified"), d_msmsScan)) return()
      stopifnot(unique(d_msmsScan$identified) %in% c("-","+",""))

      ## modified subset
      d_msmsScan$hasDP = (d_msmsScan$dp.modification != "") & (tolower(d_msmsScan$dp.modification) != "unmodified")
      d_dp = data.table::as.data.table(d_msmsScan[d_msmsScan$hasDP,])
      
      ## pick global top-5 modifications
      d_dp.mods.top = sort(table(d_dp$dp.modification), decreasing = TRUE)[1:5]
      names(d_dp.mods.top)
      ## set all other DPs to 'other'
      d_dp$dp.modification[!(d_dp$dp.modification %in% names(d_dp.mods.top))] = "other"
      dp.names = c(names(d_dp.mods.top), "other")
      
      ## pick top 3 AA's per mod
      aa.annot = NULL
      for (m in names(d_dp.mods.top)) {
        #m = names(d_dp.mods.top)[1]
        d_dp.mods.aa = d_dp[ dp.modification == m,]$dp.aa
        d_dp.mods.aa.sort = sort(table(unlist(strsplit(paste(d_dp.mods.aa, collapse=";"), ";"))), decreasing = TRUE)
        d_dp.mods.aa.sort.top3 = round(d_dp.mods.aa.sort / sum(d_dp.mods.aa.sort) * 100)[1:3]
        r = paste(names(d_dp.mods.aa.sort.top3), d_dp.mods.aa.sort.top3, sep="(", collapse=") ")
        aa.annot[m] = paste0(m, "\n", r, ")")
      }

      ## use data.table for aggregation, its MUCH faster than ddply() and uses almost no extra memory
      d_dp.mods = d_dp[, list(n=.N), by=c("fc.raw.file", "dp.modification")]
      tail(d_dp.mods)
      ## sort mods by # of occurences
      d_dp.mods.sort = plyr::ddply(d_dp.mods, "dp.modification", function(x) sum(x$n))
      d_dp.mods.sort.name = d_dp.mods.sort$dp.modification[order(d_dp.mods.sort$V1)]
      d_dp.mods$dp.modification = factor(d_dp.mods$dp.modification, levels = d_dp.mods.sort.name)

      d_noDP_id = data.table::as.data.table(d_msmsScan[d_msmsScan$identified=="+" & !d_msmsScan$hasDP,])[, list(n=.N), by=c("fc.raw.file")]
      
      d_dp.mods$n_noDP = d_noDP_id$n[match(d_dp.mods$fc.raw.file, d_noDP_id$fc.raw.file)]
      d_dp.mods$n_percent = d_dp.mods$n / d_dp.mods$n_noDP * 100
      

      plotDPMods = function(d_dp.mods){
        p = ggplot(d_dp.mods) +
              geom_line(aes(fc.raw.file, n_percent, col=dp.modification, group=dp.modification), linewidth = 1.5) +
              scale_color_manual(values = brewer.pal.Safe(n = length(dp.names), palette = "Accent"),
                                 labels = aa.annot) +
              guides(color = guide_legend(title = "modification\n(sites in percent)")) +
              ggtitle("Dependent peptides by modification type") +
              xlab("Raw file") +
              ylab("#DP / #non-DP [%]") +
              scale_x_discrete_reverse(d_dp.mods$fc.raw.file) +
              coord_flip()
        return(p)
      }

      lpl = byXflex(d_dp.mods, d_dp.mods$fc.raw.file, 15, plotDPMods, sort_indices = FALSE)

      ## QC measure: NA
      #qc_TopN_ID = ddply(df.ratio, "fc.raw.file", function(x) data.frame(val = qualUniform(x$ratio, x$count)))
      #colnames(qc_TopN_ID)[colnames(qc_TopN_ID) == "val"] = .self$qcName
      
      return(list(plots = lpl))
    }, 
    qcCat = "MS", 
    qcName = "MS^2*Scans:~Dependent Peps", 
    orderNr = 0383
  )
    return(.self)
  })
)

qcMetric_PG_Cont =  setRefClass(
  "qcMetric_PG_Cont",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpText = 
      "External protein contamination should be controlled for, therefore MaxQuant ships with a 
comprehensive, yet customizable protein contamination database, which is searched by MaxQuant by default. PTXQC 
generates a contamination plot derived from the proteinGroups (PG) table showing the fraction of total 
protein intensity attributable to contaminants. The plot employs transparency to discern differences in
the group-wise summed protein abundance. This allows to delineate a high contamination in high complexity samples from a high 
contamination in low complexity samples (e.g. from in-gel digestion). If you see only one 
abundance class ('mid'), this means all your groups have roughly
the same summed protein intensity.
Note that this plot is based on experimental groups, and therefore may not correspond 1:1 to Raw files. 
    
Heatmap score: none (since data source proteinGroups.txt is not related 1:1 to Raw files)
",
    workerFcn=function(.self, df_pg, int_cols, MAP_pg_groups)
    {
      ## completeness check
      if (!checkInput(c(int_cols, "contaminant"),df_pg)) return()
      
      df.con_stats = plyr::adply(int_cols, .margins=1, function(group) {
        #cat(group)
        total_int = sum(as.numeric(df_pg[, group]), na.rm = TRUE)
        return(data.frame(group_long = as.character(group),
                          log10_int  = log10(total_int),
                          cont_pc    = sum(as.numeric(df_pg[df_pg$contaminant, group]), na.rm = TRUE) /
                            total_int * 100
        ))
      })
      df.con_stats$group = MAP_pg_groups$short[match(df.con_stats$group_long, MAP_pg_groups$long)]
      df.con_stats$logAbdClass = getAbundanceClass(df.con_stats$log10_int)
      df.con_stats
      
      pg_plots_cont = byXflex(df.con_stats, 1:nrow(df.con_stats), 90, plot_ContsPG, sort_indices = FALSE)
      
      return(list(plots = pg_plots_cont))
    }, 
    qcCat = "Prep", 
    qcName = "PG:~Contaminants", 
    orderNr = 0110  ## should not show up in heatmap
  )
    return(.self)
  })
)

#####################################################################

qcMetric_PG_RawInt =  setRefClass(
  "qcMetric_PG_RawInt",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpText = 
      "Intensity boxplots by experimental groups. Groups are user-defined during MaxQuant configuration.
This plot displays a (customizable) threshold line for the desired mean intensity of proteins. Groups
which underperform here, are likely to also suffer from a worse MS/MS id rate and higher contamination due to
the lack of total protein loaded/detected. If possible, all groups should show a high and consistent amount
of total protein.
The height of the bar correlates to the number of proteins with non-zero abundance.

Contaminants are shown as overlayed yellow boxes, whose height corresponds to the number of contaminant proteins.
The position of the box gives the intensity distribution of the contaminants.

Heatmap score: none (since data source proteinGroups.txt is not related 1:1 to Raw files)
",
    workerFcn=function(.self, df_pg, int_cols, MAP_pg_groups, thresh_intensity)
    {
      ## completeness check
      if (!checkInput(c(int_cols, "contaminant"),df_pg)) return()
      
      
      ## some stats (for plot title)
      medians_no0 = sort(apply(log2(df_pg[, int_cols, drop = FALSE]), 2, function(x) quantile(x[x>0], probs=0.5, na.rm = TRUE))) # + c(0,0,0,0,0,0))
      int_dev_no0 = RSD(medians_no0)
      ## do not remove zeros (but add +1 since RSD is 'NA' when 'inf' is included in log-data)
      medians = sort(apply(log2(df_pg[, int_cols, drop = FALSE]+1), 2, quantile, na.rm = TRUE, probs=0.5)) # + c(0,0,0,0,0,0))
      int_dev = RSD(medians)
      int_dev.s = pastet("INT RSD [%]", round(int_dev, 3))
      lpl = boxplotCompare(data = reshape2::melt(df_pg[, c(int_cols, "contaminant"), drop = FALSE], id.vars=c("contaminant"))[,c(2,3,1)],
                           log2 = TRUE, 
                           mainlab = "PG: intensity distribution",
                           ylab = expression(log[2]*" intensity"),
                           sublab = paste0("RSD ", round(int_dev_no0, 1),"% (w/o zero int.; expected < 5%)\n",
                                           "RSD ", round(int_dev, 1),"% [high RSD --> few peptides])"),
                           abline = thresh_intensity,
                           names = MAP_pg_groups)
      
      return(list(plots = lpl))
    }, 
    qcCat = "prep", 
    qcName = "PG:~Raw~intensity", 
    orderNr = 0032
  )
    return(.self)
  })
)

#####################################################################

qcMetric_PG_LFQInt =  setRefClass(
  "qcMetric_PG_LFQInt",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpText = 
      "Label-free quantification (LFQ) intensity boxplots by experimental groups. Groups are user-defined during MaxQuant configuration.
This plot displays a (customizable) threshold line for the desired mean of LFQ intensity of proteins. Raw files
which underperform in *Raw* intensity, are likely to show an *increased* mean here, since
only high-abundance proteins are recovered and quantifyable by MaxQuant in this Raw file. The remaining proteins
are likely to receive an LFQ value of 0 (i.e. do not contribute to the distribution).
The height of the bar correlates to the number of proteins with non-zero abundance.

Contaminants are shown as overlayed yellow boxes, whose height corresponds to the number of contaminant proteins.
The position of the box gives the intensity distribution of the contaminants.

Heatmap score: none (since data source proteinGroups.txt is not related 1:1 to Raw files)
",
    workerFcn=function(.self, df_pg, int_cols, MAP_pg_groups, thresh_intensity)
    {
      ## completeness check
      if (!checkInput(c(int_cols, "contaminant"),df_pg)) return()
      
      
      ## some stats (for plot title)
      medians_no0 = sort(apply(log2(df_pg[, int_cols, drop = FALSE]), 2, function(x) quantile(x[x>0], probs=0.5, na.rm = TRUE))) # + c(0,0,0,0,0,0))
      int_dev_no0 = RSD(medians_no0)
      ## do not remove zeros (but add +1 since RSD is 'NA' when 'inf' is included in log-data)
      medians = sort(apply(log2(df_pg[, int_cols, drop = FALSE]+1), 2, quantile, na.rm = TRUE, probs=0.5)) # + c(0,0,0,0,0,0))
      int_dev = RSD(medians)
      lpl = boxplotCompare(data = reshape2::melt(df_pg[, c(int_cols, "contaminant"), drop = FALSE], id.vars=c("contaminant"))[,c(2,3,1)],
                           log2 = TRUE, 
                           mainlab = "PG: LFQ intensity distribution",
                           ylab = expression(log[2]*" intensity"),
                           sublab = paste0("RSD ", round(int_dev_no0, 1),"% (w/o zero int.; expected < 5%)\n",
                                           "RSD ", round(int_dev, 1),"% [high RSD --> few peptides])"),
                           abline = thresh_intensity,
                           names = MAP_pg_groups)
      
      return(list(plots = lpl))
    }, 
    qcCat = "prep",
    qcName = "PG:~LFQ~intensity", 
    orderNr = 0033  ## should not show up in heatmap
  )
    return(.self)
  })
)

#####################################################################

qcMetric_PG_ReporterInt =  setRefClass(
  "qcMetric_PG_ReporterInt",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpText = 
      "ITRAQ/TMT reporter intensity boxplots by experimental groups. Groups are user-defined during MaxQuant configuration.
This plot displays a (customizable) threshold line for the desired mean of reporter ion intensity of proteins.
The height of the bar correlates to the number of proteins with non-zero abundance.

Contaminants are shown as overlayed yellow boxes, whose height corresponds to the number of contaminant proteins.
The position of the box gives the intensity distribution of the contaminants.
Contaminants should be lower compared to label-free samples, since all contaminants introduced after the labeling 
should not be identified by Andromeda (since they lack the isobaric tag).
    
There is a similar 'Raw file' based metric/plot based on evidence.txt.

Heatmap score: none (since data source proteinGroups.txt is not related 1:1 to Raw files)
",
    workerFcn=function(.self, df_pg, int_cols, MAP_pg_groups, thresh_intensity)
    {
      ## completeness check
      if (!checkInput(c(int_cols, "contaminant"),df_pg)) return()
      
      
      ## some stats (for plot title)
      medians_no0 = sort(apply(log2(df_pg[, int_cols, drop = FALSE]), 2, function(x) quantile(x[x>0], probs=0.5, na.rm = TRUE))) # + c(0,0,0,0,0,0))
      reprt_dev_no0 = RSD(medians_no0)
      ## do not remove zeros (but add +1 since RSD is 'NA' when 'inf' is included in log-data)
      medians = sort(apply(log2(df_pg[, int_cols, drop = FALSE]+1), 2, quantile, probs=0.5, na.rm = TRUE)) # + c(0,0,0,0,0,0))
      reprt_dev = RSD(medians)
      lpl = boxplotCompare(   data = reshape2::melt(df_pg[, c(int_cols, "contaminant"), drop = FALSE], id.vars=c("contaminant"))[,c(2,3,1)],
                              log2 = TRUE, 
                              ylab = expression(log[2]*" reporter intensity"),
                              mainlab = "PG: reporter intensity distribution",
                              sublab = paste0("RSD ", round(reprt_dev_no0, 1),"% (w/o zero int.; expected < 5%)\n",
                                              "RSD ", round(reprt_dev, 1),"% [high RSD --> few peptides])"),
                              abline = thresh_intensity,
                              names = MAP_pg_groups)
      
      return(list(plots = lpl))
    }, 
    qcCat = "prep", 
    qcName = "PG:~Reporter~intensity", 
    orderNr = 0034  ## should not show up in heatmap
  )
    return(.self)
  })
)

#####################################################################

qcMetric_PG_PCA =  setRefClass(
  "qcMetric_PG_PCA",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpText = 
      "Principal components plots of experimental groups (as defined during MaxQuant configuration).

This plot is shown only if more than one experimental group was defined.
If LFQ was activated in MaxQuant, an additional PCA plot for LFQ intensities is shown. Similarly, if iTRAQ/TMT
reporter intensities are detected.

Since experimental groups and Raw files do not necessarily correspond 1:1, this plot cannot use the abbreviated
Raw file names, but instead must rely on automatic shortening of group names.

Heatmap score: none (since data source proteinGroups.txt is not related 1:1 to Raw files)
",
    workerFcn=function(.self, df_pg, lst_cols, MAP_pg_groups)
    {
      ## completeness check
      if (!checkInput(c(unlist(lst_cols), "contaminant"),df_pg)) return()
      
      lpl = list()
      for (cond in names(lst_cols))
      {
        #cond = names(lst_cols)[3]
        #print(lst_cols[cond])
        if (length(lst_cols[[cond]]) <= 1) 
        { ## only one condition.. PCA does not make sense (and will not work)
          next;
        }
        ## remove contaminants
        data = t(df_pg[!df_pg$contaminant, unlist(lst_cols[cond]), drop = FALSE])
        ## remove constant/zero columns (== dimensions == proteins)
        data = data[, colSums(data, na.rm = TRUE) > 0, drop = FALSE]
        rownames(data) = MAP_pg_groups$short[match(rownames(data), MAP_pg_groups$long)]
        pl = try(getPCA(data = data,
                        gg_layer = ggtitle(paste0("PG: PCA of '", sub(".", " ", cond, fixed = TRUE), "'"), "(excludes contaminants)")
                       )[["plots"]])
        #print(pl)
        if (!inherits(pl, "try-error")) lpl = append(lpl, pl);
      }
      
      return(list(plots = lpl))
    }, 
    qcCat = "General", 
    qcName = "PG:~Principal~Component", 
    orderNr = 0003
  )
    return(.self)
  })
)

#####################################################################

qcMetric_PG_Ratio =  setRefClass(
  "qcMetric_PG_Ratio",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpText = 
      "This plot shows log<sub>2</sub> ratios for SILAC-like experiments (whenever MaxQuant reports a set of 'ratio.*' columns).
Useful to spot unequal channel mixing during sample preparation. If equal mixing is expected, the 
distribution should be unimodal and its mode close to 1 (i.e., a 1:1 ratio), as indicated by a visual 
guidance line. Multimodal distributions are flagged as such automatically. If PTXQC detects ratios 
deviating strongly from 1:1 (parameterized by default beyond the range between 1:4 and 4:1), PTXQC 
automatically assumes a pulsed experiment and reports the label incorporation in percent for all groups. 

Heatmap score: none (since data source proteinGroups.txt is not related 1:1 to Raw files)
",
    workerFcn = function(.self, df_pg, ratio_cols, thresh_LabelIncorp, GL_name_min_length)
    {
      ## completeness check
      if (!checkInput(c(ratio_cols, "contaminant", "reverse"),df_pg)) return()
      
      ## remove reverse and contaminants (might skew the picture)
      idx_row = !df_pg$contaminant & !df_pg$reverse
      d_sub = log2(df_pg[idx_row, ratio_cols, drop = FALSE])
      ## rename "ratio.h.l" to "h.l" (same for m.l in tripleSILAC)
      idx_globalRatio = grep("ratio\\.[hm]\\.l$", colnames(d_sub))
      if (length(idx_globalRatio)) colnames(d_sub)[idx_globalRatio] = gsub("^ratio\\.", "", colnames(d_sub)[idx_globalRatio])
      ## simplify the rest
      if (ncol(d_sub) > length(idx_globalRatio))
      {
        idx_other = setdiff(1:ncol(d_sub), idx_globalRatio)
        colnames(d_sub)[idx_other] = shortenStrings(simplifyNames(delLCP(colnames(d_sub)[idx_other], 
                                                                         min_out_length = GL_name_min_length, 
                                                                         add_dots = TRUE), 
                                                                  min_out_length = GL_name_min_length))
      }
      #summary(d_sub)
      # 
      # plot(density(d_sub[,1], bw = "SJ", adjust=1, na.rm = TRUE, n=128))
      # plot(d_sub[,1])
      # h = density(d_sub[,1], bw = "SJ", adjust=1, na.rm = TRUE, n=128)
      
      
      ## get ranges to fix breaks for density intervals
      # breaks = seq(min(d_sub, na.rm = TRUE), max(d_sub, na.rm = TRUE), length.out=(max(dd, na.rm = TRUE)-min(dd, na.rm = TRUE))/0.5)
      # mid = hist(d_sub[, 1], breaks = breaks)$mids
      ratio.densities = do.call(rbind, (lapply(1:ncol(d_sub), function(x) {
        name = colnames(d_sub)[x]
        ## density estimation can fail if not enough data
        h = try(density(na.omit(d_sub[ ,x]), bw = "SJ", adjust=2, na.rm = TRUE), silent = TRUE)
        if (inherits(h, "try-error")) return (NULL)
        count = sum(getMaxima(h$y))
        if (count > 1) name = paste(name, "*")
        df = data.frame(x = h$x, y = h$y, col = name, multimodal = (count>1))
        return (df)
      })))
      if (is.null(ratio.densities)) {
        pl_cont = ggText("PG: ratio density",
                         paste0("No data for plotting!"),
                         "red")
        return(list(plots = list(pl_cont)))
      } 
      
      ratio.densities$alpha = c(0.8, 1)[ratio.densities$multimodal+1]
      ratio.densities$ltype = c("dotted", "solid")[ratio.densities$multimodal+1]
      #head((ratio.densities))
      
      
      ## compute label incorporation?
      ratio.mode = plyr::ddply(ratio.densities, "col", .fun = function(x) {
        mode = x$x[which.max(x$y)]
        return (data.frame(mode = mode))
      })
      
      ## on more than ratio 1:4 or 4:1 ratio, report label incorporation
      # thresh_LabelIncorp == 4 by default
      if (max(abs(ratio.mode$mode)) > abs(log2(thresh_LabelIncorp))) {
        cat(paste0("Maximum ratio (log2) was ", round(max(abs(ratio.mode$mode)),1) , ", reaching the threshold of ", abs(log2(thresh_LabelIncorp)), " for label-incorporation computation.\nComputing ratios ...\n"))
        ## back to normal scale
        ratio.norm = 2^ratio.mode$mode
        ## compute incorporation
        ratio.inc =  ratio.norm / (ratio.norm+1) * 100
        ## round
        ratio.mode$li = round(ratio.inc)
        ## new label
        ratio.mode$col_new = ratio.mode$col %+% " (" %+% ratio.mode$li %+% "%)"
        ## replace column names in data.frame
        ratio.densities$col = ratio.mode$col_new[match(ratio.densities$col, ratio.mode$col)]
        ## notify user via legend title
        legend_title = "group\n(with label inc)"
      } else
      {
        cat(paste0("Maximum ratio (log2) was ", max(abs(ratio.mode$mode)), ", NOT reaching the threshold of ", abs(log2(thresh_LabelIncorp)), " for label-incorporation computation.\nSkipping ratios ...\n"))
        legend_title = "group"
      }
      
      main_title = "PG: ratio density\n(w/o contaminants)"
      main_col = "black"
      if (any(ratio.densities$multimodal))
      {
        main_title = paste0(main_title, "\nWarning: multimodal densities detected")
        main_col = "red"
      }
      ##
      ## plot ratios
      ##
      lpl = 
        byXflex(ratio.densities, ratio.densities$col, 5, plot_RatiosPG, sort_indices = FALSE, d_range = range(d_sub, na.rm=TRUE), main_title, main_col, legend_title)
      
      return(list(plots = lpl))
    },
    qcCat = "prep", 
    qcName = "PG:~Ratio", 
    orderNr = 0019
  )
    return(.self)
  })
)

#####################################################################

#' Creates a yaml file storing the parameters that are used for creating the PTXQC report 
#' and returns these parameters as well as a list of available qc-Metrics objects.
#' 
#' Valid parameters are: 
#'    param_useMQPAR, add_fs_col, id_rate_bad, id_rate_great , pg_ratioLabIncThresh , param_PG_intThresh,
#'    param_EV_protThresh , param_EV_intThresh, param_EV_pepThresh , yaml_contaminants, param_EV_MatchingTolerance,
#'    param_evd_mbr , param_EV_PrecursorTolPPM, param_EV_PrecursorOutOfCalSD , param_EV_PrecursorTolPPMmainSearch, 
#'    param_MSMSScans_ionInjThresh, param_OutputFormats and param_PageNumbers 
#'    
#'    Please provide them as a list() of this format: list$parameter_name
#'
#'
#' @param yc A yaml class object created by YAMLClass$new()
#' @param param list of parameters sorted by names; if empty, will be populated with defaults
#' @param DEBUG_PTXQC print some debugging information; default FALSE
#' @param txt_files list of paths to MaxQuant files; if NULL, it is assumed that the parameters are for mzTab-mode
#' @param metrics list of metric names that should be plotted; if NULL, will be populated with defaults
#' @return list of parameters used for creating the report and list of qc-Metrics objects
#' @export
#'
#'
createYaml <- function(yc, param = list(), DEBUG_PTXQC = FALSE, txt_files = NULL, metrics = NULL){

  ##
  ## YAML default config
  ##
  default_param <- list()
  default_param$param_useMQPAR <- TRUE
  default_param$add_fs_col <- 14 
  default_param$id_rate_bad <- 20
  default_param$id_rate_great <- 35
  default_param$pg_ratioLabIncThresh <- 4
  default_param$param_PG_intThresh <- 25
  default_param$param_EV_protThresh <- 3500
  default_param$param_EV_intThresh <- 23
  default_param$param_EV_pepThresh <- 15000
  default_param$yaml_contaminants <- list("cont_MYCO" = c(name="MYCOPLASMA", threshold=1)) # name (FASTA), threshold for % of unique peptides
  default_param$param_EV_MatchingTolerance <- 1
  default_param$param_evd_mbr <- "auto"
  default_param$param_EV_PrecursorTolPPM <- 20
  default_param$param_EV_PrecursorOutOfCalSD <- 2
  default_param$param_EV_PrecursorTolPPMmainSearch <- NA
  default_param$param_MSMSScans_ionInjThresh <- 10
  default_param$param_OutputFormats <- c("html", "plainPDF")
  default_param$param_PageNumbers <- "on"
  
  
  ##
  ##add missing parameters from default parameter list
  ##
  for(i in c(1:length(default_param))){
    if(!names(default_param)[i] %in% names(param)) param[names(default_param)[i]] <- default_param[i]
  }

  ##
  ## check for invalid parameters
  ##
  for(i in c(length(param):1)){
    if(!names(param)[i] %in% c(NA, names(default_param))) {
      warning(paste0("Invalid parameter detected and removed: ", names(param)[i]))
      param <- param[-i] 
      }
  }
  
  
  ## determines if a local mqpar.xml should be used to grep all YAML parameters whose name starts with "MQpar_" from the
  ## original mqpar.xml instead of the yaml.config. The "MQpar_..." param from the config
  ## will be ignored and the newly written yaml.config will contain the values from mqpar.xml.
  param$param_useMQPAR = yc$getYAML("PTXQC$UseLocalMQPar", param$param_useMQPAR)
  
  param$add_fs_col = yc$getYAML("PTXQC$NameLengthMax_num", param$add_fs_col)
  
  param$id_rate_bad = yc$getYAML("File$Summary$IDRate$Thresh_bad_num", param$id_rate_bad, 0, 100)
  param$id_rate_great = yc$getYAML("File$Summary$IDRate$Thresh_great_num", param$id_rate_great, 0, 100)
  
  param$GL_name_min_length = 8
  
  param$pg_ratioLabIncThresh = yc$getYAML("File$ProteinGroups$RatioPlot$LabelIncThresh_num", param$pg_ratioLabIncThresh)
  ## default median intensity in log2 scale
  param$param_PG_intThresh = yc$getYAML("File$ProteinGroups$IntensityThreshLog2_num", param$param_PG_intThresh, 1, 100)
  
  ## get scoring threshold (upper limit)
  param$param_EV_protThresh = yc$getYAML("File$Evidence$ProteinCountThresh_num", param$param_EV_protThresh, 1, 1e5)
  
  ## default median intensity in log2 scale
  param$param_EV_intThresh = yc$getYAML("File$Evidence$IntensityThreshLog2_num", param$param_EV_intThresh, 1, 100)
  
  ## get scoring threshold (upper limit)
  param$param_EV_pepThresh = yc$getYAML("File$Evidence$PeptideCountThresh_num", param$param_EV_pepThresh, 1, 1e6)
  
  ### warn of special contaminants!
  ## these need to be in FASTA headers (description is not enough)!
  ## syntax:  list( contaminant1 = c(name, threshold), contaminant2 = c(...), ...)
  ##
  ##  if within the YAML file
  ##    SpecialContaminants: no
  ##  is set, then 'yaml_contaminants' will be 'FALSE'
  ##
  ##contaminant_default = FALSE ## to switch it off by default
  
  param$yaml_contaminants = yc$getYAML("File$Evidence$SpecialContaminants", param$yaml_contaminants)
  
  param$param_EV_MatchingTolerance = yc$getYAML("File$Evidence$MQpar_MatchingTimeWindow_num", param$param_EV_MatchingTolerance)
  if (param$param_useMQPAR && !is.null(txt_files)) {
    v = getMQPARValue(txt_files$mqpar, "//matchingTimeWindow") ## will also warn() if file is missing
    if (!is.null(v)) {
      param$param_EV_MatchingTolerance = yc$setYAML("File$Evidence$MQpar_MatchingTimeWindow_num", as.numeric(v))
    }
  }
  param$param_evd_mbr = yc$getYAML("File$Evidence$MatchBetweenRuns_wA", param$param_evd_mbr)
  
  param$param_EV_PrecursorTolPPM = yc$getYAML("File$Evidence$MQpar_firstSearchTol_num", param$param_EV_PrecursorTolPPM)
  if (param$param_useMQPAR && !is.null(txt_files)) {
    v = getMQPARValue(txt_files$mqpar, "//firstSearchTol") ## will also warn() if file is missing
    if (!is.null(v)) {
      param$param_EV_PrecursorTolPPM = yc$setYAML("File$Evidence$MQpar_firstSearchTol_num", as.numeric(v))
    }
  }
  
  param$param_EV_PrecursorOutOfCalSD = yc$getYAML("File$Evidence$firstSearch_outOfCalWarnSD_num", param$param_EV_PrecursorOutOfCalSD)
  
  ## we do not dare to have a default, since it ranges from 6 - 4.5 ppm across MQ versions
  param$param_EV_PrecursorTolPPMmainSearch = yc$getYAML("File$Evidence$MQpar_mainSearchTol_num", param$param_EV_PrecursorTolPPMmainSearch)
  if (param$param_useMQPAR && !is.null(txt_files)) {
    v = getMQPARValue(txt_files$mqpar, "//mainSearchTol") ## will also warn() if file is missing
    if (!is.null(v)) {
      param$param_EV_PrecursorTolPPMmainSearch = yc$setYAML("File$Evidence$MQpar_mainSearchTol_num", as.numeric(v))
    }
  }
  if (is.na(param$param_EV_PrecursorTolPPMmainSearch))
  {
    #warning("PTXQC: Cannot draw valid range borders for calibrated mass error, since neither 'File$Evidence$MQpar_mainSearchTol_num' is set nor a mqpar.xml file is present!", immediate. = TRUE)
  }
  
  param$param_MSMSScans_ionInjThresh = yc$getYAML("File$MsMsScans$IonInjectionThresh_num", param$param_MSMSScans_ionInjThresh, 0, 200)
  
  param$param_OutputFormats = yc$getYAML("PTXQC$OutputFormats", param$param_OutputFormats)
  
  param$param_PageNumbers = yc$getYAML("PTXQC$PlainPDF$AddPageNumbers", param$param_PageNumbers)
  
  
  ####
  ####  prepare the metrics
  ####
  lst_qcMetrics = getMetricsObjects(DEBUG_PTXQC)
  df.meta = getMetaData(lst_qcMetrics = lst_qcMetrics)
  df.meta
  ## reorder metrics (required for indexing below!)
  lst_qcMetrics_ord = lst_qcMetrics[df.meta$.id]
  
  ## write/update order from YAML
  for (i in 1:nrow(df.meta))
  {
    #cat(paste("meta id: ", df.meta$.id[i], "\n"))
    pname = paste0("order$", df.meta$.id[i])
    pval = df.meta$order[i]
    
    ##check for shiny metrics input
    if(!is.null(metrics)){
      if(!(df.meta$.id[i] %in% metrics)) pval = -1; ## omit the metric if not listed`
    }
    
    param_v = yc$getYAML(pname, pval) ## read value (if present), or return `pval`
    ## update
    if (is.numeric(param_v)) {
      lst_qcMetrics_ord[[i]]$orderNr = param_v  # for some reason, lst_qcMetrics[[df.meta$.id]] does not work
    } else {
      stop("YAML param '", pname, "' is not numeric (", param_v, "). Please fix the YAML configuration!")
    }
  }
  ## re-read meta (new ordering?)
  df.meta = getMetaData(lst_qcMetrics = lst_qcMetrics)
  ## reorder metrics (again; after param update)
  lst_qcMetrics_ord = lst_qcMetrics[df.meta$.id]
  
  return(list("yc" = yc, "param" = param, "lst_qcMetrics" = lst_qcMetrics_ord))
  
  
}

library(PTXQC)
context("createYaml.R")

test_that("createYaml", {

  ##
  ##test empty yc object, no parameter or metric specifications
  ##
  
  yc <- YAMLClass$new(list())
  expect_equal(length(createYaml(yc)$param), 19)

  ##
  ##test invalid parameter input
  ##
  
  yc <- YAMLClass$new(list())
  
  parameter <- list()
  parameter$nonsense1 <- c(1,4)
  parameter$nonsense2 <- "test"
  parameter$param_PG_intThresh <- 30
  
  expect_null(createYaml(yc, param = parameter)$param$nonsense1)
  expect_null(createYaml(yc, param = parameter)$param$nonsense2)
  expect_equivalent(createYaml(yc, param = parameter)$param$param_PG_intThresh, 30)
  expect_equal(length(createYaml(yc)$param), 19)
  
  ##
  ##test valid parameter input
  ##
  
  yc <- YAMLClass$new(list())
  
  parameter$add_fs_col <- 14
  parameter$param_OutputFormats <- "txt"
  
  expect_equivalent(createYaml(yc, param = parameter)$param$add_fs_col, 14)
  expect_equivalent(createYaml(yc, param = parameter)$param$param_OutputFormats, "txt")
  
  ##test default values
  yc <- YAMLClass$new(list())
  
  expect_equivalent(createYaml(yc)$param$param_PG_intThresh, 25)
  expect_equivalent(createYaml(yc)$param$param_OutputFormats, c("html", "plainPDF"))

  
  ##
  ##test metrics deactivation (all except qcMetric_PAR)
  ##
  yc <- YAMLClass$new(list())
  
  mets <- "qcMetric_PAR" 
  expect_equal(createYaml(yc, param = parameter, metrics = mets)$yc$yamlObj$order$qcMetric_EVD_UpSet, -1)
  expect_equal(createYaml(yc, param = parameter, metrics = mets)$yc$yamlObj$order$qcMetric_PAR, 1)
  
  ##test no deactivation 
  yc <- YAMLClass$new(list())
  expect_false(any(createYaml(yc)$yc$yamlObj$order < 1))
  
})

#'
#'  Make sure to call $readMappingFile(some_file) if you want to support a user-defined file mapping.
#'  Otherwise, calls to $getShortNames() will create/augment the mapping for filenames.
#'
#'
#'
#' @field raw_file_mapping Data.frame with columns 'from', 'to' and maybe 'best.effort' (if shorting was unsuccessful)
#' @field mapping.creation how the current mapping was obtained (user or auto)
#' @field external.mapping.file Filename of user-defined mapping file; only defined if readMappingFile() was called
#'
#' @import ggplot2
#'
#' @exportClass FilenameMapper
#' @export FilenameMapper
#'
#' @examples
#' a = FilenameMapper$new()
#' a$readMappingFile('filenamemapping.txt')
#'
FilenameMapper = setRefClass("FilenameMapper",

                       fields = list(raw_file_mapping = "data.frame", ## with cols 'from', to' and maybe 'best.effort' (if shorting was unsuccessful)
                                     mapping.creation = "character",  ## how the current mapping was obtained (user or auto)
                                     external.mapping.file = "character" ## filename of user-defined mapping file; only defined if readMappingFile() was called
                       ),
                       methods = list(

initialize=function() {
 .self$raw_file_mapping = data.frame()
 .self$mapping.creation = NA_character_
 .self$external.mapping.file = NA_character_
 return(.self)
},

specrefToRawfile = function(.self, specrefs)
{
  "Return a DF with 'ms_run', 'raw.file' and 'fc.raw.file' given a vector of spectraReferences, e.g. 'ms_run[1]:...', '...'"

  res = data.frame(ms_run = sub("[.]*:.*", "\\1", specrefs))
  return (cbind(res, .self$msrunToRawfile(res$ms_run)))
},

msrunToRawfile = function(.self, ms_runs)
{
  "Given a vector of ms_runs, c('ms_run[1]', ...), return a data.frame of identical length with columns 'raw.file' and 'fc.raw.file'."

  if (!"ms.run" %in% colnames(.self$raw_file_mapping)) stop("Mapping is missing 'ms.run' from mzTab!")

  res = .self$getRawfm()[ match(ms_runs, .self$raw_file_mapping$ms.run), c("from", "to")]
  colnames(res) = c("raw.file", "fc.raw.file")
  return (res)
},

getShortNames = function(.self, raw_filenames, max_length = 10, ms_runs = NULL)
{
  "Uses the internal mapping (or re-creates it if current one is incomplete) and maps the input raw names to shorter output names.
    Returns a vector of the same length."
  #rf <<- raw_filenames
  #raw_filenames = rf

  if (!is.null(ms_runs) && length(ms_runs) != length(raw_filenames)) stop("raw_filenames and ms_runs do not have the same length!")

  ## for MQ-DIA data, the summary.txt is empty, thus we may receive an empty list here.
  ## --> do nothing and wait for the next .txt file which has proper names
  if (length(raw_filenames) == 0) {
    warning("Empty raw file list in .txt file. Skipping mapping for now.")
    return (list())
  }

  cat(paste0("Adding fc.raw.file column ..."))
  ## if there is no mapping, or if its incomplete (outdated mapping file)
  has_mapping = (nrow(.self$raw_file_mapping) != 0)
  incomplete_mapping = has_mapping && any(is.na(match(raw_filenames, .self$raw_file_mapping$from)))
  if (!has_mapping || incomplete_mapping)
  {
    ## if the mapping is 'auto', we got handed an incomplete/different txt file before, which does not match
    ## the current file. Some files got mixed up, so we stop!
    if (incomplete_mapping)
    {
      if (is.na(.self$mapping.creation))
      {
        stop("mapping.creation member not properly initialized!")
      }
      ## we had NA's in auto mode ... bad
      if (.self$mapping.creation == .self$getMappingCreation()['auto'])
      { ## mapping is incomplete
        missing = unique(raw_filenames[is.na(match(raw_filenames, .self$raw_file_mapping$from))])
        stop(paste0("Hithero unknown Raw files: ", paste(missing, collapse=", ", sep=""), " encountered which were not present in previous data files.\nDid you mix output files from different analyses?"))
      }
    }
    ## --> redo
    rfm = .self$getShortNamesStatic(unique(raw_filenames), max_length)
    if (!is.null(ms_runs)) {
      rfm$ms.run = ms_runs[ match(rfm$from, raw_filenames)  ]
    }
    ## and remember it
    .self$raw_file_mapping = rfm
    cat("Created a new filename mapping:\n")
    print(rfm)

    ## indicate to outside that a new table is ready
    .self$mapping.creation = .self$getMappingCreation()['auto']
  }
  ## do the mapping
  v.result = as.factor(.self$raw_file_mapping$to[match(raw_filenames, .self$raw_file_mapping$from)])

  cat(paste0(" done\n"))
  return (v.result)
},

getShortNamesStatic = function(raw.files, max_len, fallbackStartNr = 1)
{
  "Static method: Shorten a set of Raw file names and return a data frame with the mappings.
   Mapping will have: $from, $to and optionally $best.effort (if shorting was unsuccessful and numbers had to be used)
   \\itemize{
     \\item{\\verb{raw.files}  Vector of Raw files.}
     \\item{\\verb{max_len} Maximal length of shortening results, before resorting to canonical names (file 1,...).}
     \\item{\\verb{fallbackStartNr} Starting index for canonical names.}
   }
   \\subsection{Return Value}{ data.frame with mapping.}
  "
  rf_name = raw.files
  ## remove prefix
  rf_name_s = delLCP(rf_name,
                     min_out_length = 8,
                     add_dots = TRUE)
  ## remove infix (2 iterations)
  rf_name_s = simplifyNames(rf_name_s,
                            2,
                            min_LCS_length = 7,
                            min_out_length = 8)

  ## check if shorter filenames are still unique (they should be.. if not we have a problem!!)
  if (length(rf_name) != length(unique(rf_name_s)))
  {
    cat("\nOriginal names:\n")
    cat(rf_name)
    cat("\nShort names:\n")
    cat(rf_name_s)
    cat("\n")
    stop("While loading MQ data: shortened raw filenames are not unique! This should not happen. Please contact the developers and provide the above names!")
  }
  df.mapping = data.frame(from = rf_name, to = rf_name_s, stringsAsFactors = FALSE)

  ## always include 'best.effort' column
  df.mapping[, "best.effort"] = df.mapping$to

  ## check if the minimal length was reached
  if (max(nchar(df.mapping$to)) > max_len)
  { ## resort to short naming convention
    cat("Filenames are longer than the maximal allowed size of '" %+% max_len %+% "'. Resorting to short versions 'file X'.\n\n")
    maxl = length(raw.files) - 1 + fallbackStartNr
    df.mapping$to = paste("file", sprintf(paste0("%0", nchar(maxl), "d"), fallbackStartNr:maxl)) ## with leading 0's if required
  }
  return(df.mapping)
},



plotNameMapping = function(.self)
{
  "Plots the current mapping of Raw file names to their shortened version.

   Convenience function to plot the mapping (e.g. to a PDF device for reporting).
   The data frame can be accessed directly via \\verb{.self$raw_file_mapping}.
   If no mapping exists, the function prints a warning to console and returns NULL (which is safe to use in print(NULL)).

   @return if mapping is available, returns a list of plots 'plots' and a Html table string 'htmlTable' ; 'NULL' otherwise.

  "
  if (nrow(.self$raw_file_mapping) == 0)
  {
    cat("No mapping found. Omitting plot.")
    return (NULL);
  }

  table_header = c("original", "short\nname")
  xpos = c(9, 11)
  extra = ""
  has_best_effort = FALSE
  if ("best.effort" %in% colnames(.self$raw_file_mapping))
  {
    has_best_effort = TRUE
    table_header = c(table_header, "best\neffort")
    xpos = c(9, 11, 13)
    if (all(.self$raw_file_mapping$to != .self$raw_file_mapping$best.effort)) {
      extra = "\n(automatic shortening of names was not sufficient - see 'best effort')"
    }

  }

  #mq_mapping = mq$raw_file_mapping
  mq_mapping = .self$raw_file_mapping
  pl_title = "Mapping of Raw files to their short names\nMapping source: " %+% .self$mapping.creation %+% extra;

  mappingChunk = function(mq_mapping)
  {
    mq_mapping$ypos = -(1:nrow(mq_mapping))
    head(mq_mapping)
    ## convert factors to string, because they will all end up in a common 'value' column
    mq_mapping.s = data.frame(lapply(mq_mapping, function(x) if (is.factor(x)) as.character(x) else {x}), stringsAsFactors= FALSE)
    mq_mapping.long = reshape2::melt(mq_mapping.s, id.vars = c("ypos"), value.name = "value")
    head(mq_mapping.long)
    mq_mapping.long$variable = as.character(mq_mapping.long$variable)
    mq_mapping.long$col = "#000000";
    mq_mapping.long$col[mq_mapping.long$variable=="to"] = "#5F0000"
    mq_mapping.long$variable[mq_mapping.long$variable=="from"] = xpos[1]
    mq_mapping.long$variable[mq_mapping.long$variable=="to"] = xpos[2]
    mq_mapping.long$variable[mq_mapping.long$variable=="best.effort"] = xpos[3]
    mq_mapping.long$variable = as.numeric(mq_mapping.long$variable)
    mq_mapping.long$size = 2;

    df.header = data.frame(ypos = 1, variable = xpos, value = table_header, col = "#000000", size=3)
    mq_mapping.long2 = rbind(mq_mapping.long, df.header)
    mq_mapping.long2$hpos = 0 ## left aligned,  1=right aligned
    mq_mapping.long2$hpos[mq_mapping.long2$variable==xpos[1]] = 1
    mq_mapping.long2$hpos[mq_mapping.long2$variable==xpos[2]] = 0

    mqmap_pl = ggplot(mq_mapping.long2, aes(x = .data$variable, y = .data$ypos))  +
      geom_text(aes(label = .data$value), color = mq_mapping.long2$col, hjust = mq_mapping.long2$hpos, size = mq_mapping.long2$size) +
      coord_cartesian(xlim=c(0,20)) +
      theme_bw() +
      theme(plot.margin = grid::unit(c(1,1,1,1), "cm"), line = element_blank(),
            axis.title = element_blank(), panel.border = element_blank(),
            axis.text = element_blank(), strip.text = element_blank(), legend.position = "none") +
      ggtitle(pl_title)
    return(mqmap_pl)
  }
  l_plots = byXflex(mq_mapping, 1:nrow(mq_mapping), 20, mappingChunk, sort_indices = FALSE);
  return (list(plots = l_plots, htmlTable = getHTMLTable(.self$raw_file_mapping, pl_title)))

},

getRawfm = function(.self)
{
  "Wrapper function for member 'raw_file_mapping', ensuring that $to is a factor"

  tmp = .self$raw_file_mapping
  tmp$to = factor(tmp$to)
  return(tmp)
},


readMappingFile = function(.self, filename)
{
  "Reads a mapping table of full Raw file names to shortened names.

  The internal structure \\verb{raw_file_mapping} is created using this file.
  If the file is missing, nothing is done and FALSE is returned.
  If the file contains contradictory information (different set of $from files) compared to
  the current mapping (if present), the internal mapping wins (filemapping is ignored) and FALSE is returned.

  The file must have two columns named: 'orig.Name' and 'new.Name' and use Tab as separator.
  This file can be used to manually substitute Raw file names within the report.
  The ordering of Raw files in the report can be changed by re-arranging the rows.
  I.e.
  \\preformatted{
  orig.Name  new.Name
  2011_05_30_ALH_OT_21_VIL_TMT_FR01   myfile A
  2011_05_30_ALH_OT_22_VIL_TMT_FR02   another B
  }

  @param filename  Source filename to read.
  @return Returns \\verb{TRUE} if file was read, \\verb{FALSE} if it does not exist.
"

  if (file.exists(filename))
  {
    message(paste0("Reading mapping file '", filename, "'\n"))
    dfs = read.delim(filename, comment.char="#", stringsAsFactors = FALSE)
    colnames(dfs) = gsub("_", ".", colnames(dfs)) ## legacy support for old "best_effort" column (now "best.effort")
    req_cols = c(from = "orig.Name", to = "new.Name")
    if (!all(req_cols %in% colnames(dfs)))
    {
      stop("Input file '", filename, "' does not contain the columns '", paste(req_cols, collapse="' and '"), "'.",
           " Please fix and re-run PTXQC!")
    }
    req_cols = c(req_cols, best.effort = "best.effort", ms.run = "ms.run") ## augment
    colnames(dfs) = names(req_cols)[match(colnames(dfs), req_cols)]

    rows_which_are_named_Total = tolower(dfs$from) == "total"
    if (any(rows_which_are_named_Total)) { ## work around bug in PTXQC which treats the 'Total' row in summary.txt as Raw file
      warning(paste0("Removing the 'Total' raw file row (all other rows are unaffected) while reading of name mapping file '", filename, "', to work around a bug in older PTX-QC versions."))
      dfs = dfs[!rows_which_are_named_Total,]
    }

    if (any(duplicated(dfs$from)) | any(duplicated(dfs$to)))
    {
      dups = c(dfs$from[duplicated(dfs$from)], dfs$to[duplicated(dfs$to)])
      stop("Input file '", filename_sorting, "' has duplicate entries ('", paste(dups, collapse=", "), ")'!",
           " Please fix and re-run PTXQC!")
    }
    dfs
    dfs$to = factor(dfs$to, levels = unique(dfs$to), ordered = TRUE) ## keep the order
    dfs$from = factor(dfs$from, levels = unique(dfs$from), ordered = TRUE) ## keep the order
    ## set internal mapping
    if (nrow(.self$raw_file_mapping) > 0 &  ## was initialized before...
        !setequal(.self$raw_file_mapping$from, dfs$from)) ## .. and has different data
    {
      print(paste0("Raw filename mapping in file '", filename, "' has different set of raw files than current data. Mapping file will be ignored and overwritten!",
                   "\nold filenames in mapping:\n  ", paste(dfs$from, collapse="\n  "),
                   "\nnew filenames from data:\n  ", paste(.self$raw_file_mapping$from, collapse="\n  ")))
      return (FALSE)
    }
    .self$raw_file_mapping = dfs
    ## set who defined it
    .self$mapping.creation = .self$getMappingCreation()['user']
    .self$external.mapping.file = filename; ## remember filename for later error messages
    return (TRUE)
  }
  return (FALSE)
},


writeMappingFile = function(.self, filename)
{
  "Writes a mapping table of full Raw file names to shortened names.

  The internal structure \\verb{raw_file_mapping} is written to the
  file specified.
  File is only created if mapping exists (in .self$raw_file_mapping).

  @param filename  Target filename to create.
  @return Returns NULL.
  "
  if (nrow(.self$raw_file_mapping) == 0)
  {
    cat("No mapping found. Writing mapping file '", filename, "' not possible!")
    return (FALSE)
  }

  dfs = data.frame(orig.Name = .self$raw_file_mapping$from, new.Name = .self$raw_file_mapping$to)
  if (nrow(dfs) == 0) return(NULL)

  if ("best.effort" %in% colnames(.self$raw_file_mapping)) {
    dfs$best.effort = .self$raw_file_mapping[, "best.effort"]
  }

  if ("ms.run" %in% colnames(.self$raw_file_mapping)) {
    dfs$ms.run = .self$raw_file_mapping[,"ms.run"]
  }

  ## use a file handle to avoid warning from write.table() when appending
  ## a table with column names 'Warning(): appending column names to file'
  FH = file(filename, "w")
  cat(file = FH,
      "# This file can be used to manually substitute Raw file names within the report.",
      "# The ordering of Raw files in the report can be changed by re-arranging the rows.",
      sep = "\n")
  write.table(x = dfs, file = FH, quote = FALSE, sep="\t", row.names = FALSE)
  close(FH) ## flush
  return (TRUE)
},



getMappingCreation = function(.self)
{
  "A static function"
  return(c(user = 'file (user-defined)', auto = 'automatic'))
}





) ## end methods list
) ## end RefClass

## default value for NA's in heatmap
HEATMAP_NA_VALUE = -Inf

#'
#' Generate a Heatmap from a list of QC measurements.
#' 
#' Each list entry is a data.frame with two columns. 
#' The first one contains the Raw file name (or the short version).
#' and should be named 'raw.file' (or 'fc.raw.file'). 
#' The second column's name must be an expression (see ?plotmath)
#' and contains quality values in the range [0,1]. If values are outside this range, 
#' a warning is issued and values are cut to the nearest allowed value (e.g. '1.2' becomes '1').
#' List entries are merged and columns are ordered by name.
#'   
#' All substrings enclosed by 'X[0-9]*X.' will be removed (can be used for sorting columns).
#' The resulting string is evaluated as an expression. 
#' E.g. parse(text = <colname>)
#' 
#' To judge the overall quality of each raw file a summary column is added, 
#' values being the mean of all other columns per row.
#'        
#' @param lst_qcMetrics List of QCMetric objects
#' @param raw_file_mapping Data.frame with 'from' and 'to' columns for name mapping to unify names from list entries
#' @return A ggplot object for printing
#' 
#' @import ggplot2
#' 
getQCHeatMap = function(lst_qcMetrics, raw_file_mapping)
{
  if (length(lst_qcMetrics) == 0) stop("Heatmap: List of Qc metrics is empty!")
  lst.QCM = lapply(lst_qcMetrics, function(qcm) {
    qcm_sc = qcm$qcScores
    if (plyr::empty(qcm_sc)) return(NULL) ## if metric was not computed, default DF is empty
    if ("raw.file" %in% colnames(qcm_sc)) {
      qcm_sc$fc.raw.file = renameFile(qcm_sc$raw.file, raw_file_mapping)  ## create short name column
      qcm_sc = qcm_sc[, !(colnames(qcm_sc) %in% "raw.file")]  ## remove raw.file column
    }
    if (!("fc.raw.file" %in% colnames(qcm_sc))) {
      cat(paste("columns:", paste0(colnames(qcm_sc), collapse=", ", sep="")))
      stop("Internal error in getQCHeatMap(): 'fc.raw.file' column missing from QC measure.")
    } 
    ## check if fc.raw.filenames are known (e.g. when column was named fc.raw.file but values are from raw.file)
    if (!(all(qcm_sc$fc.raw.file %in% raw_file_mapping$to))) {
      cat("An error occured. Current filename mapping is:\n")
      print(raw_file_mapping)
      stop("Internal error in getQCHeatMap(): 'fc.raw.file' column has invalid entries for metric '", qcm$qcName, "' with fc.raw.file names: [", paste(qcm_sc$fc.raw.file, collapse=","),"]!")
    }
    return(qcm_sc)
  })
  lst.QCM = plyr::compact(lst.QCM) ## remove 'NULL' entries
  ## final heat map of QC metrics
  df.QCM = Reduce(function(a,b) merge(a,b,all = TRUE), lst.QCM)

  ## no metrics...  
  if (is.null(df.QCM))
  {
    p = ggText("HeatMap unavailable", "No metrics were computed (not enough data)");
    return(list(plot = p, table = NULL))
  }
  ## create summary column
  lst_qcMetrics[["qcMetric_AverageQualOverall"]]$setData(df.QCM)
  ## ... add it
  df.AverageQual = lst_qcMetrics[["qcMetric_AverageQualOverall"]]$qcScores
  if (plyr::empty(df.AverageQual)) df.QCMa = df.QCM  else df.QCMa = merge(df.QCM, df.AverageQual)

  ## get order and names for each metric
  df.meta = getMetaData(lst_qcMetrics)
  head(df.meta)
  
  ## reorder rows (file names)
  df.QCMa = df.QCMa[match(raw_file_mapping$to, df.QCMa$fc.raw.file), ]
  ## ... fix factor levels
  df.QCMa$fc.raw.file = factor(df.QCMa$fc.raw.file, levels = df.QCMa$fc.raw.file)
  
  
  ## reorder columns (metrics)
  head(df.QCMa)
  qc_names_all = as.character(df.meta$name) ## contains metrics which have no scores (but the order is ok)
  qc_names_all_scores = qc_names_all[qc_names_all %in% colnames(df.QCMa)]
  df.QCMa = df.QCMa[, c("fc.raw.file", qc_names_all_scores)]
  ## add column numbering (ignore first column, which is 'fc.raw.file')
  df.QCMan = df.QCMa
  idx = 2:(ncol(df.QCMan))
  colnames(df.QCMan)[idx] = paste0(colnames(df.QCMa)[idx], "~\"[", idx-1, "]\"")
  colnames_wNum_map = data.frame(name = colnames(df.QCMa), nameWnum = colnames(df.QCMan))
  
  QCM_final.m = reshape2::melt(df.QCMan, id.vars="fc.raw.file")
  QCM_final.m$variable = factor(QCM_final.m$variable, ordered = TRUE)
  
  ## some files might not be in the original list (will receive 'bad' score in table)
  QCM_final.m$value[is.na(QCM_final.m$value)] = 0
  ## some other files might be missing on purpose
  QCM_final.m$value[QCM_final.m$value == HEATMAP_NA_VALUE] = NA
  
  if (any(QCM_final.m$value > 1, na.rm = TRUE))
  {
    warning("getQCHeatMap() received quality data values larger than one! This can be corrected here, but should be done upstream.")
    QCM_final.m$value[QCM_final.m$value > 1] = 1
  }
  if (any(QCM_final.m$value < 0, na.rm = TRUE))
  {
    warning("getQCHeatMap() received quality data values smaller than zero! This can be corrected here, but should be done upstream.")
    QCM_final.m$value[QCM_final.m$value < 0] = 0
  }
  
  ##
  ## color names by category
  ##
  col_axis = rep(c("#606060", "black"), 10) ## grey-black as tic-toc
  head(QCM_final.m)
  QCM_final.m$name = colnames_wNum_map$name[match(QCM_final.m$variable, colnames_wNum_map$nameWnum)]
  QCM_final.m$axisCat = tolower(df.meta$cat[match(QCM_final.m$name, df.meta$name)])
  QCM_final.m$axisCol = col_axis[cumsum(!duplicated(QCM_final.m$axisCat))]
  
  QCM_final.m$variable2 = factor(QCM_final.m$variable, levels = unique(QCM_final.m$variable))
  if (any(is.na(QCM_final.m$value))) QCM_final.m$dummy_col = "NA" ## use color legend for missing values
  
  ## replace the x-axis labels by expressions
  QCM_final.m$variable3 = sapply(as.character(QCM_final.m$variable2), function(x) {
    parse(text=x)
  })
  
  p = ggplot(QCM_final.m, aes(y = .data$fc.raw.file, x = .data$variable2))
  if (any(is.na(QCM_final.m$value)))
  {
    p = p + geom_tile(aes(fill = .data$value, colour = .data$dummy_col)) +
      scale_colour_manual(name="Missing", values=c("NA" = "grey50")) +
      guides(colour = guide_legend(override.aes = list(fill = 'grey50')))
  } else {
    p = p + geom_tile(aes(fill = .data$value))
  }  
  p = p + scale_y_discrete_reverse(QCM_final.m$fc.raw.file, breaks = ggAxisLabels) + 
    scale_x_discrete(labels = QCM_final.m$variable3) +
    scale_fill_gradientn(colours = c("red", "black", "green"),
                         na.value = "grey50",
                         limits=c(0, 1), 
                         guide = guide_legend(title = "score"),
                         breaks=c(1,0.5,0),
                         labels=c("best","under\nperforming","fail")) +
    theme(axis.text.x = suppressWarnings(element_text(angle = 90, 
                                     vjust = 0.5, 
                                     hjust = 1, 
                                     colour=QCM_final.m$axisCol[!duplicated(QCM_final.m$variable2)]))) +
    ggtitle("Performance overview") +
    xlab("") +
    ylab("Raw file")
  #print(p)
  return(list(plot = p, table = reshape2::dcast(QCM_final.m, fc.raw.file ~ variable)))
}  

#'
#' Extract meta information (orderNr, metric name, category)
#' from a list of Qc metric objects
#' 
#' @param lst_qcMetrics List of qcMetrics
#' @return data.frame with columns 'name', 'order' and 'cat' (category)
#'
getMetaData = function(lst_qcMetrics)
{
  df.meta = plyr::ldply(lst_qcMetrics, function(qcm) {
    #qq <<- qcm
    qcm_sc = qcm$qcScores
    if (plyr::empty(qcm_sc)) {
      ## if metric was not computed, default DF is empty
      name = qcm$qcName
    } else {
      # Name of metric (incl. filled placeholders)
      name = grepv("raw.file", colnames(qcm_sc), invert = TRUE)
    }
    data.frame(name = name, order = qcm$orderNr, cat = qcm$qcCat)
  })
  ## order meta
  if (!plyr::empty(df.meta)) df.meta = df.meta[order(df.meta$order), ]
  return(df.meta)
}

#'
#' Quality metric with linear response to input, reaching the maximum score at the given threshold.
#'
#' Ranges between 0 (worst score) and 1 (best score).
#' Useful for performance measures where reaching a certain reference threshold 't'
#' will be enough to reach 100\%.
#' The input range from [0, t] is scored from 0-100\%.
#'
#' @param x Numeric value(s) between [0, inf]
#' @param t Threshold value, which indicates 100\%
#' @return Value between [0, 1]
#'
#' @export
#'
qualLinThresh = function(x, t = 1)
{
  if (length(x) == 0) stop("Error: qualLinThresh() received empty data!")

  x[is.na(x)] = 0; ## replace NA with 0
  if (any(x < 0)) stop("qualLinThresh(): negative input 'x' not allowed!")
  if (t < 0) stop("qualLinThresh(): negative threshold 't' not allowed!")

  q = pmin(1, x / t)
  return (q)
}


#'
#' Quality metric for 'centeredness' of a distribution around zero.
#'
#' Ranges between 0 (worst score) and 1 (best score).
#' A median of zero gives the best score of 1.
#' The closer the median is to the most extreme value of the distribution, the smaller the score (until reaching 0).
#' Can be used for calibrated mass errors, as a measure of how well they are centered around 0.
#' E.g. if the median is 0.1, while the range is [-0.5,0.5], the score will be 0.8 (punishing the 20% deviation).
#' If the range of data is asymmetric, e.g. [-1.5,-0.5] and does not include zero, the score cannot reach 1,
#' since the median can never be zero.
#'
#' @param x Numeric values (e.g. ppm errors)
#' @return Value between [0, 1]
#'
#' @export
#'
qualCentered = function(x)
{
  if (length(x) == 0) stop("Error: qualCentered() received empty data!")
  if (any(is.na(x))) stop("Error: qualCentered() received NA data!")

  q = 1 - (abs(median(x)) / max(abs(x)))
  return (q)
}

#'
#' Quality metric for 'centeredness' of a distribution around zero with a user-supplied range threshold.
#'
#' Ranges between 0 (worst score) and 1 (best score).
#' The best score is achieved when the median of 'x' is close to the center of the interval [-tol, tol].
#' If median of 'x' is close to the border (on either side), the score decreases linearly to zero.
#' Can be used for uncalibrated mass errors, as a measure of how well they are centered around 0.
#'
#' NA's are removed for all computations.
#'
#' @param x  Vector of values (hopefully in interval [-tol, tol])
#' @param tol Border of interval (must be positive)
#' @return Value between [0, 1]
#'
#' @export
#'
qualCenteredRef = function(x, tol)
{
  x = na.omit(x)
  if (length(x) == 0) stop("qualCenteredRef(): input is empty or consists of NA only!")
  if (tol <= 0) stop("qualCenteredRef(): negative or zero interval border not allowed!")
  m = abs(median(x, na.rm = TRUE))
  if (m > tol) warning("qualCenteredRef(): Median of x is outside of interval. Score will be set to 0.")
  q = 1 - (m / tol)
  q = max(0, q) ## avoid negative scores if abs(x)>tol
  return (q)
}

#'
#' Scales a vector of values linearly to [0, 1]
#' If all input values are equal, returned values are all 0
#' @param X Vector of values
#' @return Scaled vector
scale01linear = function(X)
{
  rng = (max(X) - min(X));
  if (rng == 0) rng = 1; ## no range, so divide by 1
  X2 = (X - min(X)) / rng
  #summary(X2)
  return(X2)
}


#'
#' Quality metric which measures the absolute distance from median.
#'
#' Ranges between 0 (worst score) and 1 (best score).
#' Input must be between [0,1].
#' Deviations from the median of the sample represent the score for each sample point.
#'
#' @param x A vector numeric values between [0,1]
#' @return A vector of the same size as x, with quality values between [0, 1]
#'
#' @export
#'
qualMedianDist = function(x)
{
  if (length(x) == 0) stop("Error: qualMedianDist() received empty data!")
  if (any(is.na(x))) stop("Error: qualMedianDist() received NA data!")
  if (any(x < 0 | x > 1)) stop("qualMedianDist(): x values out of range [0,1]!")
  q = 1 - abs(x - median(x, na.rm = TRUE))
  return (q)
}


#'
#' Compute deviation from uniform distribution
#'
#' The score ranges between 0 (worst score) and 1 (best score).
#' Input 'x' is a vector of counts (or probabilities) for equally spaced bins in a histogram.
#' A uniform distribution (e.g. c(3,3,3) will get a score of 1. The worst possible case (e.g. c(4,0,0)), will get a score of 0,
#' and a linear increasing function (e.g. c(1,2,3)) will get something in between (0.585 here)
#'
#' In addition, bin values can be weighted (e.g. by their confidence). The total sum of weights is normalized to 1 internally.
#'
#' The distance function used is the square root of the absolute difference between a uniform distribution and the input 'x'
#' (summed for each element of 'x').
#' This distance is normalized to the worst possible input (e.g. one bin with 100% counts, all other bins being empty).
#'
#' @param x Vector of numeric intensity/count values (e.g. ID's per RT bin); bins are assumed to have equal widths
#' @param weight Vector of weights for values in 'x' (same length as 'x').
#' @return Value between [0, 1]
#'
#' @export
#'
#' @examples
#'  stopifnot(qualUniform(c(3,3,3))==1)
#'  stopifnot(qualUniform(c(4,0,0))==0)
#'
#'  ## how 'uniform' is a vector where only a single index has weight?-- answer: very
#'  stopifnot(qualUniform(c(4,0,0), c(1,0,0))==1)
#'  stopifnot(qualUniform(c(4,0,0), c(0,1,0))==1)
#'  stopifnot(qualUniform(c(0,4,0))==0)
#'  stopifnot(abs(qualUniform(c(3,2,1))-0.58578) < 0.0001)
#'  stopifnot(abs(qualUniform(c(1,2,3))-0.58578) < 0.0001)
#'  stopifnot(qualUniform(c(1,2,3), c(0,1,0))==1)
#'  stopifnot(abs(qualUniform(c(1,2,3))-0.58578) < 0.0001)
#'  stopifnot(abs(qualUniform(c(1,2,3), c(0,1,1))- 0.590316) < 0.0001)
#'  stopifnot(abs(qualUniform(c(2,3), c(1,1))-0.552786) < 0.0001)
#'  stopifnot(abs(qualUniform(1:120)-0.38661) < 0.0001)
#'
qualUniform = function(x, weight=vector())
{
  if (length(x) == 0 || any(is.na(x) | (x<0))) stop("Error: qualUniform() received negative or missing data!")

  if(length(weight)==0) {
    weight=rep(1, length(x))
  } else if (length(weight)!=length(x)) {
    stop("Error: qualUniform() received weights which do not match length of input!")
  }
  weight = weight/sum(weight)  ## normalize to 1
  x = x/sum(x)

  n = length(x)
  #y = 1/n ## expected intensity (no weights)
  y = sum(x*weight) ## expected weighted intensity
  y
  p = 1/2
  ## for worst possible score, assume bin with highest weight has all the counts
  ## i.e. error is 1-expected == 1-y for 100% bin, and abs(0-y)=y for all other bins of zero intensity
  idx_maxWeight = which(weight == max(weight))[1]
  worst = ((1-y)^p)*weight[idx_maxWeight] + (y^p)*sum(weight[-idx_maxWeight])
  ## worst ==0, iff weight-vector has only one non-zero entry, e.g. c(0,1,0)
  #worst = ((1-y)^p) + (n-1)*((1/n)^p)
  worst
  ## score of current distribution
  sc = sum( (abs(x-y)^p) * weight)
  sc
  q = ifelse(worst==0, 1, (worst - sc) / worst)
  return (q)
}


#'
#' Compute probability of Gaussian (mu=m, sd=s) at a position 0, with reference
#' to the max obtainable probability of that Gaussian at its center.
#'
#' Measure for centeredness around 0.
#' Highest score is 1, worst score is 0.
#'
#' @param mu Center of Gaussian
#' @param sd SD of Gaussian
#' @return quality, ranging from 0 (bad agreement) to 1 (perfect, i.e. centered at 0)
#'
#' @export
#'
qualGaussDev = function(mu, sd)
{
  q = dnorm(0, mean = mu, sd = sd) / dnorm(mu, mean = mu, sd = sd)
  return (q)
}

#'
#' Score an empirical density distribution of values, where the best possible distribution is right-skewed.
#'
#' The score is computed according to
#'
#' q = ((N-1) - sum_i( ((N-i-1)*x_i) ) / (N-1)
#'
#' Scores range from 0 (worst), to 1 (best).
#' E.g. c(0,0,0,16) would yield a score of 1.
#' c(16,0,0,0,0) gives a score of 0.
#'
#' @param x Vector of numeric values (e.g. height of histogram bins)
#' @param N Length of x (just a precaution currently)
#' @return Quality score in the range of [0,1]
#'
#' @export
#'
#' @examples
#'  qualHighest(c(0,0,0,16), 4)   ## 1
#'  qualHighest(c(16,0,0,0), 4)   ## 0
#'  qualHighest(c(1,1,1,1), 4)    ## 0.5
#'  qualHighest(c(0,16,0,0), 4)   ## 1/3
#'
qualHighest = function(x, N)
{
  if (length(x) == 0) stop("Error: qualHighest() received empty data!")
  if (any(is.na(x))) stop("Error: qualHighest() received NA data!")
  if (length(x)!=N) stop("Error in qualHighest(): length of x not equal to N")

  # normalize
  x = x / sum(x)
  p = 1 ## exponent needs to be >=1, using larger values will only mildly penalize intermediate values
  worst = (N-1)^p
  penalty = sum( (x*(N:1 - 1))^p )
  q = (worst - penalty) / worst
  return (q)
}


#'
#' From a list of vectors, compute all vs. all Kolmogorov-Smirnoff distance statistics (D)
#'
#' ... and report the row of the matrix which has maximum sum (i.e the best "reference" distribution).
#' The returned data.frame has as many rows as distributions given and two columns.
#' The first column 'name' gives the name of the list element, the second column 'ks_best' gives '1-statistic' of the
#' Kolmogorov-Smirnoff test to the "reference" distribution (which was picked by maximising the sum of 'ks_best').
#' Thus, the row with a 'ks_best' of 1 is the reference distribution.
#'
#' @param x List of vectors, where each vector holds a distribution
#' @return A data.frame with ks-test values of the "reference" to all other distributions (see Details)
#'
#' @export
#'
qualBestKS = function(x) {

  if (!inherits(x, "list")) stop("Function bestKS() expects a list!")
  if (length(x) == 0) stop("Error: qualBestKS() received empty data!")

  if (is.null(names(x))) names(x) = 1:length(x)

  ## result matrix
  rr = matrix(0, nrow=length(x), ncol=length(x))
  ## compute upper diagonal
  for (i in 1:length(x))
  {
    for (j in (i+1):length(x))
    {
      if (j>length(x)) next;
      rr[i,j] = 1 - suppressWarnings(  ## brags about '-value will be approximate in the presence of ties'
                      stats::ks.test(x[[i]], x[[j]])$statistic
                    )
    }
  }
  ## fill diagonal with 1's
  diag(rr) = 1
  ## add up all values for each reference distribution (equivalent to just rowSums on full matrix)
  rr_sums = rowSums(rr, na.rm = TRUE) + colSums(rr, na.rm = TRUE)
  ## pick best
  r_max = which.max(rr_sums)
  ## get values
  v = pmax(rr[r_max,], rr[, r_max])

  return (data.frame(name = names(x), ks_best = v))

}

library(PTXQC)
context("fcn_qualities.R")

test_that("qualLinThresh", {
  expect_error(qualLinThresh(-3))
  expect_error(qualLinThresh(c(1,2,-3)))
  expect_equal(qualLinThresh(3, 10), 0.3)
  expect_equal(qualLinThresh(3), 1)
  expect_equal(qualLinThresh(3, 3), 1)
  expect_equal(qualLinThresh(c(0,1,3,NA), 3), c(0, 1/3, 1, 0))
})

test_that("qualCentered", {
  expect_equal(qualCentered(3), 0)
  ## median is 0 -- perfect 
  expect_equal(qualCentered(-5:5), 1)
  ## some outliers, but still median is 0 -- perfect
  expect_equal(qualCentered(c(-5:5, -9999, 10)), 1)
  ## median 2, max=3: 1 - 2/3
  expect_equal(qualCentered( c(1,2,3)), 1 - 2/3)
  expect_equal(qualCentered(-c(1,2,3)), 1 - 2/3)
  ## further off ...
  expect_equal(qualCentered( c(5,6,7)), 1 - 6/7)
  expect_equal(qualCentered(-c(5,6,7)), 1 - 6/7)
})

test_that("qualCenteredRef", {
  expect_error(qualCenteredRef(3, 0))
  expect_error(qualCenteredRef(3, -1))
  
  ## median is 13
  expect_equal(qualCenteredRef(c(10:16), 13), 0)
  expect_equal(qualCenteredRef(c(10:16), 13*2), 0.5)
  expect_equal(qualCenteredRef(c(10:16), 13*4), 0.75)
  
  ## median is 13, ref-interval is smaller -- worst score
  expect_equal(qualCenteredRef(c(10:16), 12), 0)
  ## median is perfectly 0 -- ref-interval does not really matter
  expect_equal(qualCenteredRef(-5:5, 1), 1)
  expect_equal(qualCenteredRef(-5:5, 100), 1)
})

test_that("qualMedianDist", {
  ## values need be in [0,1]
  expect_error(qualMedianDist(-3))
  expect_error(qualMedianDist(c(0.1,0.3,-0.3)))
  
  ## median: 0.5
  x = (1:9)/10
  expect_equal(qualMedianDist(x), 1 - abs(0.5 - x))
})

test_that("qualUniform", {
 expect_equal(qualUniform(c(3,3,3)), 1)
 expect_equal(qualUniform(c(4,0,0)), 0)         

 ## how 'uniform' is a vector where only a single index has weight?-- answer: very
 expect_equal(qualUniform(c(4,0,0), c(1,0,0)), 1)
 expect_equal(qualUniform(c(4,0,0), c(0,1,0)), 1)
 expect_equal(qualUniform(c(0,4,0)), 0)
 expect_lt(abs(qualUniform(c(3,2,1))-0.58578), 0.0001)
 expect_lt(abs(qualUniform(c(1,2,3))-0.58578), 0.0001)
 expect_equal(qualUniform(c(1,2,3), c(0,1,0)), 1)
 expect_lt(abs(qualUniform(c(1,2,3))-0.58578), 0.0001)
 expect_lt(abs(qualUniform(c(1,2,3), c(0,1,1))- 0.590316), 0.0001)
 expect_lt(abs(qualUniform(c(2,3), c(1,1))-0.552786), 0.0001)
 expect_lt(abs(qualUniform(1:120)-0.38661), 0.0001)
})


test_that("qualGaussDev", {
  expect_error(qualGaussDev(-3))
  expect_equal(qualGaussDev(0, -4), NaN)
  
  ## perfect
  expect_equal(qualGaussDev(0, 1), 1)
  expect_equal(qualGaussDev(0, 3), 1)
  
  ## shifted slighty
  expect_equal(qualGaussDev(1, 1), 0.6065307, tolerance = .001)
  ## shifted more...
  expect_equal(qualGaussDev(2, 1), 0.1353353, tolerance = .001)
  
})

test_that("qualHighest", {
  expect_equal(qualHighest(c(0,0,0,16), 4), 1)
  expect_equal(qualHighest(c(16,0,0,0), 4), 0)
  expect_equal(qualHighest(c(1,1,1,1), 4) , 0.5)
  expect_equal(qualHighest(c(0,16,0,0), 4), 1/3)
})


test_that("qualBestKS", {
  ## sample some Gaussians
  g1 = rnorm(100)
  g2 = rnorm(100)
  g3 = rnorm(100)
  go = rnorm(100, mean=4) ## this will be our outlier distribution
  x = list(g1,g2,g3,go)
  
  r = qualBestKS(x)
    
  expect_equal(max(r$ks_best[1:3]), 1)
  expect_true(all(r$ks_best[1:3] > 0.7)) ## the three similar ones, should score good (one of them the the reference)
  expect_lt(r$ks_best[4], 0.2) ## outlier should score badly
})

#'
#' Get the longest common prefix from a set of strings.
#' 
#' Input is converted to character (e.g. from factor) first.
#' 
#' 
#' @param strings Vector of strings
#' @return Single string - might be empty ("")
#' 
#' @export
#' 
#' @examples 
#'   longestCommonPrefix(c("CBA.321", "CBA.77654", ""))    ## ""
#'   longestCommonPrefix(c("CBA.321", "CBA.77654", "CB"))  ## "CB"
#'   longestCommonPrefix(c("ABC.123", "ABC.456"))          ## "ABC."
#'   longestCommonPrefix(c("nothing", "in", "common"))     ## ""
#' 
#' 
longestCommonPrefix = function(strings) {
  strings = as.character(strings)
  
  if (length(strings) == 0) return("")
  if (length(strings) == 1) return(strings[1])
  
  min_len = min(sapply(strings, nchar))
  
  if (min_len == 0) return ("")
  i = 2
  for(i in 1:(min_len+1)) ## +1 is required to gauge match for last char
  {
    if (i > min_len) next; ## passed the end
    chars = substring(strings, i, i)
    if (length(unique(chars)) > 1) {
      ## mismatch at pos i
      break;
    }
  }
  if (i==1) return("")
  
  return(substr(strings[1], 1, i-1))
}


#'
#' Like longestCommonPrefix(), but on the suffix.
#' 
#' @param strings Vector of strings
#' @return Single string - might be empty ("")
#' 
#' @export
#' 
#' @examples 
#' 
#'  longestCommonSuffix(c("123.ABC", "45677.ABC", "BC"))  ## "BC"
#'  longestCommonSuffix(c("123.ABC", "", "BC"))           ## ""
#'  longestCommonSuffix(c("123.ABC", "45677.ABC"))        ## ".ABC"
#'  longestCommonSuffix(c("nothing", "in", "common"))     ## ""
#'  
longestCommonSuffix = function(strings)
{
  strings_rev = sapply(lapply(strsplit(strings, NULL), rev), paste, collapse="")
  r = longestCommonPrefix(strings_rev)
  r_rev = sapply(lapply(strsplit(r, NULL), rev), paste, collapse="")
  return (r_rev)
}





#' Removes the longest common prefix (LCP) from a vector of strings.
#' 
#' You should provide only unique strings (to increase speed).
#' If only a single string is given, the empty string will be returned unless \code{minOutputLength} is set.
#' 
#' @param x Vector of strings with common prefix
#' @param min_out_length Minimal length of the shortest element of x after LCP removal [default: 0, i.e. empty string is allowed] . If the output would be shorter, the last part of the LCP is kept.
#' @param add_dots Prepend output with '..' if shortening was done.
#' @return Shortened vector of strings
#' 
#' @examples
#'   delLCP(c("TK12345_H1"), min_out_length=0)
#'   ## ""
#' 
#'   delLCP(c("TK12345_H1"), min_out_length=4)
#'   ## "5_H1"
#' 
#'   delLCP(c("TK12345_H1"), min_out_length=4, add_dots = TRUE)
#'   ## "..5_H1"
#' 
#'   delLCP(c("TK12345_H1", "TK12345_H2"), min_out_length=4)
#'   ## "5_H1" "5_H2"
#' 
#'   delLCP(c("TK12345_H1", "TK12345_H2"), min_out_length=4, add_dots = TRUE)
#'   ## "..5_H1" "..5_H2"
#' 
#'   delLCP(c("TK12345_H1", "TK12345_H2"), min_out_length=8)
#'   ## "12345_H1", "12345_H2"
#' 
#'   delLCP(c("TK12345_H1", "TK12345_H2"), min_out_length=8, add_dots = TRUE)
#'   ## "TK12345_H1", "TK12345_H2" (unchanged, since '..' would add another two)
#' 
#'   delLCP(c("TK12345_H1", "TK12345_H2"), min_out_length=60)
#'   ## "TK12345_H1", "TK12345_H2" (unchanged)
#' 
#'   delLCP(c("TK12345_H1", "TK12345_H2"), min_out_length=60, add_dots = TRUE)
#'   ## "TK12345_H1", "TK12345_H2" (unchanged)
#' 
#' 
#' @export
delLCP <- function(x, min_out_length = 0, add_dots = FALSE)
{
  x = as.character(x)
  lcp = nchar(longestCommonPrefix( x ))   # shorten string (remove common prefix)
  
  if (min_out_length > 0)
  {
    minLen = min(nchar(x))
    if (lcp >= minLen - min_out_length) ## removing lcp would shorten the string too much
    {
      lcp = max(minLen - min_out_length, 0)
      if (lcp <= 2 & add_dots) return(x) ## we can shorten by at most 2 chars .. don't do it
    }
  }
    
  x = sapply(x, substring, lcp+1)
  if (add_dots) x = paste0("..", x)
  return (x)
}

#' Removes the longest common suffix (LCS) from a vector of strings.
#' 
#' @param x Vector of strings with common suffix
#' @return Shortened vector of strings
#' 
#' @examples
#' delLCS(c("TK12345_H1"))                     ## ""
#' delLCS(c("TK12345_H1", "TK12345_H2"))       ## "TK12345_H1" "TK12345_H2" 
#' delLCS(c("TK12345_H1", "TK12!45_H1"))       ## "TK123"    "TK12!" 
#'  
#' @export
delLCS <- function(x)
{
  lcs = nchar(longestCommonSuffix( x ))   # shorten string (remove common suffix)
  x = sapply(x, function(v) substr(v, 1, nchar(v)-lcs))
  return (x)
}

#' Count the number of chars of the longest common prefix
#' 
#' @param x Vector of strings with common prefix
#' @return Length of LCP
#' 
#' @export
lcpCount <- function(x)
{
  lcp = nchar(longestCommonPrefix( x ))   # number of common prefix chars
  return (lcp)
}

#' Count the number of chars of the longest common suffix
#' 
#' @param x Vector of strings with common suffix
#' @return Length of LCS
#' 
#' @export
lcsCount <- function(x)
{
  lcs = nchar(longestCommonSuffix( x ))   # number of common suffix chars
  return(lcs)
}

#' Compute longest common substring of two strings.
#' 
#' Implementation is very inefficient (dynamic programming in R)
#' --> use only on small instances
#' 
#' @param s1 String one
#' @param s2 String two
#' @return String containing the longest common substring
#' 
#' @export
LCS <- function(s1, s2) 
{
  if (nchar(s1)==0 || nchar(s2)==0) return("")

  
  v1 <- unlist(strsplit(s1,split=""))
  v2 <- unlist(strsplit(s2,split=""))
  
  
  num <- matrix(0, nchar(s1), nchar(s2))    
  maxlen <- 0
  pstart = 0
  
  for (i in 1:nchar(s1)) {
    for (j in 1:nchar(s2)) {
      if (v1[i] == v2[j]) {
        if ((i==1) || (j==1)) { 
          num[i,j] <- 1
        } 
        else {
          num[i,j] <- 1+num[i-1,j-1]
        }
        if (num[i,j] > maxlen) {
          maxlen <- num[i,j]
          pstart = i-maxlen+1
        }
      }
    }
  }
  
  ## return the substring found
  return (substr(s1, pstart, pstart+maxlen-1))
}

#' Find longest common substring from 'n' strings.
#' 
#' Warning: greedy heuristic! This is not guaranteed to find the best solution (or any solution at all), since its done pairwise with the shortest input string as reference.
#' 
#' @param strings A vector of strings in which to search for LCS
#' @param min_LCS_length Minimum length expected. Empty string is returned if the result is shorter
#' @return longest common substring (or "" if shorter than \code{min_LCS_length})
#' 
#' @examples
#' LCSn(c("1_abcde...",
#'        "2_abcd...",
#'        "x_abc..."))  ## --> "_abc"
#' LCSn(c("16_IMU008_CISPLA_E5_R11", 
#'        "48_IMU008_CISPLA_P4_E7_R31",
#'        "60_IMU008_CISPLA_E7_R11"), 3) ## -->"_IMU008_CISPLA_"
#' LCSn(c("AAAAACBBBBB", 
#'        "AAAAADBBBBB",
#'        "AAAABBBBBEF",
#'        "AAABBBBBDGH")) ## -->  "BBBBB"
#' LCSn(c("AAAXXBBB",
#'        "BBBXXDDD",
#'        "XXAAADDD")) ## --> fails due to greedy approach; should be "XX"
#' 
#' @export
#' 
LCSn = function(strings, min_LCS_length = 0)
{
  ## abort if there is no chance of finding a suitably long substring
  if (min(nchar(strings)) < min_LCS_length) return("");
  
  if (length(strings) <= 1) return (strings)
  
  ## apply LCS to all strings, using the shortest as reference
  idx_ref = which(nchar(strings)==min(nchar(strings)))[1]
  strings_other = strings[-idx_ref]  
  candidates = unique(sapply(strings_other, LCS, strings[idx_ref]))
  candidates
  ## if only one string remains, we're done
  if (length(candidates) == 1)
  {
    if (nchar(candidates[1]) < min_LCS_length) return("") else return (candidates[1])
  }
  ## if its more, call recursively until a solution is found
  ## continue with the shortest candidates (since only they can be common to all)
  cand_short = candidates[nchar(candidates) == min(nchar(candidates))]
  solutions = unique(sapply(cand_short, function(cand) LCSn(c(cand, strings_other))))
  ## get the longest solution
  idx_sol = which(nchar(solutions)==max(nchar(solutions)))[1];
  
  return (solutions[idx_sol])
}


#'
#' Removes common substrings (infixes) in a set of strings.
#' 
#' Usually handy for plots, where condition names should be as concise as possible.
#' E.g. you do not want names like 
#' 'TK20130501_H2M1_010_IMU008_CISPLA_E3_R1.raw' and 
#' 'TK20130501_H2M1_026_IMU008_CISPLA_E7_R2.raw' 
#' but rather 'TK.._010_I.._E3_R1.raw' and
#'            'TK.._026_I.._E7_R2.raw'
#'            
#' If multiple such substrings exist, the algorithm will remove the longest first and iterate
#' a number of times (two by default) to find the second/third etc longest common substring.
#' Each substring must fulfill a minimum length requirement - if its shorter, its not considered worth removing
#' and the iteration is aborted.
#' 
#' @param strings          A vector of strings which are to be shortened
#' @param infix_iterations Number of successive rounds of substring removal
#' @param min_LCS_length   Minimum length of the longest common substring (default:7, minimum: 6)
#' @param min_out_length   Minimum length of shortest element of output (no shortening will be done which causes output to be shorter than this threshold)
#' @return A list of shortened strings, with the same length as the input                       
#'
#' @examples
#' #library(PTXQC)
#' simplifyNames(c('TK20130501_H2M1_010_IMU008_CISPLA_E3_R1.raw',
#'                 'TK20130501_H2M1_026_IMU008_CISPLA_E7_R2.raw'), infix_iterations = 2)
#' # --> "TK.._010_I.._E3_R1.raw","TK.._026_I.._E7_R2.raw"
#' 
#' try(simplifyNames(c("bla", "foo"), min_LCS_length=5))
#' # --> error, since min_LCS_length must be >=6
#' 
#' @export
#' 
simplifyNames = function(strings, infix_iterations = 2, min_LCS_length = 7, min_out_length = 7)
{
  if (min_LCS_length<6) stop( "simplifyNames(): param 'min_LCS_length' must be 6 at least.")
  
  for (it in 1:infix_iterations)
  {
    lcs = LCSn(strings, min_LCS_length=min_LCS_length)
    if (nchar(lcs)==0) return (strings) ## no infix of minimum length found
    ## replace infix with 'ab..yz'
    strings_i = sub(paste0("(.*)", gsub("([.|()\\^{}+$*?]|\\[|\\])", "\\\\\\1", lcs), "(.*)"), 
                    paste0("\\1"
                           , ifelse(substring(lcs,1,2) == "..", "", substring(lcs,1,2)) ## dont keep ".."
                           , ".."
                           , ifelse(substring(lcs, nchar(lcs)-1) == "..", "", substring(lcs, nchar(lcs)-1)) ## dont keep ".."
                           , "\\2"), 
                    strings)
    if (min(nchar(strings_i)) < min_out_length) return(strings) ## result got too short...
    strings = strings_i
  }
  return (strings)
}

#' Shorten a string to a maximum length and indicate shorting by appending '..'
#' 
#' Some axis labels are sometimes just too long and printing them will either
#' squeeze the actual plot (ggplot) or make the labels disappear beyond the margins (graphics::plot)
#' One ad-hoc way of avoiding this is to shorten the names, hoping they are still meaningful to the viewer.
#' 
#' This function should be applied AFTER you tried more gentle methods, such as \code{\link{delLCP}} or \code{\link{simplifyNames}}.
#' 
#' @param x                Vector of input strings
#' @param max_len          Maximum length allowed
#' @param verbose          Print which strings were shortened
#' @param allow_duplicates If shortened strings are not discernible any longer, consider the short version valid (not the default), otherwise (default) return the full string (--> no-op)
#' @return  A vector of shortened strings
#' 
#' @examples
#' r = shortenStrings(c("gamg_101", "gamg_101230100451", "jurkat_06_100731121305", "jurkat_06_1"))
#' all(r == c("gamg_101", "gamg_101230100..", "jurkat_06_1007..", "jurkat_06_1"))
#' 
#' @seealso \code{\link{delLCP}}, \code{\link{simplifyNames}}
#' 
#' @export
#' 
shortenStrings = function(x, max_len = 20, verbose = TRUE, allow_duplicates = FALSE)
{
  if (any(duplicated(x)) & !allow_duplicates) stop("Duplicated input given to 'shortenStrings'!\n  ", paste(x, collapse="\n  "))

  idx = nchar(x) > max_len
  xr = x
  xr[idx] = paste0(substr(x[idx], 1, max_len-2), "..")
  
  if (!allow_duplicates & any(duplicated(xr)))
  {
    ## try cutting the prefix instead of the suffix...
    xr[idx] = paste0("..", sapply(x[idx], function(s) substr(s, nchar(s)-max_len+2, nchar(s))))
    if (any(duplicated(xr)))
    {
      warning("Duplicated output produced by 'shortenStrings'!\n  ", paste(xr, collapse="\n  "), "\nPlease rename your items (make them shorter and less similar)")
      ## fail generously and return the original input
      return(x)
    }
  }
  
  if (verbose & any(idx))
  {
    cat("The following labels will be shortened to ease plotting:\n")
    cat(paste0("  ", paste(x[idx], collapse="\n  ")))
    cat("\n")
  }

  return (xr)
}

#' Compute shortest prefix length which makes all strings in a vector uniquely identifyable.
#'
#' If there is no unique prefix (e.g. if a string is contained twice), then the length
#' of the longest string is returned, i.e. if the return value is used in a call to substr, nothing happens
#' e.g.  substr(x, 1, supCount(x)) == x
#' 
#' @param x        Vector of strings
#' @param prefix_l Starting prefix length, which is incremented in steps of 1 until all prefixes are unique (or maximum string length is reached)
#' @return Integer with minimal prefix length required
#' 
#' @examples
#'   supCount(c("abcde...", "abcd...", "abc..."))  ## 5
#'
#'   x = c("doubled", "doubled", "aLongDummyString")
#'   all( substr(x, 1, supCount(x)) == x )   
#'   ## TRUE (no unique prefix due to duplicated entries)
#' 
#' @export
#' 
supCount <- function(x, prefix_l=1)
{
  max_length = max(nchar(x))
  while (prefix_l < max_length)
  {
    dups = duplicated(sapply(x, substr, 1, prefix_l), NA)
    if (sum(dups)==0)
    {
      break;
    }
    prefix_l = prefix_l + 1
  }
  return (prefix_l)
}


#' Re-estimate a new set size to split a number of items into equally sized sets.
#'
#' This is useful for plotting large datasets where multiple pages are needed.
#' E.g. you know that you need 101 barplots, but you only want to fit about 25 per page.
#' Naively one would now do five plots, with the last one only containing a single barplot.
#' Using this function with correctSetSize(101, 25) would tell you to use 26 barplots per page,
#' so you end up with four plots, all roughly equally filled.
#' It also works the other extreme case, where your initial size is chosen slightly too high, e.g.
#' Sets of size 5 for just 8 items is too much, because we can reduce the set size to 4 and still
#' need two sets but now they are much more equally filled (correctSetSize(8, 5) == 4).
#' 
#' We allow for up to set sizes of 150\% from default, to avoid the last set being sparse (we remove it and distribute to the other bins)
#' 150%\ oversize is the extreme case, which only happens with sets of size two. With more sets the overhead is much smaller (1/X).
#' Once the number of sets is fixed, we distribute all items equally.
#' 
#' E.g. 6 items & initial_set_size=5, would result in 2 bins (5 items, 1 item), but we'd rather have one bin of 6 items
#' or 8 items & initial_set_size=5, would result in 2 bins (5+3 items), since the last set is more than half full, but we'd rather have 4+4
#'
#' @param item_count Known number of items which need to assigned to sets
#' @param initial_set_size Desired number of items a single set should hold
#' @return re-estimated set size which a set should hold in order to avoid underfilled sets
#' 
#' @examples
#'  stopifnot(
#'    correctSetSize(8, 5) == 4
#'  )
#'  stopifnot(
#'    correctSetSize(101, 25) == 26
#'  )
#'  
#' @export
#'  
correctSetSize = function(item_count, initial_set_size)
{
  blocks = seq(from = 1, to = item_count, by = initial_set_size)
  blockcount = length(blocks)
  lastblocksize = item_count - tail(blocks, n = 1) + 1
  #cat(paste("Naively, last set has size", lastblocksize, "\n"))
  if (lastblocksize < initial_set_size * 0.5 & blockcount > 1)
  { ## last block is not full, reduce block count if more than one block
    blockcount = blockcount - 1
    #cat(paste("reducing number of sets to", blockcount, "\n"))
  }
  ## distribute equally among fixed number of sets
  set_size = ceiling(item_count / blockcount)
  return (set_size)
}


#' Calls FUN on a subset of data in blocks of size 'subset_size' of unique indices.
#' 
#' One subset consists of 'subset_size' unique groups and thus of all rows which
#' in 'data' which have any of these groups.
#' The last subset might have less groups, if the number of unique groups is not dividable by subset_size.
#' 
#' FUN is applied on each subset.
#' 
#' @param data         Data.frame whose subsets to use on FUN
#' @param indices      Vector of group assignments, same length as nrow(data)
#' @param subset_size  Number of groups to use in one subset
#' @param FUN          Function applied to subsets of data
#' @param sort_indices Sort groups (by their sorted character(!) names) before building subsets
#' @param ...          More arguments to FUN
#'
#' @return list of function result (one entry for each subset)
#' 
#' @examples
#'  byX(data.frame(d=1:10), 1:10, 2, sum)
#'  
#' @export  
#' 
byX <- function(data, indices, subset_size = 5, FUN, sort_indices = TRUE, ...)
{
  stopifnot(subset_size > 0)
  
  stopifnot(nrow(data) == length(indices))
  
  groups = unique(indices)
  if (sort_indices)
  {
    #cat(paste0("Sorting indices (", length(groups), ")...\n"))
    groups = sort(groups)
    #cat(paste0(groups))
    #cat()
  }
  blocks = seq(from = 1, to = length(groups), by = subset_size)
  result = lapply(blocks, function(x) {
    #cat(paste("block", x, " ... "))
    range = x:(min(x+subset_size-1, length(groups)))
    lns = indices %in% groups[range];
    subset = (data[lns, , drop = FALSE])
    rownames(subset) = rownames(data)[lns]
    #cat(paste("call\n"))
    r = FUN(subset, ...)    
    flush.console()
    return (r)
  })
  return (result)
}

#'
#' Same as \code{\link{byX}}, but with more flexible group size, to avoid that the last group has only a few entries (<50\% of desired size).
#' 
#' The 'subset_size' param is internally optimized using \code{\link{correctSetSize}} and
#' then \code{\link{byX}} is called.
#' 
#' @param data         Data.frame whose subset to use on FUN
#' @param indices      Vector of group assignments, same length as nrow(data)
#' @param subset_size  Ideal number of groups to use in one subset -- this can be changed internally, from 75\%-150\%
#' @param FUN function Applied to subsets of data
#' @param sort_indices Groups are formed by their sorted character(!) names
#' @param ...          More arguments to FUN
#'
#' @return list of function result (one entry for each subset)
#' 
#' @examples 
#'  stopifnot(
#'    byXflex(data.frame(d=1:10), 1:10, 2, sum, sort_indices = FALSE) ==
#'    c(3, 7, 11, 15, 19)
#'  )
#'  
#' @export
#'  
byXflex <- function(data, indices, subset_size = 5, FUN, sort_indices = TRUE, ...)
{
  stopifnot(subset_size>0)
  #indices=1:24
  #subset_size=5
  groups = unique(indices)
  subset_size = correctSetSize(length(groups), subset_size)
  return (byX(data, indices, subset_size, FUN, sort_indices, ...))
}

#' Assign set numbers to a vector of values.
#'
#' Each set has size set_size (internally optimized using \code{\link{correctSetSize}}), holding values from 'values'.
#' This gives n such sets and the return value is just the set index for each value.
#' 
#' @param values       Vector of values
#' @param set_size     Number of distinct values allowed in a set
#' @param sort_values  Before assigning values to sets, sort the values?
#'
#' @return Vector (same length as input) with set numbers
#' 
#' @examples
#'  #library(PTXQC)
#'  assignBlocks(c(1:11, 1), set_size = 3, sort_values = FALSE)
#'  ## --> 1 1 1 2 2 2 3 3 3 4 4 1
#'  
#' @export
#'  
assignBlocks = function(values, set_size = 5, sort_values = TRUE)
{
  groups = unique(values)
  ## correct
  set_size = correctSetSize(length(groups), set_size)
  ## sort
  if (sort_values)
  {
    #cat(paste0("Sorting values (", length(groups), ")..."))
    groups = sort(groups)
  }
  
  blocks = seq(from=1, to=length(groups), by=set_size)
  result = rep(NA, length(values))
  for (x in 1:length(blocks))
  {
    range = blocks[x]:(min(blocks[x]+set_size-1, length(groups)))
    lns = values %in% groups[range];
    result[lns] = x
  }
  #cat("done\n")
  return (result)
}




#'
#' Grep with values returned instead of indices.
#' 
#' The parameter 'value' should not be passed to this function since it is 
#' passed internally already.
#' 
#' @param reg  regex param
#' @param data container 
#' @param ... other params forwarded to grep()
#'
#' @return values of data which matched the regex
#' 
#' @examples 
#'   grepv("x", c("abc", "xyz"))
#'   ## --> "xyz"
#'   
#' @export
#'    
grepv = function(reg, data, ...)
{
  return (grep(reg, data, value = TRUE, ...))  
}

#' paste with tab as separator
#' 
#' @param ... Arguments forwarded to paste()
#' @return return value of paste()
#' 
#' @examples
#'   pastet("tab","separated")
#'   ## --> "tab\tseparated"
#'   
#' @export
#'    
pastet = function(...)
{
  paste(..., sep="\t")
}

#' paste with newline as separator
#' 
#' @param ... Arguments forwarded to paste()
#' @return return value of paste()
#' 
#' @examples
#'   pasten("newline","separated")
#'   ## --> "newline\nseparated"
#'   
#' @export
#'    
pasten = function(...)
{
  paste(..., sep="\n")
}

#' Coefficient of variation (CV)
#' 
#' Computes sd(x) / mean(x)
#' 
#' @param x Vector of numeric values
#' @return CV
CV = function(x){sd(x) / mean(x)} 

#' Relative standard deviation (RSD)
#' 
#' Simply \code{\link{CV}}*100
#' 
#' @param x Vector of numeric values
#' @return RSD
#' 
RSD = function(x){CV(x)*100}

#' Replace 0 with NA in a vector
#' 
#' @param x A numeric vector
#' @return Vector of same size as 'x', with 0's replaced by NA
#' 
del0 = function(x)
{
  x[x==0] = NA
  return(x)
}

#'
#' Repeat each element x_i in X, n_i times.
#' 
#' @param x Values to be repeated
#' @param n Number of repeat for each x_i (same length as x)
#' @return Vector with values from x, n times
#' 
#' @examples 
#' 
#'   repEach(1:3, 1:3)  ## 1, 2, 2, 3, 3, 3
#' 
#' @export
#' 
repEach = function(x, n)
{
  stopifnot(length(x) == length(n))
  
  r = unlist(
        sapply(1:length(x), function(i) rep(x[i], each=n[i]))
      )
  
  return(r)
}


#' Find the local maxima in a vector of numbers.
#' 
#' A vector of booleans is returned with the same length as input
#' which contains TRUE when there is a maximum.
#' Simply sum up the vector to get the number of maxima.
#' 
#' @param x           Vector of numbers
#' @param thresh_rel  Minimum relative intensity to maximum intensity of 'x' required
#'                    to be a maximum (i.e., a noise threshold). Default is 20\%.
#' @return Vector of bool's, where TRUE indicates a local maximum.
#' 
#' @examples
#'     r = getMaxima(c(1,0,3,4,5,0))                                
#'     all(r == c(TRUE,FALSE,FALSE,FALSE,TRUE,FALSE))
#'     
#'     getMaxima(c(1, NA, 3, 2, 3, NA, 4, 2, 5))
#'     
#' @export     
#'     
getMaxima = function(x, thresh_rel = 0.2)
{
  pos = rep(FALSE, length(x))
  thresh_abs = max(x, na.rm = TRUE) * thresh_rel
  x_noNA = na.omit(x)
  if (length(x_noNA) == 0) return(pos)
  last = x_noNA[1]
  ## find index of first non-NA in 'x'
  i = which(last == x)[1]
  last_nonNA_i = i
  up = TRUE ## we start going up (if the next point is lower, the first point is a maximum)
  while (i < length(x))
  {
    i = i + 1;
    if (is.na(x[i])) next;
    
    if (last < x[i] && !up) 
    { # ascending in down mode
      up = !up
    } else if (last > x[i] && up)
    { # going down in up mode
      up = !up
      if (x[last_nonNA_i]>=thresh_abs) pos[last_nonNA_i] = TRUE
    }
    last = x[i]
    last_nonNA_i = i
  }
  if (up) pos[last_nonNA_i] = TRUE ## if we ended going up, the last point is a maximum
  return (pos)
}

#'
#' Prepare a Mosaic plot of two columns in long format.
#' 
#' Found at http://stackoverflow.com/questions/19233365/how-to-create-a-marimekko-mosaic-plot-in-ggplot2
#' Modified (e.g. to pass R check)
#' 
#' Returns a data frame, which can be used for plotting and has the following columns:
#' 'Var1' - marginalized values from 1st input column
#' 'Var2' - marginalized values from 2nd input column
#' 'Freq' - relative frequency of the combination given in [Var1, Var2]
#' 'margin_var1' - frequency of the value given in Var1
#' 'var2_height' - frequency of the value given in Var2, relative to Var1
#' 'var1_center' - X-position when plotting (large sets get a larger share)
#' 
#' @param data A data.frame with exactly two columns
#' @return Data.frame
#' 
#' @examples
#'   data = data.frame(raw.file = c(rep('file A', 100), rep('file B', 40)),
#'                     charge = c(rep(2, 60), rep(3, 30), rep(4, 10),
#'                                rep(2, 30), rep(3, 7), rep(4, 3)))
#'   mosaicize(data)                             
#'   
#' @export   
#' 
mosaicize = function(data)
{
  stopifnot(ncol(data)==2)
  lev_var1 = length(unique(data[, 1]))

  ct.data = as.data.frame(prop.table(table(data[, 1], data[, 2])))
  ct.data$Margin_var1 = prop.table(table(data[, 1]))
  ct.data$Var2_height = ct.data$Freq / ct.data$Margin_var1
  ct.data$Var1_center = c(0, cumsum(ct.data$Margin_var1)[(1:lev_var1) -1]) + ct.data$Margin_var1 / 2
  #ct.data = ct.data[ order(ct.data$Var1), ]
  return (ct.data)
}

#' A string concatenation function, more readable than 'paste()'.
#' 
#' @param a Char vector
#' @param b Char vector
#' @return Concatenated string (no separator)
#'
`%+%` = function(a, b)
{
  return (paste(a, b, sep=""))
} 

#'
#' Given a vector of (short/long) filenames, translate to the (long/short) version
#' 
#' @param f_names Vector of filenames
#' @param mapping A data.frame with from,to columns
#' @return A vector of translated file names as factor (ordered by mapping!)
#'
#' @export
#'
renameFile = function(f_names, mapping)
{
  if (all(f_names %in% mapping$from)) {      ## from -> to
    f_new = mapping$to[match(f_names, mapping$from)]
    f_new = factor(f_new, levels=mapping$to)
  } else if (all(f_names %in% mapping$to)) { ## to -> from
    f_new = mapping$from[match(f_names, mapping$to)]
    f_new = factor(f_new, levels=mapping$from)
  } else {
    stop("Error in renameFile: filenames cannot be found in mapping!")
  }
  
  return (f_new)
} 

#'
#' Add the value of a variable to an environment (fast append)
#'
#' The environment must exist, and its name must be given as string literal in 'env_name'!
#' The value of the variable 'v' will be stored under the name given in 'v_name'.
#' If 'v_name' is not given, a variable name will be created by increasing an internal counter
#' and using the its value padded with zeros as name (i.e., "0001", "0002" etc).
#' 
#' @param env_name String of the environment variable
#' @param v Value to be inserted
#' @param v_name String used as variable name. Automatically generated if omitted.
#' @return Always TRUE
#'
#'
appendEnv = function(env_name, v, v_name = NULL)
{
  e = eval.parent(as.name(env_name), n=3)
  e$.counter <- e$.counter + 1
  
  e[[sprintf("%04d",e$.counter)]] <- v ## pad with 0's, to ensure correct order when calling ls() on env_name
  return(TRUE)
}

#'
#' Thin out a data.frame by removing rows with similar numerical values in a certain column.
#' 
#' All values in the numerical column 'filterColname' are assigned to bins of width 'binsize'.
#' Only one value per bin is retained. All other rows are removed and the reduced
#' data frame will all its columns is returned.
#' 
#' @param data The data.frame to be filtered
#' @param filterColname Name of the filter column as string
#' @param binsize Width of a bin
#' @return Data.frame with reduced rows, but identical input columns
#'
#'
thinOut = function(data, filterColname, binsize)
{
  nstart = nrow(data)
  ## first remove duplicates (they cannot possibly pass the filter)
  data = data[!duplicated(data[, filterColname]), ]
  ## sort (expensive)
  if (is.unsorted(data[, filterColname])) data = data[order(data[, filterColname]), ]
  ## binning...
  local_deltas = c(0, diff(data[, filterColname]))
  ld_cs = cumsum(local_deltas)
  data = data[!duplicated(round(ld_cs / binsize)),]
  #print("Saved " %+% round(100 - nrow(x)/nstart*100) %+% "% data")
  return(data)
}

#'
#' Apply 'thinOut' on all subsets of a data.frame, split by a batch column
#' 
#' The binsize is computed from the global data range of the filter column by dividing the range
#' into binCount bins.
#' 
#' @param data The data.frame to be split and filtered(thinned)
#' @param filterColname Name of the filter column as string
#' @param batchColname Name of the split column as string
#' @param binCount Number of bins in the 'filterColname' dimension.
#' @return Data.frame with reduced rows, but identical input columns
#' 
thinOutBatch = function(data, filterColname, batchColname, binCount = 1000)
{
  binsize = (max(data[, filterColname], na.rm=TRUE) - min(data[, filterColname], na.rm=TRUE)) / binCount
  r = plyr::ddply(data, batchColname, thinOut, filterColname, binsize)
  return (r)  
}


#'
#' Assign a relative abundance class to a set of (log10) abundance values
#' 
#' Abundances (should be logged already) are grouped into different levels,
#' starting from the smallest values ("low") to the highest values ("high").
#' Intermediate abundances are either assigned as "mid", or "low-mid".
#' If the range is too large, only "low" and "high" are assigned, the intermediate values
#' are just numbers.
#' 
#' Example:
#'  getAbundanceClass(c(12.4, 17.1, 14.9, 12.3)) ## --> factor(c("low", "high", "mid", "low"))
#' 
#' 
#' @param x Vector of numeric values (in log10)
#' @return Vector of factors corresponding to input with abundance class names (e.g. low, high)
#'
getAbundanceClass = function(x) {
  r = data.frame(x = x)
  r$x_diff = round(r$x - min(r$x))
  r$x_diff_fac = as.numeric(factor(r$x_diff)) ## make a factor, to get ordering (lowAbd=1, ...)
  ## assign names to abundance classes
  cc = length(unique(r$x_diff))
  if (cc==1) lvls = "mid"
  if (cc==2) lvls = c("low", "high")
  if (cc==3) lvls = c("low", "mid", "high")
  if (cc==4) lvls = c("low", "low-mid", "mid-high", "high")
  if (cc>4) {
    lvls = 1:max(r$x_diff_fac)
    lvls[1] = "low"
    lvls[length(lvls)] = "high"
  }
  r$x_cl = lvls[r$x_diff_fac]
  r$x_cl = factor(r$x_cl, levels=lvls, ordered=TRUE)
  r
  return(r$x_cl)
}


#' 
#' Assembles a list of output file names, which will be created during reporting.
#' 
#' You can combine **report_name_has_folder** (and **mzTab_filename** for mzTab files) to obtain report filenames which are even more
#' robust to moving around (since they contain infixes of the mzTab filename and the folder),
#' e.g. `@em `report_HEK293-study_myProjects.html``, where the input 
#'      was `mzTab_filename='HEK293-study.mzTab` and `folder='c:/somePath/myProjects/`.
#' 
#' @param folder Directory where the MaxQuant output (txt folder) or the mzTab file resides
#' @param report_name_has_folder Boolean: Should the report files (html, pdf) contain the name
#'        of the deepest(=last) subdirectory in **txt_folder** which is not `txt`?
#'        Useful for discerning different reports in a PDF viewer.
#'        E.g. when flag is FALSE: `report_v0.91.0.html`; and `report_v0.91.0_bloodStudy.html` when flag is TRUE (and the
#'        txt folder is `.../bloodStudy/txt/` or `...bloodStudy/`)
#' @param mzTab_filename If input is an mzTab, specify its name, so that the filenames can use its basename as infix
#'        E.g. when `mzTab_filename = 'HEK293-study.mzTab'` then the output will be
#'             `report_HEK293-study.html`.
#'        This allows to get reports on multiple mzTabs in the same folder without overwriting report results.
#'        
#' @return List of output file names (just names, no file is created) 
#'         with list entries: 
#'         **yaml_file**, **heatmap_values_file**, **R_plots_file**, **filename_sorting**, **mzQC_file**, **log_file**, **report_file_prefix**, **report_file_PDF**, **report_file_HTML**
#' 
#' @export
#' 
getReportFilenames = function(folder, report_name_has_folder = TRUE, mzTab_filename = NULL)
{
  ## package version: added to output filename
  pv = try(utils::packageVersion("PTXQC"))
  if (inherits(pv, "try-error")) pv = "_unknown"
  report_version = paste0("v", pv)  
  
  ## amend report filename with a foldername where it resides, to ease discerning different reports in a PDF viewer
  extra_folderRef = ""
  folders = rev(unlist(strsplit(normalizePath(folder, winslash = .Platform$file.sep), split=.Platform$file.sep)))
  while (length(folders)){
    if (!grepl(":", folders[1]) & folders[1]!="txt") {
      extra_folderRef = paste0("_", folders[1])
      break;
    } else folders = folders[-1]
  }
  
  ## complete path + report_vXY
  report_file_simple = paste0(folder, .Platform$file.sep, "report_", report_version)
  ## .. + myMzTab [optional]
  if (!is.null(mzTab_filename) && nchar(mzTab_filename) > 0) {
    report_file_simple = paste0(report_file_simple, "_", gsub("\\.mzTab$", "", basename(mzTab_filename), ignore.case = TRUE))
  }
  
  
  ## include folder-name at the end
  if (report_name_has_folder)
    report_file_prefix = paste0(report_file_simple, extra_folderRef)
  else
    report_file_prefix = report_file_simple
  
  
  fh = list(yaml_file = paste0(report_file_prefix, ".yaml"),
            heatmap_values_file = paste0(report_file_prefix, "_heatmap.txt"),
            R_plots_file = paste0(report_file_prefix, "_plots.Rdata"),
            filename_sorting = paste0(report_file_prefix, "_filename_sort.txt"),
            mzQC_file = paste0(report_file_prefix, ".mzQC"),
            log_file = paste0(report_file_prefix, ".log"),
            report_file_prefix = report_file_prefix,
            report_file_PDF = paste0(report_file_prefix, ".pdf"),
            report_file_HTML = paste0(report_file_prefix, ".html")
           )
  return (fh)
}


#'
#' Extract the number of protein groups observed per Raw file
#' from an evidence table.
#'
#' Required columns are "protein.group.ids", "fc.raw.file" and "is.transferred".
#' 
#' If match-between-runs was enabled during the MaxQuant run,
#' the data.frame returned will contain separate values for 'transferred' evidence
#' plus an 'MBRgain' column, which will give the extra MBR evidence in percent.
#' 
#' @param df_evd Data.frame of evidence.txt as read by MQDataReader
#' @return Data.frame with columns 'fc.raw.file', 'counts', 'category', 'MBRgain'
#'
#'
getProteinCounts = function(df_evd) {
    
  required_cols = c("protein.group.ids", "fc.raw.file", "is.transferred")
  if (!all(required_cols %in% colnames(df_evd))) {
    stop("getProteinCounts(): Missing columns!")
  }

  ## report Match-between-runs data only if if it was enabled
  reportMTD = any(df_evd$is.transferred)
  
  prot_counts = plyr::ddply(df_evd, "fc.raw.file", .fun = function(x, reportMTD)
  {
    ## proteins
    x$group_mtdinfo = paste(x$protein.group.ids, x$is.transferred, sep="_")
    # remove duplicates (since strsplit below is expensive) -- has no effect on double counting of PROTEINS(!), since it honors MTD
    xpro = x[!duplicated(x$group_mtdinfo),]
    # split groups
    p_groups = lapply(as.character(xpro$protein.group.ids), function(x) {
      return (strsplit(x, split=";", fixed=TRUE))
    })
    # get number of unique proteins
    pg_set_genuineUnique = unique(unlist(p_groups[!xpro$is.transferred]))
    # MBR will contribute peptides of already known proteins
    pg_set_allMBRunique = unique(unlist(p_groups[xpro$is.transferred]))
    pg_count_GenAndMBR = length(intersect(pg_set_allMBRunique, pg_set_genuineUnique))
    # ... and peptides of proteins which are new in this Raw file (few)
    pg_count_newMBR = length(pg_set_allMBRunique) - pg_count_GenAndMBR
    # ... genuine-only proteins:
    pg_count_onlyGenuine = length(pg_set_genuineUnique) - pg_count_GenAndMBR
    
    ## return values in MELTED array
    if (reportMTD)
    {
      df = data.frame(counts = c(pg_count_onlyGenuine, pg_count_GenAndMBR, pg_count_newMBR),
                      category = c("genuine (exclusive)", "genuine + transferred", "transferred (exclusive)"),
                      MBRgain = c(NA, NA, pg_count_newMBR / (pg_count_onlyGenuine + pg_count_GenAndMBR) * 100))
    } else {
      df = data.frame(counts = c(pg_count_onlyGenuine),
                      category = c("genuine"),
                      MBRgain = NA)
    }  
    
    return (df)
  }, reportMTD = reportMTD)
  
  
  return (prot_counts)
}


#'
#' Extract the number of peptides observed per Raw file
#' from an evidence table.
#'
#' Required columns are "fc.raw.file", "modified.sequence" and "is.transferred".
#' 
#' If match-between-runs was enabled during the MaxQuant run,
#' the data.frame returned will contain separate values for 'transferred' evidence
#' plus an 'MBRgain' column, which will give the extra MBR evidence in percent.
#' 
#' @param df_evd Data.frame of evidence.txt as read by MQDataReader
#' @return Data.frame with columns 'fc.raw.file', 'counts', 'category', 'MBRgain'
#'
#'
getPeptideCounts = function(df_evd) {
  
  required_cols = c("fc.raw.file", "modified.sequence", "is.transferred")
  if (!all(required_cols %in% colnames(df_evd))) {
    stop("getPeptideCounts(): Missing columns!")
  }
  
  ## report Match-between-runs data only if if it was enabled
  reportMTD = any(df_evd$is.transferred)
  
  pep_counts = plyr::ddply(df_evd, "fc.raw.file", .fun = function(x, reportMTD)
  {
    x <<- x
    #pep_count_genuineAll = sum(!x$is.transferred) # (we count double sequences... could be charge +2, +3,... or oversampling)
    pep_set_genuineUnique = unique(x$modified.sequence[!x$is.transferred]) ## unique sequences (discarding PTM's)
    pep_set_allMBRunique = unique(x$modified.sequence[x$is.transferred])
    pep_count_GenAndMBR = length(intersect(pep_set_genuineUnique, pep_set_allMBRunique)) ## redundant, i.e. both Genuine and MBR
    pep_count_newMBR = length(pep_set_allMBRunique) - pep_count_GenAndMBR ## new MBR peptides
    pep_count_onlyGenuine = length(pep_set_genuineUnique) - pep_count_GenAndMBR ## genuine-only peptides
    
    
    ## return values in MELTED array
    if (reportMTD)
    {
      df = data.frame(counts = c(pep_count_onlyGenuine, pep_count_GenAndMBR, pep_count_newMBR),
                      category = c("genuine (exclusive)", "genuine + transferred", "transferred (exclusive)"),
                      MBRgain = c(NA, NA, pep_count_newMBR / (pep_count_onlyGenuine + pep_count_GenAndMBR) * 100))
    } else {
      df = data.frame(counts = c(pep_count_onlyGenuine),
                      category = c("genuine"),
                      MBRgain = NA)
    }  
    
    return (df)
  }, reportMTD = reportMTD)
  
  
  return (pep_counts)
}

#'
#' Check if a file is writable and blocks an interactive session, waiting for user input.
#'
#' This functions gives the user a chance to make the output file writeable before a write attempt
#' is actually made by R to avoid having run the whole program again upon write failure.
#'
#' Note: The file will not be overwritten or changed by this function.
#' 
#' @param filename The file to test for writable
#' @param prompt_text If not writable, show this prompt text to the user
#' @param abort_answer If the user enters this string into the prompt, this function will stop()
#' @return TRUE if writable, FALSE if aborted by user or (not-writeable and non-interactive)
#'
wait_for_writable = function(filename,
                             prompt_text = paste0("The file '", filename, "' is not writable. Please close all applications using this file. Press '", abort_answer, "' to abort!"),
                             abort_answer = "n")
{
  while(TRUE)
  {
    is_writeable = try({
                       f = file(filename, open="a"); ## test for append, which will not overwrite the file
                       close(f)
                       }, silent = TRUE
                       )
                       
    if (inherits(is_writeable, "try-error")) {
      if (interactive()) {
        r = invisible(readline(prompt = prompt_text))
        if (r == abort_answer)
        { ## manual abort by user
          return (FALSE)
        }
      } else
      { ## not writable and non-interactive .. we cannot wait
        return(FALSE)  
      }
    } else {
      return (TRUE)
    }
  }
  
  ## unreachable
  return (FALSE)
}



#'
#' Estimate the empirical density and return it
#'
#'
#' @param samples Vector of input values (samples from the distribution)
#' @param y_eval Vector of points where CDF is evaluated (each percentile by default)
#' @return Data.frame with columns 'x', 'y'
#'
#' @examples
#'   plot(getECDF(rnorm(1e4)))
#'   
#' @export
#'    
getECDF = function(samples, y_eval = (1:100)/100)
{
  ## internal little helper
  inv_ecdf <- function(f){ 
    x <- environment(f)$x 
    y <- environment(f)$y 
    approxfun(y, x) 
  } 
  ## compute ECDF
  ec = ecdf(samples)
  ## sample it at certain y-positions
  iec = inv_ecdf(ec)
  x_eval = iec(y_eval)
  r = data.frame(x = x_eval, y = y_eval)
  return (r)
}

#'
#' Discretize RT peak widths by averaging values per time bin.
#'
#' Should be applied for each Raw file individually.
#'
#' Returns a data.frame, where 'bin' gives the index of each bin, 'RT' is 
#' the middle of each bin and 'peakWidth' is the averaged peak width per bin.
#'
#' @param data Data.frame with columns 'retention.time' and 'retention.length'
#' @param RT_bin_width Bin size in minutes
#' @return Data.frame with columns 'bin', 'RT', 'peakWidth'
#' 
#' @export
#' 
#' @examples
#'   data = data.frame(retention.time = seq(30,200, by=0.001)) ## one MS/MS per 0.1 sec
#'   data$retention.length = seq(0.3, 0.6, length.out = nrow(data)) + rnorm(nrow(data), 0, 0.1)
#'   d = peakWidthOverTime(data)
#'   plot(d$RT, d$peakWidth)
#'   
#'   
peakWidthOverTime = function(data, RT_bin_width = 2) 
{      
  r = range(data$retention.time, na.rm = TRUE)
  brs = seq(from = r[1], to = r[2] + RT_bin_width, by = RT_bin_width)
  data$bin = findInterval(data$retention.time, brs, all.inside = TRUE) ## faster than cut(..., labels = FALSE)
  #data$bin = cut(data$retention.time, breaks = brs, include.lowest = TRUE, labels = FALSE)
  retLStats = plyr::ddply(data, "bin", .fun = function(xb) {
    data.frame(RT = brs[xb$bin[1]], peakWidth = median(xb$retention.length, na.rm = TRUE))
  })
  return(retLStats)
}

#' Get all currently available metrics
#'  
#' @param DEBUG_PTXQC Use qc objects from the package (FALSE) or from environment (TRUE/DEBUG)
#' @return List of matric objects
#' 
getMetricsObjects = function(DEBUG_PTXQC = FALSE)
{
  if (DEBUG_PTXQC) {
    lst_qcMetrics_str = ls(sys.frame(which = 0), pattern="qcMetric_") ## executed outside of package, i.e. not loaded...
  } else {
    lst_qcMetrics_str = ls(name = getNamespace("PTXQC"), pattern="qcMetric_") 
  }
  if (length(lst_qcMetrics_str) == 0) stop("getMetricsObjects(): No metrics found! Very weird!")
  
  lst_qcMetrics = sapply(lst_qcMetrics_str, function(m) {
    q = get(m)
    if (inherits(q, "refObjectGenerator")) {
      return(q$new())
    }
    return(NULL)
  })
  return(lst_qcMetrics)
}

#'
#' Make a color (given as name or in RGB) darker by factor x = [0 = black, 1=unchanged]
#' @param color A color as understood by col2rgb
#' @param factor Between 0 (make black) and 1 (leave color as is)
#' @return darkened color
#' 
darken = function(color, factor=0.8){
  dark_col = grDevices::rgb(t(grDevices::col2rgb(color) * factor), maxColorValue=255)
  return(dark_col)
}


#'
#' When MaxQuant is run with a wrong locale (i.e. the decimal separator is not a '.', but a ','), 
#' then MaxQuant results are plainly wrong and broken. The can be detected by, e.g. checking for negative charge annotation
#'
#' @param df_evd Evidence table from which we only need the 'charge' column
#'
checkEnglishLocale = function(df_evd) {
  if (!checkInput(c("charge"), df_evd)) return(TRUE)
  
  
  if (any(df_evd$charge < 0)){
    return (FALSE)
  }
  else return (TRUE)
}

library(PTXQC)
context("fcn_misc.R")

test_that("delLCP", {
  ## using expect_equivalent(), since delLCP returns a named vector, but we only check the values
  expect_equivalent(delLCP(c("TK12345_H1"), min_out_length=0), "")
  expect_equivalent(delLCP(c("TK12345_H1"), min_out_length=4), "5_H1")
  expect_equivalent(delLCP(c("TK12345_H1"), min_out_length=4, add_dots = TRUE), "..5_H1")
  expect_equivalent(delLCP(c("TK12345_H1", "TK12345_H2"), min_out_length=4), c("5_H1", "5_H2"))
  expect_equivalent(delLCP(c("TK12345_H1", "TK12345_H2"), min_out_length=4, add_dots = TRUE), c("..5_H1", "..5_H2"))
  expect_equivalent(delLCP(c("TK12345_H1", "TK12345_H2"), min_out_length=8), c("12345_H1", "12345_H2"))
  
  ## (unchanged, even though its 10 chars, since '..' would add another two)
  expect_equivalent(delLCP(c("TK12345_H1", "TK12345_H2"), min_out_length=8, add_dots = TRUE), c("TK12345_H1", "TK12345_H2"))
  ## (unchanged)
  expect_equivalent(delLCP(c("TK12345_H1", "TK12345_H2"), min_out_length=60), c("TK12345_H1", "TK12345_H2"))
  ## (unchanged)
  expect_equivalent(delLCP(c("TK12345_H1", "TK12345_H2"), min_out_length=60, add_dots = TRUE), c("TK12345_H1", "TK12345_H2"))

})

test_that("delLCS", {
  ## using expect_equivalent(), since delLCP returns a named vector, but we only check the values
  expect_equivalent(delLCS(c("TK12345_H1")), "")
  expect_equivalent(delLCS(c("TK12345_H1", "TK12345_H2")), c("TK12345_H1", "TK12345_H2"))
  expect_equivalent(delLCS(c("TK12345_H1", "TK12!45_H1")), c("TK123", "TK12!"))
})

test_that("lcpCount", {
  expect_equal(lcpCount(c("TK12345_H1")), 10)
  expect_equal(lcpCount(c("TK12345_H1", "TK12345!_H2")), 7)
  expect_equal(lcpCount(c("TK12345_H1", "TK12345_H2")), 9)
})

test_that("lcsCount", {
  expect_equal(lcsCount(c("TK12345_H1")), 10)
  expect_equal(lcsCount(c("TK12345_H1", "TK12345!_H2")), 0)
  expect_equal(lcsCount(c("TK12345!_H2", "TK12345_H2")), 3)
})

test_that("LCS", {
  expect_equal(LCS("TK12345_H1", "TK12345!_H2"), "TK12345")
  expect_equal(LCS("PRE1_TK12345!_H2", "PRE2_TK12345_H2"), "_TK12345")
  expect_equal(LCS("PRE1_TK12345_H2", "PRE2_TK12345_H2"), "_TK12345_H2")
  expect_equal(LCS("TK12345_H1", ""), "")
  expect_equal(LCS("", "TK12345_H1"), "")
})

test_that("LCSn", {
  expect_equal(LCSn(c("1_abcde...")), "1_abcde...")
  expect_equal(LCSn(c("123")), "123")
  expect_equal(LCSn(c("123"), min_LCS_length= 7), "")
  expect_equal(LCSn(c("1_abcde...", "2_abcd...", "x_abc...")), "_abc")
  expect_equal(LCSn(c("1_abcde...", "2_abcd...", "x_abc..."), min_LCS_length= 7), "")
})


test_that("simplifyNames", {
  expect_equal(simplifyNames(c('TK20130501_H2M1_010_IMU008_CISPLA_E3_R1.raw')),
               c("TK20130501_H2M1_010_IMU008_CISPLA_E3_R1.raw"))
  expect_equal(simplifyNames(c('TK20130501_H2M1_010_IMU008_CISPLA_E3_R1.raw', 'TK20130501_H2M1_026_IMU008_CISPLA_E7_R2.raw'), infix_iterations = 2),
               c("TK.._010_I.._E3_R1.raw","TK.._026_I.._E7_R2.raw"))
  expect_equal(simplifyNames(c('TK20130501_H2M1_010_IMU008_CISPLA_E3_R1.raw', 'TK20130501_H2M1_026_IMU008_CISPLA_E7_R2.raw'), infix_iterations = 1),
               c("TK.._010_IMU008_CISPLA_E3_R1.raw","TK.._026_IMU008_CISPLA_E7_R2.raw"))
  
  ## could be shortened to c("1ab..gh", "2ab..gh"), but thats too short...
  expect_equal(simplifyNames(c("1abcdefgh", "2abcdefgh"), min_out_length=99), c("1abcdefgh", "2abcdefgh"))
  expect_equal(simplifyNames(c("1abcdefgh", "2abcdefgh"), min_out_length=3), c("1ab..gh", "2ab..gh"))
  
  inp = unlist(strsplit("3265_Lumik3_Shakya_150IS_HCD_11A_1 3265_Lumik3_Shakya_150IS_HCD_11B_1 3265_Lumik3_Shakya_150IS_HCD_11C_1 3265_Lumik3_Shakya_150IS_HCD_11D_1 3265_Lumik3_Shakya_150IS_HCD_12A_1 3265_Lumik3_Shakya_150IS_HCD_12B_1 3265_Lumik3_Shakya_150IS_HCD_12C_1 3265_Lumik3_Shakya_150IS_HCD_12D_1 3265_Lumik3_Shakya_150IS_HCD_13A_1 3265_Lumik3_Shakya_150IS_HCD_13B_1 3265_Lumik3_Shakya_150IS_HCD_13C_1 3265_Lumik3_Shakya_150IS_HCD_13D_1 3265_Lumik3_Shakya_150IS_HCD_14A_1 3265_Lumik3_Shakya_150IS_HCD_14B_1 3265_Lumik3_Shakya_150IS_HCD_14C_1 3265_Lumik3_Shakya_150IS_HCD_14D_1", " "))
  expc = unlist(strsplit("32.._11A_1,32.._11B_1,32.._11C_1,32.._11D_1,32.._12A_1,32.._12B_1,32.._12C_1,32.._12D_1,32.._13A_1,32.._13B_1,32.._13C_1,32.._13D_1,32.._14A_1,32.._14B_1,32.._14C_1,32.._14D_1", ","))
  expect_equal(simplifyNames(inp, infix_iterations = 2), expc)

  
  expect_equal(simplifyNames(c("bla", "foo"), min_LCS_length=99), c("bla", "foo"))
  ## internally, min_LCS_length>=6 is required
  expect_error(simplifyNames(c("bla", "foo"), min_LCS_length=5))
  
})

test_that("shortenStrings", {
  ## no not allow duplicates (since they are already ambiguous, even without shortening)
  dups = c("someString", "aDuplicate", "aDuplicate")
  expect_error(shortenStrings(dups))
  expect_equivalent(shortenStrings(dups, allow_duplicates = TRUE), dups)
  
  expect_equivalent(shortenStrings(c("gamg_101", "gamg_101230100451", "jurkat_06_100731121305", "jurkat_06_1"), max_len = 20),
                    c("gamg_101", "gamg_101230100451", "jurkat_06_10073112..", "jurkat_06_1"))

})

test_that("supCount", {
  expect_equal(supCount(c("abcde...", "abcd...", "abc...")), 5)

  x = c("doubled", "doubled", "aLongDummyString")
  ## due to duplicated entries there is no prefix which makes them unique
  expect_equal(substr(x, 1, supCount(x)), x)
})



test_that("correctSetSize", {
  expect_equal(correctSetSize(8, 5), 4)
  expect_equal(correctSetSize(101, 25), 26)
})

test_that("byX", {
  expect_equivalent(byX(data.frame(d=1:10), 1:10, 1, sum, sort_indices = FALSE), as.list(1:10))
  expect_equivalent(byX(data.frame(d=1:10), 1:10, 1, sum, sort_indices = TRUE), as.list(1:10))
  expect_equivalent(byX(data.frame(d=1:10), as.character(1:10), 1, sum, sort_indices = TRUE), as.list(c(1, 10, 2:9)))
  expect_equivalent(byX(data.frame(d=1:10), 1:10, 2, sum, sort_indices = FALSE), as.list(c(1+2,3+4,5+6,7+8,9+10)))
  expect_equivalent(byX(data.frame(d=1:10), 1:10, 7, sum, sort_indices = FALSE), as.list(c(sum(1:7), sum(8:10))))
})

test_that("byXflex", {
  expect_equivalent(byXflex(data.frame(d=1:10), 1:10, 1, sum, sort_indices = FALSE), as.list(1:10))
  expect_equivalent(byXflex(data.frame(d=1:10), 1:10, 1, sum, sort_indices = TRUE), as.list(1:10))
  expect_equivalent(byXflex(data.frame(d=1:10), as.character(1:10), 1, sum, sort_indices = TRUE), as.list(c(1,10, 2:9)))
  expect_equivalent(byXflex(data.frame(d=1:10), 1:10, 2, sum, sort_indices = FALSE), as.list(c(1+2,3+4,5+6,7+8,9+10)))
  ## within 150% (7*1.5), just create one group
  expect_equivalent(byXflex(data.frame(d=1:10), 1:10, 7, sum, sort_indices = FALSE), as.list(c(sum(1:10))))
  ## two groups are maintained, but instead of 6, 4 we use 5,5 for more balanced group size
  expect_equivalent(byXflex(data.frame(d=1:10), 1:10, 6, sum, sort_indices = FALSE), as.list(c(sum(1:5), sum(6:10))))
})

test_that("getMaxima", {
  expect_equivalent(getMaxima(c(1,0,3,4,5,0)), c(TRUE,FALSE,FALSE,FALSE,TRUE,FALSE))
  expect_equivalent(getMaxima(c(1,0,3,4,5,0), thresh_rel=0.5), c(FALSE,FALSE,FALSE,FALSE,TRUE,FALSE))
})

thinOut = function(data, filterColname, binsize)
test_that("thinOut", {
  data = data.frame(x = c(1:100), y = 1:10)
  expect_equal(dim(thinOut(data, "x", 5)), c(21,2))
  # duplicates in 'x' column are irrelevant
  data = data.frame(x = c(1:100, 1:10), y = 1:110)
  expect_equal(dim(thinOut(data, "x", 5)), c(21,2))
})

qcMetric_EVD_modTable =  setRefClass(
  "qcMetric_EVD_modTable",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(    
    helpTextTemplate = 
      "Compute an occurence table of modifications (e.g. Oxidation (M)) for all peptides, including the unmodified.

The plot will show percentages, i.e. is normalized by the total number of peptide sequences (where different charge state counts as a separate peptide) per Raw file.

The sum of frequencies may exceed 100% per Raw file, since a peptide can have multiple modifications.
E.g. given three peptides in a single Raw file                 <br>
1. _M(Oxidation (M))LVLDEADEM(Oxidation (M))LNK_               <br>
2. _(Acetyl (Protein N-term))M(Oxidation (M))YGLLLENLSEYIK_    <br>
3. DPFIANGER                                                   <br>

, the following frequencies arise:

* 33% of 'Acetyl (Protein N-term)' <br>
* 33% of 'Oxidation (M)'           <br>
* 33% of '2 Oxidation (M)'         <br>
* 33% of 'Unmodified'              <br>

Thus, 33% of sequences are unmodified, implying 66% are modified at least once. 
If a modification, e.g. Oxidation(M), occurs multiple times in a single peptide it's listed as a separate modification (here '2 Oxidation (M)').

Heatmap score [EVD: Pep ModTable]: Deviation of (unmodified peptides fraction) when compared to a representative Raw file ('qualMedianDist' function).
",
    workerFcn = function(.self, df_evd)
    {
      ## completeness check
      if (!checkInput(c("fc.raw.file", "modifications"), df_evd)) return()

      name_unmod = "Unmodified"
      name_unmod_inverse = "Modified (total)"
      tbl = modsToTableByRaw(df_evd, name_unmod, name_unmod_inverse)
      lpl = byXflex(tbl, tbl$fc.raw.file, FUN = plot_peptideMods)
      #for (pl in lpl) print(pl)
      
      ## QC measure: deviation from representative sample (in terms of unmodified peptide count)
      qc_mods = tbl[tbl$modification_names %in% c(name_unmod, name_unmod_inverse),]
      qc_mods[, .self$qcName] = qualMedianDist(qc_mods$Freq / 100)
      
      return(list(plots = lpl, qcScores = qc_mods[, c("fc.raw.file", .self$qcName)]))
    }, 
    qcCat = "prep", 
    qcName = "EVD:~Peptide~VarMod", 
    orderNr = 0103
  )
    return(.self)
  })
)

#' 
#' Convert list of (mixed)modifications to a frequency table
#' 
#' @param mod_list A vector with modifications, each for a specific peptide. Multiple mods per entry are allowed, each separated by comma.
#' @return A data.frame with 'modification_names' and 'Freq' (0-100)
#'
#' @export
#' 
#' @examples
#' modsToTable(c("Ox (M)", 
#'               "Unmodified",
#'               "Ox (M),Acetyl (Prot N-term)",
#'               "2 Ox (M)",
#'               "Unmodified",
#'               "Unmodified"))
#' 
#'
modsToTable = function(mod_list)
{
  modification_names = unlist(strsplit(mod_list, ",", fixed=TRUE))
  tt = data.frame(table(modification_names) / length(mod_list) * 100)
  if (! ("modification_names" %in% colnames(tt))) stop("table has wrong column names!")
  return(tt)
}

context("modstoTable.R")

test_that("modstoTable", {
  r = modsToTable(c("A", "Unmodified", "A,B", "C", "Unmodified", "Unmodified"))
 

  exp = data.frame(modification_names = factor(c("A", "B", "C", "Unmodified")), 
                   Freq = c(200/6, 100/6, 100/6, 300/6))
  testthat::expect_equal(r, exp)
})

#'
#' Plot a text as graphic using ggplot2.
#'
#' @param title The title of the plot
#' @param text Centered text, can contain linebreaks
#' @param col Colour of text (excluding the title)
#' @return ggplot object
#'
#'
ggText = function(title, text, col = "black") {
  pl = ggplot(data.frame(text = text, ypos=1, xpos=1),
                       aes(x = .data$xpos, y = .data$ypos))  +
    geom_text(aes(label = .data$text), colour = col, family="mono") +
    theme_bw() +
    theme(plot.margin = grid::unit(c(1,1,1,1), "cm"), line = element_blank(), axis.title = element_blank(), panel.border = element_blank(),
                   axis.text = element_blank(), strip.text = element_blank(), legend.position="none") +
    ggtitle(title)
  return(pl)
}

#'
#' Augment a ggplot with footer text
#'
#' @param gg_obj ggplot2 object to be printed
#' @param bottom_left Footer text for bottom left side
#' @param bottom_right Footer text for bottom right side
#' @return -
#'
printWithFooter = function(gg_obj, bottom_left = NULL, bottom_right = NULL)
{
  if ("gtable" %in% class(gg_obj)) plot(gg_obj) else print(gg_obj)

  if (!is.null(bottom_left))
  {
    label = grid::textGrob(bottom_left,
                           x = 0.02,  # left side
                           y = 0.0,   # bottom
                           just = "left",
                           hjust = NULL,
                           vjust = -.5,
                           gp = grid::gpar(fontsize=7, col="#333333"))
    grid::grid.draw(label)
  }
  if (!is.null(bottom_right))
  {
    label = grid::textGrob(bottom_right,
                           x = 0.98,  # right side
                           y = 0.0,   # bottom
                           just = "right",
                           hjust = NULL,
                           vjust = -.5,
                           gp = grid::gpar(fontsize=7, col="#333333"))
    grid::grid.draw(label)
  }
}

#'
#' Inverse the order of items on the x-axis (for discrete scales)
#'
#' @param values The vector of values as given to the x aestetic
#' @param ... Other arguments forwarded to 'scale_y_discrete()'
#' @return ggplot object, concatenatable with '+'
#'
#' @import ggplot2
#'
scale_x_discrete_reverse = function(values, ...)
{
  if (inherits(values, "factor")) values = droplevels(values)
  return (scale_x_discrete(limits = rev(levels(values)), ... ))
}

#'
#' Inverse the order of items on the y-axis (for discrete scales)
#'
#' @param values The vector of values as given to the y aestetic
#' @param ... Other arguments forwarded to 'scale_y_discrete()'
#' @return ggplot object, concatenatable with '+'
#'
#' @import ggplot2
#'
scale_y_discrete_reverse = function(values, ...)
{
  if (inherits(values, "factor")) values = droplevels(values)
  return (scale_y_discrete(limits = rev(levels(values)), ... ))
}

#'
#' Function to thin out the number of labels shown on an axis in GGplot
#'
#' By default, 20 labels (or up to 40 see below) are shown.
#' If the number of items is less than twice the number of desired labels,
#' all labels will be shown (to avoid irregular holes for some labels).
#' I.e. if n=20, and x has 22 entries, there would be only two labels removed, giving a very irregular picture.
#' It only becomes somewhat regular if after any label there is at least one blank, i.e. at most
#' half the entries are labeled.
#' #'
#' Example:
#'   ## p is any ggplot object
#'   p + scale_y_discrete(breaks = ggAxisLabels)
#'   ## customize 'n'
#'   my.ggAxisLabels = function(x) ggAxisLabels(x, n = 4)
#'   p + scale_y_discrete(breaks = my.ggAxisLabels)
#'
#' @param x Vector of labels (passed by GGplot)
#' @param n Number of labels to show
#' @return Shortened version of 'x'
#'
#'
ggAxisLabels = function(x, n = 20)
{
  ## less than 20 items: nothing to do
  if (length(x) <= n) return (x)

  ## avoid irregular holes in labels if |x| < 2n (some neighbouring labels would be present, while others not)
  ## giving patterns like 1101011010110. So we might as well use all labels, ie. 11111111111111
  if (length(x) <= 2*n) return (x)  ## --> return all labels

  ## just pick 'n' out of |x|
  breaks = x[seq(1, length(x), length.out = n)]
  names(breaks) = x[breaks]
  return (breaks)
}


#'
#' Distribute a set of points with fixed y-values on a stretch of the x-axis.
#'
#' #'
#' Usage:
#'   ggplot(...) + geom_X(...) + pointsPutX(...)
#'
#' @param x_range [min,max] valid range of x-values
#' @param x_section [min,max] fraction in which to distribute the values (in [0,1] for min,max, e.g. c(0.03,0.08) for 3-8\%)
#' @param y Y-values
#' @param col Colour of the points (used as argument to aes(colour=))
#' @return ggplot object with new geom_point
#'
#' @import ggplot2
#' @export
#'
pointsPutX = function(x_range, x_section, y, col = NA){
  #require(ggplot2)
  x_dist = dist(x_range)
  x_pos = x_range[1] + x_section*x_dist
  d = data.frame(x_ = seq(x_pos[1], x_pos[2], length.out=length(y)), y_ = y, col_= col)

  if (is.na(d$col_[1])) pl = geom_point(data = d, aes(x = .data$x_, y = .data$y_))
  else pl = geom_point(data = d, aes(x = .data$x_, y = .data$y_, col = .data$col_))

  return (pl)
}


#'
#' A blank theme (similar to the deprecated theme_blank())
#'
#' @return A ggplot2 object, representing an empty theme
#'
#' @import ggplot2
#' @export
#'
theme_blank = function()
{
  theme_blank = theme_bw()
  theme_blank$line = element_blank()
  theme_blank$rect = element_blank()
  theme_blank$strip.text = element_blank()
  theme_blank$axis.title = element_blank()
  return (theme_blank)
}

#'
#' Return color brew palettes, but fail hard if number of requested colors
#' is larger than the palette is holding.
#'
#' Internally calls 'brewer.pal(n, palette)', checking 'n' beforehand.
#'
#' @param n Number of colours
#' @param palette Name of palette (e.g. "set1")
#' @return character vector of colors
#'
brewer.pal.Safe = function(n = 3, palette = "Set1")
{
  idx = which(rownames(RColorBrewer::brewer.pal.info) == palette)
  if (length(idx) != 1) stop("Palette ", palette," unknown!")
  if (RColorBrewer::brewer.pal.info$maxcolors[idx] < n) stop("Palette ", palette, " provides ", RColorBrewer::brewer.pal.info$maxcolors[idx], " colors, but not ", n, " as requested!")
  ## avoid warning about less than 3 colors, if n<3
  return (RColorBrewer::brewer.pal(max(3,n), palette)[1:n])
}

#' Retrieve a parameter value from a mqpar.xml file
#' 
#' If the file has the param, then return it as string.
#' If the file is missing, warning is shown and NULL is returned.
#' If the param (i.e. XML tag) is unknown or cannot be extracted, the program will quit (since this is a hard error).
#' When multiple occurrences of the param are found (usually due to parameter groups), we test if the values are all identical.
#' If so, the value is returned. If the values are different, a warning is emitted and NULL is returned unless 'allow_multiple = TRUE'
#' 
#' E.g. calling getMQPARValue("mqpar.xml", "//firstSearchTol")
#' will look up the line
#' <firstSearchTol>20</firstSearchTol>
#' and return "20" (string!).
#' 
#' 
#' @param mqpar_filename Filename (incl. absolute or relative path) to the mqpar.xml file
#' @param xpath          An XPath to extract the content of XML tag(s), e.g. '//firstSearchTol'
#' @param allow_multiple If the XPath expression returns more than one value, all values must be identical (not allowing multiple different values) or 'stop()' is called
#' 
#' @return The stored value as string(!)
#'
#' @export
#'
getMQPARValue = function(mqpar_filename, xpath, allow_multiple = FALSE)
{
  #xpath = "//firstSearchTol"
  #mqpar_filename = txt_files$mqpar
  if (!file.exists(mqpar_filename)) {
    message("Info: The file '", mqpar_filename, "' was not found. MaxQuant parameters could not be extracted. Will fall back to default value, which however is only an approximation.",
            " Please either: a) copy the mqpar.xml which was used for this MQ analysis into your TXT folder or,",
            " b) make sure that you configure all YAML parameters whose name starts with 'MQpar_' correctly.", immediate. = TRUE)
    return (NULL)
  }
  
  lines = xml2::read_xml(mqpar_filename)
  ## is the tag present multiple times? (if yes, we found parameter groups)
  results = xml2::xml_text(xml2::xml_find_all(lines, xpath))

  if (length(results) == 0) {
    stop("getMQPARValue(): The XPath '", xpath, "' was not found in the file '", mqpar_filename, "'. Please contact the package support.", call. = FALSE);
  }
  
  if ((allow_multiple == FALSE)) {
    
    if (length(unique(results)) == 1) {
      return (results[1]);
    }
    
    warning("getMQPARValue(): The XPath '", xpath, "' was found more than once in the file '", mqpar_filename, "' with different values (probably due to usage of parameter groups).",
            " PTXQC currently cannot deal with that -- the YAML param is going to be used. Sorry.", immediate. = TRUE);
    return (NULL);
  }
  
  ## all tests passed, return the result(s)
  return (results)
}

#####################################################################

qcMetric_EVD_UserContaminant =  setRefClass(
  "qcMetric_EVD_UserContaminant",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate =
      "User defined contaminant plot based on peptide intensities and counts.
Usually used for Mycoplasma detection, but can be used for an arbitrary (set of) proteins.

All proteins (and their peptides) which contain the search string from the YAML file are considered contaminants.
The contaminant's search string is searched in the full FASTA header in proteinGroups.txt.
If proteinGroups.txt is not available/found,
only protein identifiers can be considered. The search realm used is given in the plot subtitle.
You should choose the contaminant name to be distinctive.
Only peptides belonging to a single protein group are considered when computing the fractions (contaminant vs. all),
since peptides shared across multiple groups are potentially false positives.

Two abundance measures are computed per Raw file:

   - fraction of contaminant intensity (used for scoring of the metric)
   - fraction of contaminant spectral counts (as comparison; both should be similar)

If the intensity fraction exceeds the threshold (indicated by the dashed horizontal line) a contamination is assumed.

For each Raw file exceeding the threshold an additional plot giving cumulative Andromeda peptide
score distributions is shown.
This allows to decide if the contamination is true. Contaminant scores
should be equally high (or higher), i.e. to the right, compared to the sample scores.
Each graph's subtitle is augmented with a p-value of the Kologorov-Smirnoff test of this data
(Andromeda scores of contaminant peptides vs. sample peptides).
If the p-value is high, there is no score difference between the two peptide populations.
In particular, the contaminant peptides are not bad-scoring, random hits.
These p-values are also shown in the first figure for each Raw file. Note that the p-value is purely based
on Andromeda scores and is independent of intensity or spectral counts.


Heatmap score [EVD: Contaminant <name>]: boolean score, i.e. 0% (fail) if the intensity threshold was exceeded; otherwise 100% (pass).
",
    workerFcn = function(.self, df_evd, df_pg, lst_contaminants)
    {
      #lst_contaminants = yaml_contaminants
      ## completeness check
      ## PG is either missing, or has the correct data
      if (!is.null(df_pg) | !checkInput(c("id", "fasta.headers"), df_pg)) return()
      ## "score" might not be present (e.g. missing in MQ 1.0.13.13)
      if (!checkInput(c("protein.group.ids", "type", "intensity", "fc.raw.file"),df_evd)) return()

      local_qcScores = data.frame()

      lpl = list()

      ca_entry = lst_contaminants[[1]]
      for (ca_entry in lst_contaminants)
      {
        ca = ca_entry[1]
        ##
        if (ca == FALSE) {
          cat("No special contaminants requested!\n")
          break;
        }

        ca_thresh = as.numeric(ca_entry[2])

        not_found = TRUE

        if (is.null(df_pg)) {
          ## only search in protein IDs
          pg_id = df_evd$protein.group.ids[grep(ca, df_evd$proteins)]
          ## this could be multiple PGs ("PG1; PG2") per cell, but we require unique peptides below, so its not a problem
          search_realm = "protein name only"
        } else {
          ## search in FASTA headers and protein IDs
          pg_id = df_pg$id[c(grep(ca, df_pg$fasta.headers, ignore.case = TRUE),
                             grep(ca, df_pg$protein.ids, ignore.case = TRUE))]
          search_realm = "full FASTA header"
        }


        if (length(pg_id) > 0)
        {
          ## we might or might not have found something... we plot it anyways, so the user can be sure that we searched for it
          not_found = FALSE

          ## find peptides which only have one group (ignoring razor peptides where we cannot be sure)
          evd_uniqueGroup = !grepl(";", df_evd$protein.group.ids)
          ## do not trust MBR here. We want real evidence!
          evd_realMS = !grepl("MATCH", df_evd$type)
          ## for each Raw file: find unique peptides of our contaminant
          cont_data.l = plyr::dlply(df_evd[evd_uniqueGroup & evd_realMS, ], "fc.raw.file",
                              function(x) {
                                if (length(grep(";", x$protein.group.ids))) stop("more than one proteinGroup for supposedly unique peptide...")

                                x$idx_cont = x$protein.group.ids %in% pg_id

                                sc = sum(x$idx_cont) / nrow(x) * 100
                                int = sum(as.numeric(x$intensity[x$idx_cont]), na.rm = TRUE) / sum(as.numeric(x$intensity), na.rm = TRUE) * 100

                                above.thresh = (sc > ca_thresh) | (int > ca_thresh)
                                cont_scoreECDF = NULL;
                                if ("score" %in% colnames(x)) {
                                  cont_scoreECDF = plyr::ddply(x, "idx_cont", function(xx) {
                                    if (length(unique(xx$score)) < 2) return(NULL) ## not enough data for ECDF
                                    r = getECDF(xx$score)
                                    r$condition = c("sample", "contaminant")[xx$idx_cont[1]+1]
                                    return(r)
                                  })
                                }
                                if (!any(x$idx_cont)){
                                  ks_p = NA
                                } else { ## no contaminant peptide
                                  ks_p = suppressWarnings(  ## brags about '-value will be approximate in the presence of ties'
                                    ks.test(x$score[x$idx_cont], x$score[!x$idx_cont], alternative = "greater")$p.value
                                  )
                                }
                                return (list(cont_data = data.frame(spectralCount = sc, intensity = int,
                                                                    above.thresh = above.thresh, fc.raw.file = x$fc.raw.file[1],
                                                                    score_KS = ks_p),
                                             cont_scoreECDF = cont_scoreECDF))
                              })
          head(cont_data.l)

          ## melt
          cont_data = plyr::ldply(cont_data.l, function(l) { l$cont_data })
          cont_data.long = reshape2::melt(cont_data, id.vars="fc.raw.file")

          #
          # old: not_found = all(cont_data.long$value[cont_data.long$variable == "above.thresh"] == FALSE)
        }

        if (not_found)
        { ## identifier was not found in any sample
          pl_cont = ggText("EVD: Contaminants",
                           paste0("Contaminant '", ca, "' was not found in any sample.\n\nDid you use the correct database?"),
                           "red")
          lpl = append(lpl, list(pl_cont))
        } else {
          ## plot User-Contaminants
          lpl_i = byXflex(data = cont_data.long, indices = cont_data.long$fc.raw.file, subset_size = 120,
                          FUN = plot_ContUser, sort_indices = TRUE,
                          name_contaminant = ca, extra_limit = ca_thresh, subtitle = paste("search realm:", search_realm))
          lpl = append(lpl, lpl_i)

          ## plot Andromeda score distribution of contaminant vs. sample
          pl_andr = plyr::llply(cont_data.l, function(l)
          {
            if (l$cont_data$above.thresh == FALSE ||
                is.null(l$cont_scoreECDF))
            {
              return(NULL) ## some entries might be skipped (not significant)
            }
            p = plot_ContUserScore(l$cont_scoreECDF, l$cont_data$fc.raw.file, l$cont_data$score_KS)
            #print(p)
            return(p)
          })
          pl_andr_nonNull = plyr::compact(pl_andr) ## remove 'NULL' entries from plot list
          lpl = append(lpl, pl_andr_nonNull)

          ## add heatmap column
          cname = sprintf(.self$qcName, ca)
          cont_data[,cname] = as.numeric(!cont_data$above.thresh) ## inverse (0 is 'bad')

          qcScore = cont_data[, c("fc.raw.file", cname)]
          if (ncol(local_qcScores) == 0){
            local_qcScores = qcScore
          } else {
            local_qcScores = merge(local_qcScores, qcScore)
          }

        }
      } ## contaminant loop

      return(list(plots = lpl, qcScores = local_qcScores))
    },
    qcCat = "Prep",
    qcName = "EVD:User~Contaminant~(%s)",
    orderNr = 0020
  )
    return(.self)
  })
)


#####################################################################

qcMetric_EVD_PeptideInt =  setRefClass(
  "qcMetric_EVD_PeptideInt",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate =
      "Peptide precursor intensity per Raw file from evidence.txt WITHOUT match-between-runs evidence.
Low peptide intensity usually goes hand in hand with low MS/MS identifcation rates and unfavourable signal/noise ratios,
which makes signal detection harder. Also instrument acquisition time increases for trapping instruments.

Failing to reach the intensity threshold is usually due to unfavorable column conditions, inadequate
column loading or ionization issues. If the study is not a dilution series or pulsed SILAC experiment, we
would expect every condition to have about the same median log-intensity (of 2<sup>%1.1f</sup>).
The relative standard deviation (RSD) gives an indication about reproducibility across files and should be below 5%%.

Depending on your setup, your target thresholds might vary from PTXQC's defaults.
Change the threshold using the YAML configuration file.

Heatmap score [EVD: Pep Intensity (>%1.1f)]:
  Linear scale of the median intensity reaching the threshold, i.e. reaching 2<sup>21</sup> of 2<sup>23</sup> gives score 0.25.
",
    workerFcn = function(.self, df_evd, thresh_intensity)
    {
      ## completeness check
      if (!checkInput(c("fc.raw.file", "intensity", "contaminant"), df_evd)) return()

      ## update helpText
      .self$helpText = sprintf(.self$helpTextTemplate, thresh_intensity, thresh_intensity)

      medians_pep = plyr::ddply(df_evd[ , c("fc.raw.file", "intensity")], "fc.raw.file",
                          function(x) data.frame(med = log2(quantile(x$intensity, probs=0.5, na.rm = TRUE))))

      int_dev_pep = RSD((medians_pep$med))
      int_dev.s = pastet("INT RSD [%]", round(int_dev_pep, 3))
      lpl = boxplotCompare(data = df_evd[, c("fc.raw.file", "intensity", "contaminant")],
                           log2 = TRUE,
                           mainlab = "EVD: peptide intensity distribution",
                           ylab = expression(log[2]*" intensity"),
                           sublab = paste0("RSD ", round(int_dev_pep, 1),"% (expected < 5%)\n"),
                           abline = thresh_intensity)
      #for (pl in lpl) print(pl)

      ## QC measure for peptide intensity
      qc_pepint = medians_pep
      cname = sprintf(.self$qcName, thresh_intensity)
      qc_pepint[,cname] = qualLinThresh(2^qc_pepint$med, 2^thresh_intensity) ## use non-log space
      qcScore = qc_pepint[, c("fc.raw.file", cname)]

      return(list(plots = lpl, qcScores = qcScore))
    },
    qcCat = "prep",
    qcName = "EVD:~Peptide~Intensity~(\">%1.1f\")",
    orderNr = 0030
  )
    return(.self)
  })
)


#####################################################################

qcMetric_EVD_ReporterInt =  setRefClass(
  "qcMetric_EVD_ReporterInt",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpText =
      "ITRAQ/TMT reporter intensity violin plots of all PSMs for each channel and Raw file.
The second subplot shows labeling efficiency (LE), i.e the fraction of PSMs with non-zero abundance (100% = full labeling of all PSMs; 0% = no reporter ions at all). This is used for heatmap scoring. See below.

There is a similar 'Experimental Group' based metric/plot based on proteins.txt.

PTXQC uses isotope-corrected intensities (eliminating channel carry-over) to allow for detection of empty channels, e.g. due to mis-labeling.
If MaxQuant did no isotope correction (i.e. corrected and uncorrected channels are equal),
the plot title will show a warning. The scores are too optimistic in this case (since carry-over will be mistaken for actual signal).

Note: global labelling efficiency can only be judged indirectly with this metric, since isobaric reporters where set as
      fixed modification. Thus, MaxQuant. will only identify labeled peptides in the first place.
      Observing only very few peptides (see peptide count metric), is a good indicator.
      However, if only the labeling of a few channels failed, this will be noticable here!

Labeling can still be poor, even though identification was successful.

A labeling efficiency (LE) is computed per Raw file AND channel as: the percentage of PSMs which have non-zero reporter intensity.
Ideally LE reaches 100 percent (all peptides have an intensity in the channel; biological missingness ignored).

Heatmap score: minimum labeling efficiency per Raw file across all channels.
I.e. for 4-plex ITRAQ and two Raw files, there will be 8 labeling efficiency (LE) values.
Each Raw file is now scored by the minimum LE of all its 4 channels.
",
    workerFcn=function(.self, df_evd)
    {
      ## completeness check
      if (!checkInput(c("fc.raw.file"), df_evd)) return()
      ## check if reporter.intensity.0... is present
      cols_reporter = grepv("^reporter.intensity.corrected.[0-9]", colnames(df_evd));
      cols_reporter.nc = grepv("^reporter.intensity.[0-9]", colnames(df_evd));
      if(length(cols_reporter) <= 1 || length(cols_reporter.nc) <= 1) {
        message("Info: Two reporter.intensity and two reporter.intensity.corrected columns are needed for metric ReporterIntensity.")
        return()}
      ## check if correction was done at all
      if (all(df_evd[1:1000, cols_reporter] == df_evd[1:1000, cols_reporter.nc], na.rm = TRUE))
      {
        title_subtext = "Warning: MaxQuant did NO isotope correction";
        title_color = "red"
      } else {
        title_subtext = "";
        title_color = "black"
      }
      g_title = "EVD: Reporter Intensities"
      ## use data.table for aggregation, its MUCH faster than ddply() and uses almost no extra memory
      df_reps = reshape2::melt(df_evd[, c("fc.raw.file", cols_reporter)],
                     id.vars ="fc.raw.file",
                     value.name = "intensity",
                     variable.name = "channel")
      head(df_reps)
      dt_reps = data.table::data.table(df_reps)

      ## do NOT remove -inf and NA's and 0's -- we need them to count labeling-efficiency (#entries with intensity > 0 vs. ALL)

      ## rename 'reporter.intensity.corrected.0' to '0'
      dt_reps$channel = substring(dt_reps$channel, nchar('reporter.intensity.corrected.') + 1)
      ## invert the channel order (so that channel 0 is highest, i.e. appears on top in plot)
      dt_reps$channel = factor(dt_reps$channel, levels = sort(unique(dt_reps$channel), decreasing = TRUE))
      head(dt_reps)

      ylims_minmax = range(dt_reps$intensity[dt_reps$intensity>0])
      if (is.na(ylims_minmax[1])) {ylims_minmax = range(1)} ## data is all 0! Make range at least 1, so log-range does not crash

      fcn_boxplot_internal = function(data, title_subtext = title_subtext, title_color = title_color)
      {
        ### first subplot (distribution of intensities)
        data_noZero = data[data$intensity!=0,]
        pl = ggplot(data=data_noZero) +
          geom_violin(aes(x = .data$fc.raw.file,
                          y = .data$intensity,
                          color = .data$channel,
                          fill = .data$channel
                         )) +
          xlab("") +
          ylab("reporter intensity (zeros removed)") +
          guides(#alpha = guide_legend(title="Label Eff"),
                 fill = guide_legend(reverse = TRUE), ## inverse label order, so that channel 0 is on top
                 color = guide_none()) +
          theme(axis.text.x = element_text(angle = 45, vjust = 0.5),
                legend.position = "right",
                plot.title = element_text(color = title_color)) +
          ggtitle(g_title, title_subtext) +
          PTXQC:::scale_x_discrete_reverse(unique(data$fc.raw.file)) +
          scale_y_log10(limits = ylims_minmax) +
          coord_flip()
        #pl

        ylims = data[, { #limits = boxplot.stats(intensity, coef = 0.7)$stats;
                          list(labEff_PC = sum(intensity > 0, na.rm = TRUE) / (.N))
                       }, by = c("fc.raw.file", "channel")]


        ### second subplot (labeling efficiency)
        pl_eff = ggplot(data = ylims) + geom_bar(aes(x = .data$fc.raw.file,
                                                     y = .data$labEff_PC * 100,
                                                     fill = .data$channel),
                                                 stat = "identity",
                                                 position = "dodge") +
          xlab("") +
          ylab("labelling efficiency (%)") +
          ylim(0, 100) +
          guides(fill = guide_legend(reverse = TRUE), ## inverse label order, so that channel 0 is on top
                 color = guide_none()) +
          theme(legend.position = "right") +
          ggtitle("Fraction of Non-Zero Intensities", "") +
          PTXQC:::scale_x_discrete_reverse(unique(ylims$fc.raw.file)) +
          coord_flip()
        #pl_eff
        pl_both = gridExtra::grid.arrange(pl, pl_eff, ncol=2)
        #print(pl)
        return(pl_both)
      }
      channel_count = length(cols_reporter)
      lpl = byXflex(data = dt_reps, indices = dt_reps$fc.raw.file, subset_size = round(40 / channel_count),
                    sort_indices = TRUE, FUN = fcn_boxplot_internal, title_subtext = title_subtext, title_color = title_color)
      lpl
      # heatmap scoring
      ## .. take min score over all channels
      ylims = dt_reps[, { #limits = boxplot.stats(intensity, coef = 0.7)$stats;
        list(labEff_PC = sum(intensity > 0, na.rm = TRUE) / (.N))
      }, by = c("fc.raw.file", "channel")]
      qcScore = ylims[, list(score_min = min(labEff_PC)), by=c("fc.raw.file")]
      colnames(qcScore) = c("fc.raw.file", .self$qcName)

      ## add manual title, since we return a grid.arrange() where automatic extraction is hard
      return(list(plots = lpl, qcScores = qcScore, title = rep(list(g_title), length(lpl))))
    },
    qcCat = "prep",
    qcName = "EVD:~Reporter~intensity",
    orderNr = 0031  ## should not show up in heatmap
  )
    return(.self)
  })
)


#####################################################################

qcMetric_EVD_ProteinCount =  setRefClass(
  "qcMetric_EVD_ProteinCount",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate =
      "Number of Protein groups (after FDR) per Raw file. A configurable target threshold is indicated as dashed line.

If MBR was enabled, three categories ('genuine (exclusive)', 'genuine + transferred', 'transferred (exclusive)'
are shown, so the user can judge the gain that MBR provides. Here, 'transferred (exclusive)' means that this protein group
has peptide evidence which originates only from transferred peptide IDs. The quantification is (of course) always from the
local Raw file.
Proteins in the 'genuine + transferred' category have peptide evidence from within the Raw file by MS/MS, but at the same time
also peptide IDs transferred to this Raw file using MBR were used. It is not unusual to see the 'genuine + transferred' category be the
rather large, since a protein group usually has peptide evidence from both sources.
To see of MBR worked, it is better to look at the two MBR-related metrics.

If MBR would be switched off, you can expect to see the number of protein groups corresponding to 'genuine (exclusive)' + 'genuine + transferred'.
In general, if the MBR gain is low and the MBR scores are bad (see the two MBR-related metrics),
MBR should be switched off for the Raw files which are affected (could be a few or all).

Heatmap score [EVD: Prot Count (>%1.0f)]: Linear scoring from zero. Reaching or exceeding the target threshold gives a score of 100%%.
",
    workerFcn = function(.self, df_evd, df_evd_tf, thresh_protCount)
    {
      ## completeness check

      req_cols = c("fc.raw.file", "protein.group.ids", "is.transferred")

      if (!checkInput(req_cols, df_evd)) return()

      .self$helpText = sprintf(.self$helpTextTemplate, thresh_protCount)

      protC = getProteinCounts(rbind(df_evd[,req_cols], df_evd_tf[, req_cols]))

      protC$block = factor(assignBlocks(protC$fc.raw.file, 30))

      max_prot = max(unlist(plyr::dlply(protC, "fc.raw.file", function(x) sum(x$counts))))
      ## average gain in percent
      reportMTD = nrow(df_evd_tf) > 0
      gain_text = ifelse(reportMTD, sprintf("MBR gain: +%.0f%%", mean(protC$MBRgain, na.rm = TRUE)), "")

      lpl = plyr::dlply(protC, "block", .fun = function(x)
      {
        p = plot_CountData(data = x,
                           y_max = max(thresh_protCount, max_prot)*1.1,
                           thresh_line = thresh_protCount,
                           title = c("EVD: ProteinGroups count", gain_text))
        #print(p)
        return (p)
      })

      ## QC measure for protein ID performance
      qc_protc = plyr::ddply(protC, "fc.raw.file", function(x){
        if (nrow(x) == 3 && length(grep("^genuine", x$category))!= 2){
          stop("expected two categories to start with 'genuine...'")
        }
        r = data.frame(genuineAll = sum(x$counts[grep("^genuine", x$category)]))
        return (r)
      })
      cname = sprintf(.self$qcName, thresh_protCount)
      qc_protc[,cname] = qualLinThresh(qc_protc$genuineAll, thresh_protCount)
      qcScore = qc_protc[, c("fc.raw.file", cname)]

      ## add mzQC metric
      mzqc = lapply(1:nrow(qc_protc), function(row){
          out = rmzqc::getQualityMetricTemplate("MS:1002406") # count of identified clusters
          out$value = qc_protc$genuineAll[row];
          #cat(row, " ", qc_protc$genuineAll[row], " ",out$value, "\n");
          return(out) })
      names(mzqc) = qc_protc$fc.raw.file
      ## done
      return(list(plots = lpl, qcScores = qcScore, mzQC = mzqc))
    },
    qcCat = 'general',
    qcName = "EVD:~Protein~Count~(\">%1.0f\")",
    orderNr = 0450
  )
    return(.self)
  })
)


#####################################################################

qcMetric_EVD_PeptideCount =  setRefClass(
  "qcMetric_EVD_PeptideCount",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate =
      "Number of unique (i.e. not counted twice) peptide sequences including modifications (after FDR) per Raw file. A configurable target threshold is indicated as dashed line.

If MBR was enabled, three categories ('genuine (exclusive)', 'genuine + transferred', 'transferred (exclusive)'
are shown, so the user can judge the gain that MBR provides.
Peptides in the 'genuine + transferred' category were identified within the Raw file by MS/MS, but at the same time
also transferred to this Raw file using MBR. This ID transfer can be correct (e.g. in case of different charge states),
or incorrect -- see MBR-related metrics to tell the difference.
Ideally, the 'genuine + transferred' category should be rather small, the other two should be large.

If MBR would be switched off, you can expect to see the number of peptides corresponding to 'genuine (exclusive)' + 'genuine + transferred'.
In general, if the MBR gain is low and the MBR scores are bad (see the two MBR-related metrics),
MBR should be switched off for the Raw files which are affected (could be a few or all).

Heatmap score [EVD: Pep Count (>%1.0f)]: Linear scoring from zero. Reaching or exceeding the target threshold gives a score of 100%%.
",
    workerFcn = function(.self, df_evd, df_evd_tf, thresh_pepCount)
    {
      ## completeness check

      req_cols = c("fc.raw.file", "modified.sequence", "is.transferred")
      if (!checkInput(req_cols, df_evd)) return()
      if (nrow(df_evd_tf)>0 & !checkInput(req_cols, df_evd_tf)) return()

      .self$helpText = sprintf(.self$helpTextTemplate, thresh_pepCount)

      pepC = getPeptideCounts(rbind(df_evd[, req_cols], df_evd_tf[, req_cols]))
      pepC$block = factor(assignBlocks(pepC$fc.raw.file, 30))

      max_pep = max(unlist(plyr::dlply(pepC, "fc.raw.file", function(x) sum(x$counts))))
      ## average gain in percent
      reportMTD = any(df_evd$is.transferred)
      gain_text = ifelse(reportMTD, sprintf("MBR gain: +%.0f%%", mean(pepC$MBRgain, na.rm = TRUE)), "")

      lpl = plyr::dlply(pepC, "block", .fun = function(x)
      {
        p = plot_CountData(data = x,
                           y_max = max(thresh_pepCount, max_pep)*1.1,
                           thresh_line = thresh_pepCount,
                           title = c("EVD: Peptide ID count", gain_text))
        #print(p)
        return (p)
      })

      ## QC measure for peptide ID performance
      qc_pepc = plyr::ddply(pepC, "fc.raw.file", function(x){
        if (nrow(x) == 3 && length(grep("^genuine", x$category))!= 2){
          stop("expected two categories to start with 'genuine...'")
        }
        r = data.frame(genuineAll = sum(x$counts[grep("^genuine", x$category)]))
        return (r)
      })
      cname = sprintf(.self$qcName, thresh_pepCount)
      qc_pepc[,cname] = qualLinThresh(qc_pepc$genuineAll, thresh_pepCount)
      qcScore = qc_pepc[, c("fc.raw.file", cname)]

      return(list(plots = lpl, qcScores = qcScore))
    },
    qcCat = 'general',
    qcName = "EVD:~Peptide~Count~(\">%1.0f\")",
    orderNr = 0400
  )
    return(.self)
  })
)


#####################################################################

qcMetric_EVD_RTPeakWidth =  setRefClass(
  "qcMetric_EVD_RTPeakWidth",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate =
      "One parameter of optimal and reproducible chromatographic separation is the distribution of widths of
peptide elution peaks, derived from the evidence table. Ideally, all Raw files show a similar
distribution, e.g. to allow for equal conditions during dynamic precursor exclusion, RT alignment or
peptide quantification.

Heatmap score [EVD: RT Peak Width]: Scored using BestKS function, i.e. the D statistic of a Kolmogoriv-Smirnoff test.
",
    workerFcn = function(.self, df_evd)
    {
      ## completeness check
      if (!checkInput(c("retention.time", "retention.length", "fc.raw.file"), df_evd)) return()

      ## compute some summary stats before passing data to ggplot (performance issue for large experiments)
      df_evd.m.d = plyr::ddply(df_evd[,c("retention.time", "retention.length", "fc.raw.file")], "fc.raw.file", .fun = peakWidthOverTime)
      head(df_evd.m.d)
      ## median peak width
      df_evd.m.d_avg = plyr::ddply(df_evd[,c("retention.length","fc.raw.file")], "fc.raw.file", .fun = function(x) {
        #fcr = as.character(x$fc.raw.file[1])
        #cat(fcr)
        m = median(x$retention.length, na.rm = TRUE);
        return(data.frame(median = m))
      })
      df_evd.m.d_avg$fc.raw.file_aug = paste0(df_evd.m.d_avg$fc.raw.file, " (~", round(df_evd.m.d_avg$median, 1)," min)")
      .self$outData[["avg_peak_width"]] = df_evd.m.d_avg

      ## augment Raw filename with avg. RT peak width
      df_evd.m.d$fc.raw.file = plyr::mapvalues(df_evd.m.d$fc.raw.file, df_evd.m.d_avg$fc.raw.file, df_evd.m.d_avg$fc.raw.file_aug)
      df_evd.m.d$block = factor(assignBlocks(df_evd.m.d$fc.raw.file, 6)) ## color set is 9, so do not increase this (6*150%)
      ## identical limits for all plots
      df_evd.xlim = range(df_evd.m.d$RT, na.rm = TRUE)
      ## ignore top peaks, since they are usually early non-peptide eluents
      df_evd.ylim = c(0, quantile(df_evd.m.d$peakWidth, 0.99, na.rm = TRUE))

      ## plot peak width
      lpl = list()
      for (bl in unique(df_evd.m.d$block))
      { ## needs to be within a function, otherwise rep_data$add and print() somehow have delayed eval's which confused ggplot...
        lpl[[bl]] = plot_RTPeakWidth(data = df_evd.m.d[df_evd.m.d$block==bl,], x_lim = df_evd.xlim, y_lim = df_evd.ylim)
      }

      ## QC measure for reproducibility of peak shape
      ##.. create a list of distributions
      l_dists = plyr::dlply(df_evd[,c("retention.length", "fc.raw.file")], "fc.raw.file", function(x) return(x$retention.length))
      qc_evd_PeakShape = qualBestKS(l_dists)
      colnames(qc_evd_PeakShape) = c("fc.raw.file", .self$qcName)

      return(list(plots = lpl, qcScores = qc_evd_PeakShape))
    },
    qcCat = "LC",
    qcName = "EVD:~RT~Peak~Width",
    orderNr = 0170
  )
    return(.self)
  })
)


#####################################################################

qcMetric_EVD_MBRAlign =  setRefClass(
  "qcMetric_EVD_MBRAlign",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate =
      "MBR Alignment: First of two steps (1=align, 2=transfer) during Match-between-runs.
This plot is based purely on real MS/MS ids. Ideally, RTs of identical peptides should be equal (i.e. very small residual RT delta)
across Raw files after alignment.

MaxQuants RT correction is shown in blue -- it should be well within the alignment search window (20min by default) set
during MaxQuant configuration.
The resulting residual RT delta after RT alignment (compared to a reference Raw file), is shown as green/red dots. One dot represents
one peptide (incl. charge). Every dot (peptide) outside an allowed residual delta RT (1min by default) is colored red.
All others are green.
The ratio of 'green' vs. 'green+red' peptides is annotated using 'sc: ' (for 'score') in the plot subtitles. High values are better (green peptides dominate).

If moving 'red' dots to the horizontal zero-line (to make them green) requires large RT shifts, then increasing the alignment
search window might help MaxQuant to find a better alignment.


Heatmap score [EVD: MBR Align]: ratio of 'green' vs. 'green+red' peptides
",
    workerFcn = function(.self, df_evd, tolerance_matching, raw_file_mapping)
    {
      ## completeness check
      if (!checkInput(c("type", "calibrated.retention.time", "retention.time.calibration", "id", "raw.file", "modified.sequence", "charge"), df_evd)) return()

      ## find reference
      if (('fraction' %in% colnames(df_evd)) && (length(unique(df_evd$fraction)) > 1)) {
        ## fractions: there must be more than one, otherwise MQ will treat the samples as unfractionated
        refRaw = NA
        col_fraction = "fraction"
        txt_subtitle = "fraction: neighbour comparison"
        evd_has_fractions = TRUE
        df_evd$fraction[is.na(df_evd$fraction)] = 32000
      } else {
        refRaw = findAlignReference(df_evd)
        col_fraction = c()
        txt_subtitle = paste("alignment reference:", gsub("\\", "/", refRaw, fixed = TRUE)) ## subtitles in ggplot must not contain '\'
        evd_has_fractions = FALSE
      }

      lpl = list()
      qcScore = .self$qcScores

      if (!evd_has_fractions && length(refRaw) == 0) {
        lpl[[1]] = ggText("EVD: Alignment check", paste0("Cannot find a reference Raw file!\nPlease report this as a 'bug'!"))
      } else {
        if (!evd_has_fractions & (length(refRaw) != 1))
        {
          refRaw = refRaw[1] ## take the first
          warning(paste0("Cannot find a unique reference Raw file (files: ", paste(refRaw, collapse=", "), "). Picking the first."), immediate. = TRUE)
        }
        ## find RT curve based on genuine 3D peaks (should be flat)
        d_alignQ = alignmentCheck(df_evd[(df_evd$type %in% c("MULTI-MSMS")),
                                         c("calibrated.retention.time",
                                           "id", "raw.file", col_fraction, "modified.sequence", "charge")],
                                  referenceFile = refRaw)
        ## augment more columns
        d_alignQ$retention.time.calibration = df_evd$retention.time.calibration[match(d_alignQ$id, df_evd$id)]

        if (diff(range(na.omit(d_alignQ$retention.time.calibration))) < 1e-5)
        {
          txt_subtitle = paste0(txt_subtitle, " || WARNING: MaxQuant did not correct RTs in any way!");
          warning("EVD MBRAlign: MaxQuant did not correct RTs in any way, despite MBR=on")
        }

        if (nrow(d_alignQ)==0)
        { ## very unusual case: reference contains no evidence -- e.g. pull-down experiment
          lpl[[1]] = ggText("EVD: RT Distance of peptides from reference after alignment", "Alignment cannot be verfied -- no data.")
        } else {
          ## filter data (reduce PDF file size)
          evd_RT_t = thinOutBatch(d_alignQ,
                                  "calibrated.retention.time",
                                  "raw.file")

          evd_RT_t$fc.raw.file = renameFile(evd_RT_t$raw.file, raw_file_mapping)

          ## QC measure for alignment quality
          ## compute % of matches within matching boundary (1 min by default)
          qcAlign = ScoreInAlignWindow(d_alignQ, tolerance_matching)
          if (!is.na(refRaw)) { ## rescue reference file (it will not show up in fraction-less data, and would otherwise be scored 'red')
            qcAlign = rbind(qcAlign, data.frame(raw.file=refRaw, withinRT=1))
          }
          qcAlign[, .self$qcName] = qcAlign$withinRT
          qcScore = qcAlign[, c("raw.file", .self$qcName)]

          qcAlign$fc.raw.file = renameFile(qcAlign$raw.file, raw_file_mapping)
          qcAlign$newlabel = qcAlign$fc.raw.file
          if (evd_has_fractions)
          { ## amend fc.raw.file with fraction number
            qcAlign$fraction = df_evd$fraction[match(qcAlign$fc.raw.file, df_evd$fc.raw.file)]
            qcAlign$newlabel = paste0(qcAlign$fc.raw.file, " - frc", qcAlign$fraction)
          }
          ## amend fc.raw.file with % good ID pairs
          qcAlign$newlabel = paste0(qcAlign$newlabel, " (sc: ", round(qcAlign$withinRT*100), "%)")
          evd_RT_t$fc.raw.file_ext = plyr::mapvalues(evd_RT_t$fc.raw.file, qcAlign$fc.raw.file, qcAlign$newlabel)

          evd_RT_t$RTdiff_in = c("green", "red")[(abs(evd_RT_t$rtdiff) > tolerance_matching)+1]

          ## plot alignment result
          y_lim = quantile(c(evd_RT_t$rtdiff, evd_RT_t$retention.time.calibration), probs = c(0.01, 0.99), na.rm = TRUE) * 1.1
          lpl =
            byX(evd_RT_t, evd_RT_t$fc.raw.file, 3*3, plot_MBRAlign, sort_indices = FALSE,
                y_lim = y_lim, title_sub = txt_subtitle, match_tol = tolerance_matching)

        } ## no data
      } ## ambigous reference file


      return(list(plots = lpl, qcScores = qcScore))
    },
    qcCat = "LC",
    qcName = "EVD:~MBR~Align",
    orderNr = 0210
  )
    return(.self)
  })
)


#####################################################################

qcMetric_EVD_MBRIdTransfer =  setRefClass(
  "qcMetric_EVD_MBRIdTransfer",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate =
      "MBR Transfer: Last of two steps (1=align, 2=transfer) during Match-between-runs.
If MaxQuant only transfers peptide ID's which are not present in the target file, then
each Raw file should not have any duplicates of identical peptides (incl. charge).
Sometimes, a single or split 3D-peak gets annotated multiple times, that's ok. However, the same peptide should not
be annotated twice (or more) at vastly different points in RT.

This plot shows three columns:
 - left: the 'genuine' situation (pretending that no MBR was computed)
 - middle: looking only at transferred IDs
 - right: combined picture (a mixture of left+middle, usually)

Each peptide falls into three categories (the colors):
 - single (good, because it has either one genuine OR a transferred ID).
 - in-group (also good, because all ID's are very close in RT)
 - out-group (bad, spread across the RT gradient -- should not be possible; a false ID)

Heatmap score [EVD: MBR ID-Transfer]: The fraction of non-out-group peptides (i.e. good peptides) in the middle column.
This score is 'pessimistic' because if few ID's were transferred, but all of them are bad, the score is bad, even though
the majority of peptides is still ok (because they are genuine). However, in this case MBR
provides few (and wrong) additional information, and should be disabled.
",
    workerFcn = function(.self, df_evd, df_evd_tf, avg_peak_width)
    {
      ## completeness check
      #stopifnot(c("...") %in% colnames(df_evd))
      if (!checkInput(c("modified.sequence"), df_evd)) return()

      df_evd_all = merge(df_evd, df_evd_tf, all = TRUE)

      ## increase of segmentation by MBR:
      ## three values returned: single peaks(%) in genuine, transferred and all(combined)
      qMBR = peakSegmentation(df_evd_all)
      head(qMBR)
      ## for groups: get their RT-spans
      ## ... genuine ID's only (as 'rtdiff_genuine')
      ##  or genuine+transferred (as 'rtdiff_mixed'))
      ## Could be empty (i.e. no groups, just singlets) if data is really sparse ..
      qMBRSeg_Dist = idTransferCheck(df_evd_all)
      #head(qMBRSeg_Dist)
      #head(qMBRSeg_Dist[qMBRSeg_Dist$fc.raw.file=="file 13",])


      ## Check which fraction of ID-pairs belong to the 'in-width' group.
      ## The allowed RT delta is given in 'avg_peak_width' (estimated from global peak width for each file)
      qMBRSeg_Dist_inGroup = inMatchWindow(qMBRSeg_Dist, df.allowed.deltaRT = avg_peak_width)
      ## puzzle together final picture
      scoreMBRMatch = computeMatchRTFractions(qMBR, qMBRSeg_Dist_inGroup)
      #head(scoreMBRMatch)
      #scoreMBRMatch[scoreMBRMatch$fc.raw.file=="file 3",]

      ## plot ID-transfer
      lpl =
        byX(scoreMBRMatch, scoreMBRMatch$fc.raw.file, 12, plot_MBRIDtransfer, sort_indices = FALSE)

      ##
      ##  Quality
      ##
      qualMBR.m = merge(scoreMBRMatch[scoreMBRMatch$sample=="genuine",],
                        scoreMBRMatch[scoreMBRMatch$sample=="transferred",], by="fc.raw.file")
      qualMBR.m = merge(qualMBR.m, scoreMBRMatch[scoreMBRMatch$sample=="all",], by="fc.raw.file")
      cname = .self$qcName
      qualMBR.m[, cname] = 1 - qualMBR.m$multi.outRT.y # could be NaN if: no-transfer at all, or: no groups but only singlets transferred
      qualMBR.m[is.na(qualMBR.m$multi.outRT.y) & !is.na(qualMBR.m$single.y), cname] = 1 ## only singlets transferred, wow...
      qualMBR.m[is.na(qualMBR.m[, cname]), cname] = HEATMAP_NA_VALUE
      qcScore = qualMBR.m[, c("fc.raw.file", cname)]

      return(list(plots = lpl, qcScores = qcScore))
    },
    qcCat = "LC",
    qcName = "EVD:~MBR~ID-Transfer",
    orderNr = 0220
  )
    return(.self)
  })
)


#####################################################################

qcMetric_EVD_MBRaux =  setRefClass(
  "qcMetric_EVD_MBRaux",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate =
      "Auxililiary plots -- experimental -- without scores.

Return a tree plot with a possible alignment tree.
This allows the user to judge which Raw files have similar corrected RT's (i.e. where aligned successfully).
If there are clear sub-clusters, it might be worth introducing artifical fractions into MaxQuant,
to avoid ID-transfer between these clusters (use the MBR-Align and MBR-ID-Transfer metrics to support the decision).

If the input contains fractions, leaf nodes will be colored accordingly.
Distinct sub-clusters should have their own color.
If not, MaxQuant's fraction settings should be optimized.
Note that introducing fractions in MaxQuant will naturally lead to a clustering here (it's somewhat circular).

Heatmap score: none.
",
    workerFcn = function(.self, df_evd)
    {
      ## completeness check

      if (!checkInput(c("type", "is.transferred", "calibrated.retention.time", "fc.raw.file", "modified.sequence", "charge"), df_evd)) return()

      if (('fraction' %in% colnames(df_evd)) && (length(unique(df_evd$fraction)) > 1)) {
        ## fractions: there must be more than one, otherwise MQ will treat the samples as unfractionated
        col_fraction = "fraction"
      } else {
        col_fraction = c()
      }

      lpl = list()
      lpl[["tree"]] =
        RTalignmentTree(df_evd[(df_evd$type %in% c("MULTI-MSMS")),
                               c("calibrated.retention.time", "fc.raw.file", col_fraction, "modified.sequence", "charge")],
                        col_fraction = col_fraction)

      ## MBR: additional evidence by matching MS1 by AMT across files
      if (any(df_evd$is.transferred)) {
        ## gain for each raw file: absolute gain, and percent gain
        mtr.df = plyr::ddply(df_evd, "fc.raw.file", function(x) {
          match_count_abs = sum(x$is.transferred)
          ## if only matched IDs are present, this would be 'Inf' -- we limit that to 1e4
          match_count_pc  = min(1e4, round(100*match_count_abs/(nrow(x)-match_count_abs))) ## newIDs / oldIDs
          return (data.frame(abs = match_count_abs, pc = match_count_pc))
        })
        lpl[["gain"]] =
          plot_MBRgain(data = mtr.df, title_sub = "")
      }

      return(list(plots = lpl))
    },
    qcCat = "LC",
    qcName = "EVD:~MBR~auxilliary",
    orderNr = 0221
  )
    return(.self)
  })
)



#####################################################################

qcMetric_EVD_Charge =  setRefClass(
  "qcMetric_EVD_Charge",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate =
      "Charge distribution per Raw file. For typtic digests, peptides of charge 2
(one N-terminal and one at tryptic C-terminal R or K residue) should be dominant.
Ionization issues (voltage?), in-source fragmentation, missed cleavages and buffer irregularities can
cause a shift (see Bittremieux 2017, DOI: 10.1002/mas.21544).
The charge distribution should be similar across Raw files.
Consistent charge distribution is paramount for comparable 3D-peak intensities across samples.

Heatmap score [EVD: Charge]: Deviation of the charge 2 proportion from a representative Raw file ('qualMedianDist' function).
",
    workerFcn = function(.self, df_evd)
    {
      ## completeness check
      if (!checkInput(c("is.transferred", "fc.raw.file", "charge"), df_evd)) return()

      d_charge = mosaicize(df_evd[!df_evd$is.transferred, c("fc.raw.file", "charge")])
      lpl =
        byXflex(d_charge, d_charge$Var1, 30, plot_Charge, sort_indices = TRUE)

      ## QC measure for charge centeredness
      qc_charge = plyr::ddply(df_evd[!df_evd$is.transferred, c("charge",  "fc.raw.file")], "fc.raw.file", function(x) data.frame(c = (sum(x$charge==2)/nrow(x))))
      qc_charge[, .self$qcName] = qualMedianDist(qc_charge$c)

      return(list(plots = lpl, qcScores = qc_charge[, c("fc.raw.file", .self$qcName)]))
    },
    qcCat = "prep",
    qcName = "EVD:~Charge",
    orderNr = 0100
  )
    return(.self)
  })
)


#####################################################################

qcMetric_EVD_IDoverRT =  setRefClass(
  "qcMetric_EVD_IDoverRT",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate =
      "Judge column occupancy over retention time.
Ideally, the LC gradient is chosen such that the number of identifications (here, after FDR filtering) is
uniform over time, to ensure consistent instrument duty cycles. Sharp peaks and uneven distribution of
identifications over time indicate potential for LC gradient optimization.
See Moruz 2014, DOI: 10.1002/pmic.201400036 for details.

Heatmap score [EVD: ID rate over RT]: Scored using 'Uniform' scoring function, i.e. constant receives good score, extreme shapes are bad.
",
    workerFcn = function(.self, df_evd)
    {
      ## completeness check
      if (!checkInput(c("retention.time", "fc.raw.file"), df_evd)) return()

      raws_perPlot = 6

      rt_range = range(df_evd$retention.time, na.rm = TRUE)
      df_idRT = plyr::ddply(df_evd, "fc.raw.file", function(x) {
        h = hist(x$retention.time, breaks=seq(from=rt_range[1]-3, to=rt_range[2]+3, by=3), plot = FALSE)
        return(data.frame(RT = h$mid, counts = h$counts))
      })
      lpl =
        byXflex(df_idRT, df_idRT$fc.raw.file, raws_perPlot, plot_IDsOverRT, sort_indices = TRUE)

      ## QC measure for uniform-ness
      qcScore = plyr::ddply(df_evd[, c("retention.time",  "fc.raw.file")], "fc.raw.file",
                      function(x) data.frame(metric = qualUniform(na.omit(x$retention.time))))
      colnames(qcScore)[colnames(qcScore)=="metric"] = .self$qcName

      return(list(plots = lpl, qcScores = qcScore))
    },
    qcCat = "LC",
    qcName = "EVD:~ID~rate~over~RT",
    orderNr = 0150
  )
    return(.self)
  })
)



#####################################################################

qcMetric_EVD_PreCal =  setRefClass(
  "qcMetric_EVD_PreCal",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate =
      "Mass accurary before calibration. Outliers are marked as such ('out-of-search-tol') using ID rate and standard deviation as additional information (if available).
If any Raw file is flagged 'failed', increasing MaxQuant's first-search tolerance (20ppm by default, here: %1.1f ppm) might help
to enable successful recalibration.
A bug in MaxQuant sometimes leads to excessively high ppm mass errors (>10<sup>4</sup>) reported in the output
data. However, this can sometimes be corrected for by re-computing the delta mass error from other data. If this is
the case, a warning ('bugfix applied') will be shown.

Heatmap score [EVD: MS Cal Pre (%1.1f)]: the centeredness (function CenteredRef) of uncalibrated masses in relation to the search window size.
",
    workerFcn = function(.self, df_evd, df_idrate, tolerance_pc_ppm, tolerance_sd_PCoutOfCal)
    {
      ## completeness check
      #stopifnot(c("...") %in% colnames(df_pg))

      .self$helpText = sprintf(.self$helpTextTemplate, tolerance_pc_ppm, tolerance_pc_ppm)

      if (!checkInput(c("fc.raw.file", "uncalibrated.mass.error..ppm."), df_evd)) return()

      ## for some mzTab (not recalibrated) 'mass.error..ppm.' is not there... but we only need a dummy
      if (!("mass.error..ppm." %in% colnames(df_evd))) df_evd$mass.error..ppm. = 0

      fix_cal = fixCalibration(df_evd, df_idrate, tolerance_sd_PCoutOfCal)
      if (is.null(fix_cal)) {
        warning("Internal error. Data missing. Skipping metric!", immediate. = TRUE)
        return()
      }
      ## some outliers can have ~5000ppm, blowing up the plot margins
      ## --> remove outliers
      ylim_g = range(boxplot.stats(fix_cal$df_evd$uncalibrated.mass.error..ppm.)$stats[c(1, 5)], c(-tolerance_pc_ppm, tolerance_pc_ppm) * 1.05)
      ## PLOT
      lpl =
        byXflex(fix_cal$df_evd, fix_cal$df_evd$fc.raw.file, 20, plot_UncalibratedMSErr, sort_indices = TRUE,
                MQBug_raw_files = fix_cal$affected_raw_files,
                y_lim = ylim_g,
                stats = fix_cal$stats,
                extra_limit = tolerance_pc_ppm,
                title_sub = fix_cal$recal_message)

      ## scores
      qc_MS1deCal = plyr::ddply(fix_cal$df_evd, "fc.raw.file",
                          function(x) {
                            xd = na.omit(x$uncalibrated.mass.error..ppm.)
                            if (length(xd)==0) {
                              r = HEATMAP_NA_VALUE ## if empty, give the Raw file an 'NA' score
                            } else if (fix_cal$stats$outOfCal[fix_cal$stats$fc.raw.file == x$fc.raw.file[1]]) {
                              r = 0 ## if we suspect out-of-calibration, give lowest score
                            } else {
                              r = qualCenteredRef(xd, tolerance_pc_ppm)
                            }
                            return (data.frame(med_rat = r))
                          })

      cname = sprintf(.self$qcName, tolerance_pc_ppm)
      colnames(qc_MS1deCal) = c("fc.raw.file", cname)


      return(list(plots = lpl, qcScores = qc_MS1deCal))
    },
    qcCat = "MS",
    qcName = "EVD:~MS~Cal-Pre~(%1.1f)",
    orderNr = 0260
  )
    return(.self)
  })
)


#####################################################################

qcMetric_EVD_PostCal =  setRefClass(
  "qcMetric_EVD_PostCal",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate =
      "Precursor mass accuracy after calibration. Failed samples from precalibration data are still marked here.
Ppm errors should be centered on zero and their spread is expected to be significantly smaller than before calibration.

Heatmap score [EVD: MS Cal-Post]: The variance and centeredness around zero of the calibrated distribution (function GaussDev).
",
    workerFcn = function(.self, df_evd, df_idrate, tolerance_pc_ppm, tolerance_sd_PCoutOfCal, tol_ppm_mainSearch)
    {
      ## completeness check
      #stopifnot(c("...") %in% colnames(df_pg))
      if (!checkInput(c("uncalibrated.mass.error..ppm.", "mass", "mass.error..ppm."), df_evd)) return()

      fix_cal = fixCalibration(df_evd, df_idrate, tolerance_sd_PCoutOfCal)

      ylim_g = range(na.rm = TRUE, boxplot.stats(fix_cal$df_evd$mass.error..ppm.)$stats[c(1, 5)], c(-tol_ppm_mainSearch, tol_ppm_mainSearch) * 1.05)
      ## PLOT
      lpl =
        byXflex(fix_cal$df_evd, fix_cal$df_evd$fc.raw.file, 20, plot_CalibratedMSErr, sort_indices = TRUE,
                MQBug_raw_files = fix_cal$affected_raw_files,
                y_lim = ylim_g,
                stats = fix_cal$stats,
                extra_limit = tol_ppm_mainSearch,
                title_sub = fix_cal$recal_message_post)

      ## QC measure for post-calibration ppm error
      ## .. assume 0 centered and StdDev of observed data
      obs_par = plyr::ddply(fix_cal$df_evd[, c("mass.error..ppm.", "fc.raw.file")], "fc.raw.file",
                      function(x) data.frame(mu = mean(x$mass.error..ppm., na.rm = TRUE),
                                             sd = sd(x$mass.error..ppm., na.rm = TRUE)))
      qc_MS1Cal = data.frame(fc.raw.file = obs_par$fc.raw.file,
                             val = sapply(1:nrow(obs_par), function(x) qualGaussDev(obs_par$mu[x], obs_par$sd[x])))
      ## if we suspect out-of-calibration, give lowest score
      qc_MS1Cal$val[qc_MS1Cal$fc.raw.file %in% fix_cal$stats$fc.raw.file[ fix_cal$stats$outOfCal ]] = 0
      ## MQ mass bugfix will not work for postCalibration, since values are always too low
      qc_MS1Cal$val[qc_MS1Cal$fc.raw.file %in% fix_cal$stats$fc.raw.file[ fix_cal$stats$hasMassErrorBug ]] = HEATMAP_NA_VALUE
      colnames(qc_MS1Cal)[colnames(qc_MS1Cal) == "val"] = .self$qcName

      return(list(plots = lpl, qcScores = qc_MS1Cal))
    },
    qcCat = "MS",
    qcName = "EVD:~MS~Cal-Post",
    orderNr = 0270
  )
    return(.self)
  })
)

#####################################################################

qcMetric_EVD_Top5Cont =  setRefClass(
  "qcMetric_EVD_Top5Cont",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate =
      "PTXQC will explicitly show the five most abundant external protein contaminants
(as detected via MaxQuant's contaminants FASTA file) by Raw file, and summarize the
remaining contaminants as 'other'. This allows to track down which proteins exactly contaminate your sample.
Low contamination is obviously better.
The 'Abundance class' models the average peptide intensity in each Raw file and is visualized using varying degrees of
transparency. It is not unusual to see samples with low sample content to have higher contamination.
If you see only one abundance class ('mid'), this means all your Raw files have roughly
the same peptide intensity distribution.

If you see less than 5 contaminants, it either means there are actually less, or that one (or more) of the shortened contaminant names
subsume multiple of the top5 contaminants (since they start with the same prefix).

Heatmap score [EVD: Contaminants]: as fraction of summed intensity with 0 = sample full of contaminants; 1 = no contaminants

",
    workerFcn = function(.self, df_evd)
    {
      ## completeness check
      if (!checkInput(c("intensity", "contaminant", "fc.raw.file", "proteins"), df_evd)) return()

      ##
      ## elaborate contaminant fraction per Raw.file (this is not possible from PG, since raw files could be merged)
      ## find top 5 contaminants (globally)
      ##
      ## if possible, work on protein names (since MQ1.4), else use proteinIDs
      if ("protein.names" %in% colnames(df_evd))
      {
        evd_pname = "protein.names"
      } else if ("proteins" %in% colnames(df_evd)) {
        evd_pname = "proteins"
      } else {
        stop("Top5-Contaminants: Neither 'protein.names' nor 'proteins' column was found in data but is required.")
      }

      ## protein.names are sometimes not unique, e.g. if a contaminant is involved:
      ## "P02768;CON__P02768-1" and "P02768" will both give the same name (since contaminant name is empty)
      ## Thus, the distribution of bars will look slightly different (but summed percentages are identical)

      ## some protein.names are empty (usually the CON__ ones) ... so we substitute with ID
      df_evd$pname = df_evd[, evd_pname];
      df_evd$pname[df_evd$pname==""] = df_evd$proteins[df_evd$pname==""] ## a NOP if it already is 'proteins', but ok

      df_evd.totalInt = sum(as.numeric(df_evd$intensity), na.rm = TRUE)
      df_evd.cont.only = df_evd[df_evd$contaminant > 0,]

      cont.top = by(df_evd.cont.only, df_evd.cont.only$pname, function(x) sum(as.numeric(x$intensity), na.rm = TRUE) / df_evd.totalInt*100)
      cont.top.sort = sort(cont.top, decreasing = TRUE)
      #head(cont.top.sort)
      cont.top5.names = names(cont.top.sort)[1:5]

      lpl = list()
      if (is.null(cont.top5.names))
      {
        lpl[["noCont"]] = ggText("EVD: Top5 Contaminant per Raw file",
                                 "No contaminants found in any sample.\n\nIncorporating contaminants during search is highly recommended!",
                                 "red")
      } else {
        lpl =
          byXflex(df_evd[, c("intensity", "pname", "fc.raw.file", "contaminant")], df_evd$fc.raw.file, 40, sort_indices = TRUE,
                  plot_ContEVD, top5=cont.top5.names)
      }

      ## QC measure for contamination
      qc_cont = plyr::ddply(df_evd[, c("intensity", "contaminant", "fc.raw.file")], "fc.raw.file",
                      function(x) {
                        val = ifelse(is.null(cont.top5.names),
                                     HEATMAP_NA_VALUE, ## use NA in heatmap if there are no contaminants
                                     1-qualLinThresh(sum(as.numeric(x$intensity[x$contaminant]), na.rm = TRUE)/
                                                       sum(as.numeric(x$intensity), na.rm = TRUE)))
                        return(data.frame(val = val, check.names = FALSE))
                      }
      )
      colnames(qc_cont)[colnames(qc_cont) == "val"] = .self$qcName

      return(list(plots = lpl, qcScores = qc_cont))
    },
    qcCat = "Prep",
    qcName = "EVD:~Contaminants",
    orderNr = 0010
  )
    return(.self)
  })
)

#####################################################################

qcMetric_EVD_MS2OverSampling =  setRefClass(
  "qcMetric_EVD_MS2OverSampling",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate =
      "An oversampled 3D-peak is defined as a peak whose peptide ion (same sequence and same charge
state) was identified by at least two distinct MS<sup>2</sup> spectra in the same Raw file.
    For high complexity samples, oversampling of individual 3D-peaks automatically leads to undersampling
    or even omission of other 3D-peaks, reducing the number of identified peptides. Oversampling occurs in
    low-complexity samples or long LC gradients, as well as undersized dynamic exclusion windows for data
    independent acquisitions.

    If DIA-Data: this metric is skipped

Heatmap score [EVD: MS<sup>2</sup> Oversampling]: The percentage of non-oversampled 3D-peaks.
",
    workerFcn = function(.self, df_evd)
    {
      ## completeness check
      if (!checkInput(c("fc.raw.file", "ms.ms.count"), df_evd)) return()

      ## this is DIA data -- omit metric
      if (all(range(df_evd$ms.ms.count, na.rm = TRUE) == c(0,0))) return()

      d_dups = plyr::ddply(df_evd, "fc.raw.file", function(x) {
        tt = as.data.frame(table(x$ms.ms.count), stringsAsFactors = FALSE)
        tt$Count = as.numeric(tt$Var1)
        ## remove "0", since this would be MBR-features
        tt = tt[tt$Count!=0,]
        ## summarize everything above 3 counts
        if (any(tt$Count >= 3)) {
          tt$Count[tt$Count >= 3] = "3+"
          tt = plyr::ddply(tt, "Count", function(x) data.frame(Freq=sum(x$Freq)))
        }
        ## make counts relative
        fraction = tt$Freq / sum(tt$Freq) * 100
        return (data.frame(n=as.character(tt$Count), fraction = fraction))
      })

      lpl =
        byXflex(d_dups, d_dups$fc.raw.file, 30, plot_MS2Oversampling, sort_indices = TRUE)

      ## QC measure for how many peaks were fragmented only once
      qc_evd_twin = d_dups[d_dups$n==1,]
      cname = .self$qcName
      qc_evd_twin[, cname] = qualLinThresh(qc_evd_twin$fraction/100)

      return(list(plots = lpl, qcScores = qc_evd_twin[, c("fc.raw.file", cname)]))
    },
    qcCat = "MS",
    qcName = "EVD:~MS^2~Oversampling",
    orderNr = 0250
  )
    return(.self)
  })
)

#####################################################################

qcMetric_EVD_CarryOver =  setRefClass(
  "qcMetric_EVD_CarryOver",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate =
      "Sample carryover ... not ready yet...

Heatmap score [EVD: Carryover]: The percentage of peptide identifications whose sequence gives rise to a large 'retention span'.
",
    workerFcn = function(.self, df_evd)
    {
      ## completeness check
      if (!checkInput(c("fc.raw.file", "ms.ms.count", "retention.length"), df_evd)) return()

      raws_perPlot = 6

      aboveThresh_fn = function(data) {
        #t = quantile(data, probs=0.75, na.rm=TRUE) + 3*IQR(data, na.rm = TRUE)
        t = median(data, na.rm=TRUE) * 10
        return(t)
      }

      rt_range = range(df_evd$retention.time, na.rm = TRUE)
      df_carry = plyr::ddply(df_evd, "fc.raw.file", function(x) {
        thresh = round(aboveThresh_fn(x$retention.length), 1)
        weirds = x[x$retention.length > thresh,]
        ## remove "0", since this would be MBR-features
        weirds = weirds[weirds$ms.ms.count!=0,]
        if (nrow(weirds) == 0) return(data.frame())
        weirds$fc.raw.file = paste0(weirds$fc.raw.file, " (>", thresh, " min)")
        h = hist(weirds$retention.time, breaks=seq(from=rt_range[1]-3, to=rt_range[2]+3, by=3), plot = FALSE)
        result = data.frame(RT = h$mid, counts = h$counts, fn = weirds$fc.raw.file[1])
        return(result)
      })
      #df_carry = result
      df_carry$fc.raw.file = df_carry$fn
      lpl =
        byXflex(df_carry, df_carry$fc.raw.file, raws_perPlot, plot_DataOverRT, sort_indices = TRUE,
                title = "EVD: Peptides with wide RT span", y_lab = "# of Peptide Sequences")
      lpl


      qc_evd_carry = plyr::ddply(df_evd, "fc.raw.file", function(x) {
        thresh = aboveThresh_fn(x$retention.length);
        pc = sum(x$retention.length > thresh, na.rm = TRUE) / nrow(x)
        return (data.frame(larger_pc=pc))
      })

      ## QC measure for how many IDs are part of a large span
      cname = .self$qcName
      qc_evd_carry[, cname] = qualLinThresh(1 - qc_evd_carry$larger_pc)

      return(list(plots = lpl, qcScores = qc_evd_carry[, c("fc.raw.file", cname)]))
    },
    qcCat = "MS",
    qcName = "EVD:~CarryOver",
    orderNr = 0250
  )
    return(.self)
  })
)

#####################################################################

qcMetric_EVD_MissingValues =  setRefClass(
  "qcMetric_EVD_MissingValues",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate =
      "Missing peptide intensities per Raw file from evidence.txt.
This metric shows the fraction of missing peptides compared to all peptides seen in the whole experiment.
The more Raw files you have, the higher this fraction is going to be (because there is always going
to be some exotic [low intensity?] peptide which gets [falsely] identified in only a single Raw file).
A second plot shows how many peptides (Y-axis) are covered by at least X Raw files.
A third plot shows the density of the observed (line) and the missing (filled area) data.
To reconstruct the distribution of missing values, an imputation strategy is required, so the argument is somewhat
circular here. If all Raw files are (technical) replicates, i.e. we can expect that missing peptides are indeed
present and have an intensity similar to the peptides we do see, then the median is a good estimator.
This method performs a global normalization across Raw files (so their observed intensitiy distributions have the same mean),
before computing the imputed values. Afterwards, the distributions are de-normalized again (shifting them back to their)
original locations -- but this time with imputed peptides.

Peptides obtained via Match-between-run (MBR) are accounted for (i.e. are considered as present = non-missing).
Thus, make sure that MBR is working as intended (see MBR metrics).

<b>Warning:</b> this metric is meaningless for fractionated data!
<b>TODO:</b> compensate for lower scores in large studies (with many Raw files), since peptide FDR is accumulating!?

Heatmap score [EVD: Pep Missing]: Linear scale of the fraction of missing peptides.
",
    workerFcn = function(.self, df_evd)
    {
      ## completeness check
      if (!checkInput(c("fc.raw.file", "modified.sequence", "intensity"), df_evd)) return()

      if (('fraction' %in% colnames(df_evd)) && (length(unique(df_evd$fraction)) > 1)) {
        lpl = list(ggText("Missing Values Skipped", "Missing values calculation skipped. Fractionated data detected!"))
        return(list(plots = lpl))
      }

      if (length(unique(df_evd$fc.raw.file)) < 2) {
        lpl = list(ggText("Missing Values Skipped", "Need more than one Raw file!"))
        return(list(plots = lpl))
      }

      ## make peptides unique per Raw file
      df_u = plyr::ddply(df_evd[ , c("fc.raw.file", "modified.sequence")], "fc.raw.file",
                   function(x) {
                     return(x[!duplicated(x$modified.sequence),])
                  })

      global_peps = unique(df_u$modified.sequence)
      global_peps_count = length(global_peps)

      ## percent identified in each Raw file
      pep_set = plyr::ddply(df_u[ , c("fc.raw.file", "modified.sequence")], "fc.raw.file",
                      function(x) {
                        score = 100*length(intersect(global_peps, x$modified.sequence)) / global_peps_count
                        return(data.frame(idFraction = score))
                      })

      lpl = byXflex(pep_set, pep_set$fc.raw.file, subset_size = 50, FUN = function(dx) {
        p = ggplot(dx) +
          geom_bar(aes(x = .data$fc.raw.file, y = .data$idFraction), stat = "identity") +
          ggtitle("[experimental] EVD: Non-Missing Peptides", "compared to all peptides seen in experiment") +
          xlab("") +
          ylab("Fraction of total peptides [%]") +
          ylim(0, 100) +
          scale_x_discrete_reverse(dx$fc.raw.file) +
          coord_flip()
        return(p)
      })

      #for (pl in lpl) print(pl)

      tbl = table(df_u$modified.sequence)
      head(tbl)
      tbl_smry = as.data.frame(table(tbl))
      tbl_smry$FreqRel = tbl_smry$Freq / global_peps_count
      tbl_smry = tbl_smry[nrow(tbl_smry):1,] ## invert
      tbl_smry$FreqCum = cumsum(tbl_smry$FreqRel) * 100
      tbl_smry$x = as.numeric(tbl_smry$tbl)

      p = ggplot(tbl_smry, aes(x = .data$x, y = .data$FreqCum)) +
        geom_line() +
        geom_point() +
        ggtitle("[experimental] EVD: Non-missing by set") +
        xlab("Minimum # Raw files") +
        ylab("Fraction of total peptides [%]") +
        ylim(0, 100)
      lpl[["missingCul"]] = p

      ## intensity distribution of missing values
      df_evd$logInt = log2(df_evd$intensity)

      lpl_dens = byXflex(df_evd[, c("modified.sequence", "fc.raw.file", "logInt")], df_evd$fc.raw.file,
                         subset_size = 5, FUN = function(dx) {
        d_mat = reshape2::dcast(dx, modified.sequence ~ fc.raw.file, fun.aggregate = mean, value.var = "logInt")

        ## ... normalization factors
        d_mat_mult = sapply(2:ncol(d_mat), function(x) {
          mult = mean(d_mat[, x] / d_mat[, 2], na.rm = TRUE)
          return(mult)
        })
        df_mult = data.frame(fc.raw.file = colnames(d_mat)[-1], mult = d_mat_mult)
        ## .. normalize data
        d_mat_n = d_mat
        d_mat_n[, -1] = sweep( d_mat_n[, -1, drop=FALSE], 2, d_mat_mult, '/')
        ##
        head(d_mat_n)
        ## find impute value
        pep_mean = rowMeans(d_mat_n[, -1, drop=FALSE], na.rm = TRUE)
        df_missing = plyr::ddply(df_mult, "fc.raw.file", function(x) {
          ## get set of missing values
          values = pep_mean[is.na(d_mat_n[, as.character(x$fc.raw.file)])]
          ## de-normalize (back to old intensity range)
          values = values * x$mult
          return(data.frame(missingVals = values))
        })
        head(df_missing)

        pl = ggplot(df_missing, aes(x = .data$missingVals, col = .data$fc.raw.file, fill = .data$fc.raw.file)) +
          geom_area(position = position_dodge(width=0), binwidth = 0.5, stat="bin", alpha=0.5) +
          geom_freqpoly(data = dx, aes(x = .data$logInt, col = .data$fc.raw.file), binwidth = 0.5, linewidth = 1.2) +
          xlab("Intensity [log2]") +
          ggtitle(" [experimental] EVD: Imputed Peptide Intensity Distribution of Missing Values") +
          scale_fill_manual(values = rep(RColorBrewer::brewer.pal(6,"Accent"), times=40), guide = guide_legend("")) +
          scale_colour_manual(values = rep(RColorBrewer::brewer.pal(6,"Accent"), times=40), guide = "none")
        return(pl)
      })

      lpl = append(lpl, lpl_dens)


      ## QC measure for fraction of missing values
      cname = .self$qcName
      pep_set[, cname] = qualLinThresh(pep_set$idFraction, 100) ## a no-op, just for clarity
      qcScore = pep_set[, c("fc.raw.file", cname)]

      return(list(plots = lpl, qcScores = qcScore))
    },
    qcCat = "prep",
    qcName = "EVD:~Pep~Missing~Values",
    orderNr = 0390  # just before peptide count
  )
    return(.self)
  })
)

#####################################################################

qcMetric_EVD_UpSet =  setRefClass(
  "qcMetric_EVD_UpSet",
  contains = "qcMetric",
  methods = list(initialize=function() {  callSuper(
    helpTextTemplate =
      'The metric shows an upSet plot based on the number of modified peptide sequences per Raw file, intersected or merged with other Raw files (see below for details).<br>

If the number of Raw files is >=6, only the `distinct` plot is generated (the other two are skipped for performance reasons).
<a href="https://raw.githubusercontent.com/cbielow/PTXQC/master/inst/reportTemplate/modes_UpSet.png" target="_blank" rel="noopener"><span>See here for an example plot showing how the set size is computed</span> </a>.

Definition: An `active set` is the set of black dots in a column of the plot -- as opposed to the grey dots (you will understand when you see it).

<p>
<b>distinct:</b> shows the number of sequences that are present in ALL active sets. For three Raw files and active sets A and B, this would mean all sequences which occur in A and B (intersect), but not in C (setdiff).<br>
<b>intersection:</b> shows the number of sequences that occurs in all active sets (intersection).<br>
<b>union:</b> shows the number of sequences that occurs in total. For two files that are all sequences that occurs either in A or in B (union).<br>
<p>
Heatmap score [EVD: UpSet]: The proportion of sequences that the file has in common with all other files.
',
    workerFcn = function(.self, df_evd)
    {
      if (!checkInput(c("modified.sequence", "fc.raw.file"), df_evd)) return()

      getOutputWithMod = function(dl, mode){
        unlist(sapply(1:length(dl), function(numElem){
          comb = combn(names(dl),numElem)
          sapply(1:ncol(comb), function(x){
              sets = comb[,x]
              exp = as.expression(paste(sets, collapse = "&"))
              value = length(Reduce(mode, dl[sets]))
              names(value) = exp
              return(value)
          })
        }))
      }

      lf = tapply(df_evd$modified.sequence, df_evd$fc.raw.file, function(x){return(list(unique(x)))})
      # get rid of rawfiles without any PepIDs
      lf = Filter(function(l) length(l)>0 && any(!is.na(l)), lf)
      if (length(lf) <= 1)
      {
        lpl = list(ggText("UpSetR", "Only single Raw file detected. Cannot compute unions/intersections."))
        return(list(plots = lpl, title = list("EVD: UpSet")))
      }


      lpl = list(UpSetR::upset(UpSetR::fromList(lf), nsets = min(20, length(lf)), keep.order = TRUE, mainbar.y.label = "distinct size"))
      if (length(lf) < 6)
      { ## performance for enumerating all supersets forbids doing it on larger sets until we make this code smarter...
        lpl[[2]] = UpSetR::upset(UpSetR::fromExpression(getOutputWithMod(lf, intersect)), mainbar.y.label = "intersection size")
        lpl[[3]] = UpSetR::upset(UpSetR::fromExpression(getOutputWithMod(lf, union)), mainbar.y.label = "union size")
      }
      titles = list("EVD: UpSet distinct",
                    "EVD: UpSet intersect",
                    "EVD: UpSet union")[1:length(lpl)]

      score = sapply(1:length(names(lf)), function(x){
        union = unique(unlist(lf[-x]))
        inters = intersect(lf[[x]], union)
        score = length(inters)/length(union)
        return(score)
      })

      qcScore = data.frame(fc.raw.file = names(lf), score = score)
      colnames(qcScore)[2] = .self$qcName

      return(list(plots = lpl, title = titles, qcScores = qcScore))
    },
    qcCat = "LC",
    qcName = "EVD:~UpSet",
    orderNr = 0500  # just before peptide count
  )
    return(.self)
  })
)

#'
#' Class which can compute plots and generate mzQC output (usually for a single metric).
#'
#' Reference class which is instanciated with a metric description and a
#' worker function (at initialization time, i.e. in the package)
#' and can produce plots and mzQC values (at runtime, when data is provided) using setData().
#' 
#' All derived classes need to implement a 'workerFcn()' function, which returns a list with
#' elements: c("plots", "mzQC", "htmlTable", "qcScores", "title"),
#' where 'plots' is required; all others are optional.
#'
#' @field helpText  Description (lengthy) of the metric and plot elements
#' @field workerFcn Function which generates a result (usually plots). Data is provided using setData().
#' @field plots     List of plots (after setData() was called)
#' @field htmlTable A table for display in the HTML report (preferred over a plot in Html mode)
#' @field qcScores  Data.frame of scores from a qcMetric (computed within workerFcn())
#' @field mzQC      An named list of mzQC MzQCqualityMetric's (named by their fc.raw.file for runQuality or concatenated fc.raw.files for setQualities (e.g. "file 1;file4")) (valid after setData() was called)
#' @field qcCat     QC category (LC, MS, or prep)
#' @field qcName    Name of the qcScore in the heatmap
#' @field orderNr   Column index during heatmap generation and for the general order of plots
#'
#' @exportClass qcMetric
#' @export qcMetric
#' 
#' @examples 
#'
#' require(ggplot2)
#' dd = data.frame(x=1:10, y=11:20)
#' a = qcMetric$new(helpText="small help text", 
#'                  ## arbitrary arguments, matched during setData()
#'                  workerFcn=function(.self, data, gtitle)
#'                  {
#'                    ## usually some code here to produce ggplots
#'                    pl = lapply(1:2, function(xx) {
#'                        ggplot(data) +
#'                          geom_point(aes(x=x*xx,y=y)) +
#'                          ggtitle(gtitle)
#'                      })
#'                      ## add mzQC metric for count of identified clusters
#'                      mzqc = lapply(1:3, function(id){
#'                        out = rmzqc::getQualityMetricTemplate("MS:1002406")
#'                        out$value = id;
#'                        return(out) })
#'                      names(mzqc) = paste0("file", 1:3);
#'                    return(list(plots = pl, mzQC = mzqc))
#'                  }, 
#'                  qcCat="LC", 
#'                  qcName="MS/MS Peak shape", 
#'                  orderNr = 30)
#' ## test some output
#' a$setData(dd, "my title")
#' a$plots  ## the raw plots
#' a$getPlots(TRUE) ## same as above
#' a$getPlots(FALSE) ## plots without title
#' a$getTitles()  ## get the titles of the all plots
#' a$helpText
#' a$qcName
#' a$mzQC
#' 
#' 
qcMetric = setRefClass("qcMetric",
                       
   fields = list(helpText = "character",         ## identical to helpTextTemplate by default, but can be modified in workerFcn() if desired
                 helpTextTemplate = "character", ## a template (with placeholders for update at runtime) or a static text
                 workerFcn = "function",  ## returns list(plots =, [qcScores=])
                 plots = "list",
                 htmlTable = "character",
                 title = "list",
                 ## the following members are related to the heatmap only
                 qcScores = "data.frame", ## with columns "raw.file", "score"
                 mzQC = "list", ## of MzQCbaseQuality objects
                 qcCat = "character", ## one of "prep", "LC", "MS" or empty (e.g. for PG)
                 qcName = "character", ## expression e.g. "MS^2~ID~Rate"
                 orderNr = "numeric", ## ordering of plots -- also applies to Heatmap; gaps are ignored
                 outData = "list" ## optional auxiliary output data generated in workerFcn
                 ),
   methods = list(
       initialize = function(helpTextTemplate = NA_character_,
                           workerFcn = function(){},
                           qcCat = NA_character_,
                           qcName = NA_character_,
                           orderNr = NaN) {
           .self$helpTextTemplate = helpTextTemplate;
           .self$helpText = helpTextTemplate;
           .self$workerFcn = workerFcn;
           .self$plots = list();  ## obtained from worker
           .self$htmlTable = NA_character_;  ## obtained from worker
           .self$qcScores = data.frame();  ## obtained from worker
           .self$mzQC = list();  ## obtained from worker
           .self$title = list();
           .self$qcCat = qcCat;
           .self$qcName = qcName;
           .self$orderNr = orderNr;
           .self$outData = list();
           return(.self)
       },
       setData = function(df, ...) {
         "Internally calls the workerFcn(), which computes the actual plots, metric scores and supporting data
          (e.g. mzQC metrics) of the derived class; the resulting data is checked and stored in the members of this class.

          Arguments:
          \\describe{
            \\item{\\code{df}}{The expected data, usually a data frame. If empty, this function will return immediately without failure.}
            \\item{\\code{...}}{Additional arguments passed to the workerFcn().}
          }

          Returns NULL.
         "
         cat("Starting to work on", gsub("~", " ", .self$qcName), "...\n")
         if (.self$orderNr < 0)
         {
           cat("  Metric disabled. Skipping...\n")
           return (NULL)
         }
         
         if (is.null(df))
         {
           cat("  No data available. Skipping...\n")
           return (NULL)
         }
         
         ## GC stats
         mem_before = gc(verbose = FALSE, reset = TRUE) ## resetting Max to see how much this metric uses
         t_before = proc.time()
         
         r = workerFcn(.self, df, ...)
         
         ## clean memory to get a clean picture of each metrics memory footprint
         ## to enable the user to skip expensive metrics
         mem_after = gc(verbose = FALSE)
         cat(paste0("\nMemory [MB] prior|after|max(diff) : ", 
                    sum(mem_before[, 2]), " | ",
                    sum(mem_after[, 2]), " | ",
                    sum(mem_after[, 6]), " (",
                    round(sum(mem_after[, 6])-sum(mem_before[, 2]),0) ,")\n",
                    "\nDuration: ", round((proc.time() - t_before)[3]), " s\n\n"))
         if (is.null(r)) {
           message(c("Worker of '", .self$qcName, "' returned prematurely due to missing data! Skipping metric!"))
           return(NULL);         
         }
         
         valid_elements = c("plots", "mzQC", "htmlTable", "qcScores", "title")
         required_elements = c("plots")
         
         if (!(all(required_elements %in% names(r)))) stop(c("Worker of '", .self$qcName, "' did not return all required fields (", paste(required_elements, collapse = ","), ")!"))
         if (!inherits(r[["plots"]], "list")) stop(c("Worker of '", .self$qcName, "' did not return plots in list format!"))
         
         if (!all(names(r) %in% valid_elements)) stop(c("Worker of '", .self$qcName, "' return invalid fields (", paste(setdiff(names(r), valid_elements), collapse = ","), ")!"))
         

         lpl = flattenList(r[["plots"]])
         #print(lpl) ## debug
         .self$plots = lpl;
         #.self$plots = r[["plots"]]
         
         if ("mzQC" %in% names(r)) .self$mzQC = r[["mzQC"]];
         if ("htmlTable" %in% names(r)) .self$htmlTable = r[["htmlTable"]];
         if ("qcScores" %in% names(r)) .self$qcScores = r[["qcScores"]];
         if ("title" %in% names(r)) .self$title = r[["title"]]
         
         
         cat("...", gsub("~", " ", .self$qcName), " done\n")
         return(NULL)
       },
       
       getPlots = function(withTitle = TRUE) {
         if (!withTitle) { ## no title
           r = lapply(.self$plots, function(p) {
             ## delete title 
             if (inherits(p, "ggplot")) p = p + ggtitle(NULL)
             return (p)
           })
           return(r)
         };
         return(.self$plots)
       },
       
       getTitles = function(stopOnMissing = TRUE, subtitle_sep = " - ") {
         labels = sapply(1:length(.self$plots), 
                         function(idx) {
                           if (length(.self$title) != 0){
                             return(.self$title[[idx]])
                           }
                           else if ("title" %in% names(.self$plots[[idx]]$labels)){
                             titles = .self$plots[[idx]]$labels$title
                             #title = 'atop("PG: PCA of 'reporter intensity'", scriptstyle("(excludes contaminants)"))'
                             regex = 'atop\\("(.*)", scriptstyle\\("(.*)"\\)\\)'
                             m = regexpr(regex, titles, perl = TRUE)
                             if (m == 1) { ## hit!
                               text = substring(titles, attr(m, "capture.start"), attr(m, "capture.start") + attr(m, "capture.length") - 1)
                               return(paste0(text[1], subtitle_sep, text[2]))
                             }
                             return (titles)
                           } else if (stopOnMissing) {
                             stop(c("getTitles() for ", .self$qcName, ": No title found in ggplot object at index ", idx, "!"))
                           } else return("")
                         })
         return(labels)
       }
         
     )
   ) ## refClass

#########################################################################################


#'
#' Flatten lists of lists with irregular depths to just a list of items,
#' i.e. a list of the leaves (if you consider the input as a tree).
#' 
#' @param x List of 'stuff' (could be lists or items or a mix)
#' @return A flat list
#'
#'
#'
flattenList = function(x) {
  repeat {
    idx_list = sapply(x, function(arg) {return(all(class(arg) == "list"))})
    if (!any(idx_list)) return(x)
    r_list = Reduce(append, x[idx_list])
    items = x[!idx_list]
    x = Reduce(append, list(r_list, items))
  }
}

checkInput = function(required_columns, given_df)
{
  if (!all(required_columns %in% colnames(given_df)))
  {
    warning(paste0("Input check failed: columns '", 
                   paste(setdiff(required_columns, colnames(given_df)), collapse="', '", sep=""),
                   "' are not present in input data!"),
            immediate. = TRUE)
    return (FALSE)
  }
  return (TRUE)
}

qcMetric_AverageQualOverall = 
  setRefClass("qcMetric_AverageQualOverall",
              contains = "qcMetric",
              methods = list(
                initialize=function() {
                  callSuper(
                    helpTextTemplate = "Internal metric to compute the average quality across all other metrics",
                    workerFcn = function(.self, df.QCM)
                    {
                      if (plyr::empty(df.QCM)) stop("AverageQual_qc::workerFcn(): input empty!")
                      lpl = list() ## empty...
                      qcScore = plyr::ddply(df.QCM, "fc.raw.file", function(df.row) {
                        df.row.raw = unlist(df.row[,!grepl("fc.raw.file", colnames(df.row))])
                        df.row.raw[is.infinite(df.row.raw)] = NA  ## mask explicitly missing values, since it will bias the mean otherwise
                        return (data.frame(val = mean(df.row.raw, na.rm = TRUE)))
                      })
                      colnames(qcScore)[colnames(qcScore)=="val"] = .self$qcName
                      
                      return(list(plots = lpl, qcScores = qcScore))
                    },
                    qcCat = "General",
                    qcName = "Average~Overall~Quality",
                    orderNr = 9999
                  )
                  return(.self)
                })
              )

#qcA = qcMetric_AverageQualOverall_$new()
#qcA$setData(data.frame(fc.raw.file = letters[1:5], qual1 = 1:5, qual2 = 5:9))
#qcA$qcScores
#qcA$helpText

library(PTXQC)
#library(testthat)
library(ggplot2)
context("qcMetric.R")

test_that("qcMetric", {
  
  dd = data.frame(x=1:10, y=11:20)
  
  a = qcMetric$new(helpText="small help text", 
                   workerFcn=function(.self, data, gtit)
                   {
                     # usually, plots are produced, but they are hard to check. So we generate simple values
                     pl = lapply(1:2, function(xx) ggplot(data) + geom_point(aes(x=x*xx,y=y)) + ggtitle(paste(gtit, xx)))
                     qcScores = data.frame(r=1, z=9:10)
                     return(list(plots = pl, qcScores = qcScores))
                   }, 
                   qcCat="LC", 
                   qcName="MS/MS Peak shape", 
                   orderNr = 30)
  
  a$setData(dd, "title assigned by worker")
  s = a$qcScores
  expect_equivalent(s, data.frame(r=1, z=9:10))

  p = a$plots
  expect_equal(length(p), 2)
  a$getPlots(TRUE) ## just check if it works
  a$getPlots(FALSE) ## just check if it works
  expect_equal(paste("title assigned by worker", 1:2), a$getTitles())
  
  expect_equal("small help text", a$helpText)
  expect_equal("MS/MS Peak shape", a$qcName)
  
  expect_equal(30, a$orderNr)
})

@echo OFF
ECHO.
ECHO This batch file is used to invoke the QC script by drag'n'dropping of
ECHO a MaxQuant txt folder (or any file within) onto this .bat file.
ECHO Within the same txt folder, a QC report in Html/PDF format will be created.
ECHO.
ECHO First time installation of this script [for admins only]:
ECHO The script expects a certain structure to be present.
ECHO This .bat file expects a subfolder named '_internal'
ECHO which holds:
ECHO      'R-3.1.0'             Holds a complete R installation 
ECHO                            (including all packages required to run the 
ECHO                             PTXQC package)
ECHO      'compute_QC_report.R' The R script that is invoked by this .bat
ECHO.
REM configure this directory according to above instructions
REM 
REM  --- config done --- 
REM

REM change this, if your R installation has a different version and you do not want to rename the folder
set R_VERSION=R-3.1.0

REM A usage message with multiline (i.e. keep the empty line!)
setlocal EnableDelayedExpansion
set USAGE= ^

Usage: ^

%0 ^<txt-folder^> [config.yaml] ^

Please try again.
REM end of USAGE

REM check number of arguments (must be 1 [txt folder] or 2 [txt folder + yaml])
set argC=0
for %%x in (%*) do Set /A argC+=1
if %argC%==0 (
  ECHO Not enough arguments^^! (No arguments provided^)
  ECHO !USAGE!
  goto end
)
if !argC! gtr 2 (
  ECHO Too many arguments^^! (At most two are allowed^)
  ECHO Args:
  for %%x in (%*) do echo %%x
  ECHO !USAGE!
  goto end
)

REM usually, path without quotes are needed, e.g. for set PATH=!I!;%PATH%  (spaces are no problem)
set I=%~dp0\_internal
REM sometimes quoted strings are required (i.e. when calling rscript)
set Iqt="%~dp0\_internal"

REM currently, we do not allow Drag'n'drop PTXQC to be installed on a path with spaces, since
REM this breaks pandoc.exe (and the temp directories it creates) for some reason
for /f "tokens=2" %%a IN ("%I%") do goto NoSpaces

REM use drive+path of this .bat file 
ECHO.
ECHO Found QC directory at '%I%'
ECHO.

REM Check upon first argument (the txt folder - or any file in it)
REM use bare %1 not "%1", it works fine, even with spaces
if not exist %1 (
  goto not_found
) 
set is_dir=0
REM PushD+PopD is the only reliable way I found to check if %1 is a directory or a file
REM Everything else proposed (e.g., %1\. or %1\nul does not work on Win7)
PUSHD %1 && POPD || goto is_file
set is_dir=1
:is_file
if %is_dir%==1 ( 
  echo -- '%1' is a directory
  REM this will be quoted automatically
  set txt=%1
) else ( 
  echo -- '%1' is a file
  REM ~dp (drive,path); the final '\' is required!; manual quoting required as well
  set txt="%~dp1\"
) 

ECHO Txt folder is at '%txt%'

if %argC%==2 (
  REM We use ~dpnx (drive,path,name,extension) instead of %2 to get rid of potential extra surrounding quotes
  set yaml_file=%~dpnx2
  ECHO.
  ECHO Yaml file is: !yaml_file!
  ECHO.
  REM check file extension. We could also use %~x2 to get the file extension
  if !yaml_file:~-5! NEQ .yaml (
    ECHO Error: Second argument '!yaml_file!' does not seem to be a '.yaml' file.
    ECHO Got: %~x2
    ECHO Please rename or use proper file.
    goto end
  )
  if exist !yaml_file! (
    REM Same as above, just using quotes
    set yaml_file="%~dpnx2"
    ECHO Using YAML configuration file !yaml_file!
    ECHO.
  ) else (
    ECHO Error: Yaml file !yaml_file! given as second argument does not exist.
    ECHO Every QC script run will create a default .yaml file in the respective txt folder, which you can COPY
    ECHO and use as input to any subsequent run.
    ECHO.
    goto end
  )
)

REM use R_LIBS, not R_LIBS_USER, since the latter will APPEND to the search path, i.e. if the package is installed locally, it will take precedence and mess
REM up versioning
set R_LIBS=!I!\!R_VERSION!\library
REM echo !R_LIBS!

REM check for existence of rscript.exe
set R_SCRIPT_EXE=!R_VERSION!\bin\x64\rscript.exe
dir !Iqt!\!R_SCRIPT_EXE! > NUL 2>&1
if ERRORLEVEL 1 (
  ECHO.
  ECHO Warning: Could not find 64bit R installation at '!Iqt!\!R_SCRIPT_EXE!'.
  ECHO Falling back to 32bit. Be aware that this might lead to 'Out-of-memory' errors for larger projects.
  ECHO.
  set R_SCRIPT_EXE=!R_VERSION!\bin\rscript.exe
  dir !Iqt!\!R_SCRIPT_EXE! > NUL 2>&1
  if ERRORLEVEL 1 (
    ECHO.
    ECHO Error: Could not find a 32bit R installation at '!Iqt!\!R_SCRIPT_EXE!'.
    ECHO.
    ECHO No R found. Please check your PTXQC installation and read the installation instructions! Quitting now.
    ECHO.
    PAUSE
    EXIT 1
  )
)
ECHO.
ECHO Found R at '!Iqt!\!R_SCRIPT_EXE!'.
ECHO.

REM check for PANDOC
set PATH=!I!\!R_VERSION!\bin\pandoc;!PATH!
where pandoc > NUL 2>&1
if ERRORLEVEL 1 (
  ECHO.
  ECHO 'Pandoc.exe not found in '!I!\!R_VERSION!\bin\pandoc\pandoc.exe'.
  ECHO.
  ECHO Please install Pandoc from 'https://github.com/jgm/pandoc/releases'
  ECHO and copy it to the folder mentioned above.
  ECHO.
  goto end
)


REM works WITH and WITHOUT spaces in !txt!
if !yaml_file! NEQ "" (
  ECHO Using YAML !yaml_file! file and calling R now ...
  ECHO.
  ECHO Calling '!Iqt!\!R_SCRIPT_EXE! --vanilla !Iqt!\compute_QC_report.R !txt! !yaml_file!'
  REM using Iqt to guard against path with spaces. Note that using a manually quoted !I!, i.e. "!I!", does not work
  REM However, quoting the arguments is ok.
  !Iqt!\!R_SCRIPT_EXE! --vanilla !Iqt!\compute_QC_report.R !txt! !yaml_file!
) else (
REM 
  ECHO.
  ECHO Calling '!Iqt!\!R_SCRIPT_EXE! --vanilla !Iqt!\compute_QC_report.R !txt!'
  REM using Iqt to guard against path with spaces. Note that using a manually quoted !I!, i.e. "!I!", does not work
  REM However, quoting the arguments is ok.
  !Iqt!\!R_SCRIPT_EXE! --vanilla !Iqt!\compute_QC_report.R !txt!
)


set myscriptEL=%errorlevel%
REM report error, if any
if %myscriptEL% GTR 0 (
  ECHO Failed folder %txt%. Does it exist and is writeable^^!?
  ECHO Failed folder %txt%. Does it exist and is writeable^^!? >> \error.log
  ECHO.
  ECHO The error should be indicated above.
  ECHO Common errors are ^(yours might be different^)
  ECHO   - your R installation runs out of memory; especially if its 32bit
  ECHO   - the file '%I%\compute_QC_report.R' is missing
  ECHO.
  ECHO If the error persists, follow the
  ECHO instructions on https://github.com/cbielow/PTXQC
  ECHO in the section 'Bug reporting'.
  ECHO.
  ECHO Leave this window open until the bug report is submitted!
  ECHO.
  ECHO You can copy the text from this window by 
  ECHO   1^) right-clicking here
  ECHO   2^) choose 'Select all'
  ECHO   3^) press the 'Enter' key
  ECHO   4^) paste it somewhere ^(issue ticket, email^)
  ECHO This ensures that all text is available to debug the problem.
  ECHO Thanks^^!
)

goto end

:noSpaces
ECHO Drag'n'Drop PTXQC is installed in a path containing spaces ("!I!")
ECHO This is not allowed (since it breaks HTML report generation). Please move it to a path without spaces.
ECHO.
goto end

:not_found
echo Could not find the given (network) folder '%1'^^! Please contact your admin/bioinformatician^^!
goto end

:end
pause

@echo OFF
REM Here we hardcode the YAML file which is forwarded to 'createQC_dragNdrop.bat'
REM The '%~dp0' just prefixes the YAML file with the path to this .bat file
REM i.e. the YAML file should be right next to it.
REM You can also put the .yaml file somewhere else and hardcode the path
REM e.g. set yaml_file=c:\temp\my.yaml
REM Just make sure the file ending remains '.yaml'
set yaml_file=%~dp0\config.yaml

ECHO.
ECHO This batch file allows invoking 'createQC_dragNdrop.bat'
ECHO with a YAML config file which resides in this directory.

REM check number of arguments (must be exactly 1 [txt folder])
set argC=0
for %%x in (%*) do Set /A argC+=1
if %argC% NEQ 1 (
  ECHO.
  ECHO Wrong number of arguments^!
  ECHO.
  ECHO Exactly one argument (containing the txt folder or any file within^) is expected^!
  ECHO.
  goto end
)


ECHO.
ECHO Using '%yaml_file%' as YAML config
ECHO.
ECHO.
REM Calling original bat file with two arguments
REM We use %~dp0 to get the path of this bat file, since the working directory depends on the calling environment
"%~dp0\createQC_dragNdrop.bat" %1 "%yaml_file%"


:end
pause;  

---
title: "Input/Output data for PTXQC"
author: "Chris Bielow <chris.bielow@fu-berlin.de>"
date: '`r Sys.Date()`'
output:
  html_document: default
  pdf_document: null
vignette: >
  %\VignetteIndexEntry{Input/Output data for PTXQC}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

# Input/Output data for PTXQC

PTXQC generates quality control reports for [MaxQuant] txt data, and [OpenMS] mzTab data.

## Input

When a full report is generated, PTXQC requires the following **input files**.
  
MaxQuant:  

  - parameters.txt
  - summary.txt
  - proteinGroups.txt
  - evidence.txt
  - msms.txt
  - msmsScans.txt
  
These files can be found in the `<baseDir>/combined/txt` folder after MaxQuant analysis successfully finished.
Here, `<baseDir>` is your main MaxQuant project folder containing your Raw files.
  
OpenMS:

  - an .mzTab file from the QualityControl TOPP tool
  
  Note: PTX-QC will compute all metrics for which input data is present. MzTab files from other software packages (even mzTab from MaxQuant) will be missing some data and thus only produce a small subset of metrics. 


If you are not interested in the full report, PTXQC can be configured to look only at some of these files by editing the YAML file. See the [PTX-CustomizeReport][Ref_VignCust] vignette
```{r, eval=FALSE}
vignette("PTXQC-CustomizeReport", package = "PTXQC")
```
for details. Alternatively (for .txt input), you can rename any .txt file you want PTXQC to ignore.

## Output

By default, PTXQC will generate a report in both **Html** and **PDF** format.

To enable/disable a certain format (Html/PDF), just set it in the YAML config file (see above for the vignette hint on report customization).

Further output files (with the same prefix name as the Html/PDF report), include

  - `report_vXXX_heatmap.txt` (values of the overview heatmap)
  - `report_vXXX_filename_sort.txt` (customize the order and naming of Raw files in the report)
  - `report_v0XXX.yaml` (the configuration file)
  
The YAML configuration file is created the first time you run PTXQC on a txt-folder or mzTab file. Customize it as you see fit.

Read more about `report_vXXX_filename_sort.txt` and `report_vXXX.yaml` in the [PTX-CustomizeReport][Ref_VignCust] vignette.
  

  [MaxQuant]: https://maxquant.org
  [OpenMS]: https://openms.de/
  [Ref_VignCust]: https://github.com/cbielow/PTXQC/blob/master/vignettes/PTXQC-CustomizeReport.Rmd

---
title: "Customizing PTXQC reports"
author: "Chris Bielow <chris.bielow@fu-berlin.de>"
date: '`r Sys.Date()`'
output:
  html_document: default
  pdf_document: null
vignette: >
  %\VignetteIndexEntry{Customizing PTXQC reports}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

# Customizing PTXQC reports

Reports rely on a small set of parameters, which do two things:

  - set optimal values for the PTXQC metrics and scoring functions
    (usually tighly connected to the data and MS instrument)
  - report length: number of metrics shown in the report or partial evaluation of the txt-folder

The first point can partially be addressed via a **mqpar.xml file**.
However, not all thresholds are known or even interesting to MaxQuant, e.g. the number of expected proteins.
For this, PTXQC comes with a **YAML config file**. This file also addresses the report length.

## Automatically importing MaxQuant parameters via mqpar.xml

Some settings in MaxQuant, e.g. the MS/MS search tolerance are important for PTXQC in order to
correctly score the mass recalibration or show expected error margins.

So how does PTXQC know about the settings used during the MaxQuant run? 
Copy the `mqpar.xml` which MaxQuant generated in the analysis directory 
into your txt folder, e.g. `c:\MQ_Diabetes_study\mqpar.xml` -> `c:\MQ_Diabetes_study\combined\txt\mqpar.xml`. 
This will enable PTXQC to extract some parameter settings and results in more accurate reports. 
You will see a warning message during report generation if `mqpar.xml` is missing.
The FAQ vignette has additional information on why this manual copying is senseful.

Alternatively, you can set the MaxQuant parameters manually in the YAML file (see next section) to match your MaxQuant configuration.

## More customization using the YAML config file

A configuration file allows you to enable/disable certain plots, and to modify most target thresholds used for the scoring metrics and within plots, e.g. the number of proteins you expect. This will differ by platform and instrument; so adapt it to your needs. The defaults are meant for a Thermo Orbitrap Velos, a 4h LC gradient for a complex matrix (e.g. human cells).
Disabling certain report sections will decrease report size and hide metrics which you may not need. The time required to create the report will also reduce, but usually this is not a bottleneck.

PTXQC uses YAML format for its parameter configuration. YAML is short for 'YAML Aint Markup Language'; yes, it's an endless recursion.
There is a nice tutorial for YAML at [https://rhnh.net/2011/01/31/yaml-tutorial] -- or just Google to find more.
Also, any PTXQC YAML file will start with a comment section which explains the format.

PTXQC will create a `report_*.yaml` file within your txt folder upon its first invocation.
If you did not run PTXQC before on this folder, you can take any other YAML config file
or run PTXQC with default settings (see above) to create such a file.
Edit the YAML file as you desire using a text editor of your choice (e.g. Notepad++). **Each YAML file has a leading comment section which will describe the basics of how YAML looks like and what you should/not change.**

> Note: PTXQC will only read the YAML file which **matches its own version exactly**, i.e. `report_v0.82.1.yaml` will be
> ignored if PTXQC has version `0.83.1` to avoid compatibility problems. A new YAML file matching the current PTXQC 
> version will automatically be generated. If you want to use your old settings, rename the file manually.

The YAML shown below is shortened and lacks parts of the introductory comment section.
But in general, the YAML file will look like this:

    # This is a configuration file for PTXQC reporting.
    # One such file is generated automatically every time a report PDF is created.
    # You can make a copy of this file, then modify its values and use the copy as an input to another round of report generation,
    # e.g., to exclude/include certain plots or change some global settings.
    #
    # << more comment text in original YAML file ... >>
    #
    PTXQC:
      UseLocalMQPar: yes
      ReportFilename:
        extended: yes
      NameLengthMax_num: 10.0
      AddPageNumbers: 'on'
    order:
      qcMetric_PAR: 1.0
      qcMetric_PG_PCA: 3.0
      qcMetric_EVD_Top5Cont: 10.0
      qcMetric_PG_Ratio: 19.0
      qcMetric_EVD_UserContaminant: 20.0
      << more metrics here... >>
    File:
      Parameters:
        enabled: yes
      Summary:
        enabled: yes
        IDRate:
          Thresh_bad_num: 20.0
          Thresh_great_num: 35.0
      ProteinGroups: ...
      Evidence:
        enabled: yes
        ProteinCountThresh_num: 3500.0
        PeptideCountThresh_num: 15000.0
        MatchBetweenRuns_wA: auto

To disable certain metrics/plots, just change the **order** of the metric in question to a **negative** value,
e.g. `qcMetric_PG_PCA: -3.0`, will disable the PCA plot.
This is helpful to disable uninteresting metrics, but also a temporary solution if your PC cannot cope with RAM demands of PTXQC
or a metric is broken (you should open a bug report in this case).
        

## Rename Raw files in the report

PTXQC will try to shorten Raw file names to make the plot axis as compact as possible, giving more area to show the data rather than overly long Raw file names.
It will use prefix and infix removal of strings which are common to all Raw file names. Name your files in a systematic to make this approach as effective as possible. 
Should this shortening not reach a predefined threshold of 10 characters in length for the longest filename after shortening, 
PTXQC will simply use consecutive numbers, i.e. **file 01**, **file 02** to name your files on the axis of the plots.

If you do not want that, you can either:

  - Increase this length threshold in the YAML config to force usage of the shortened versions. 
    But be advised that using long filenames might result in ugly axis with overlapping text etc.
  - Use a manual mapping table: a file named `report_vXXX_filename_sort.txt` will be created in your txt-folder,
    which contains the file name mapping that PTXQC computed. You can edit this file with any text editor
    and upon running PTXQC again, this new mapping will be applied.

## Reordering Raw files in the report

By default, the order of Raw files in the report is _identical_ to the order specified within MaxQuant.

You already read about the `report_vXXX_filename_sort.txt` file above.
It can not only be used to assign custom names, but also to re-arrange the order in which Raw files are shown in the plots and heatmap.
Open it with a text editor and simply rearrange the rows in the file. Easy!

---
title: "Drag'n'drop Usage Guide for PTXQC"
author: "Chris Bielow <chris.bielow@fu-berlin.de>"
date: '`r Sys.Date()`'
output:
  html_document: default
  pdf_document: null
vignette: >
  %\VignetteIndexEntry{Drag'n'drop Usage Guide for PTXQC}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

# Drag'n'drop Usage Guide for PTXQC

**Requirements:**

  - any Windows System 
  - R and PTXQC should be installed using the installation instructions
    at https://github.com/cbielow/PTXQC/tree/master/inst/dragNdrop.

## Generating a quick QC report without asking questions!

  - open a Windows Explorer and find your MaxQuant txt-output folder
  - open another Windows Explorer window and navigate to the (network) folder where the QC software
    was copied to during installation, e.g. `P:\Chris\Exchange\QC` (where drive P could be some network share).
  - drag'n'drop the txt folder onto `createQC_dragNdrop.bat` (you can also drag any file within the
    txt folder, e.g. `evidence.txt`)

<img src="DragNDrop_Action.png" width="600" alt="Drag'n'drop action" />

 - A console will pop up and show stuff. Wait... if all went well, one of the last lines is `Time elapsed: ... min`
   and a pdf report within your txt folder. It has a version and a folder name attached to it
   (e.g. `report_v0.63_diabetes.pdf`). The version number will increase every time there is an 
   update in PTXQC (big change in version number means big change in functionality). 
   The name is taken from the path where the txt is located in - it helps you to distinguish
   report files when they are open in a PDF viewer or copied somewhere else. 
   The console window will stay up until you press a key.

<img src="DragNDrop_BatWindow.png" width="677" alt="The command window, visible during report generation" />

## Generating a QC report using a configuration file

There are two ways to customize parameters for a report:

  - PTXQC will search for a YAML file within the txt-folder for which it generates the report. 
    If there is one (with the correct name, i.e. `report_vXXX.yaml`, where XXX is the PTXQC version number),
    this report file will be used. Thus, if you already have txt-folder containing a report pdf
    plus the `report_*.yaml` inside, just edit the YAML file and re-run 
    PTXQC.
  - If your txt-folder has not been analysed by PTXQC before *or* you want to try another configuration, 
    you can use the `createQC_dragNdrop_withYAML.bat`. This .bat file expects a YAML configuration file
    right next to itself. This YAML file you need to provide. But no worries. One such YAML file 
    with default parameters is generated each time you invoke a QC report generation. 
    Just take any YAML file from a previous run in any other txt-folder. Copy the YAML file next to the .bat file, and name it `config.yaml`.
    Modify it to your needs using a text editor such as Notepad.
    From now on, you can use the `createQC_dragNdrop_withYAML.bat` for any report which should use this configuration.
    The usage is identical to the one you already know - just target `createQC_dragNdrop_withYAML.bat` instead of `createQC_dragNdrop.bat`
    when dragging a txt-folder.

**Note**: The second option will take precedence over the first.

---
title: "List of Metrics"
author: "Chris Bielow <chris.bielow@fu-berlin.de>"
date: '`r Sys.Date()`'
output:
  html_document:
    mathjax: null
    number_sections: yes
    toc: no
  pdf_document:
    toc: no
vignette: >
  %\VignetteIndexEntry{List of Metrics}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---


<!-- the TOC (table of content) would go here, but we create it manually to use this as a vignette
      - - > use 'toc: no' above to disable the TOC 
 -->

<a id="TOP"/>

<style type="text/css">
  p { 
    padding: 5px 0;
  }

  .helpText {
    clear: left;
  }
  
</style>

```{r setup, include=TRUE, echo=FALSE, results="asis"}
## global options
knitr::opts_chunk$set(echo=FALSE, warning=FALSE, error=FALSE, message=FALSE, fig.width=10)

lst_qcMetrics_ord = PTXQC:::getMetricsObjects()

txt_TOC = "# Table of Contents
  - [Overview](#Overview)
  - [Metrics](#Metrics)
"
txt_BODY = ""
for (qcm in lst_qcMetrics_ord)
{
  newname = gsub("(.*)\\(.*" , "\\1", gsub("[\\*~%]" , " ", gsub("[\\^\">]" , "", qcm$qcName))) 
  ## remove weird symbols (to serve as link)
  newname_lnk = gsub("[^a-zA-Z0-9]", "", newname)
  txt_TOC = paste0(txt_TOC, paste0("    - [", newname, "](#", newname_lnk, ")\n"))
  txt_BODY = paste0(txt_BODY, '\n
## ', newname, ' <a name="', newname_lnk, '"/>
\n
  <div class="helpText">', qcm$helpTextTemplate, '</div>
<p style="text-align: right"><a href="#TOP">-- back to top --</a></p>\n\n')
}

## print the TOC
cat(txt_TOC)

```

# Overview <a name="Overview"/>

The following metrics are implemented in PTXQC.<br>

Reasons why metrics might not appear in every report
 
 * applicable only to certain types of data, e.g. SILAC or TMT
 * metric was disabled manually using the YAML config file
 * missing input data (incomplete tables), i.e. some .txt files are missing

Metric names are prefixed by the data-source, e.g. "EVD: Charge" displays and scores the charge distribution, based on data from 'evidence.txt' (EVD).

 * EVD: evidence.txt
 * SM: summary.txt
 * PAR: parameters.txt
 * PG: proteingroups.txt
 * MSMS: msms.txt
 * MS2Scans: msmsscans.txt
 

Some metrics are incompletely named here, since they might contain (hithero unknown) thresholds, which are taken from mqpar.xml (when available) or defaulted in real reports.

# Metrics <a name="Metrics"/>
```{r metrics, echo=FALSE, results="asis"}
  cat(txt_BODY)
```

If the above list is empty, the vignette was not compiled (e.g. when viewing on GitHub).
Please see https://CRAN.R-project.org/package=PTXQC --> Vignettes for a compiled version.

---
title: "Basic R-Usage Guide for PTXQC"
author: "Chris Bielow <chris.bielow@fu-berlin.de>"
date: '`r Sys.Date()`'
output:
  html_document: default
  pdf_document: null
vignette: >
  %\VignetteIndexEntry{Basic R-Usage Guide for PTXQC}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

# Basic R-Usage Guide for PTXQC

This vignette serves as a quickstart guide for R users to create a report
from within R using PTXQC.

**Target Audience:** R users

Reports can be generated using the default configuration of PTXQC. 
See
```{r, eval=FALSE}
vignette("PTXQC-InputData", package = "PTXQC")
```
how a txt folder should look like.

## Generate a default report 

This is the most simple invokation of PTXQC::createReport().

```{r, eval=FALSE}
require(PTXQC)

## the next require() is needed to prevent a spurious error in certain R versions (might be a bug in R or a package)
## error message is:
##    Error in Scales$new : could not find function "loadMethod"
require(methods)

## specify a path to a MaxQuant txt folder
## Note: This folder needs to be complete (see 'vignette("PTXQC-InputData", package = "PTXQC")')
if (1) {
  ## we will use an example dataset from PRIDE (dataset 2 of the PTXQC publication)
  local_zip = tempfile(fileext=".zip")
  tryCatch({
    download.file("ftp://ftp.pride.ebi.ac.uk/pride/data/archive/2015/11/PXD003133/txt_20min.zip", destfile = local_zip)
  }, error = function(err) {
    ## on Windows, one can get a 'cannot open URL' using the default method. So we try another
    download.file("ftp://ftp.pride.ebi.ac.uk/pride/data/archive/2015/11/PXD003133/txt_20min.zip", destfile = local_zip, method= "internal")
    
  })
  unzip(local_zip, exdir = tempdir()) ## extracts content
  txt_folder = file.path(tempdir(), "txt_20min")
} else {
  ## if you have local MaxQuant output, just use it
  txt_folder = "c:/Proteomics/MouseLiver/combined/txt"
}

r = createReport(txt_folder)

cat(paste0("\nReport generated as '", r$report_file, "'\n\n"))

```

## Generate a customized report

The report can be customized via a YAML configuration file.
For details, see

```{r, eval=FALSE}
vignette("PTXQC-CustomizeReport", package = "PTXQC")
```

After editing the YAML configuration to your needs, run the code below.

```{r, eval=FALSE}
require(PTXQC)
require(yaml)

## the next require() is needed to prevent a spurious error in certain R versions (might be a bug in R or a package)
## error message is:
##    Error in Scales$new : could not find function "loadMethod"
require(methods)

## specify a path to a MaxQuant txt folder
## Note: This folder can be incomplete, depending on your YAML config
if (1) {
  ## we will use an example dataset from PRIDE (dataset 2 of the PTXQC publication)
  local_zip = tempfile(fileext=".zip")
  download.file("ftp://ftp.pride.ebi.ac.uk/pride/data/archive/2015/11/PXD003133/txt_20min.zip", destfile = local_zip)
  unzip(local_zip, exdir = tempdir()) ## extracts content
  txt_folder = file.path(tempdir(), "txt_20min")
} else {
  ## if you have local MaxQuant output, just use it
  txt_folder = "c:/Proteomics/MouseLiver/combined/txt"
}

## use a YAML config inside the target directory if present
fh_out = getReportFilenames(txt_folder)
if (file.exists(fh_out$yaml_file))
{
  cat("\nUsing YAML config already present in target directory ...\n")
  yaml_config = yaml.load_file(input = fh_out$yaml_file)
} else {
  cat("\nYAML config not found in folder '", txt_folder, "'. The first run of PTXQC will create one for you.", sep="")
  yaml_config = list()
}

r = createReport(txt_folder, mztab_file = NULL, yaml_obj = yaml_config)

cat(paste0("\nReport generated as '", r$report_file, "'\n\n"))

```

That's it.

---
title: "ProTeomiX (PTX) Quality Control (QC) Report"
output:
  html_document:
    mathjax: null
    number_sections: yes
    toc: yes
  pdf_document:
    toc: yes
---


<!-- the TOC (table of content) will go here - use 'toc: no' above to disable the TOC -->


<style type="text/css">
  p { 
    padding: 5px 0;
  }

  .helpSymbol {
    border: 1px solid #AAAAAA;
    padding: 3px;
    text-align: center;
    width: 100px;
  }
  .helpText {
    clear: left;
  }
  
  .tooltipPTXQC {
    position: relative;
    display: inline-block;
    border-bottom: 1px dotted black;
  }
  
  .tooltipPTXQC .tooltiptext {
    visibility: hidden;
    width: 120px;
    background-color: black;
    color: #fff;
    text-align: center;
    border-radius: 6px;
    padding: 5px 0;
  
    /* Position the tooltip */
    position: absolute;
    z-index: 1;
  }
  
  .tooltipPTXQC:hover .tooltiptext {
    visibility: visible;
  }
</style>

```{r setup, include=FALSE}
## global options
knitr::opts_chunk$set(echo=FALSE, warning=FALSE, error=FALSE, message=FALSE, fig.width=10)
```

<a name="Overview"></a>

# Overview

## Checks

<!-- locale static -->
<div style="color:#008800">
  <input type="checkbox" disabled="disabled" checked="checked"> MaxQuant had an english locale (dot as decimal separator; otherwise you would receive an error when generating the report)
</div>

<!-- locale dynamic -->
<div class="tooltipPTXQC" id="locale_short"><span id="locale_shorttext"></span> (hover here for details)
  
<div class="tooltiptext" style="width:700px; border: 4px solid #73AD21; margin: auto; padding: 5px;" id="locale_box">
<font size="+2">Dynamic (at this very point in time) check of <b>this</b> system's locale (region) settings</font><br>
    (independent of how the txt-folder for this report was generated!)
<div id="locale_check">
<p>Decimal separator: <span id="locale_decimal"></span></p>
<span id="locale_explanation"></span>
</div>
</div>
</div>


<script type="text/javascript">
 /**
   * Returns an object containing decimal and thousand separator
   * based on user's locale
   */
  function getNumberSeparators() {
    
    // default
    var res = {
      "decimal": "",
      "thousand": ""
    };
    
    // convert a number formatted according to locale
    var str = parseFloat(1234.56).toLocaleString();
    
    // if the resulting number does not contain previous number
    // (i.e. in some Arabic formats), return defaults
    if (!str.match("1"))
      return res;
    
    // get decimal and thousand separators
    res.decimal = str.replace(/.*4(.*)5.*/, "$1");
    res.thousand = str.replace(/.*1(.*)2.*/, "$1");
    
    // return results
    return res;
  }
  
  function checkString(str) {
    if (str === " ")
      return "[space]";
    else if(str === ".")
      return "[dot]"
    else if(str === ",")
      return "[comma]"
    else if(str === "")
      return "<error>"
    else
      return str;
  }
  
  /**
    * Running the locale function
    */
  var separators = getNumberSeparators();
  
  document.getElementById("locale_decimal").innerHTML = checkString(separators.decimal);
  
  var sep_MQ_ok = (separators.decimal == ".")
  
  var bg_col = (sep_MQ_ok ? 'lightgreen' : '#FA8072')
  var border_col = (sep_MQ_ok ? 'green' : 'red')
  var text_col = (sep_MQ_ok ? '#009900' : '#FF0000')
  document.getElementById( 'locale_box' ).style.backgroundColor = bg_col
  document.getElementById( 'locale_box' ).style.borderColor  = border_col
  document.getElementById( 'locale_short' ).style.color = text_col
  var explanatation = (sep_MQ_ok ? "<b>All good</b>: The decimal separator is '.' [dot], which would allow MaxQuant to run just fine on <b>this</b> system!" :
                                   "<b>Warning</b>: your decimal separator is NOT '.' [dot]!<br>Running MaxQuant on this system would lead to broken MaxQuant txt output files!")
  document.getElementById("locale_explanation").innerHTML = explanatation;
  var shorttext = (sep_MQ_ok ? '<input type="checkbox" disabled="disabled" checked="checked"> Dynamic locale check for <b>this</b> system is dot! (MaxQuant would run fine)' :
                               '<input type="checkbox" disabled="disabled"                  > Dynamic locale check for <b>this</b> system is <b>not</b> dot! (MaxQuant would fail)')
document.getElementById("locale_shorttext").innerHTML = shorttext;


</script>

**Quick guide**

  - navigate using the 'Table of Content' above or just scroll down
  - collapse sections by just clicking their name
  - click on the "&darr; Show Help" symbol to see a description for each metric and its scoring
  - edit the YAML config <a href="`r paste0("file:./", basename(rprt_fns$yaml_file))`">`r paste0("file:./", basename(rprt_fns$yaml_file))`</a> to
     - set the report format (HTML, plainPDF, or both)
     - disable metrics
     - customize the order of metrics
     - set individual target thresholds
  - edit `r paste0("file:./", basename(rprt_fns$filename_sorting))` to customize
     - the order of Raw files in plots and heatmap
     - abbreviations of Raw file names
  - edit the Html template `r paste0("file:./", basename(html_template))` to
     - alter the look and feel of this report (e.g. colors, font and image sizes, help texts, etc.)
  - report bugs and file requests via the [GitHub issue tracker](https://github.com/cbielow/PTXQC/issues)
  - full PTXQC manual available [here](https://github.com/cbielow/PTXQC)


## HeatMap
```{r MQparams, echo=FALSE, results="asis"}
hm[["plot"]]
```

## Name Mapping
```{r nameMapping, echo=FALSE, results="asis"}
if(!is.null(pl_nameMapping) && (!any(is.na(pl_nameMapping)))) {
  cat(pl_nameMapping$htmlTable)
}
```


## Metrics
```{r metrics, echo=FALSE, results="asis"}
  #for (qcm in lst_qcMetrics[1:3])
  for (qcm in lst_qcMetrics)
  {
    if (length(qcm$plots) == 0) next; # skip empty metrics
    #debug: 
    #cat(c("processing ", qcm$qcName))
    v_titles = qcm$getTitles(subtitle_sep = "\n<b>")
    plots = qcm$getPlots(withTitle = FALSE)
    if (length(v_titles) != length(plots)) {
      ## this is weird...
      stop("Number of titles is unequal to number of plots ...")
    }
    last_title = ""
    for (i in 1:length(v_titles))
    {
      if (last_title != v_titles[i]) {  ## print only new titles
        cat(paste0("\n\n### ", gsub("\n", "\n<br>", v_titles[i]), "</b>\n\n"))
      }
      if (i == 1) { ## print helpText only once
        cat(paste0('\n
<div>
  <div class="helpSymbol">&darr; Show Help</div>
  <div class="helpText">', qcm$helpText, "</div>
</div>"))
      }
      if (!is.na(qcm$htmlTable))
      { ## if there is an HTML table, prefer it over the plots
        cat(qcm$htmlTable)
        break;
      }
      print(plots[[i]])
      cat("\n<br>\n")
      last_title = v_titles[i]
    }
    cat('\n<p style="text-align: right">[back to top](#Overview)</p>\n\n')
  }
```



<script type="text/javascript">
jQuery(document).ready(function()
{
 // initially hide the help text
  jQuery(".helpText").hide();
  
  // toggle help text by clicking on the symbol
  jQuery(".helpSymbol").click(function(event)
  {
    var target = $( event.target );
    // "Show Help &darr;" is the first item in a <div> which encloses their section
    // so toggling its sibling (the actual text) will just leave the "Show Help &darr;" visible
    if ( target.is( "div" ) )
    {
      target.siblings().slideToggle(200);
      event.stopPropagation()
    }
  });
  
  
  // enable shortening the report
  jQuery(".section").click(function(event)
  {
    var target = $( event.target );
    // headings are the first item in a <div> which encloses their section
    // so toggling their siblings will just leave the heading visible
    if ( target.is( "h1, h2, h3, h4" ) )
    {
      target.siblings().slideToggle(200);
      event.stopPropagation()
    }
  });
});

</script>

---
title: "Frequently asked questions for PTXQC"
author: "Chris Bielow <chris.bielow@fu-berlin.de>"
date: '`r Sys.Date()`'
output:
  html_document:
    toc: true
  pdf_document: null
vignette: >
  %\VignetteIndexEntry{Frequently asked questions for PTXQC}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

# FAQ (Frequently asked questions) for PTXQC

## Installation
_______

### Q: During installation an error pops up: '... cannot open URL 'ftp://ftp.pride.ebi.ac.uk/pride/ ..'. ###

**Answer:** When R builds the vignettes for PTXQC, it needs to download a test dataset from a public PRIDE repository. On rare occasions, the server
might not be reachable. Either wait a little until it comes back up and try again, OR simply skip building the vignettes using

```install_github("cbielow/PTXQC", build_vignettes = FALSE, dependencies = TRUE)```

This omits the part where the error occurs. Downside: no offline
vignettes for PTXQC -- but you can always look them up on GitHub.

Alternatively, use CRAN: `install.packages("PTXQC")`.

### Q: During installation an error pops up: '... cannot open the connection...'. ###

**Answer:** One common error is that the CRAN mirror hosting the R packages is incomplete/buggy and installation of packages fails.
   This **could** look like this:
   ```
   Error in XXXX(file.path(pkgname, "DESCRIPTION"), c("Package", "Type")) : 
   cannot open the connection
   In addition: Warning messages:
   ...
   ```
   or
   ```
   Warning message:
   In download.file(url, destfile, method, mode = "wb", ...) :
   downloaded length 228962 != reported length 229751
   ```
   
   Try another CRAN mirror (type `chooseCRANmirror()`) and then re-run the failed command.
  
   If you're still stuck, please open a [ticket][issuetracker] (see 'Bug Reporting' on the main PTXQC homepage for the details you should provide).

### Q: During vignette building, I see an error related to Pandoc and 'It seems you should call rmarkdown::render() instead of knitr::knit2html()'.   

**Answer:** The full error message could look like:
```ERROR Warning in engine$weave(file, quiet = quiet, encoding = enc) : Pandoc (>= 1.12.3) and/or pandoc-citeproc is not available. Please install both. Warning in readLines(con) : incomplete final line found on 'PTXQC-Basic_Guide_for_R_users.Rmd' Error: processing vignette 'PTXQC-Basic_Guide_for_R_users.Rmd' failed with diagnostics: It seems you should call rmarkdown::render() instead of knitr::knit2html() because PTXQC-Basic_Guide_for_R_users.Rmd appears to be an R Markdown v2 document. Execution halted Error: Command failed (1)```

Knitr (which is the engine which builds the vignettes), relies on Pandoc. If you called `install_github(..., build_vignettes = TRUE)` then Pandoc is required!
See the installation instructions on where to get it. Make sure to restart R once you have installed Pandoc!
Pandoc is also required to build the report in Html format -- which the default. PDF works just fine without Pandoc.

If you really don't have or want Pandoc, you can skip vignettes and install PTXQC using `install_github(..., build_vignettes = FALSE)` and
configure your YAML parameter file to omit Html reports and just get plain PDF (not recommended though).
_______

## Report Creation

### Q: A warning **"The file '<...>/mqpar.xml' was not found..."** in the console popped up. ###

**Answer:** PTXQC uses the mqpar.xml file from MaxQuant to extract certain parameters (like mass accuracy search tolerances) automatically. By default the mqpar.xml is created by MaxQuant in the root folder of each analysis (where the Raw files reside). The warning now tells you to **copy this mqpar.xml into the combined/txt/ folder**, where PTXQC can read it.
    
The reason why PTXQC does not directly use the mqpar.xml from the root folder is two-fold:

1. the mqpar.xml at the root folder might be outdated/changed, especially of you run MaxQuant multiple times, and rename the txt folders (to compare multiple parameter settings)
2. users often copy the txt folder from the MaxQuant 'server' to a remote machine 
            for result interpretation. Usually only the txt folder is copied, but not the Raw files. 
            If PTXQC is run on the remote machine, the mqpar.xml needs to reside within the txt folder.

_______

## Report Interpretation

### Q: How do I know what the figures in the report mean? ###

**Answer:** Most metrics and figures are described in the publication (see 'Citation' on the main homepage of PTXQC). Updated and new metrics will be described in the documentation of the package. A short description of each plot and the scoring function is given in each Html report (click the "Show Help" box just below each section heading). See [examples][https://github.com/cbielow/PTXQC/tree/master/inst/examples] for more.

_______
   
  [issuetracker]: https://github.com/cbielow/PTXQC/issues

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MzTabReader.R
\docType{class}
\name{MzTabReader-class}
\alias{MzTabReader-class}
\alias{MzTabReader}
\title{Class to read an mzTab file and store the tables internally.}
\description{
The 'sections' field is initialized after $readMzTab was called.
The 'fn_map' fields should be initialized via ...$fn_map$readMappingFile(...) manually if user-defined filename mappings are desired
and is automatically updated/queried when $readMzTab is called.
}
\section{Fields}{

\describe{
\item{\code{sections}}{MzTab sections as list. Valid list entries are: "MTD", "PRT", "PEP", "PSM", "SML", "filename" and "comments"}

\item{\code{fn_map}}{FilenameMapper which can translate raw filenames into something shorter}
}}

\section{Methods}{

\describe{
\item{\code{RTUnitCorrection(dt)}}{Convert all RT columns from seconds (OpenMS default) to minutes (MaxQuant default)}

\item{\code{getEvidence()}}{Basically the PSM table and additionally columns named 'raw.file' and 'fc.raw.file'.}

\item{\code{getMSMSScans(identified_only = FALSE)}}{Basically the PSM table (partially renamed columns) and additionally two columns 'raw.file' and 'fc.raw.file'. 
If identified_only is TRUE, only MS2 scans which were identified (i.e. a PSM) are returned -- this is equivalent to msms.txt in MaxQuant.}

\item{\code{getParameters()}}{Converts internal mzTab metadata section to a two column key-value data.frame similar to MaxQuants parameters.txt.}

\item{\code{getProteins()}}{Basically the PRT table ...}

\item{\code{getSummary()}}{Converts internal mzTab metadata section to a two data.frame with columns 'fc.raw.file', 'ms.ms.identified....'
similar to MaxQuants summary.txt.}

\item{\code{renameColumns(dt, namelist)}}{Renames all columns and throws a warning if a column does not exist in the data}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plotTableRaw}
\alias{plotTableRaw}
\title{Colored table plot.}
\usage{
plotTableRaw(data, colours = "black", fill = NA, just = "centre")
}
\arguments{
\item{data}{Table as Data.frame}

\item{colours}{Single or set of colours (col-wise)}

\item{fill}{Cell fill (row-wise)}

\item{just}{(ignored)}
}
\value{
gTable
}
\description{
Code taken from http://stackoverflow.com/questions/23819209/change-text-color-for-cells-using-tablegrob-in-r
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_UncalibratedMSErr}
\alias{plot_UncalibratedMSErr}
\title{A boxplot of uncalibrated mass errors for each Raw file.}
\usage{
plot_UncalibratedMSErr(
  data,
  MQBug_raw_files,
  stats,
  y_lim,
  extra_limit,
  title_sub
)
}
\arguments{
\item{data}{A data.frame with columns 'fc.raw.file', 'uncalibrated.mass.error..ppm.'}

\item{MQBug_raw_files}{List of Raw files with invalid calibration values}

\item{stats}{A data.frame with columns 'fc.raw.file', 'sd', 'outOfCal'}

\item{y_lim}{Range of y-axis}

\item{extra_limit}{Position where a v-line is plotted (for visual guidance)}

\item{title_sub}{Subtitle}
}
\value{
GGplot object
}
\description{
Boxes are optionally colored to indicate that a MQ bug was detected or
if PTXQC detected a too narrow search window.
}
\examples{
  n = c(150, 1000, 1000, 1000)
  data = data.frame(fc.raw.file = repEach(letters[4:1], n),
                    uncalibrated.mass.error..ppm. = c(rnorm(n[1], 13, 2.4),
                                                      rnorm(n[2], 1, 0.5),
                                                      rnorm(n[3], 3, 0.7),
                                                      rnorm(n[4], 4.5, 0.8)))
  stats = data.frame(fc.raw.file = letters[4:1],
                     sd_uncal = c(2.4, 0.5, 0.7, 0.8),
                     outOfCal = c(TRUE, FALSE, FALSE, FALSE))
  plot_UncalibratedMSErr(data, MQBug_raw_files = letters[1],
                         stats, y_lim = c(-20,20), 15, "subtitle")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_MQalign.R
\name{alignmentCheck}
\alias{alignmentCheck}
\title{Verify an alignment by checking the retention time differences of identical peptides across Raw files}
\usage{
alignmentCheck(data, referenceFile)
}
\arguments{
\item{data}{A data.frame with columns 'calibrated.retention.time', 'id', 'modified.sequence', 'charge', 'raw.file' and 'fraction' (if present)}

\item{referenceFile}{A raw file name as occuring in data$raw.file, serving as alignment reference (when no fractions are used).}
}
\value{
A data.frame containing the RT diff for each feature found in a Raw file and the reference.
}
\description{
The input is a data frame containing feature evidence with corrected retention times,
e.g. a 'calibrated.retention.time' column.
}
\details{
Note that this function must be given real MS/MS identifications only (type "MULTI-MSMS")
in order to work correctly!

For each peptide sequence (and charge) in the reference Raw file, this function looks up the
already calibrated retention time difference of the same feature in all other files. For every comparison made,
we report the RT difference. If alignment worked perfectly, the differences are very small (<1 min).

An 'id' column must be present, to enable mapping the result of this function to the original data frame.

A reference Raw file can be identified using 'findAlignReference()'. If Maxquants experimental design included
pre-fractionation, a column named 'fraction' should be given and 'referenceFile' should be empty. This function will
pick the one Raw file for each fraction (the first in order) to use as reference. Only the immediately neighbouring
fractions will be matched to this reference.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{checkEnglishLocale}
\alias{checkEnglishLocale}
\title{When MaxQuant is run with a wrong locale (i.e. the decimal separator is not a '.', but a ','), 
then MaxQuant results are plainly wrong and broken. The can be detected by, e.g. checking for negative charge annotation}
\usage{
checkEnglishLocale(df_evd)
}
\arguments{
\item{df_evd}{Evidence table from which we only need the 'charge' column}
}
\description{
When MaxQuant is run with a wrong locale (i.e. the decimal separator is not a '.', but a ','), 
then MaxQuant results are plainly wrong and broken. The can be detected by, e.g. checking for negative charge annotation
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PTXQC-package.R
\docType{package}
\name{PTXQC-package}
\alias{PTXQC-package}
\title{PTXQC: A package for computing Quality Control (QC) metrics for Proteomics (PTX)}
\description{
The following sections describe the main components:
}
\section{Input}{

Valid input data are either the files from MaxQuant's .txt folder (all versions from MaxQuant >= 1.0 upwards are supported)
or a single mzTab file. All mzTab files will work, but most metrics can be obtained from OpenMS' mzTab as produced
by the QualityControl TOPP tool (from OpenMS 2.5 onwards).
}

\section{Important functions}{

The central function of this package is called \code{\link{createReport}} and it accepts either MaxQuant or mzTab data, along with
a configuration (optional).
There is a parser for mzTab \code{\link{MzTabReader}} and MaxQuant txt files \code{\link{MQDataReader}}, as well as a plethora of QC metrics
derived from a common \code{\link{qcMetric}} class and scoring functions \code{qual...}, e.g. \code{\link{qualGaussDev}}.
}

\section{Configuration}{

The user can modify the behaviour of PTXQC, e.g. to enable/disable certain metrics or change scoring thresholds, via a YAML object/file.
By default a Yaml file is written automatically side-by-side to the input files upon running PTXQC for the first time on a particular input.
A custom Yaml object can be passed to the main \code{\link{createReport}} function for customization.
Use \code{yaml::yaml.load_file(input = 'myYAML.yaml')} to load an existing file and pass the Yaml object along.
}

\section{Output}{

Either a PDF and/or Html report which contains QC plots and a description of the metrics.
}

\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/cbielow/PTXQC}
  \item Report bugs at \url{https://github.com/cbielow/PTXQC/issues}
}

}
\author{
\strong{Maintainer}: Chris Bielow \email{chris.bielow@bsc.fu-berlin.de}

Other contributors:
\itemize{
  \item Juliane Schmachtenberg [contributor]
  \item Swenja Wagner [contributor]
  \item Patricia Scheil [contributor]
  \item Tom Waschischek [contributor]
  \item Guido Mastrobuoni [data contributor, reviewer]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_MissedCleavages}
\alias{plot_MissedCleavages}
\title{Plot bargraph of missed cleavages.}
\usage{
plot_MissedCleavages(data, title_sub = "")
}
\arguments{
\item{data}{A data.frame with columns 'fc.raw.file', '...' (missed cleavage classes)}

\item{title_sub}{Plot's subtitle}
}
\value{
GGplot object
}
\description{
Per Raw file, an arbitrary number of missed cleavage classes (one per column) can be given.
The total fraction of 3D-peaks must sum to 1 (=100%).
Columns are ordered by name.
}
\details{
A visual threshold line is drawn at 75% (expected MC0 count).
}
\examples{
  data = data.frame(fc.raw.file = letters[1:5],
                    MC0 = c(0.8, 0.5, 0.85, 0.2, 0.9),
                    MC1 = c(0.1, 0.4, 0.05, 0.7, 0.0),
                    "MS2+" =  c(0.1, 0.1, 0.1, 0.1, 0.1),
                    check.names = FALSE)
  plot_MissedCleavages(data, "contaminant inclusion unknown")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_MQalign.R
\name{ScoreInAlignWindow}
\alias{ScoreInAlignWindow}
\title{Compute the fraction of features per Raw file which have an acceptable RT difference after alignment}
\usage{
ScoreInAlignWindow(data, allowed.deltaRT = 1)
}
\arguments{
\item{data}{A data.frame with columns 'rtdiff' and 'raw.file'}

\item{allowed.deltaRT}{The allowed matching difference (1 minute by default)}
}
\value{
A data.frame with one row for each raw.file and columns 'raw.file' and 'withinRT' (0-1)
}
\description{
Using the result from 'alignmentCheck()', score the features of every Raw file and see if they 
have been properly aligned.
Returned value is between 0 (bad) and 1 (all aligned).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_qualities.R
\name{qualBestKS}
\alias{qualBestKS}
\title{From a list of vectors, compute all vs. all Kolmogorov-Smirnoff distance statistics (D)}
\usage{
qualBestKS(x)
}
\arguments{
\item{x}{List of vectors, where each vector holds a distribution}
}
\value{
A data.frame with ks-test values of the "reference" to all other distributions (see Details)
}
\description{
... and report the row of the matrix which has maximum sum (i.e the best "reference" distribution).
The returned data.frame has as many rows as distributions given and two columns.
The first column 'name' gives the name of the list element, the second column 'ks_best' gives '1-statistic' of the
Kolmogorov-Smirnoff test to the "reference" distribution (which was picked by maximising the sum of 'ks_best').
Thus, the row with a 'ks_best' of 1 is the reference distribution.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modsToTableByRaw.R
\name{modsToTableByRaw}
\alias{modsToTableByRaw}
\title{Convert list of (mixed)modifications to a frequency table}
\usage{
modsToTableByRaw(
  df_evd,
  name_unmod = "Unmodified",
  name_unmod_inverse = "Modified (total)"
)
}
\arguments{
\item{df_evd}{data.frame with 'fc.raw.file' and a 'modifications' column, which contains the modifications for each peptide.}

\item{name_unmod}{String in 'modifications' which represents an unmodified peptide}

\item{name_unmod_inverse}{If non-empty, then inverse the frequencies of the 'name_unmod' modifications (i.e. 100-x) IFF they are >=50\% on average (across Raw files) and rename them to this string}
}
\value{
A data.table with 'fc.raw.file', 'modification_names' (factor), and 'Freq' (0-100)
}
\description{
Convert list of (mixed)modifications to a frequency table
}
\examples{
data = data.frame(fc.raw.file = rep(c("file A", "file B"),
                                each = 3),
       modifications = c("Oxidation (M)",
                        "Unmodified",
                        "Oxidation (M),Acetyl (Protein N-term)",
                        "2 Oxidation (M)",
                        "Unmodified", "Unmodified"))
modsToTableByRaw(data)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_MBRIDtransfer}
\alias{plot_MBRIDtransfer}
\title{Plot MaxQuant Match-between-runs id transfer performance.}
\usage{
plot_MBRIDtransfer(data)
}
\arguments{
\item{data}{A data.frame with columns as described above}
}
\value{
GGplot object
}
\description{
The plots shows the different categories of peak classes
}
\details{
The input is a data.frame with columns
  'fc.raw.file' - raw file name
  'single' - fraction of peptides with are represent only once
  'multi.inRT' - fraction of peptides with are represent multiple times,
                 but within a certain RT peak width
  'multi.outRT' - fraction of peptides with are represent multiple times,
                  with large RT distance
  'sample' - raw file
where each row represents one peptide sequence.
}
\examples{
 data = data.frame(fc.raw.file = rep(c("file A", "file B"), each = 3),
                   single = c(0.9853628, 0.8323160, 0.9438375,
                              0.9825538, 0.8003763, 0.9329961),
                   multi.inRT = c(0.002927445, 0.055101018, 0.017593087,
                                  0.005636457, 0.099640044, 0.031870056),
                   multi.outRT = c(0.01170978, 0.11258294, 0.03856946,
                                   0.01180972, 0.09998363, 0.03513386),
                   sample = rep(c("genuine", "transferred", "all"), 2))
 plot_MBRIDtransfer(data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{LCSn}
\alias{LCSn}
\title{Find longest common substring from 'n' strings.}
\usage{
LCSn(strings, min_LCS_length = 0)
}
\arguments{
\item{strings}{A vector of strings in which to search for LCS}

\item{min_LCS_length}{Minimum length expected. Empty string is returned if the result is shorter}
}
\value{
longest common substring (or "" if shorter than \code{min_LCS_length})
}
\description{
Warning: greedy heuristic! This is not guaranteed to find the best solution (or any solution at all), since its done pairwise with the shortest input string as reference.
}
\examples{
LCSn(c("1_abcde...",
       "2_abcd...",
       "x_abc..."))  ## --> "_abc"
LCSn(c("16_IMU008_CISPLA_E5_R11", 
       "48_IMU008_CISPLA_P4_E7_R31",
       "60_IMU008_CISPLA_E7_R11"), 3) ## -->"_IMU008_CISPLA_"
LCSn(c("AAAAACBBBBB", 
       "AAAAADBBBBB",
       "AAAABBBBBEF",
       "AAABBBBBDGH")) ## -->  "BBBBB"
LCSn(c("AAAXXBBB",
       "BBBXXDDD",
       "XXAAADDD")) ## --> fails due to greedy approach; should be "XX"

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{repEach}
\alias{repEach}
\title{Repeat each element x_i in X, n_i times.}
\usage{
repEach(x, n)
}
\arguments{
\item{x}{Values to be repeated}

\item{n}{Number of repeat for each x_i (same length as x)}
}
\value{
Vector with values from x, n times
}
\description{
Repeat each element x_i in X, n_i times.
}
\examples{

  repEach(1:3, 1:3)  ## 1, 2, 2, 3, 3, 3

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{\%+\%}
\alias{\%+\%}
\title{A string concatenation function, more readable than 'paste()'.}
\usage{
a \%+\% b
}
\arguments{
\item{a}{Char vector}

\item{b}{Char vector}
}
\value{
Concatenated string (no separator)
}
\description{
A string concatenation function, more readable than 'paste()'.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_qualities.R
\name{qualUniform}
\alias{qualUniform}
\title{Compute deviation from uniform distribution}
\usage{
qualUniform(x, weight = vector())
}
\arguments{
\item{x}{Vector of numeric intensity/count values (e.g. ID's per RT bin); bins are assumed to have equal widths}

\item{weight}{Vector of weights for values in 'x' (same length as 'x').}
}
\value{
Value between [0, 1]
}
\description{
The score ranges between 0 (worst score) and 1 (best score).
Input 'x' is a vector of counts (or probabilities) for equally spaced bins in a histogram.
A uniform distribution (e.g. c(3,3,3) will get a score of 1. The worst possible case (e.g. c(4,0,0)), will get a score of 0,
and a linear increasing function (e.g. c(1,2,3)) will get something in between (0.585 here)
}
\details{
In addition, bin values can be weighted (e.g. by their confidence). The total sum of weights is normalized to 1 internally.

The distance function used is the square root of the absolute difference between a uniform distribution and the input 'x'
(summed for each element of 'x').
This distance is normalized to the worst possible input (e.g. one bin with 100% counts, all other bins being empty).
}
\examples{
 stopifnot(qualUniform(c(3,3,3))==1)
 stopifnot(qualUniform(c(4,0,0))==0)

 ## how 'uniform' is a vector where only a single index has weight?-- answer: very
 stopifnot(qualUniform(c(4,0,0), c(1,0,0))==1)
 stopifnot(qualUniform(c(4,0,0), c(0,1,0))==1)
 stopifnot(qualUniform(c(0,4,0))==0)
 stopifnot(abs(qualUniform(c(3,2,1))-0.58578) < 0.0001)
 stopifnot(abs(qualUniform(c(1,2,3))-0.58578) < 0.0001)
 stopifnot(qualUniform(c(1,2,3), c(0,1,0))==1)
 stopifnot(abs(qualUniform(c(1,2,3))-0.58578) < 0.0001)
 stopifnot(abs(qualUniform(c(1,2,3), c(0,1,1))- 0.590316) < 0.0001)
 stopifnot(abs(qualUniform(c(2,3), c(1,1))-0.552786) < 0.0001)
 stopifnot(abs(qualUniform(1:120)-0.38661) < 0.0001)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{peakWidthOverTime}
\alias{peakWidthOverTime}
\title{Discretize RT peak widths by averaging values per time bin.}
\usage{
peakWidthOverTime(data, RT_bin_width = 2)
}
\arguments{
\item{data}{Data.frame with columns 'retention.time' and 'retention.length'}

\item{RT_bin_width}{Bin size in minutes}
}
\value{
Data.frame with columns 'bin', 'RT', 'peakWidth'
}
\description{
Should be applied for each Raw file individually.
}
\details{
Returns a data.frame, where 'bin' gives the index of each bin, 'RT' is 
the middle of each bin and 'peakWidth' is the averaged peak width per bin.
}
\examples{
  data = data.frame(retention.time = seq(30,200, by=0.001)) ## one MS/MS per 0.1 sec
  data$retention.length = seq(0.3, 0.6, length.out = nrow(data)) + rnorm(nrow(data), 0, 0.1)
  d = peakWidthOverTime(data)
  plot(d$RT, d$peakWidth)
  
  
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_MQ.R
\name{fixCalibration}
\alias{fixCalibration}
\title{Detect (and fix) MaxQuant mass recalibration columns, since they
sometimes report wrong values.}
\usage{
fixCalibration(
  df_evd,
  df_idrate = NULL,
  tolerance_sd_PCoutOfCal = 2,
  low_id_rate = 1
)
}
\arguments{
\item{df_evd}{Evidence data.frame with columns ()}

\item{df_idrate}{Data.frame from summary.txt, giving ID rates for each raw file (cols: "ms.ms.identified....", "fc.raw.file"). Can also be NULL.}

\item{tolerance_sd_PCoutOfCal}{Maximal standard deviation allowed before considered 'failed'}

\item{low_id_rate}{Minimum ID rate in Percent before a Raw file is considered 'failed'}
}
\value{
list of data (stats, affected_raw_files, df_evd, recal_message, recal_message_post)
}
\description{
Returns a list of items for both diagnostics and possibly a fixed evidence data.frame.
Also two strings with messages are returned, which can serve as user message for
pre and post calibration status.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mzQC.R
\name{getRunQualityTemplate}
\alias{getRunQualityTemplate}
\title{Get an mzQC runQuality without actual metrics, but with full metadata}
\usage{
getRunQualityTemplate(fc.raw.file, raw_file_mapping)
}
\arguments{
\item{fc.raw.file}{For which run}

\item{raw_file_mapping}{A data.frame with cols 'from', 'to' and maybe 'best.effort' (if shorting was unsuccessful), as e.g. obtained by a FilenameMapper$raw_file_mapping}
}
\value{
An MzQCrunQuality object
}
\description{
Get an mzQC runQuality without actual metrics, but with full metadata
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MQDataReader.R
\name{getFileEncoding}
\alias{getFileEncoding}
\title{Determine if a file is 'UTF-8' or 'UTF-8-BOM' (as of MQ2.4) or 'UTF-16BE' or 'UTF-16LE'}
\usage{
getFileEncoding(filename)
}
\arguments{
\item{filename}{Relative or absolute path to a file}
}
\value{
'' if the file does not exist or is not readable
}
\description{
Determine if a file is 'UTF-8' or 'UTF-8-BOM' (as of MQ2.4) or 'UTF-16BE' or 'UTF-16LE'
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_MBRAlign}
\alias{plot_MBRAlign}
\title{Plot MaxQuant Match-between-runs alignment performance.}
\usage{
plot_MBRAlign(data, y_lim, title_sub, match_tol)
}
\arguments{
\item{data}{A data.frame with columns as described above}

\item{y_lim}{Plot range of y-axis}

\item{title_sub}{Subtitle}

\item{match_tol}{Maximal residual RT delta to reference (usually ~1 min)}
}
\value{
GGplot object
}
\description{
The plots shows the correction function applied by MaxQuant, and the
residual RT (ideally 0) of each peptide to its reference. Uncalibrated peptides
are shown in red, calibrated ones in green.
The MaxQuant RT correction which was applied prior is shown in blue. The range of this function
can give hints if the allowed RT search window (20min by default) is sufficient or if
MaxQuant should be re-run with more tolerant settings.
}
\details{
The input is a data.frame with columns
  'calibrated.retention.time' - resulting (hopefully) calibratated RT after MQ-recal (the X-axis of the plot)
  'retention.time.calibration' - delta applied by MaxQuant
  'rtdiff' - remaining RT diff to reference peptide of the same sequence
  'RTdiff_in' - is the feature aligned (within 'match_tol')?
  'fc.raw.file_ext' - raw file
where each row represents one peptide whose RT was corrected by MaxQuant.
}
\examples{

 data = data.frame(fc.raw.file_ext = "file A", ## more than one would be possible
                   calibrated.retention.time = c(20:100),
                   retention.time.calibration = 6 + sin((20:100)/10))
 data$rtdiff = rnorm(nrow(data))
 data$RTdiff_in = c("green", "red")[1 + (abs(data$rtdiff) > 0.7)]

 plot_MBRAlign(data, c(-10, 10), "fancy subtitle", 0.7)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{RSD}
\alias{RSD}
\title{Relative standard deviation (RSD)}
\usage{
RSD(x)
}
\arguments{
\item{x}{Vector of numeric values}
}
\value{
RSD
}
\description{
Simply \code{\link{CV}}*100
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_QCHeat.R
\name{getMetaData}
\alias{getMetaData}
\title{Extract meta information (orderNr, metric name, category)
from a list of Qc metric objects}
\usage{
getMetaData(lst_qcMetrics)
}
\arguments{
\item{lst_qcMetrics}{List of qcMetrics}
}
\value{
data.frame with columns 'name', 'order' and 'cat' (category)
}
\description{
Extract meta information (orderNr, metric name, category)
from a list of Qc metric objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_MQalign.R
\name{inMatchWindow}
\alias{inMatchWindow}
\title{For grouped peaks: separate them into in-width vs. out-width class.}
\usage{
inMatchWindow(data, df.allowed.deltaRT)
}
\arguments{
\item{data}{A data.frame with columns 'fc.raw.file', 'rtdiff_mixed', 'rtdiff_genuine'}

\item{df.allowed.deltaRT}{The allowed matching difference for each Raw file (as data.frame(fc.rawfile, m))}
}
\value{
A data.frame with one row for each raw.file and columns 'raw.file' and score 'withinRT' (0-1)
}
\description{
Looking at groups only: Compute the fraction of 3D-peak pair groups per Raw file which 
have an acceptable RT difference after alignment using the result from 'idTransferCheck()',
i.e. compute the fraction of groups which are within a certain RT tolerance.
}
\details{
Returned value is between 0 (bad) and 1 (all within tolerance).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_miscGGplot.R
\name{printWithFooter}
\alias{printWithFooter}
\title{Augment a ggplot with footer text}
\usage{
printWithFooter(gg_obj, bottom_left = NULL, bottom_right = NULL)
}
\arguments{
\item{gg_obj}{ggplot2 object to be printed}

\item{bottom_left}{Footer text for bottom left side}

\item{bottom_right}{Footer text for bottom right side}
}
\value{
-
}
\description{
Augment a ggplot with footer text
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{appendEnv}
\alias{appendEnv}
\title{Add the value of a variable to an environment (fast append)}
\usage{
appendEnv(env_name, v, v_name = NULL)
}
\arguments{
\item{env_name}{String of the environment variable}

\item{v}{Value to be inserted}

\item{v_name}{String used as variable name. Automatically generated if omitted.}
}
\value{
Always TRUE
}
\description{
The environment must exist, and its name must be given as string literal in 'env_name'!
The value of the variable 'v' will be stored under the name given in 'v_name'.
If 'v_name' is not given, a variable name will be created by increasing an internal counter
and using the its value padded with zeros as name (i.e., "0001", "0002" etc).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_qualities.R
\name{qualMedianDist}
\alias{qualMedianDist}
\title{Quality metric which measures the absolute distance from median.}
\usage{
qualMedianDist(x)
}
\arguments{
\item{x}{A vector numeric values between [0,1]}
}
\value{
A vector of the same size as x, with quality values between [0, 1]
}
\description{
Ranges between 0 (worst score) and 1 (best score).
Input must be between [0,1].
Deviations from the median of the sample represent the score for each sample point.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_qualities.R
\name{qualCenteredRef}
\alias{qualCenteredRef}
\title{Quality metric for 'centeredness' of a distribution around zero with a user-supplied range threshold.}
\usage{
qualCenteredRef(x, tol)
}
\arguments{
\item{x}{Vector of values (hopefully in interval [-tol, tol])}

\item{tol}{Border of interval (must be positive)}
}
\value{
Value between [0, 1]
}
\description{
Ranges between 0 (worst score) and 1 (best score).
The best score is achieved when the median of 'x' is close to the center of the interval [-tol, tol].
If median of 'x' is close to the border (on either side), the score decreases linearly to zero.
Can be used for uncalibrated mass errors, as a measure of how well they are centered around 0.
}
\details{
NA's are removed for all computations.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{getReportFilenames}
\alias{getReportFilenames}
\title{Assembles a list of output file names, which will be created during reporting.}
\usage{
getReportFilenames(
  folder,
  report_name_has_folder = TRUE,
  mzTab_filename = NULL
)
}
\arguments{
\item{folder}{Directory where the MaxQuant output (txt folder) or the mzTab file resides}

\item{report_name_has_folder}{Boolean: Should the report files (html, pdf) contain the name
of the deepest(=last) subdirectory in **txt_folder** which is not `txt`?
Useful for discerning different reports in a PDF viewer.
E.g. when flag is FALSE: `report_v0.91.0.html`; and `report_v0.91.0_bloodStudy.html` when flag is TRUE (and the
txt folder is `.../bloodStudy/txt/` or `...bloodStudy/`)}

\item{mzTab_filename}{If input is an mzTab, specify its name, so that the filenames can use its basename as infix
E.g. when `mzTab_filename = 'HEK293-study.mzTab'` then the output will be
     `report_HEK293-study.html`.
This allows to get reports on multiple mzTabs in the same folder without overwriting report results.}
}
\value{
List of output file names (just names, no file is created) 
        with list entries: 
        **yaml_file**, **heatmap_values_file**, **R_plots_file**, **filename_sorting**, **mzQC_file**, **log_file**, **report_file_prefix**, **report_file_PDF**, **report_file_HTML**
}
\description{
You can combine **report_name_has_folder** (and **mzTab_filename** for mzTab files) to obtain report filenames which are even more
robust to moving around (since they contain infixes of the mzTab filename and the folder),
e.g. `@em `report_HEK293-study_myProjects.html``, where the input 
     was `mzTab_filename='HEK293-study.mzTab` and `folder='c:/somePath/myProjects/`.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{assignBlocks}
\alias{assignBlocks}
\title{Assign set numbers to a vector of values.}
\usage{
assignBlocks(values, set_size = 5, sort_values = TRUE)
}
\arguments{
\item{values}{Vector of values}

\item{set_size}{Number of distinct values allowed in a set}

\item{sort_values}{Before assigning values to sets, sort the values?}
}
\value{
Vector (same length as input) with set numbers
}
\description{
Each set has size set_size (internally optimized using \code{\link{correctSetSize}}), holding values from 'values'.
This gives n such sets and the return value is just the set index for each value.
}
\examples{
 #library(PTXQC)
 assignBlocks(c(1:11, 1), set_size = 3, sort_values = FALSE)
 ## --> 1 1 1 2 2 2 3 3 3 4 4 1
 
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{wait_for_writable}
\alias{wait_for_writable}
\title{Check if a file is writable and blocks an interactive session, waiting for user input.}
\usage{
wait_for_writable(
  filename,
  prompt_text = paste0("The file '", filename,
    "' is not writable. Please close all applications using this file. Press '",
    abort_answer, "' to abort!"),
  abort_answer = "n"
)
}
\arguments{
\item{filename}{The file to test for writable}

\item{prompt_text}{If not writable, show this prompt text to the user}

\item{abort_answer}{If the user enters this string into the prompt, this function will stop()}
}
\value{
TRUE if writable, FALSE if aborted by user or (not-writeable and non-interactive)
}
\description{
This functions gives the user a chance to make the output file writeable before a write attempt
is actually made by R to avoid having run the whole program again upon write failure.
}
\details{
Note: The file will not be overwritten or changed by this function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{getAbundanceClass}
\alias{getAbundanceClass}
\title{Assign a relative abundance class to a set of (log10) abundance values}
\usage{
getAbundanceClass(x)
}
\arguments{
\item{x}{Vector of numeric values (in log10)}
}
\value{
Vector of factors corresponding to input with abundance class names (e.g. low, high)
}
\description{
Abundances (should be logged already) are grouped into different levels,
starting from the smallest values ("low") to the highest values ("high").
Intermediate abundances are either assigned as "mid", or "low-mid".
If the range is too large, only "low" and "high" are assigned, the intermediate values
are just numbers.
}
\details{
Example:
 getAbundanceClass(c(12.4, 17.1, 14.9, 12.3)) ## --> factor(c("low", "high", "mid", "low"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{LCS}
\alias{LCS}
\title{Compute longest common substring of two strings.}
\usage{
LCS(s1, s2)
}
\arguments{
\item{s1}{String one}

\item{s2}{String two}
}
\value{
String containing the longest common substring
}
\description{
Implementation is very inefficient (dynamic programming in R)
--> use only on small instances
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/YAMLClass.R
\docType{class}
\name{YAMLClass-class}
\alias{YAMLClass-class}
\alias{YAMLClass}
\title{Query a YAML object for a certain parameter.}
\description{
If the object has the param, then return it.
If the param is unknown, create it with the given default value and return the default.
}
\section{Fields}{

\describe{
\item{\code{yamlObj}}{A Yaml object as created by \code{\link[yaml]{yaml.load}}}
}}

\section{Methods}{

\describe{
\item{\code{getYAML(param_name, default, min = NA, max = NA)}}{Query this YAML object for a certain parameter and return its value. If it does not exist it is created with a default value.
An optional min/max range can be specified and will be enforced if the value is known (default will be used upon violation).}

\item{\code{setYAML(param_name, value)}}{Set a YAML parameter to a certain value. Overwrites the old value or creates a new entry if hithero unknown.}

\item{\code{writeYAML(filename)}}{Write YAML config (including some documentation) to a YAML file. Returns TRUE on success (always), unless writing the file generates an error.}
}}

\examples{
    yc = YAMLClass$new(list())
    val = yc$getYAML("cat$subCat", "someDefault")
    val  ## someDefault
    val = yc$setYAML("cat$subCat", "someValue")
    val  ## someValue
    yc$getYAML("cat$subCat", "someDefault") ## still 'someValue' (since its set already)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{del0}
\alias{del0}
\title{Replace 0 with NA in a vector}
\usage{
del0(x)
}
\arguments{
\item{x}{A numeric vector}
}
\value{
Vector of same size as 'x', with 0's replaced by NA
}
\description{
Replace 0 with NA in a vector
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{print.PTXQC_table}
\alias{print.PTXQC_table}
\title{helper S3 class, enabling print(some-plot_Table-object)}
\usage{
\method{print}{PTXQC_table}(x, ...)
}
\arguments{
\item{x}{Some Grid object to plot}

\item{...}{Further arguments (not used, but required for consistency with other print methods)}
}
\description{
helper S3 class, enabling print(some-plot_Table-object)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_miscGGplot.R
\name{theme_blank}
\alias{theme_blank}
\title{A blank theme (similar to the deprecated theme_blank())}
\usage{
theme_blank()
}
\value{
A ggplot2 object, representing an empty theme
}
\description{
A blank theme (similar to the deprecated theme_blank())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qcMetric_MSMSScans.R
\docType{class}
\name{qcMetric_MSMSScans_TopNoverRT-class}
\alias{qcMetric_MSMSScans_TopNoverRT-class}
\alias{qcMetric_MSMSScans_TopNoverRT}
\title{Metric for msmsscans.txt, showing TopN over RT.}
\description{
Metric for msmsscans.txt, showing TopN over RT.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_IDRate}
\alias{plot_IDRate}
\title{Plot percent of identified MS/MS for each Raw file.}
\usage{
plot_IDRate(data, id_rate_bad, id_rate_great, label_ID)
}
\arguments{
\item{data}{A data.frame with columns as described above}

\item{id_rate_bad}{Number below which the ID rate is considered bad}

\item{id_rate_great}{Number above which the ID rate is considered great}

\item{label_ID}{Named vector with colors for the categories given in data$cat}
}
\value{
GGplot object
}
\description{
Useful for a first overall impression of the data.
}
\details{
The input is a data.frame with columns
  'fc.raw.file' - name of the Raw file
  'ms.ms.identified....' - fraction of identified MS/MS spectra in percent
  'cat' - identification category as arbitrary string
where each row represents one Raw file.
}
\examples{
 id_rate_bad = 20; id_rate_great = 35;
 label_ID = c("bad (<20\%)" = "red", "ok (...)" = "blue", "great (>35\%)" = "green")
 data = data.frame(fc.raw.file = paste('file', letters[1:3]),
                   ms.ms.identified.... = rnorm(3, 25, 15))
 data$cat = factor(cut(data$ms.ms.identified....,
                       breaks=c(-1, id_rate_bad, id_rate_great, 100),
                       labels=names(label_ID)))
 plot_IDRate(data, id_rate_bad, id_rate_great, label_ID)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{grepv}
\alias{grepv}
\title{Grep with values returned instead of indices.}
\usage{
grepv(reg, data, ...)
}
\arguments{
\item{reg}{regex param}

\item{data}{container}

\item{...}{other params forwarded to grep()}
}
\value{
values of data which matched the regex
}
\description{
The parameter 'value' should not be passed to this function since it is 
passed internally already.
}
\examples{
  grepv("x", c("abc", "xyz"))
  ## --> "xyz"
  
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_qualities.R
\name{qualCentered}
\alias{qualCentered}
\title{Quality metric for 'centeredness' of a distribution around zero.}
\usage{
qualCentered(x)
}
\arguments{
\item{x}{Numeric values (e.g. ppm errors)}
}
\value{
Value between [0, 1]
}
\description{
Ranges between 0 (worst score) and 1 (best score).
A median of zero gives the best score of 1.
The closer the median is to the most extreme value of the distribution, the smaller the score (until reaching 0).
Can be used for calibrated mass errors, as a measure of how well they are centered around 0.
E.g. if the median is 0.1, while the range is [-0.5,0.5], the score will be 0.8 (punishing the 20% deviation).
If the range of data is asymmetric, e.g. [-1.5,-0.5] and does not include zero, the score cannot reach 1,
since the median can never be zero.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_miscGGplot.R
\name{pointsPutX}
\alias{pointsPutX}
\title{Distribute a set of points with fixed y-values on a stretch of the x-axis.}
\usage{
pointsPutX(x_range, x_section, y, col = NA)
}
\arguments{
\item{x_range}{[min,max] valid range of x-values}

\item{x_section}{[min,max] fraction in which to distribute the values (in [0,1] for min,max, e.g. c(0.03,0.08) for 3-8\%)}

\item{y}{Y-values}

\item{col}{Colour of the points (used as argument to aes(colour=))}
}
\value{
ggplot object with new geom_point
}
\description{
#'
Usage:
  ggplot(...) + geom_X(...) + pointsPutX(...)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PTXQC-package.R
\name{PTXQC}
\alias{PTXQC}
\title{Quality control for proteomics data}
\description{
Quality control for proteomics data
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_MS2Decal}
\alias{plot_MS2Decal}
\title{Plot bargraph of oversampled 3D-peaks.}
\usage{
plot_MS2Decal(data)
}
\arguments{
\item{data}{A data.frame with columns 'file', 'msErr', 'type'}
}
\value{
GGplot object
}
\description{
Per Raw file, at most three n's must be given, i.e.
the fraction of 3D-peaks for n=1, n=2 and n=3(or more).
The fractions must sum to 1 (=100%).
}
\examples{
  n = c(100, 130, 50)
  data = data.frame(file = repEach(paste(letters[1:3],"\nLTQ [Da]"), n),
                    msErr = c(rnorm(n[1], 0.5), rnorm(n[2], 0.0), rnorm(n[3], -0.5)),
                    type = c("forward", "decoy")[1+(runif(sum(n))>0.95)])
  plot_MS2Decal(data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{delLCS}
\alias{delLCS}
\title{Removes the longest common suffix (LCS) from a vector of strings.}
\usage{
delLCS(x)
}
\arguments{
\item{x}{Vector of strings with common suffix}
}
\value{
Shortened vector of strings
}
\description{
Removes the longest common suffix (LCS) from a vector of strings.
}
\examples{
delLCS(c("TK12345_H1"))                     ## ""
delLCS(c("TK12345_H1", "TK12345_H2"))       ## "TK12345_H1" "TK12345_H2" 
delLCS(c("TK12345_H1", "TK12!45_H1"))       ## "TK123"    "TK12!" 
 
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_ScanIDRate}
\alias{plot_ScanIDRate}
\title{Plot line graph of TopN over Retention time.}
\usage{
plot_ScanIDRate(data)
}
\arguments{
\item{data}{A data.frame with columns 'fc.raw.file', 'scan.event.number', 'ratio', 'count'}
}
\value{
GGplot object
}
\description{
Number of Raw files must be 6 at most. Function will stop otherwise.
}
\examples{
  data = data.frame(fc.raw.file = factor(rep(c("d","a","x"), each=10), levels = c("d","a","x")),
                    scan.event.number = 1:10,
                    ratio = seq(40, 20, length.out=10),
                    count = seq(400, 200, length.out=10))
  plot_ScanIDRate(data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{thinOutBatch}
\alias{thinOutBatch}
\title{Apply 'thinOut' on all subsets of a data.frame, split by a batch column}
\usage{
thinOutBatch(data, filterColname, batchColname, binCount = 1000)
}
\arguments{
\item{data}{The data.frame to be split and filtered(thinned)}

\item{filterColname}{Name of the filter column as string}

\item{batchColname}{Name of the split column as string}

\item{binCount}{Number of bins in the 'filterColname' dimension.}
}
\value{
Data.frame with reduced rows, but identical input columns
}
\description{
The binsize is computed from the global data range of the filter column by dividing the range
into binCount bins.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_MQ.R
\name{getFragmentErrors}
\alias{getFragmentErrors}
\title{Extract fragment mass deviation errors from a data.frame from msms.txt}
\usage{
getFragmentErrors(x, recurse = 0)
}
\arguments{
\item{x}{Data frame in long format with numerical expression data}

\item{recurse}{Internal usage only. Leave at 0 when calling.}
}
\value{
Data frame with mass errors ('msErr') and their 'unit' (Da or ppm) or NULL (if no fragments were given)
}
\description{
Given a data.frame as obtainable from a msms.txt with
 - a 'mass.analyzer' column which contains only a single value for the whole column
 - a 'mass.deviations..da.' and (if available) 'mass.deviations..ppm.'
 - a 'masses' column (only required if 'mass.deviations..ppm.' is unavailable and the mass.analyzer
   indicates hig-res data)
}
\details{
Mass deviations are extracted from the columns, e.g. each cell containing values separated by
semicolons is split into single values. The appropriate unit is chosen (Da or ppm, depending on
ITMS or FTMS data). Also the fragmentation type can be used: CID indicates ITMS, HCD to FTMS.
This is not 100% safe, but older MQ versions do not report the mass analyzer properly.

Sometimes, peptides are identified purely based on MS1, i.e. have no fragments. These will be ignored.

If ppm mass deviations are not available, errors in Da will be converted to ppm using the corresponding mass values.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createYaml.R
\name{createYaml}
\alias{createYaml}
\title{Creates a yaml file storing the parameters that are used for creating the PTXQC report 
and returns these parameters as well as a list of available qc-Metrics objects.}
\usage{
createYaml(
  yc,
  param = list(),
  DEBUG_PTXQC = FALSE,
  txt_files = NULL,
  metrics = NULL
)
}
\arguments{
\item{yc}{A yaml class object created by YAMLClass$new()}

\item{param}{list of parameters sorted by names; if empty, will be populated with defaults}

\item{DEBUG_PTXQC}{print some debugging information; default FALSE}

\item{txt_files}{list of paths to MaxQuant files; if NULL, it is assumed that the parameters are for mzTab-mode}

\item{metrics}{list of metric names that should be plotted; if NULL, will be populated with defaults}
}
\value{
list of parameters used for creating the report and list of qc-Metrics objects
}
\description{
Valid parameters are: 
   param_useMQPAR, add_fs_col, id_rate_bad, id_rate_great , pg_ratioLabIncThresh , param_PG_intThresh,
   param_EV_protThresh , param_EV_intThresh, param_EV_pepThresh , yaml_contaminants, param_EV_MatchingTolerance,
   param_evd_mbr , param_EV_PrecursorTolPPM, param_EV_PrecursorOutOfCalSD , param_EV_PrecursorTolPPMmainSearch, 
   param_MSMSScans_ionInjThresh, param_OutputFormats and param_PageNumbers 
   
   Please provide them as a list() of this format: list$parameter_name
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_MQalign.R
\name{idTransferCheck}
\alias{idTransferCheck}
\title{Check how close transferred ID's after alignment are to their genuine IDs within one Raw file.}
\usage{
idTransferCheck(df_evd_all)
}
\arguments{
\item{df_evd_all}{A data.frame with columns 'type', 'calibrated.retention.time', 'modified.sequence', 'charge', 'raw.file'}
}
\value{
A data.frame containing the RT diff for each ID-group found in a Raw file (bg = genuine).
}
\description{
The input is a data.frame containing feature evidence with corrected retention times,
e.g. a 'calibrated.retention.time' column.
}
\details{
Note that this function must be given MS/MS identifications of type "MULTI-MSMS" and "MSMS-MATCH".
It will stop() otherwise.
 
We compare for each peptide sequence (and charge) the RT difference within groups of either genuine as well as mixed pairs.
For every comparison made, we report the RT span If alignment worked perfectly, the span are very small (<1 min),
for the mixed group, i.e. the pairs are accidentally split 3D peaks. Alignment performance has no influence on the
genuine-only groups.

Note: We found early MaxQuant versions (e.g. 1.2.2.5) to have an empty 'modified.sequence' column for 'MULTI-MATCH' entries.
The sequence which SHOULD be present is equal to the immediate upper row. This is what we use to guess the sequence.
However, this relies on the data.frame not being subsetted before (we can sort using the 'id' column)!
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{longestCommonSuffix}
\alias{longestCommonSuffix}
\title{Like longestCommonPrefix(), but on the suffix.}
\usage{
longestCommonSuffix(strings)
}
\arguments{
\item{strings}{Vector of strings}
}
\value{
Single string - might be empty ("")
}
\description{
Like longestCommonPrefix(), but on the suffix.
}
\examples{

 longestCommonSuffix(c("123.ABC", "45677.ABC", "BC"))  ## "BC"
 longestCommonSuffix(c("123.ABC", "", "BC"))           ## ""
 longestCommonSuffix(c("123.ABC", "45677.ABC"))        ## ".ABC"
 longestCommonSuffix(c("nothing", "in", "common"))     ## ""
 
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mzQC.R
\name{QCMetaFilenames}
\alias{QCMetaFilenames}
\title{Define a Singleton class which holds the full raw filenames (+path) and their PSI-MS CV terms for usage in the mzQC metadata}
\description{
The internal data is filled using, e.g. 'getMetaFilenames()'
}
\section{Super class}{
\code{\link[R6P:Singleton]{R6P::Singleton}} -> \code{QCMetaFilenames}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{data}}{Stores the data of the singleton. Set the data once before using the singleton all over the place}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-QCMetaFilenames-clone}{\code{QCMetaFilenames$clone()}}
}
}
\if{html}{\out{
<details open><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="R6P" data-topic="Singleton" data-id="initialize"><a href='../../R6P/html/Singleton.html#method-Singleton-initialize'><code>R6P::Singleton$initialize()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-QCMetaFilenames-clone"></a>}}
\if{latex}{\out{\hypertarget{method-QCMetaFilenames-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{QCMetaFilenames$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modsToTable.R
\name{modsToTable}
\alias{modsToTable}
\title{Convert list of (mixed)modifications to a frequency table}
\usage{
modsToTable(mod_list)
}
\arguments{
\item{mod_list}{A vector with modifications, each for a specific peptide. Multiple mods per entry are allowed, each separated by comma.}
}
\value{
A data.frame with 'modification_names' and 'Freq' (0-100)
}
\description{
Convert list of (mixed)modifications to a frequency table
}
\examples{
modsToTable(c("Ox (M)", 
              "Unmodified",
              "Ox (M),Acetyl (Prot N-term)",
              "2 Ox (M)",
              "Unmodified",
              "Unmodified"))


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_PCA.R
\name{getPCA}
\alias{getPCA}
\title{Create a principal component analysis (PCA) plot for the first two dimensions.}
\usage{
getPCA(data, do_plot = TRUE, connect_line_order = NA, gg_layer)
}
\arguments{
\item{data}{Matrix(!) where each row is one high-dimensional point, with ncol dimensions, 
e.g. a mouse as an array of proteinexpressions
rownames(data) give classes for colouring (can be duplicates in matrices, as opposed to data.frames)}

\item{do_plot}{Show PCA plot? if ==2, then shows correlations plot as well}

\item{connect_line_order}{Connect points by lines, the order is given by this vector.
Default: NA (no lines)}

\item{gg_layer}{More parameters added to a ggplot object (ggplot(x) + gg_layer)}
}
\value{
[invisible] Named list with "PCA": The PCA object as returned by \code{\link[stats]{prcomp}}, access $x for PC values
                                and "plots": list of plot objects (one or two)
}
\description{
Create a principal component analysis (PCA) plot for the first two dimensions.
}
\examples{
n = 5
m = 10
data = matrix(runif(n * m), nrow = n, ncol = m)
rownames(data) = 1:n
getPCA(data, connect_line_order = 1:n, gg_layer = ggplot2::ggtitle("test"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_TIC}
\alias{plot_TIC}
\title{Plot Total Ion Count over time}
\usage{
plot_TIC(data, x_lim, y_lim)
}
\arguments{
\item{data}{A data.frame with columns 'fc.raw.file', 'RT', 'intensity'}

\item{x_lim}{Plot range of x-axis}

\item{y_lim}{Plot range of y-axis}
}
\value{
GGplot object
}
\description{
The input is a data.frame with already averaged counts over binned RT-slices.
}
\examples{

 data = data.frame(fc.raw.file = rep(c("file A", "file B", "file C"), each=81),
                   RT = c(20:100),
                   intensity = c(rnorm(81, mean=20), rnorm(81, mean=10), rnorm(81, mean=30)))
 plot_TIC(data, c(10, 100), c(0, 40))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{supCount}
\alias{supCount}
\title{Compute shortest prefix length which makes all strings in a vector uniquely identifyable.}
\usage{
supCount(x, prefix_l = 1)
}
\arguments{
\item{x}{Vector of strings}

\item{prefix_l}{Starting prefix length, which is incremented in steps of 1 until all prefixes are unique (or maximum string length is reached)}
}
\value{
Integer with minimal prefix length required
}
\description{
If there is no unique prefix (e.g. if a string is contained twice), then the length
of the longest string is returned, i.e. if the return value is used in a call to substr, nothing happens
e.g.  substr(x, 1, supCount(x)) == x
}
\examples{
  supCount(c("abcde...", "abcd...", "abc..."))  ## 5

  x = c("doubled", "doubled", "aLongDummyString")
  all( substr(x, 1, supCount(x)) == x )   
  ## TRUE (no unique prefix due to duplicated entries)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{pastet}
\alias{pastet}
\title{paste with tab as separator}
\usage{
pastet(...)
}
\arguments{
\item{...}{Arguments forwarded to paste()}
}
\value{
return value of paste()
}
\description{
paste with tab as separator
}
\examples{
  pastet("tab","separated")
  ## --> "tab\tseparated"
  
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_TopN}
\alias{plot_TopN}
\title{Plot line graph of TopN over Retention time.}
\usage{
plot_TopN(data)
}
\arguments{
\item{data}{A data.frame with columns 'fc.raw.file', 'scan.event.number', 'n'}
}
\value{
GGplot object
}
\description{
Number of Raw files must be 6 at most. Function will stop otherwise.
}
\examples{
  data = data.frame(fc.raw.file = rep(c("d","a","x"), each=10),
                    scan.event.number = 1:10,
                    n = 11:20)
  plot_TopN(data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mzQC.R
\name{assembleMZQC}
\alias{assembleMZQC}
\title{Collects all 'mzQC' members from each entry in lst_qcMetrics and stores them in an overall mzQC object, which can be written to disk (see writeMZQC()) or augmented otherwise}
\usage{
assembleMZQC(lst_qcMetrics, raw_file_mapping)
}
\arguments{
\item{lst_qcMetrics}{A list of qcMetric objects which have their mzQC member populated with "MzQCrunQuality" and/or "MzQCsetQuality" objects}

\item{raw_file_mapping}{A data.frame with cols 'from', to' and maybe 'best.effort' (if shorting was unsuccessful), as e.g. obtained by a FilenameMapper$raw_file_mapping}
}
\value{
An MzQCmzQC object (root object of an mzQC document)
}
\description{
Collects all 'mzQC' members from each entry in lst_qcMetrics and stores them in an overall mzQC object, which can be written to disk (see writeMZQC()) or augmented otherwise
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_qualities.R
\name{qualHighest}
\alias{qualHighest}
\title{Score an empirical density distribution of values, where the best possible distribution is right-skewed.}
\usage{
qualHighest(x, N)
}
\arguments{
\item{x}{Vector of numeric values (e.g. height of histogram bins)}

\item{N}{Length of x (just a precaution currently)}
}
\value{
Quality score in the range of [0,1]
}
\description{
The score is computed according to
}
\details{
q = ((N-1) - sum_i( ((N-i-1)*x_i) ) / (N-1)

Scores range from 0 (worst), to 1 (best).
E.g. c(0,0,0,16) would yield a score of 1.
c(16,0,0,0,0) gives a score of 0.
}
\examples{
 qualHighest(c(0,0,0,16), 4)   ## 1
 qualHighest(c(16,0,0,0), 4)   ## 0
 qualHighest(c(1,1,1,1), 4)    ## 0.5
 qualHighest(c(0,16,0,0), 4)   ## 1/3

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{longestCommonPrefix}
\alias{longestCommonPrefix}
\title{Get the longest common prefix from a set of strings.}
\usage{
longestCommonPrefix(strings)
}
\arguments{
\item{strings}{Vector of strings}
}
\value{
Single string - might be empty ("")
}
\description{
Input is converted to character (e.g. from factor) first.
}
\examples{
  longestCommonPrefix(c("CBA.321", "CBA.77654", ""))    ## ""
  longestCommonPrefix(c("CBA.321", "CBA.77654", "CB"))  ## "CB"
  longestCommonPrefix(c("ABC.123", "ABC.456"))          ## "ABC."
  longestCommonPrefix(c("nothing", "in", "common"))     ## ""


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{darken}
\alias{darken}
\title{Make a color (given as name or in RGB) darker by factor x = [0 = black, 1=unchanged]}
\usage{
darken(color, factor = 0.8)
}
\arguments{
\item{color}{A color as understood by col2rgb}

\item{factor}{Between 0 (make black) and 1 (leave color as is)}
}
\value{
darkened color
}
\description{
Make a color (given as name or in RGB) darker by factor x = [0 = black, 1=unchanged]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mzQC.R
\name{getMetaFilenames}
\alias{getMetaFilenames}
\title{Parses the given mqpar.xml file (or, if not found, tries the 'txt_folder' + '/../../' folder (i.e. where the raw data should be)) to extract the full filepaths for all Raw files}
\usage{
getMetaFilenames(mqpar_file, txt_folder)
}
\arguments{
\item{mqpar_file}{Location of the mqpar.xml (can be empty, if unknown)}

\item{txt_folder}{Fallback option: path to the txt folder (which contains evidence.txt, etc)}
}
\value{
May return 'NULL' if no mqpar.xml could be found.
        Otherwise: data.frame with columns:
        \itemize{
          \item 'file' (no path), 'path' (full path incl. names)
          \item 'file_no_suffix' (as 'file' but without suffix) 
          \item 'CV' (CV term for filetype, e.g. for Thermo Raw)
        }
}
\description{
Parses the given mqpar.xml file (or, if not found, tries the 'txt_folder' + '/../../' folder (i.e. where the raw data should be)) to extract the full filepaths for all Raw files
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MQDataReader.R
\name{read.MQ}
\alias{read.MQ}
\title{Convenience wrapper for MQDataReader when only a single MQ file should be read
and file mapping need not be stored.}
\usage{
read.MQ(
  file,
  filter = "",
  type = "pg",
  col_subset = NA,
  add_fs_col = 10,
  LFQ_action = FALSE,
  ...
)
}
\arguments{
\item{file}{see \code{MQDataReader::readMQ()}}

\item{filter}{see \code{MQDataReader::readMQ()}}

\item{type}{see \code{MQDataReader::readMQ()}}

\item{col_subset}{see \code{MQDataReader::readMQ()}}

\item{add_fs_col}{see \code{MQDataReader::readMQ()}}

\item{LFQ_action}{see \code{MQDataReader::readMQ()}}

\item{...}{see \code{MQDataReader::readMQ()}}
}
\value{
see \code{MQDataReader::readMQ()}
}
\description{
For params, see \code{MQDataReader::readMQ()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_ContsPG}
\alias{plot_ContsPG}
\title{Plot contaminants from proteinGroups.txt}
\usage{
plot_ContsPG(data)
}
\arguments{
\item{data}{A data.frame with columns 'group', 'cont_pc', 'logAbdClass'}
}
\value{
GGplot object
}
\description{
Plot contaminants from proteinGroups.txt
}
\examples{

 data = data.frame( 'group' = letters[1:10], 'cont_pc' = 2:11, 'logAbdClass' = c("low","high"))
 plot_ContsPG(data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_miscGGplot.R
\name{ggAxisLabels}
\alias{ggAxisLabels}
\title{Function to thin out the number of labels shown on an axis in GGplot}
\usage{
ggAxisLabels(x, n = 20)
}
\arguments{
\item{x}{Vector of labels (passed by GGplot)}

\item{n}{Number of labels to show}
}
\value{
Shortened version of 'x'
}
\description{
By default, 20 labels (or up to 40 see below) are shown.
If the number of items is less than twice the number of desired labels,
all labels will be shown (to avoid irregular holes for some labels).
I.e. if n=20, and x has 22 entries, there would be only two labels removed, giving a very irregular picture.
It only becomes somewhat regular if after any label there is at least one blank, i.e. at most
half the entries are labeled.
#'
Example:
  ## p is any ggplot object
  p + scale_y_discrete(breaks = ggAxisLabels)
  ## customize 'n'
  my.ggAxisLabels = function(x) ggAxisLabels(x, n = 4)
  p + scale_y_discrete(breaks = my.ggAxisLabels)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{getHTMLTable}
\alias{getHTMLTable}
\title{Create an HTML table with an extra header row}
\usage{
getHTMLTable(data, caption = NA)
}
\arguments{
\item{data}{A data.frame which serves as table}

\item{caption}{A set of headlines, e.g. c("top line", "bottom line")}
}
\value{
table as html character string for cat()'ing into an html document
}
\description{
Create an HTML table with an extra header row
}
\examples{
  data = data.frame(raw.file = letters[1:4],
                    id.rate = 3:6)
  getHTMLTable(data,
               caption = "some header line")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createReport.R
\name{createReport}
\alias{createReport}
\title{Create a quality control report (in PDF format).}
\usage{
createReport(
  txt_folder = NULL,
  mztab_file = NULL,
  yaml_obj = list(),
  report_filenames = NULL,
  enable_log = FALSE
)
}
\arguments{
\item{txt_folder}{Path to txt output folder of MaxQuant (e.g. "c:/data/Hek293/txt")}

\item{mztab_file}{Alternative to **txt_folder**, you can provide a single mzTab file which contains PSM, PEP and PRT tables}

\item{yaml_obj}{A nested list object with configuration parameters for the report.
Useful to switch off certain plots or skip entire sections.}

\item{report_filenames}{Optional list with names (as generated by \code{\link{getReportFilenames}}).
If not provided, will be created internally by calling \code{\link{getReportFilenames}}.}

\item{enable_log}{If TRUE all console output (including warnings and errors) is logged to the file given in **report_filenames$log_file**.
Note: warnings/errors can only be shown in either the log **or** the console, not both!}
}
\value{
List with named filename strings, e.g. $yaml_file, $report_file etc..
}
\description{
This is the main function of the package and the only thing you need to call directly if you are
just interested in getting a QC report.
}
\details{
You need to provide either
a) the folder name of the 'txt' output, as generated by MaxQuant or an mzTab file
or b) an mzTab file as generated by the OpenMS QualityControl TOPP tool (other mzTab files will probably not work)

Optionally, provide a YAML configuration object, which allows to (de)activate certain plots and holds other parameters.
The yaml_obj is complex and best obtained by running this function once using the default (empty list).
A full YAML configuration object will be written in the 'txt' folder you provide and can be loaded using
\code{\link[yaml]{yaml.load}}.

The PDF and the config file will be stored in the given txt folder.
}
\note{
You need write access to the txt/mzTab folder!

For updates, bug fixes and feedback please visit \url{https://github.com/cbielow/PTXQC}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qcMetric.R
\docType{class}
\name{qcMetric-class}
\alias{qcMetric-class}
\alias{qcMetric}
\title{Class which can compute plots and generate mzQC output (usually for a single metric).}
\description{
Reference class which is instanciated with a metric description and a
worker function (at initialization time, i.e. in the package)
and can produce plots and mzQC values (at runtime, when data is provided) using setData().
}
\details{
All derived classes need to implement a 'workerFcn()' function, which returns a list with
elements: c("plots", "mzQC", "htmlTable", "qcScores", "title"),
where 'plots' is required; all others are optional.
}
\section{Fields}{

\describe{
\item{\code{helpText}}{Description (lengthy) of the metric and plot elements}

\item{\code{workerFcn}}{Function which generates a result (usually plots). Data is provided using setData().}

\item{\code{plots}}{List of plots (after setData() was called)}

\item{\code{htmlTable}}{A table for display in the HTML report (preferred over a plot in Html mode)}

\item{\code{qcScores}}{Data.frame of scores from a qcMetric (computed within workerFcn())}

\item{\code{mzQC}}{An named list of mzQC MzQCqualityMetric's (named by their fc.raw.file for runQuality or concatenated fc.raw.files for setQualities (e.g. "file 1;file4")) (valid after setData() was called)}

\item{\code{qcCat}}{QC category (LC, MS, or prep)}

\item{\code{qcName}}{Name of the qcScore in the heatmap}

\item{\code{orderNr}}{Column index during heatmap generation and for the general order of plots}
}}

\section{Methods}{

\describe{
\item{\code{setData(df, ...)}}{Internally calls the workerFcn(), which computes the actual plots, metric scores and supporting data
          (e.g. mzQC metrics) of the derived class; the resulting data is checked and stored in the members of this class.

          Arguments:
          \describe{
            \item{\code{df}}{The expected data, usually a data frame. If empty, this function will return immediately without failure.}
            \item{\code{...}}{Additional arguments passed to the workerFcn().}
          }

          Returns NULL.
         }
}}

\examples{

require(ggplot2)
dd = data.frame(x=1:10, y=11:20)
a = qcMetric$new(helpText="small help text", 
                 ## arbitrary arguments, matched during setData()
                 workerFcn=function(.self, data, gtitle)
                 {
                   ## usually some code here to produce ggplots
                   pl = lapply(1:2, function(xx) {
                       ggplot(data) +
                         geom_point(aes(x=x*xx,y=y)) +
                         ggtitle(gtitle)
                     })
                     ## add mzQC metric for count of identified clusters
                     mzqc = lapply(1:3, function(id){
                       out = rmzqc::getQualityMetricTemplate("MS:1002406")
                       out$value = id;
                       return(out) })
                     names(mzqc) = paste0("file", 1:3);
                   return(list(plots = pl, mzQC = mzqc))
                 }, 
                 qcCat="LC", 
                 qcName="MS/MS Peak shape", 
                 orderNr = 30)
## test some output
a$setData(dd, "my title")
a$plots  ## the raw plots
a$getPlots(TRUE) ## same as above
a$getPlots(FALSE) ## plots without title
a$getTitles()  ## get the titles of the all plots
a$helpText
a$qcName
a$mzQC


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_QCHeat.R
\name{getQCHeatMap}
\alias{getQCHeatMap}
\title{Generate a Heatmap from a list of QC measurements.}
\usage{
getQCHeatMap(lst_qcMetrics, raw_file_mapping)
}
\arguments{
\item{lst_qcMetrics}{List of QCMetric objects}

\item{raw_file_mapping}{Data.frame with 'from' and 'to' columns for name mapping to unify names from list entries}
}
\value{
A ggplot object for printing
}
\description{
Each list entry is a data.frame with two columns. 
The first one contains the Raw file name (or the short version).
and should be named 'raw.file' (or 'fc.raw.file'). 
The second column's name must be an expression (see ?plotmath)
and contains quality values in the range [0,1]. If values are outside this range, 
a warning is issued and values are cut to the nearest allowed value (e.g. '1.2' becomes '1').
List entries are merged and columns are ordered by name.
  
All substrings enclosed by 'X[0-9]*X.' will be removed (can be used for sorting columns).
The resulting string is evaluated as an expression. 
E.g. parse(text = <colname>)
}
\details{
To judge the overall quality of each raw file a summary column is added, 
values being the mean of all other columns per row.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_DataOverRT}
\alias{plot_DataOverRT}
\title{Plot some count data over time for each Raw file.}
\usage{
plot_DataOverRT(
  data,
  title,
  y_lab,
  x_lim = range(data$RT),
  y_max = max(data$counts)
)
}
\arguments{
\item{data}{A data.frame with columns as described above}

\item{title}{The plot title}

\item{y_lab}{Label of y-axis}

\item{x_lim}{Limits of the x-axis (2-tuple)}

\item{y_max}{Maximum of the y-axis (single value)}
}
\value{
GGplot object
}
\description{
The input is a data.frame with columns
  'RT' - RT in seconds, representing one bin
  'counts' - number of counts at this bin
  'fc.raw.file' - name of the Raw file
where each row represents one bin in RT.
}
\details{
At most nine(!) Raw files can be plotted. If more are given,
an error is thrown.
}
\examples{
 data = data.frame(fc.raw.file = rep(paste('file', letters[1:3]), each=30),
                            RT = seq(20, 120, length.out = 30),
                        counts = c(rnorm(30, 400, 20), rnorm(30, 250, 15), rnorm(30, 50, 15)))
 plot_DataOverRT(data, "some title", "count data")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{getMaxima}
\alias{getMaxima}
\title{Find the local maxima in a vector of numbers.}
\usage{
getMaxima(x, thresh_rel = 0.2)
}
\arguments{
\item{x}{Vector of numbers}

\item{thresh_rel}{Minimum relative intensity to maximum intensity of 'x' required
to be a maximum (i.e., a noise threshold). Default is 20\%.}
}
\value{
Vector of bool's, where TRUE indicates a local maximum.
}
\description{
A vector of booleans is returned with the same length as input
which contains TRUE when there is a maximum.
Simply sum up the vector to get the number of maxima.
}
\examples{
    r = getMaxima(c(1,0,3,4,5,0))                                
    all(r == c(TRUE,FALSE,FALSE,FALSE,TRUE,FALSE))
    
    getMaxima(c(1, NA, 3, 2, 3, NA, 4, 2, 5))
    
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{renameFile}
\alias{renameFile}
\title{Given a vector of (short/long) filenames, translate to the (long/short) version}
\usage{
renameFile(f_names, mapping)
}
\arguments{
\item{f_names}{Vector of filenames}

\item{mapping}{A data.frame with from,to columns}
}
\value{
A vector of translated file names as factor (ordered by mapping!)
}
\description{
Given a vector of (short/long) filenames, translate to the (long/short) version
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{CV}
\alias{CV}
\title{Coefficient of variation (CV)}
\usage{
CV(x)
}
\arguments{
\item{x}{Vector of numeric values}
}
\value{
CV
}
\description{
Computes sd(x) / mean(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FilenameMapper.R
\docType{class}
\name{FilenameMapper-class}
\alias{FilenameMapper-class}
\alias{FilenameMapper}
\title{Make sure to call $readMappingFile(some_file) if you want to support a user-defined file mapping.
 Otherwise, calls to $getShortNames() will create/augment the mapping for filenames.}
\description{
Make sure to call $readMappingFile(some_file) if you want to support a user-defined file mapping.
 Otherwise, calls to $getShortNames() will create/augment the mapping for filenames.
}
\section{Fields}{

\describe{
\item{\code{raw_file_mapping}}{Data.frame with columns 'from', 'to' and maybe 'best.effort' (if shorting was unsuccessful)}

\item{\code{mapping.creation}}{how the current mapping was obtained (user or auto)}

\item{\code{external.mapping.file}}{Filename of user-defined mapping file; only defined if readMappingFile() was called}
}}

\section{Methods}{

\describe{
\item{\code{getShortNamesStatic(raw.files, max_len, fallbackStartNr = 1)}}{Static method: Shorten a set of Raw file names and return a data frame with the mappings.
 Mapping will have: $from, $to and optionally $best.effort (if shorting was unsuccessful and numbers had to be used)
 \itemize{
   \item{\verb{raw.files}  Vector of Raw files.}
   \item{\verb{max_len} Maximal length of shortening results, before resorting to canonical names (file 1,...).}
   \item{\verb{fallbackStartNr} Starting index for canonical names.}
 }
 \subsection{Return Value}{ data.frame with mapping.}}
}}

\examples{
a = FilenameMapper$new()
a$readMappingFile('filenamemapping.txt')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_ContUserScore}
\alias{plot_ContUserScore}
\title{Plot Andromeda score distribution of contaminant peptide vs. matrix peptides.}
\usage{
plot_ContUserScore(data, raw.file, score)
}
\arguments{
\item{data}{A data.frame with columns 'x', 'y', 'condition'}

\item{raw.file}{Name of Raw file for which the data is displayed (will become part of the plot title)}

\item{score}{Score of how distinct the distributions are (will become part of the title)}
}
\value{
GGplot object
}
\description{
The data is expected to be an ECDF already, x being the Andromeda score, y being the culmulative probability.
The Score is the probability of a Kolm.-Smirnoff test that the contaminant scores are larger (i.e.
large p-values indicate true contamination).
You will only see this plot if the %-threshold (YAML config) was reached. This is a saveguard against false-positive,
but high-scoring contaminant peptides, which would erroneously give you a large p-value and make you believe
your sample is contaminated although that's not the case.
}
\examples{

 data = data.frame(x = 10:60,
                   y = c(seq(0,1,length=51), seq(0.1, 1, length=51)),
                   condition = rep(c("sample","contaminant"), each=51))
 plot_ContUserScore(data, 'test file', 0.96)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_CalibratedMSErr}
\alias{plot_CalibratedMSErr}
\title{Plot bargraph of uncalibrated mass errors for each Raw file.}
\usage{
plot_CalibratedMSErr(
  data,
  MQBug_raw_files,
  stats,
  y_lim,
  extra_limit = NA,
  title_sub = ""
)
}
\arguments{
\item{data}{A data.frame with columns 'fc.raw.file', 'mass.error..ppm.'}

\item{MQBug_raw_files}{List of Raw files with invalid calibration values}

\item{stats}{A data.frame with columns 'fc.raw.file', 'outOfCal'}

\item{y_lim}{Range of y-axis}

\item{extra_limit}{Position where a v-line is plotted (for visual guidance)}

\item{title_sub}{Subtitle}
}
\value{
GGplot object
}
\description{
Boxes are optionally colored to indicate that a MQ bug was detected or
if PTXQC detected a too narrow search window.
}
\examples{
  n = c(150, 1000, 1000, 1000)
  data = data.frame(fc.raw.file = repEach(letters[4:1], n),
                    mass.error..ppm. = c(rnorm(n[1], 1, 2.4),
                                         rnorm(n[2], 0.5, 0.5),
                                         rnorm(n[3], 0.1, 0.7),
                                         rnorm(n[4], 0.3, 0.8)))
  stats = data.frame(fc.raw.file = letters[4:1],
                     sd = c(2.4, 0.5, 0.7, 0.8),
                     outOfCal = c(TRUE, FALSE, FALSE, FALSE))
  plot_CalibratedMSErr(data, MQBug_raw_files = letters[1], stats, y_lim = c(-20,20), 15, "subtitle")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_qualities.R
\name{qualGaussDev}
\alias{qualGaussDev}
\title{Compute probability of Gaussian (mu=m, sd=s) at a position 0, with reference
to the max obtainable probability of that Gaussian at its center.}
\usage{
qualGaussDev(mu, sd)
}
\arguments{
\item{mu}{Center of Gaussian}

\item{sd}{SD of Gaussian}
}
\value{
quality, ranging from 0 (bad agreement) to 1 (perfect, i.e. centered at 0)
}
\description{
Measure for centeredness around 0.
Highest score is 1, worst score is 0.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{lcsCount}
\alias{lcsCount}
\title{Count the number of chars of the longest common suffix}
\usage{
lcsCount(x)
}
\arguments{
\item{x}{Vector of strings with common suffix}
}
\value{
Length of LCS
}
\description{
Count the number of chars of the longest common suffix
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_MQalign.R
\name{peakSegmentation}
\alias{peakSegmentation}
\title{Determine fraction of evidence which causes segmentation, i.e. sibling peaks at different RTs
confirmed either by genuine or transferred MS/MS.}
\usage{
peakSegmentation(df_evd_all)
}
\arguments{
\item{df_evd_all}{A data.frame of evidences containing the above columns}
}
\value{
A data.frame with one row per Raw file and 
        three columns: 
          1) % of native single peaks (ignoring transferred IDs)
          2) % of single peaks (group of size=1) using only groups which have one transferred evidence
          3) % of single peaks using all groups
}
\description{
Sometimes, MQ splits a feature into 2 or more if the chromatograpic conditions are not optimal and there
is a drop in RT intensity.
If both features contain successful MS/MS scans, we will find the same peptide twice (with slightly different RT)
in the same charge state. This constitutes a natively split peak and is rare (95% of all genuine peaks are unique).
}
\details{
If Match-between-runs is used and the RT alignment is not perfect, then a peptide might be inferred at a wrong
RT position, even though this Raw file already contains MS/MS evidence of this peptide.
Usually the number of peak duplicates rises drastically (e.g. only 75% of peaks are unique after MBR was used).
In most cases, the RT is too far off to be a split peak. It's rather a lucky hit with accidentally the same mass-to-charge,
and thus the intensity is random.
To find by how much these peak pairs differ in RT, use idTransferCheck() and inMatchWindow().

Required columns are 'is.transferred', 'fc.raw.file', 'modified.sequence', 'charge', 'type'.

Note that this function must be given MS/MS identifications of type "MULTI-MSMS" and "MSMS-MATCH".
It will stop() otherwise.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_CountData}
\alias{plot_CountData}
\title{Plot Protein groups per Raw file}
\usage{
plot_CountData(data, y_max, thresh_line, title)
}
\arguments{
\item{data}{A data.frame with columns 'fc.raw.file', 'counts', 'category'}

\item{y_max}{Plot limit of y-axis}

\item{thresh_line}{Position of a threshold line, indicating the usual target value}

\item{title}{Main title, and optional subtitle (if vector of length 2 is provided)}
}
\value{
GGplot object
}
\description{
The input is a data.frame with protein/peptide counts, where 'category' designates
the origin of information (genuine ID, transferred ID, or both).
}
\examples{

 data = data.frame(fc.raw.file = rep(c("file A", "file B"), each=3),
                   counts = c(3674, 593, 1120, 2300, 400, 600),
                   category = c("genuine","genuine+transferred","transferred"))
 plot_CountData(data, 6000, 4000, c("EVD: Protein Groups count", "gain: 23\%"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_miscGGplot.R
\name{ggText}
\alias{ggText}
\title{Plot a text as graphic using ggplot2.}
\usage{
ggText(title, text, col = "black")
}
\arguments{
\item{title}{The title of the plot}

\item{text}{Centered text, can contain linebreaks}

\item{col}{Colour of text (excluding the title)}
}
\value{
ggplot object
}
\description{
Plot a text as graphic using ggplot2.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{getMetricsObjects}
\alias{getMetricsObjects}
\title{Get all currently available metrics}
\usage{
getMetricsObjects(DEBUG_PTXQC = FALSE)
}
\arguments{
\item{DEBUG_PTXQC}{Use qc objects from the package (FALSE) or from environment (TRUE/DEBUG)}
}
\value{
List of matric objects
}
\description{
Get all currently available metrics
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{byXflex}
\alias{byXflex}
\title{Same as \code{\link{byX}}, but with more flexible group size, to avoid that the last group has only a few entries (<50\% of desired size).}
\usage{
byXflex(data, indices, subset_size = 5, FUN, sort_indices = TRUE, ...)
}
\arguments{
\item{data}{Data.frame whose subset to use on FUN}

\item{indices}{Vector of group assignments, same length as nrow(data)}

\item{subset_size}{Ideal number of groups to use in one subset -- this can be changed internally, from 75\%-150\%}

\item{FUN}{function Applied to subsets of data}

\item{sort_indices}{Groups are formed by their sorted character(!) names}

\item{...}{More arguments to FUN}
}
\value{
list of function result (one entry for each subset)
}
\description{
The 'subset_size' param is internally optimized using \code{\link{correctSetSize}} and
then \code{\link{byX}} is called.
}
\examples{
 stopifnot(
   byXflex(data.frame(d=1:10), 1:10, 2, sum, sort_indices = FALSE) ==
   c(3, 7, 11, 15, 19)
 )
 
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MQDataReader.R
\docType{class}
\name{MQDataReader-class}
\alias{MQDataReader-class}
\alias{MQDataReader}
\title{S5-RefClass to read MaxQuant .txt files}
\description{
This class is used to read MQ data tables using \code{MQDataReader::readMQ()} while holding
the internal raw file --> short raw file name mapping (stored in a member called
'fn_map') and updating/using it every time \code{MQDataReader::readMQ()} is called.
}
\details{
Since MaxQuant changes capitalization and sometimes even column names, it seemed convenient
to have a function which just reads a txt file and returns unified column names, irrespective of the MQ version.
So, it unifies access to columns (e.g. by using lower case for ALL columns) and ensures columns are
identically named across MQ versions:
\preformatted{
 alternative term          new term
 -----------------------------------------
 protease                  enzyme
 protein.descriptions      fasta.headers
 potential.contaminant     contaminant
 mass.deviations           mass.deviations..da.
 basepeak.intensity        base.peak.intensity
}

We also correct 'reporter.intensity.*' naming issues to MQ 1.6 convention, when 'reporter.intensity.not.corrected' is present.
MQ 1.5 uses: reporter.intensity.X and reporter.intensity.not.corrected.X
MQ 1.6 uses: reporter.intensity.X and reporter.intensity.corrected.X

Note: you must find a regex which matches both versions, or explicitly add both terms if you are requesting only a subset
      of columns!

Fixes for msmsScans.txt:
 negative Scan Event Numbers in msmsScans.txt are reconstructed by using other columns

Automatically detects UTF8-BOM encoding and deals with it (since MQ2.4).

Example of usage:
\preformatted{
  mq = MQDataReader$new()
  d_evd = mq$readMQ("evidence.txt", type="ev", filter="R", col_subset=c("proteins", "Retention.Length", "retention.time.calibration"))
}

If the file is empty, this function shows a warning and returns NULL.
If the file is present but cannot be read, the program will stop.
}
\section{Methods}{

\describe{
\item{\code{getInvalidLines()}}{Detect broken lines (e.g. due to Excel import+export)

   When editing a MQ txt file in Microsoft Excel, saving the file can cause it to be corrupted,
   since Excel has a single cell content limit of 32k characters
   (see http://office.microsoft.com/en-001/excel-help/excel-specifications-and-limits-HP010342495.aspx)
   while MQ can easily reach 60k (e.g. in oxidation sites column).
   Thus, affected cells will trigger a line break, effectively splitting one line into two (or more).

   If the table has an 'id' column, we can simply check the numbers are consecutive. If no 'id' column is available,
   we detect line-breaks by counting the number of NA's per row and finding outliers.
   The line break then must be in this line (plus the preceeding or following one). Depending on where
   the break happened we can also detect both lines right away (if both have more NA's than expected).

   Currently, we have no good strategy to fix the problem since columns are not aligned any longer, which
   leads to columns not having the class (e.g. numeric) they should have.
   (thus one would need to un-do the linebreak and read the whole file again)

   [Solution to the problem: try LibreOffice 4.0.x or above -- seems not to have this limitation]

   @return Returns a vector of indices of broken (i.e. invalid) lines
  }

\item{\code{readMQ(
  file,
  filter = "",
  type = "pg",
  col_subset = NA,
  add_fs_col = 10,
  check_invalid_lines = TRUE,
  LFQ_action = FALSE,
  ...
)}}{Wrapper to read a MQ txt file (e.g. proteinGroups.txt).

   Arguments:
   \describe{
     \item{\code{file}}{(Relative) path to a MQ txt file.}
     \item{\code{filter}}{Searched for "C" and "R". If present, [c]ontaminants and [r]everse hits are removed if the respective columns are present.
       E.g. to filter both, \code{filter = "C+R"}}
     \item{\code{type}}{Allowed values are:
       \itemize{
         \item "pg" (proteinGroups) [default], adds abundance index columns (*AbInd*, replacing 'intensity')
         \item "sm" (summary), splits into three row subsets (raw.file, condition, total)
         \item "ev" (evidence), will fix empty modified.sequence cells for older MQ versions (when MBR is active)
         \item "msms_scans", will fix invalid (negative) scan event numbers
       }
       Any other value will not add/modify any columns.}
     \item{\code{col_subset}}{A vector of column names as read by read.delim(), e.g., spaces are replaced by dot already.
       If given, only columns with these names (ignoring lower/uppercase) will be returned (regex allowed).
       E.g. \code{col_subset = c("^lfq.intensity.", "protein.name")}}
     \item{\code{add_fs_col}}{If TRUE and a column 'raw.file' is present, an additional column 'fc.raw.file' will be added with
       common prefix AND common substrings removed (\code{\link{simplifyNames}}).
       E.g. two rawfiles named 'OrbiXL_2014_Hek293_Control', 'OrbiXL_2014_Hek293_Treated' will give 'Control', 'Treated'.
       If \code{add_fs_col} is a number AND the longest short-name is still longer, the names are discarded and replaced by
       a running ID of the form 'file <x>', where <x> is a number from 1 to N.
       If the function is called again and a mapping already exists, this mapping is used.
       Should some raw.files be unknown (ie the mapping from the previous file is incomplete), they will be augmented.}
     \item{\code{check_invalid_lines}}{After reading the data, check for unusual number of NA's to detect if file was corrupted by Excel or alike.}
     \item{\code{LFQ_action}}{[For type=='pg' only] An additional custom LFQ column ('cLFQ...') is created where
       zero values in LFQ columns are replaced by the following method IFF(!) the corresponding raw intensity is >0 (indicating that LFQ is erroneusly 0):
       \itemize{
         \item "toNA": replace by NA
         \item "impute": replace by lowest LFQ value >0 (simulating 'noise')
       }}
     \item{\code{...}}{Additional parameters passed on to read.delim().}
   }

   Returns a data.frame of the respective file.
  }

\item{\code{substitute(
  colname,
  valid_entries = c(NA, "", "+"),
  replacements = c(FALSE, FALSE, TRUE)
)}}{Replaces values in the mq.data member with (binary) values.

   Most MQ tables contain columns like 'contaminants' or 'reverse', whose values are either empty strings
   or "+", which is inconvenient and can be much better represented as TRUE/FALSE.
   The params \code{valid_entries} and \code{replacements} contain the matched pairs, which determine what is replaced with what.

   Arguments:
   \describe{
     \item{\code{colname}}{Name of the column (e.g. 'contaminants') in the mq.data table.}
     \item{\code{valid_entries}}{Vector of values to be replaced (must contain all values expected in the column -- fails otherwise).}
     \item{\code{replacements}}{Vector of values inserted with the same length as \code{valid_entries}.}
   }

   Returns \code{TRUE} if successful.
  }
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_ContUser}
\alias{plot_ContUser}
\title{Plot user-defined contaminants from evidence.txt}
\usage{
plot_ContUser(data, name_contaminant, extra_limit, subtitle = NULL)
}
\arguments{
\item{data}{A data.frame with columns 'fc.raw.file', 'variable', 'value'}

\item{name_contaminant}{Name of the contaminant shown in title}

\item{extra_limit}{Position where a h-line is plotted (for visual guidance)}

\item{subtitle}{Optional subtitle for plot}
}
\value{
GGplot object
}
\description{
Kolmogorov-Smirnoff p-values are plotted on top of each group.
High p-values indicate that Andromeda scores for contaminant peptides
are equal or higher compared to sample peptide scores, i.e. the probability that
sample peptides scores are NOT greater than contaminant peptide scores.
}
\examples{

 data = data.frame(fc.raw.file = letters[1:3],
                   variable = c(rep("spectralCount", 3),
                                 rep("intensity", 3),
                                 rep("above.thresh", 3),
                                 rep("score_KS", 3)),
                   value = c(10, 20, 15, 9, 21, 14, 0, 1, 1, 0.3, 0.01, 0.04))
 plot_ContUser(data, "myco", 5, "subtitle")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_MQ.R
\name{boxplotCompare}
\alias{boxplotCompare}
\title{Boxplots - one for each condition (=column) in a data frame.}
\usage{
boxplotCompare(
  data,
  log2 = TRUE,
  ylab = "intensity",
  mainlab = ylab,
  sublab = "",
  boxes_per_page = 30,
  abline = NA,
  coord_flip = TRUE,
  names = NA
)
}
\arguments{
\item{data}{Data frame in long format with numerical expression data}

\item{log2}{Apply log2 to the data (yes/no)}

\item{ylab}{Label on Y-axis}

\item{mainlab}{Main title}

\item{sublab}{Sub title}

\item{boxes_per_page}{Maximum number of boxplots per plot. Yields multiple plots if more groups are given.}

\item{abline}{Draw a horizontal green line at the specified y-position (e.g. to indicate target median values)}

\item{coord_flip}{Exchange Y and X-axis for better readability}

\item{names}{An optional data.frame(long=.., short=..), giving a renaming scheme (long->short) for the 'name' column}
}
\value{
List of ggplot objects
}
\description{
Given a data.frame with two/three columns in long format (name, value, [contaminant]; in that order), each group (given from 1st column)
is plotted as a bar.
Contaminants (if given) are separated and plotted as yellow bars.
}
\details{
Boxes are shaded: many NA or Inf lead to more transparency. Allows to easily spot sparse groups
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_Charge}
\alias{plot_Charge}
\title{The plots shows the charge distribution per Raw file.
The output of 'mosaicize()' can be used directly.}
\usage{
plot_Charge(d_charge)
}
\arguments{
\item{d_charge}{A data.frame with columns as described above}
}
\value{
GGplot object
}
\description{
The input is a data.frame with columns
  'Var1' - name of the Raw file
  'Var2' - charge (used as fill color)
  'Var1_center' - contains X-position of the Raw file
  'Var2_height' - relative frequency of the charge
  'Margin_var1' -
where each row represents one peptide sequence.
}
\examples{
 data = data.frame(raw.file = c(rep('file A', 100), rep('file B', 40)),
                       data = c(rep(2, 60), rep(3, 30), rep(4, 10),
                                rep(2, 30), rep(3, 7), rep(4, 3)))
 plot_Charge(mosaicize(data))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{lcpCount}
\alias{lcpCount}
\title{Count the number of chars of the longest common prefix}
\usage{
lcpCount(x)
}
\arguments{
\item{x}{Vector of strings with common prefix}
}
\value{
Length of LCP
}
\description{
Count the number of chars of the longest common prefix
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{thinOut}
\alias{thinOut}
\title{Thin out a data.frame by removing rows with similar numerical values in a certain column.}
\usage{
thinOut(data, filterColname, binsize)
}
\arguments{
\item{data}{The data.frame to be filtered}

\item{filterColname}{Name of the filter column as string}

\item{binsize}{Width of a bin}
}
\value{
Data.frame with reduced rows, but identical input columns
}
\description{
All values in the numerical column 'filterColname' are assigned to bins of width 'binsize'.
Only one value per bin is retained. All other rows are removed and the reduced
data frame will all its columns is returned.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_IonInjectionTimeOverRT}
\alias{plot_IonInjectionTimeOverRT}
\title{Plot line graph of TopN over Retention time.}
\usage{
plot_IonInjectionTimeOverRT(data, stats, extra_limit)
}
\arguments{
\item{data}{A data.frame with columns 'fc.raw.file', 'rRT', 'medIIT'}

\item{stats}{A data.frame with columns 'fc.raw.file', 'mean'}

\item{extra_limit}{Visual guidance line (maximum acceptable IIT)}
}
\value{
GGplot object
}
\description{
Number of Raw files must be 6 at most. Function will stop otherwise.
}
\examples{
  data = data.frame(fc.raw.file = rep(c("d","a","x"), each=100),
                    rRT = seq(20, 120, length.out = 100),
                    medIIT = c(round(runif(100, min=3, max=5)),
                               round(runif(100, min=5, max=8)),
                               round(runif(100, min=1, max=3)))
                    )
  stats = data.frame(fc.raw.file = c("d","a","x"),
                     mean = c(4, 6.5, 2))
  plot_IonInjectionTimeOverRT(data, stats, 10)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_miscGGplot.R
\name{scale_x_discrete_reverse}
\alias{scale_x_discrete_reverse}
\title{Inverse the order of items on the x-axis (for discrete scales)}
\usage{
scale_x_discrete_reverse(values, ...)
}
\arguments{
\item{values}{The vector of values as given to the x aestetic}

\item{...}{Other arguments forwarded to 'scale_y_discrete()'}
}
\value{
ggplot object, concatenatable with '+'
}
\description{
Inverse the order of items on the x-axis (for discrete scales)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_RatiosPG}
\alias{plot_RatiosPG}
\title{Plot ratios of labeled data (e.g. SILAC) from proteinGroups.txt}
\usage{
plot_RatiosPG(df_ratios, d_range, main_title, main_col, legend_title)
}
\arguments{
\item{df_ratios}{A data.frame with columns 'x', 'y', 'col', 'ltype'}

\item{d_range}{X-axis range of plot}

\item{main_title}{Plot title}

\item{main_col}{Color of title}

\item{legend_title}{Legend text}
}
\value{
GGplot object
}
\description{
The 'x' values are expected to be log2() transformed already.
}
\examples{

 x1 = seq(-3, 3, by = 0.1)
 y1 = dnorm(x1)
 x2 = seq(-5, 1, by = 0.1)
 y2 = dnorm(x2, mean = -1)
 data = data.frame( x = c(x1,x2),
                    y = c(y1,y2),
                    col = c(rep("ok", length(x1)), rep("shifted", length(x2))),
                    ltype = c(rep("solid", length(x1)), rep("dotted", length(x2))))
 plot_RatiosPG(data, range(data$x), "Ratio plot", "red", "group")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_MS2Oversampling}
\alias{plot_MS2Oversampling}
\title{Plot bargraph of oversampled 3D-peaks.}
\usage{
plot_MS2Oversampling(data)
}
\arguments{
\item{data}{A data.frame with columns 'fc.raw.file', 'n', 'fraction'}
}
\value{
GGplot object
}
\description{
Per Raw file, at most three n's must be given, i.e.
the fraction of 3D-peaks for n=1, n=2 and n=3(or more).
The fractions must sum to 1 (=100%).
}
\examples{
  data = data.frame(fc.raw.file = rep(letters[1:3], each=3),
                    n = 1:3,
                    fraction = c(0.8, 0.1, 0.1, 0.6, 0.3, 0.1, 0.7, 0.25, 0.05))
  plot_MS2Oversampling(data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_MBRgain}
\alias{plot_MBRgain}
\title{Plot MaxQuant Match-between-runs id transfer performance as a scatterplot.}
\usage{
plot_MBRgain(data, title_sub = "")
}
\arguments{
\item{data}{A data.frame with columns as described above}

\item{title_sub}{Subtitle text}
}
\value{
GGplot object
}
\description{
Per Raw file, the absolute number of transferred IDs as well as the relative gain in percent.
}
\details{
The input is a data.frame with columns
  'fc.raw.file' - raw file name
  'abs' - absolute number of transferred ID's
  'pc' - gain on top of genuine IDs [%]
where each row represents one rawfile.
}
\examples{
 data = data.frame(fc.raw.file = paste("file", letters[1:4]),
                   abs = c(5461, 5312, 3618, 502),
                   pc = c(34, 32, 22, 2))
 plot_MBRgain(data, "MBR gain: 18\%")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_TopNoverRT}
\alias{plot_TopNoverRT}
\title{Plot line graph of TopN over Retention time.}
\usage{
plot_TopNoverRT(data)
}
\arguments{
\item{data}{A data.frame with columns 'fc.raw.file', 'rRT', 'topN'}
}
\value{
GGplot object
}
\description{
Number of Raw files must be 6 at most. Function will stop otherwise.
}
\examples{
  data = data.frame(fc.raw.file = rep(letters[1:3], each=100),
                    rRT = seq(20, 120, length.out = 100),
                    topN = c(round(runif(100, min=3, max=5)),
                             round(runif(100, min=5, max=8)),
                             round(runif(100, min=1, max=3)))
                    )
  plot_TopNoverRT(data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_ContEVD}
\alias{plot_ContEVD}
\title{Plot contaminants from evidence.txt, broken down into top5-proteins.}
\usage{
plot_ContEVD(data, top5)
}
\arguments{
\item{data}{A data.frame with columns 'fc.raw.file', 'contaminant', 'pname', 'intensity'}

\item{top5}{Name of the Top-5 Proteins (by relative intensity or whatever seems relevant)}
}
\value{
GGplot object
}
\description{
Plot contaminants from evidence.txt, broken down into top5-proteins.
}
\examples{

 data = data.frame(intensity = 1:12,
                   pname = rep(letters[1:3], 4),
                   fc.raw.file = rep(paste("f", 1:4), each=3),
                   contaminant = TRUE)
 ## providing more proteins than present... d,e will be ignored
 plot_ContEVD(data, top5 = letters[1:5])
 ## classify 'c' as 'other'
 plot_ContEVD(data, top5 = letters[1:2])

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qcMetric.R
\name{flattenList}
\alias{flattenList}
\title{Flatten lists of lists with irregular depths to just a list of items,
i.e. a list of the leaves (if you consider the input as a tree).}
\usage{
flattenList(x)
}
\arguments{
\item{x}{List of 'stuff' (could be lists or items or a mix)}
}
\value{
A flat list
}
\description{
Flatten lists of lists with irregular depths to just a list of items,
i.e. a list of the leaves (if you consider the input as a tree).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_mqpar.R
\name{getMQPARValue}
\alias{getMQPARValue}
\title{Retrieve a parameter value from a mqpar.xml file}
\usage{
getMQPARValue(mqpar_filename, xpath, allow_multiple = FALSE)
}
\arguments{
\item{mqpar_filename}{Filename (incl. absolute or relative path) to the mqpar.xml file}

\item{xpath}{An XPath to extract the content of XML tag(s), e.g. '//firstSearchTol'}

\item{allow_multiple}{If the XPath expression returns more than one value, all values must be identical (not allowing multiple different values) or 'stop()' is called}
}
\value{
The stored value as string(!)
}
\description{
If the file has the param, then return it as string.
If the file is missing, warning is shown and NULL is returned.
If the param (i.e. XML tag) is unknown or cannot be extracted, the program will quit (since this is a hard error).
When multiple occurrences of the param are found (usually due to parameter groups), we test if the values are all identical.
If so, the value is returned. If the values are different, a warning is emitted and NULL is returned unless 'allow_multiple = TRUE'
}
\details{
E.g. calling getMQPARValue("mqpar.xml", "//firstSearchTol")
will look up the line
<firstSearchTol>20</firstSearchTol>
and return "20" (string!).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_peptideMods.R
\name{plot_peptideMods}
\alias{plot_peptideMods}
\title{Plot peptide modification frequencies}
\usage{
plot_peptideMods(tbl, y_max = NA, show_missing_modification_levels = TRUE)
}
\arguments{
\item{tbl}{A data.frame with 'fc.raw.file', 'modification_names' (can be a factor), and 'Freq' (0-100)}

\item{y_max}{The upper limit of the y-axis's (==Freq); useful for multiple plots with identical limits; if 'NA' the limit is computed from the given 'tbl'}

\item{show_missing_modification_levels}{If 'tbl$modification_names' is a factor and has more (but missing) levels than actually used, should missing values be dropped or assumed as '0' frequency?}
}
\value{
GGplot object
}
\description{
The input is a data.frame, as obtained from modsToTableByRaw().
}
\examples{

 data = data.frame(fc.raw.file = rep(c("file A", "file B"), each=3),
                   modifications = c("Oxidation (M)",
                   "Unmodified",
                   "Oxidation (M), Acetyl (Protein N-term)",
                   "2 Oxidation (M)", 
                   "Unmodified", 
                   "Unmodified"))
 tbl = modsToTableByRaw(data)
 plot_peptideMods(tbl,show_missing_modification_levels = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{byX}
\alias{byX}
\title{Calls FUN on a subset of data in blocks of size 'subset_size' of unique indices.}
\usage{
byX(data, indices, subset_size = 5, FUN, sort_indices = TRUE, ...)
}
\arguments{
\item{data}{Data.frame whose subsets to use on FUN}

\item{indices}{Vector of group assignments, same length as nrow(data)}

\item{subset_size}{Number of groups to use in one subset}

\item{FUN}{Function applied to subsets of data}

\item{sort_indices}{Sort groups (by their sorted character(!) names) before building subsets}

\item{...}{More arguments to FUN}
}
\value{
list of function result (one entry for each subset)
}
\description{
One subset consists of 'subset_size' unique groups and thus of all rows which
in 'data' which have any of these groups.
The last subset might have less groups, if the number of unique groups is not dividable by subset_size.
}
\details{
FUN is applied on each subset.
}
\examples{
 byX(data.frame(d=1:10), 1:10, 2, sum)
 
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_MQalign.R
\name{findAlignReference}
\alias{findAlignReference}
\title{Return list of raw file names which were reported by MaxQuant as reference point for alignment.}
\usage{
findAlignReference(data)
}
\arguments{
\item{data}{The data.frame with columns 'retention.time.calibration' and 'raw.file'}
}
\value{
List of reference raw files (usually just one)
}
\description{
There is only one reference point which has '0' in 'retention.time.calibration' column in evidence.txt 
as corrected RT. This is true for most MaxQuant versions
and also true for fractions. However, some evidence.txt files show 0.03 as an averaged minimum per Raw file.
We use the raw.file with the smallest average as reference.
}
\details{
Note that MaxQuant uses a guide tree to align the Raw files, so the order of files does not influence the 
alignment. But the first file will always be used as reference point when reporting delta-RTs. And this file is also
used by PTXQC as reference file vs all other files to find the real calibration function (see alignmentCheck()).

This function might return multiple raw file names (if MQ decides to change its mind at some point in the future).
In this case the result should be treated with caution or (better) regarded as failure.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plotTable}
\alias{plotTable}
\title{Plot a table with row names and title}
\usage{
plotTable(
  data,
  title = "",
  footer = "",
  col_names = colnames(data),
  fill = c("grey90", "grey70"),
  col = "black",
  just = "centre"
)
}
\arguments{
\item{data}{A data.frame with columns as described above}

\item{title}{Table title}

\item{footer}{Footer text}

\item{col_names}{Column names for Table}

\item{fill}{Fill pattern (by row)}

\item{col}{Text color (by column)}

\item{just}{(ignored)}
}
\value{
gTree object with class 'PTXQC_table'
}
\description{
Restriction: currently, the footer will be cropped at the table width.
}
\examples{
  data = data.frame(raw.file = letters[1:4],
                    id.rate = 3:6)
  plotTable(data,
            title = "Bad files",
            footer = "bottom",
            col_names = c("first col", "second col"),
            col=c("red", "green"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_plots.R
\name{plot_IDsOverRT}
\alias{plot_IDsOverRT}
\title{Plot IDs over time for each Raw file.}
\usage{
plot_IDsOverRT(data, x_lim = range(data$RT), y_max = max(data$counts))
}
\arguments{
\item{data}{A data.frame with columns as described above}

\item{x_lim}{Limits of the x-axis (2-tuple)}

\item{y_max}{Maximum of the y-axis (single value)}
}
\value{
GGplot object
}
\description{
Uses plot_DataOverRT() internally.
}
\examples{
 data = data.frame(fc.raw.file = rep(paste('file', letters[1:3]), each=30),
                            RT = seq(20, 120, length.out = 30),
                        counts = c(rnorm(30, 400, 20), rnorm(30, 250, 15), rnorm(30, 50, 15)))
 plot_IDsOverRT(data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{delLCP}
\alias{delLCP}
\title{Removes the longest common prefix (LCP) from a vector of strings.}
\usage{
delLCP(x, min_out_length = 0, add_dots = FALSE)
}
\arguments{
\item{x}{Vector of strings with common prefix}

\item{min_out_length}{Minimal length of the shortest element of x after LCP removal [default: 0, i.e. empty string is allowed] . If the output would be shorter, the last part of the LCP is kept.}

\item{add_dots}{Prepend output with '..' if shortening was done.}
}
\value{
Shortened vector of strings
}
\description{
You should provide only unique strings (to increase speed).
If only a single string is given, the empty string will be returned unless \code{minOutputLength} is set.
}
\examples{
  delLCP(c("TK12345_H1"), min_out_length=0)
  ## ""

  delLCP(c("TK12345_H1"), min_out_length=4)
  ## "5_H1"

  delLCP(c("TK12345_H1"), min_out_length=4, add_dots = TRUE)
  ## "..5_H1"

  delLCP(c("TK12345_H1", "TK12345_H2"), min_out_length=4)
  ## "5_H1" "5_H2"

  delLCP(c("TK12345_H1", "TK12345_H2"), min_out_length=4, add_dots = TRUE)
  ## "..5_H1" "..5_H2"

  delLCP(c("TK12345_H1", "TK12345_H2"), min_out_length=8)
  ## "12345_H1", "12345_H2"

  delLCP(c("TK12345_H1", "TK12345_H2"), min_out_length=8, add_dots = TRUE)
  ## "TK12345_H1", "TK12345_H2" (unchanged, since '..' would add another two)

  delLCP(c("TK12345_H1", "TK12345_H2"), min_out_length=60)
  ## "TK12345_H1", "TK12345_H2" (unchanged)

  delLCP(c("TK12345_H1", "TK12345_H2"), min_out_length=60, add_dots = TRUE)
  ## "TK12345_H1", "TK12345_H2" (unchanged)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_MQalign.R
\name{RTalignmentTree}
\alias{RTalignmentTree}
\title{Return a tree plot with a possible alignment tree.}
\usage{
RTalignmentTree(df_evd, col_fraction = c())
}
\arguments{
\item{df_evd}{Evidence table containing calibrated retention times and sequence information.}

\item{col_fraction}{Empty vector or 1-values vector giving the name of the fraction column (if existing)}
}
\value{
ggplot object containing the correlation tree
}
\description{
This allows the user to judge which Raw files have similar corrected RT's (i.e. where aligned successfully).
If there are clear sub-clusters, it might be worth introducing artifical fractions into MaxQuant,
to avoid ID-transfer between these clusters (use the MBR-Align and MBR-ID-Transfer metrics to support the decision).
}
\details{
If the input contains fractions, leaf nodes will be colored accordingly.
Distinct sub-clusters should have their own color.
If not, MaxQuant's fraction settings should be optimized.
Note that introducing fractions in MaxQuant will naturally lead to a clustering here (it's somewhat circular).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_MQalign.R
\name{computeMatchRTFractions}
\alias{computeMatchRTFractions}
\title{Combine several data structs into a final picture for segmentation incurred by 'Match-between-runs'.}
\usage{
computeMatchRTFractions(qMBR, qMBRSeg_Dist_inGroup)
}
\arguments{
\item{qMBR}{A data.frame as computed by peakSegmentation()}

\item{qMBRSeg_Dist_inGroup}{A data.frame as computed by inMatchWindow()}
}
\value{
A data.frame which details the distribution of singlets and pairs (inRT and outRT) for each Raw file and genuine vs. all
}
\description{
qMBRSeg_Dist_inGroup might be empty if there are only singlets (transferred and genuine), but then the scores will be pretty
boring as well (100%).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{correctSetSize}
\alias{correctSetSize}
\title{Re-estimate a new set size to split a number of items into equally sized sets.}
\usage{
correctSetSize(item_count, initial_set_size)
}
\arguments{
\item{item_count}{Known number of items which need to assigned to sets}

\item{initial_set_size}{Desired number of items a single set should hold}
}
\value{
re-estimated set size which a set should hold in order to avoid underfilled sets
}
\description{
This is useful for plotting large datasets where multiple pages are needed.
E.g. you know that you need 101 barplots, but you only want to fit about 25 per page.
Naively one would now do five plots, with the last one only containing a single barplot.
Using this function with correctSetSize(101, 25) would tell you to use 26 barplots per page,
so you end up with four plots, all roughly equally filled.
It also works the other extreme case, where your initial size is chosen slightly too high, e.g.
Sets of size 5 for just 8 items is too much, because we can reduce the set size to 4 and still
need two sets but now they are much more equally filled (correctSetSize(8, 5) == 4).
}
\details{
We allow for up to set sizes of 150\% from default, to avoid the last set being sparse (we remove it and distribute to the other bins)
150%\ oversize is the extreme case, which only happens with sets of size two. With more sets the overhead is much smaller (1/X).
Once the number of sets is fixed, we distribute all items equally.

E.g. 6 items & initial_set_size=5, would result in 2 bins (5 items, 1 item), but we'd rather have one bin of 6 items
or 8 items & initial_set_size=5, would result in 2 bins (5+3 items), since the last set is more than half full, but we'd rather have 4+4
}
\examples{
 stopifnot(
   correctSetSize(8, 5) == 4
 )
 stopifnot(
   correctSetSize(101, 25) == 26
 )
 
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_qualities.R
\name{qualLinThresh}
\alias{qualLinThresh}
\title{Quality metric with linear response to input, reaching the maximum score at the given threshold.}
\usage{
qualLinThresh(x, t = 1)
}
\arguments{
\item{x}{Numeric value(s) between [0, inf]}

\item{t}{Threshold value, which indicates 100\%}
}
\value{
Value between [0, 1]
}
\description{
Ranges between 0 (worst score) and 1 (best score).
Useful for performance measures where reaching a certain reference threshold 't'
will be enough to reach 100\%.
The input range from [0, t] is scored from 0-100\%.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{getECDF}
\alias{getECDF}
\title{Estimate the empirical density and return it}
\usage{
getECDF(samples, y_eval = (1:100)/100)
}
\arguments{
\item{samples}{Vector of input values (samples from the distribution)}

\item{y_eval}{Vector of points where CDF is evaluated (each percentile by default)}
}
\value{
Data.frame with columns 'x', 'y'
}
\description{
Estimate the empirical density and return it
}
\examples{
  plot(getECDF(rnorm(1e4)))
  
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{getPeptideCounts}
\alias{getPeptideCounts}
\title{Extract the number of peptides observed per Raw file
from an evidence table.}
\usage{
getPeptideCounts(df_evd)
}
\arguments{
\item{df_evd}{Data.frame of evidence.txt as read by MQDataReader}
}
\value{
Data.frame with columns 'fc.raw.file', 'counts', 'category', 'MBRgain'
}
\description{
Required columns are "fc.raw.file", "modified.sequence" and "is.transferred".
}
\details{
If match-between-runs was enabled during the MaxQuant run,
the data.frame returned will contain separate values for 'transferred' evidence
plus an 'MBRgain' column, which will give the extra MBR evidence in percent.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{mosaicize}
\alias{mosaicize}
\title{Prepare a Mosaic plot of two columns in long format.}
\usage{
mosaicize(data)
}
\arguments{
\item{data}{A data.frame with exactly two columns}
}
\value{
Data.frame
}
\description{
Found at http://stackoverflow.com/questions/19233365/how-to-create-a-marimekko-mosaic-plot-in-ggplot2
Modified (e.g. to pass R check)
}
\details{
Returns a data frame, which can be used for plotting and has the following columns:
'Var1' - marginalized values from 1st input column
'Var2' - marginalized values from 2nd input column
'Freq' - relative frequency of the combination given in [Var1, Var2]
'margin_var1' - frequency of the value given in Var1
'var2_height' - frequency of the value given in Var2, relative to Var1
'var1_center' - X-position when plotting (large sets get a larger share)
}
\examples{
  data = data.frame(raw.file = c(rep('file A', 100), rep('file B', 40)),
                    charge = c(rep(2, 60), rep(3, 30), rep(4, 10),
                               rep(2, 30), rep(3, 7), rep(4, 3)))
  mosaicize(data)                             
  
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{shortenStrings}
\alias{shortenStrings}
\title{Shorten a string to a maximum length and indicate shorting by appending '..'}
\usage{
shortenStrings(x, max_len = 20, verbose = TRUE, allow_duplicates = FALSE)
}
\arguments{
\item{x}{Vector of input strings}

\item{max_len}{Maximum length allowed}

\item{verbose}{Print which strings were shortened}

\item{allow_duplicates}{If shortened strings are not discernible any longer, consider the short version valid (not the default), otherwise (default) return the full string (--> no-op)}
}
\value{
A vector of shortened strings
}
\description{
Some axis labels are sometimes just too long and printing them will either
squeeze the actual plot (ggplot) or make the labels disappear beyond the margins (graphics::plot)
One ad-hoc way of avoiding this is to shorten the names, hoping they are still meaningful to the viewer.
}
\details{
This function should be applied AFTER you tried more gentle methods, such as \code{\link{delLCP}} or \code{\link{simplifyNames}}.
}
\examples{
r = shortenStrings(c("gamg_101", "gamg_101230100451", "jurkat_06_100731121305", "jurkat_06_1"))
all(r == c("gamg_101", "gamg_101230100..", "jurkat_06_1007..", "jurkat_06_1"))

}
\seealso{
\code{\link{delLCP}}, \code{\link{simplifyNames}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{pasten}
\alias{pasten}
\title{paste with newline as separator}
\usage{
pasten(...)
}
\arguments{
\item{...}{Arguments forwarded to paste()}
}
\value{
return value of paste()
}
\description{
paste with newline as separator
}
\examples{
  pasten("newline","separated")
  ## --> "newline\nseparated"
  
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_misc.R
\name{simplifyNames}
\alias{simplifyNames}
\title{Removes common substrings (infixes) in a set of strings.}
\usage{
simplifyNames(
  strings,
  infix_iterations = 2,
  min_LCS_length = 7,
  min_out_length = 7
)
}
\arguments{
\item{strings}{A vector of strings which are to be shortened}

\item{infix_iterations}{Number of successive rounds of substring removal}

\item{min_LCS_length}{Minimum length of the longest common substring (default:7, minimum: 6)}

\item{min_out_length}{Minimum length of shortest element of output (no shortening will be done which causes output to be shorter than this threshold)}
}
\value{
A list of shortened strings, with the same length as the input
}
\description{
Usually handy for plots, where condition names should be as concise as possible.
E.g. you do not want names like 
'TK20130501_H2M1_010_IMU008_CISPLA_E3_R1.raw' and 
'TK20130501_H2M1_026_IMU008_CISPLA_E7_R2.raw' 
but rather 'TK.._010_I.._E3_R1.raw' and
           'TK.._026_I.._E7_R2.raw'
           
If multiple such substrings exist, the algorithm will remove the longest first and iterate
a number of times (two by default) to find the second/third etc longest common substring.
Each substring must fulfill a minimum length requirement - if its shorter, its not considered worth removing
and the iteration is aborted.
}
\examples{
#library(PTXQC)
simplifyNames(c('TK20130501_H2M1_010_IMU008_CISPLA_E3_R1.raw',
                'TK20130501_H2M1_026_IMU008_CISPLA_E7_R2.raw'), infix_iterations = 2)
# --> "TK.._010_I.._E3_R1.raw","TK.._026_I.._E7_R2.raw"

try(simplifyNames(c("bla", "foo"), min_LCS_length=5))
# --> error, since min_LCS_length must be >=6

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_qualities.R
\name{scale01linear}
\alias{scale01linear}
\title{Scales a vector of values linearly to [0, 1]
If all input values are equal, returned values are all 0}
\usage{
scale01linear(X)
}
\arguments{
\item{X}{Vector of values}
}
\value{
Scaled vector
}
\description{
Scales a vector of values linearly to [0, 1]
If all input values are equal, returned values are all 0
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcn_miscGGplot.R
\name{brewer.pal.Safe}
\alias{brewer.pal.Safe}
\title{Return color brew palettes, but fail hard if number of requested colors
is larger than the palette is holding.}
\usage{
brewer.pal.Safe(n = 3, palette = "Set1")
}
\arguments{
\item{n}{Number of colours}

\item{palette}{Name of palette (e.g. "set1")}
}
\value{
character vector of colors
}
\description{
Internally calls 'brewer.pal(n, palette)', checking 'n' beforehand.
}
